//
// $Id: file.cpp 111 2008-08-19 10:24:14Z mrchip $
//
// Copyright (C) 2005 STADT-LAND-FLUSS INGENIEURDIENSTE GmbH
//
// Authors:
// Ulrich Kiel <u.kiel@S-L-F.de>
//
// This program is free software under the GPL (>=v2)
// Read the file COPYING coming with WSPLGEN for details.
//

#ifdef __BORLANDC__
#pragma hdrstop
#endif

//---------------------------------------------------------------------
#include <stdio.h>

#include "shape.h"
#include "tools.h"
#include "tri.h"
#include "file.h"

//---------------------------------------------------------------------
static char	PrjIn[5000] = "";
static char	PrjOut[5000] = "";

//---------------------------------------------------------------------
bool LoadPar(std::string FileName, std::vector<std::string> *ParList)
{
	write_fortschritt("->Laden der Parameterdatei '%s' gestartet\n", FileName.c_str());

	FILE *fh = fopen(FileName.c_str(), "r");
	if (fh == 0)
	{
		write_error(2201, "Konnte '%s' nicht zum Lesen ffnen\n", FileName.c_str());
	}

	char line[1000];

	while (fgets (line, sizeof (line)-1, fh))
	{
		line[sizeof(line)-1] = '\0';
		if (strlen(line) > 0) line[strlen(line)-1] = '\0';

		if ('\n' == line[strlen(line)-1])	line[strlen(line)-1] = '\0';
		if ('\r' == line[strlen(line)-1])	line[strlen(line)-1] = '\0';
		if ('\n' == line[strlen(line)-1])	line[strlen(line)-1] = '\0';

		ParList->push_back(line);
	}

	fclose(fh);

	write_fortschritt("<-Laden der Parameterdatei '%s' beendet\n", FileName.c_str());

	return (true);
}

//---------------------------------------------------------------------
bool LoadPRJ(std::string FileName, int DebugLevel)
{
	write_fortschritt("->Laden der Projektionsdatei gestartet\n", FileName.c_str());

	std::string PRJFileName = ExchangeFileExt(FileName, ".prj");
	FILE *ph = fopen(PRJFileName.c_str(), "r");
	if (!ph)
	{
		PRJFileName = ExchangeFileExt(FileName, ".PRJ");
		ph = fopen(PRJFileName.c_str(), "r");
	}
	if (!ph)
	{
		PRJFileName = ExchangeFileExt(FileName, ".Prj");
		ph = fopen(PRJFileName.c_str(), "r");
	}

	if (ph)
	{
		if (0 == fgets(PrjIn, sizeof(PrjIn), ph))
		{
			PrjIn[0] = '\0';
			write_fortschritt("Projektionsdatei '%s' konnte nicht gelesen werden\n", PRJFileName.c_str());
		}
		else
		{
			write_fortschritt("Projektionsdatei '%s' gelesen\n", PRJFileName.c_str());
			if (DebugLevel >= 6)
			{
				write_fortschritt("Projektion: '%s'\n", PrjIn);
			}

			if (PrjOut[0] == '\0')
			{
				strncpy(PrjOut, PrjIn, sizeof(PrjOut)-1);
				PrjOut[sizeof(PrjOut)-1] = '\0';
			}
			else
			{
				if (strncmp(PrjOut, PrjIn, sizeof(PrjOut)-1) != 0)
				{
					write_warning(3115, "Nicht alle Projektionen sind gleich\nDies kann zu merkwrdigen Ergebnissen fhren\n");
				}
			}
		}
		fclose(ph);
	}
	else
	{
		write_fortschritt("Projektionsdatei '%s' konnte nicht geffnet werden\n", PRJFileName.c_str());
	}
	write_fortschritt("<-Laden der Projektionsdatei beendet\n", FileName.c_str());

	return (true);
}

//---------------------------------------------------------------------
bool LoadDGM(std::string FileName, TNodeList *NodeList, TElementList *ElementList, TXYZList *BegrenzungsPolygon, int DebugLevel)
{
	write_fortschritt("->Laden des DGM in Datei '%s' gestartet\n", FileName.c_str());

	if (NodeList == 0)		dump_error(__FILE__, __LINE__, "Die Knotenliste ist nicht definiert\n");
	if (ElementList == 0)	dump_error(__FILE__, __LINE__, "Die Elementliste ist nicht definiert\n");

	NodeList->Clear();
	ElementList->Clear();

	std::string Ext = GetFileExt(FileName);


	if (ToUpperCase(Ext) == ".ADF")
	{
		LoadDGMTIN(FileName, NodeList, ElementList, BegrenzungsPolygon, DebugLevel);
		CheckForDuplicates(NodeList, ElementList, DebugLevel);
	}
	else if (ToUpperCase(Ext) == ".2DM")
	{
		LoadDGM2DM(FileName, NodeList, ElementList, BegrenzungsPolygon, DebugLevel);
		CheckForDuplicates(NodeList, ElementList, DebugLevel);
	}
	else if (ToUpperCase(Ext) == ".GRD" || ToUpperCase(Ext) == ".ASC")
	{
		LoadDGMGRD(FileName, NodeList, BegrenzungsPolygon, DebugLevel);
		CheckForDuplicates(NodeList, ElementList, DebugLevel);
	}
	else if (ToUpperCase(Ext) == ".XYZ" || ToUpperCase(Ext) == ".TXT")
	{
		LoadDGMXYZ(FileName, NodeList, BegrenzungsPolygon, DebugLevel);
		CheckForDuplicates(NodeList, ElementList, DebugLevel);
	}
	else if (ToUpperCase(Ext) == ".SHP")
	{
		LoadDGMSHP(FileName, NodeList, BegrenzungsPolygon, DebugLevel);
		CheckForDuplicates(NodeList, ElementList, DebugLevel);
	}
	else
	{
		write_error(1222, "Ungltige Dateinamenserweiterung '%s' beim Dateinamen '%s' fr den Parameter -DGM\nErlaubt sind nur ADF, GRD, ASC, 2DM, XYZ, TXT oder SHP\n", ToUpperCase(Ext).c_str(), FileName.c_str());
	}

	write_fortschritt("<-Laden des DGM in Datei '%s' beendet\n", FileName.c_str());

	return (true);
}

//---------------------------------------------------------------------------
bool LoadXYZ(FILE* fhtnxy, FILE* fhtnz, int Nr, long* XL, long* YL, int* ZI)
{
	double X = 0.0;
	double Y = 0.0;
	float Z = (float)0.0;

	fseek (fhtnxy, (Nr-1) * (sizeof(X) + sizeof (Y)), SEEK_SET);

	if (1 != fread(&X, sizeof(X), 1, fhtnxy) || 1 != fread(&Y, sizeof(Y), 1, fhtnxy))
	{
		write_error(2226, "Konnte keine X/Y-Werte fr den Knoten %d lesen\n", Nr);
	}

	fseek (fhtnz, (Nr-1) * sizeof(Z), SEEK_SET);

	if (1 != fread(&Z, sizeof(Z), 1, fhtnz))
	{
		write_error(2227, "Konnte keinen Z-Wert fr den Knoten %d lesen\n", Nr);
	}

	Swap8Bytes(&X);
	Swap8Bytes(&Y);
	Swap4Bytes(&Z);

	if (Z > 1e30 || Z < -1e30)	return (false);

	*XL = (long)(X * 100.0 + 0.5);
	*YL = (long)(Y * 100.0 + 0.5);
	*ZI = (long)(Z * 100.0 + 0.5);

	return (true);
}

//---------------------------------------------------------------------------
bool LoadDGMTIN(std::string FileName, TNodeList *NodeList, TElementList *ElementList, TXYZList *Begrenzung, int DebugLevel)
{
	write_fortschritt("->Laden des DGM in Datei '%s' als TIN gestartet\n", FileName.c_str());

	std::string ThulFileName = GetFilePath(FileName) + "thul.adf";
	std::string TnxyFileName = GetFilePath(FileName) + "tnxy.adf";
	std::string TnzFileName = GetFilePath(FileName) + "tnz.adf";
	std::string TnodFileName = GetFilePath(FileName) + "tnod.adf";
//	std::string TedgFileName = GetFilePath(FileName) + "tedg.adf";

	FILE *fhthul = 0;
	FILE *fhtnxy = 0;
	FILE *fhtnod = 0;
	FILE *fhtnz = 0;
//	FILE *fhtedg = 0;

	fhtnxy = fopen(TnxyFileName.c_str(), "rb");
	if (fhtnxy == 0)	write_error(2201, "Konnte '%s' nicht zum Lesen ffnen\n", TnxyFileName.c_str());

	fhtnod = fopen(TnodFileName.c_str(), "rb");
	if (fhtnod == 0)	write_error(2201, "Konnte '%s' nicht zum Lesen ffnen\n", TnodFileName.c_str());

	fhtnz = fopen(TnzFileName.c_str(), "rb");
	if (fhtnz == 0)	write_error(2201, "Konnte '%s' nicht zum Lesen ffnen\n", TnzFileName.c_str());


	TXYZList Huelle;

	bool IsSetBegrenzung = false;
	if (Begrenzung && Begrenzung->size() >= 3)	IsSetBegrenzung = true;

	if (!IsSetBegrenzung)
	{
		write_fortschritt("->Laden der Umhllung gestartet\n");

		fhthul = fopen(ThulFileName.c_str(), "rb");
		if (fhthul == 0)	write_error(2201, "Konnte '%s' nicht zum Lesen ffnen\n", ThulFileName.c_str());

		int Nr = 0;

		// Zuerst kommt scheinbar die Umhllung aller Knoten
		// Also auch oder besser gesagt genau der Wchterknoten
		do
		{
			fread(&Nr, sizeof(Nr), 1, fhthul);
			Swap4Bytes(&Nr);
		} while (Nr > 0);

		// Jetzt die eigentlichen Knoten

		unsigned int AnzHul = 0;

		while (1 == fread(&Nr, sizeof(Nr), 1, fhthul))
		{
			Swap4Bytes(&Nr);

			AnzHul++;

			double X = 0.0;
			double Y = 0.0;

			fseek (fhtnxy, (Nr-1) * (sizeof(X) + sizeof (Y)), SEEK_SET);

			if (1 != fread(&X, sizeof(X), 1, fhtnxy) || 1 != fread(&Y, sizeof(Y), 1, fhtnxy))
			{
				write_error(2226, "Konnte keine 2 Werte fr den Knoten %d lesen\n", Nr);
			}

			Swap8Bytes(&X);
			Swap8Bytes(&Y);

			long XL = (long)(X * 100.0 + 0.5);
			long YL = (long)(Y * 100.0 + 0.5);

			Huelle.Add(XL, YL);

			if (DebugLevel >= 9 || (DebugLevel >= 1 && AnzHul % 100 == 0) || (AnzHul > 0 && AnzHul % 1000 == 0))
			{
				write_fortschritt("%d Sttzstellen gelesen\n", AnzHul);
			}
		}

		fclose(fhthul);

		write_fortschritt("%d Sttzstellen gelesen\n", AnzHul);

		if (DebugLevel >= 9)
		{
			std::string HuelleFileName = ExchangeFileExt(FileName, "_huelle.shp");
			SavePolygon(HuelleFileName, &Huelle, DebugLevel);
		}

		write_fortschritt("<-Laden der Umhllung beendet\n");
	}


	write_fortschritt("->Laden der Elemente gestartet\n");

	// Nur so funktioniert spter AddWithQuadTree

	long MinX = 0;
	long MinY = 0;
	long MaxX = 0;
	long MaxY = 0;

	if (IsSetBegrenzung)
	{
		MinX = Begrenzung->MinX;
		MinY = Begrenzung->MinY;
		MaxX = Begrenzung->MaxX;
		MaxY = Begrenzung->MaxY;
	}
	else
	{
		MinX = Huelle.MinX;
		MinY = Huelle.MinY;
		MaxX = Huelle.MaxX;
		MaxY = Huelle.MaxY;
	}

	NodeList->CreateQuadTree(MinX, MinY, MaxX, MaxY);

	int Nr1 = -1;
	int Nr2 = -1;
	int Nr3 = -1;

//	int Ei1 = -1;
//	int Ei2 = -1;
//	int Ei3 = -1;

	int AnzElements = 0;
//	while (1 == fread(&Nr1, sizeof(Nr1), 1, fhtnod) && fread(&Nr2, sizeof(Nr2), 1, fhtnod) && 1 == fread(&Nr3, sizeof(Nr3), 1, fhtnod) && 1 == fread(&Ei1, sizeof(Ei1), 1, fhtedg) && fread(&Ei2, sizeof(Ei2), 1, fhtedg) && 1 == fread(&Ei3, sizeof(Ei3), 1, fhtedg))
	while (1 == fread(&Nr1, sizeof(Nr1), 1, fhtnod) && fread(&Nr2, sizeof(Nr2), 1, fhtnod) && 1 == fread(&Nr3, sizeof(Nr3), 1, fhtnod))
	{
		Swap4Bytes((byte *)&Nr1);
		Swap4Bytes((byte *)&Nr2);
		Swap4Bytes((byte *)&Nr3);
//		Swap4Bytes((byte *)&Ei1);
//		Swap4Bytes((byte *)&Ei2);
//		Swap4Bytes((byte *)&Ei3);

		AnzElements++;

		if (DebugLevel >= 9 || (DebugLevel >= 1 && AnzElements % 5000 == 0) || (AnzElements > 0 && AnzElements % 50000 == 0))
		{
			write_fortschritt("%d von %d Elementen im Speicher\n", ElementList->size(), AnzElements);
		}

/*
		TTinTyp TinTyp = TTREGULAR;

		if (Ei1 < 0)
		{
			Ei1 = -Ei1;
			int Typ = Ei1 & 0x40000000;

			if (Typ == 0x00000000)	TinTyp = TTSOFT;
			else					TinTyp = TTHARD;
		}

		if (Ei2 < 0)
		{
			Ei2 = -Ei2;
			int Typ = Ei2 & 0x40000000;

			if (Typ == 0x00000000)	TinTyp = TTSOFT;
			else 					TinTyp = TTHARD;
		}

		if (Ei3 < 0)
		{
			Ei3 = -Ei3;
			int Typ = Ei3 & 0x40000000;

			if (Typ == 0x00000000)	TinTyp = TTSOFT;
			else					TinTyp = TTHARD;
		}
*/

		long X1 = 0;
		long Y1 = 0;
		int Z1 = 0;
		long X2 = 0;
		long Y2 = 0;
		int Z2 = 0;
		long X3 = 0;
		long Y3 = 0;
		int Z3 = 0;

		if (false == LoadXYZ(fhtnxy, fhtnz, Nr1, &X1, &Y1, &Z1)) continue;
		if (false == LoadXYZ(fhtnxy, fhtnz, Nr2, &X2, &Y2, &Z2)) continue;
		if (false == LoadXYZ(fhtnxy, fhtnz, Nr3, &X3, &Y3, &Z3)) continue;

		if (X1 < MinX || X1 > MaxX || Y1 < MinY || Y1 > MaxY)	continue;
		if (X2 < MinX || X2 > MaxX || Y2 < MinY || Y2 > MaxY)	continue;
		if (X3 < MinX || X3 > MaxX || Y3 < MinY || Y3 > MaxY)	continue;

		long Xm = (long)(X1 / 3.0 + X2 / 3.0 + X3 / 3.0);
		long Ym = (long)(Y1 / 3.0 + Y2 / 3.0 + Y3 / 3.0);

		TInsideTyp Inside = INSIDE;
		if (IsSetBegrenzung) Inside = Begrenzung->IsInsideXYList(Xm, Ym);
		else		   		 Inside = Huelle.IsInsideXYList(Xm, Ym);

		if (Inside == NOT_INSIDE)	continue;

		TNode *Node1 = NodeList->FindByXY(X1, Y1);
		if (0 == Node1)
		{
			Node1 = new TNode(Nr1, X1, Y1, Z1);
			NodeList->AddWithQuadTree(Node1);
		}

		TNode *Node2 = NodeList->FindByXY(X2, Y2);
		if (0 == Node2)
		{
			Node2 = new TNode(Nr2, X2, Y2, Z2);
			NodeList->AddWithQuadTree(Node2);
		}

		TNode *Node3 = NodeList->FindByXY(X3, Y3);
		if (0 == Node3)
		{
			Node3 = new TNode(Nr3, X3, Y3, Z3);
			NodeList->AddWithQuadTree(Node3);
		}

		TElement *Element = new TElement(Node1, Node2, Node3);
		ElementList->Add(Element);
	}

//	fclose(fhtedg);
	fclose(fhtnxy);
	fclose(fhtnz);
	fclose(fhtnod);

	write_fortschritt("%d von %d Elementen im Speicher\n", ElementList->size(), AnzElements);

	write_fortschritt("<-Laden der Elemente beendet\n");

	write_fortschritt("<-Laden des DGM in Datei '%s' als TIN beendet\n", FileName.c_str());

	return (true);
}

//---------------------------------------------------------------------------
bool LoadDGMGRD(std::string FileName, TNodeList *NodeList, TXYZList *XyList, int DebugLevel)
{
	write_fortschritt("->Laden des DGM in Datei '%s' als GRD gestartet\n", FileName.c_str());

	FILE *fh = fopen(FileName.c_str(), "r");
	if (fh == 0)
	{
		write_error(2201, "Konnte '%s' nicht zum Lesen ffnen\n", FileName.c_str());
	}

	char line[1000];
	std::string Line;

	int 	NCols = 0;
	int 	NRows = 0;
	bool	Center = false;
	double 	Xll = 0;
	double	Yll = 0;
	double	CellSize = 0.0;
	double	NoDataValue = 0.0;

	////////////
	// NCols
	////////////
	if (0 == fgets (line, sizeof(line)-1, fh))
	{
		write_error(2298, "Konnte die 1. Kopfzeile in der Datei '%s' nicht lesen\n", FileName.c_str());
	}
	line[sizeof(line)-1] = '\0';

	Line = line;
	Line = ToUpperCase(Line);

	if (1 != sscanf(Line.c_str(), "NCOLS%d", &NCols))
	{
		write_error(2215, "Konnte in der 1. Kopfzeile der Datei '%s'\nnicht das Schlsselwort 'ncols' gefolgt von einer Zahl lesen\n", FileName.c_str());
	}

	if (NCols <= 0)
	{
		write_error(3209, "In der Datei '%s' ist NCols (%d) kleiner gleich Null.\nNCols muss immer positiv sein.\n", FileName.c_str(), NCols);
	}

	////////////
	// NRows
	////////////
	if (0 == fgets (line, sizeof(line)-1, fh))
	{
		write_error(2298, "Konnte die 2. Kopfzeile in der Datei '%s' nicht lesen\n", FileName.c_str());
	}
	line[sizeof(line)-1] = '\0';

	Line = line;
	Line = ToUpperCase(Line);

	if (1 != sscanf(Line.c_str(), "NROWS%d", &NRows))
	{
		write_error(2215, "Konnte in der 2. Kopfzeile der Datei '%s'\nnicht das Schlsselwort 'nrows' gefolgt von einer Zahl lesen\n", FileName.c_str());
	}

	if (NRows <= 0)
	{
		write_error(3209, "In der Datei '%s' ist NRows (%d) kleiner gleich Null.\nNRows muss immer positiv sein.\n", FileName.c_str(), NRows);
	}

	////////////
	// XLLCENTER oder XLLCORNER
	////////////
	if (0 == fgets (line, sizeof(line)-1, fh))
	{
		write_error(2298, "Konnte die 3. Kopfzeile in der Datei '%s' nicht lesen\n", FileName.c_str());
	}
	line[sizeof(line)-1] = '\0';

	Line = line;
	Line = ToUpperCase(Line);

	if (1 != sscanf(Line.c_str(), "XLLCORNER%lf", &Xll))
	{
		if (1 != sscanf(Line.c_str(), "XLLCENTER%lf", &Xll))
		{
			write_error(2215, "Konnte in der 3. Kopfzeile der Datei '%s'\nweder das Schlsselwort 'xllcorner'\nnoch das Schlsselwort 'xllcenter' gefolgt von einer Zahl lesen\n", FileName.c_str());
		}
		else
		{
			Center = true;
		}
	}

	////////////
	// YLLCENTER oder YLLCORNER
	////////////
	if (0 == fgets (line, sizeof(line)-1, fh))
	{
		write_error(2298, "Konnte die 4. Kopfzeile in der Datei '%s' nicht lesen\n", FileName.c_str());
	}
	line[sizeof(line)-1] = '\0';

	Line = line;
	Line = ToUpperCase(Line);

	if (Center)
	{
		if (1 != sscanf(Line.c_str(), "YLLCENTER%lf", &Yll))
		{
			write_error(2215, "Konnte in der 4. Kopfzeile der Datei '%s'\nnicht das Schlsselwort 'yllcenter' gefolgt von einer Zahl lesen\n", FileName.c_str());
		}
	}
	else
	{
		if (1 != sscanf(Line.c_str(), "YLLCORNER%lf", &Yll))
		{
			write_error(2215, "Konnte in der 4. Kopfzeile der Datei '%s'\nnicht das Schlsselwort 'yllcorner' gefolgt von einer Zahl lesen\n", FileName.c_str());
		}
	}

	////////////
	// CellSize
	////////////
	if (0 == fgets (line, sizeof(line)-1, fh))
	{
		write_error(2298, "Konnte die 5. Kopfzeile in der Datei '%s' nicht lesen\n", FileName.c_str());
	}
	line[sizeof(line)-1] = '\0';

	Line = line;
	Line = ToUpperCase(Line);

	if (1 != sscanf(Line.c_str(), "CELLSIZE%lf", &CellSize))
	{
		write_error(2215, "Konnte in der 5. Kopfzeile der Datei '%s'\nnicht das Schlsselwort 'cellsize' gefolgt von einer Zahl lesen\n", FileName.c_str());
	}

	if (CellSize <= 0)
	{
		write_error(3211, "In der Datei '%s' ist CellSize (%.2f) kleiner gleich Null.\nCellSize muss immer positiv sein.\n", FileName.c_str(), CellSize);
	}

	////////////
	// NoDataValue
	////////////
	if (0 == fgets (line, sizeof(line)-1, fh))
	{
		write_error(2298, "Konnte die 6. Kopfzeile in der Datei '%s' nicht lesen\n", FileName.c_str());
	}
	line[sizeof(line)-1] = '\0';

	Line = line;
	Line = ToUpperCase(Line);

	if (1 != sscanf(Line.c_str(), "NODATA_VALUE%lf", &NoDataValue))
	{
		write_error(2215, "Konnte in der 6. Kopfzeile der Datei '%s'\nnicht das Schlsselwort 'nodata_value' gefolgt von einer Zahl lesen\n", FileName.c_str());
	}


	unsigned int MaxWerte = NCols * NRows;
	unsigned int AnzWerte = 0;
	unsigned int AnzKnoten = 0;
	for (unsigned int i=0; i<MaxWerte; i++)
	{
		if (1 != fscanf (fh, "%s", line))
		{
			write_error(2298, "Konnte den %d. Wert in der Datei '%s' nicht lesen\n", AnzWerte+1, FileName.c_str());
		}

		for (unsigned int i=0; i<strlen(line); i++)
		{
			if (line[i] == ',') line[i] = '.';
		}

		double Z = 0.0;
		double X = Xll + (AnzWerte % NCols) * CellSize;
		double Y = Yll + (NRows-1) * CellSize - (AnzWerte / NCols) * CellSize;
		if (false == Center)
		{
			X = X + CellSize / 2.0;
			Y = Y + CellSize / 2.0;
		}

		if (1 != sscanf (line, "%lf", &Z))
		{
			write_error(2298, "Konnte den %d. Wert in der Datei '%s' nicht lesen\n", AnzWerte+1, FileName.c_str());
		}

		AnzWerte++;

		if (DebugLevel >= 9 || (DebugLevel >= 1 && AnzWerte % 10000 == 0) || (AnzWerte > 0 && AnzWerte % 100000 == 0))
		{
			write_fortschritt("%d von %d Werten gelesen, %d von %d Knoten im Speicher\n", AnzWerte, MaxWerte, NodeList->size(), AnzKnoten);
		}

		long XL = (long)(X * 100 + 0.5);
		long YL = (long)(Y * 100 + 0.5);
		int ZI = (long)(Z * 100 + 0.5);

		TInsideTyp InsideTyp = INSIDE;
		if (XyList && XyList->size() > 2) InsideTyp = XyList->IsInsideXYList(XL, YL);
		if (InsideTyp == INSIDE || InsideTyp == ON_LINE)
		{
			AnzKnoten++;

			TNode *Node = new TNode(AnzKnoten, XL, YL, ZI);
			NodeList->Add(Node);
		}
	}
	write_fortschritt("%d von %d Werten gelesen, %d von %d Knoten im Speicher\n", AnzWerte, MaxWerte, NodeList->size(), AnzKnoten);

	fclose(fh);

	write_fortschritt("<-Laden des DGM in Datei '%s' als GRD beendet\n", FileName.c_str());

	return (true);
}

//---------------------------------------------------------------------------
bool LoadDGMXYZ(std::string FileName, TNodeList *NodeList, TXYZList *XyList, int DebugLevel)
{
	write_fortschritt("->Laden des DGM in Datei '%s' als XYZ gestartet\n", FileName.c_str());

	FILE *fh = fopen(FileName.c_str(), "r");
	if (fh == 0)
	{
		write_error(2201, "Konnte '%s' nicht zum Lesen ffnen\n", FileName.c_str());
	}

	char line[1000];

	if (0 == fgets (line, sizeof(line)-1, fh))
	{
		write_error(2298, "Konnte die Kopfzeile in der Datei '%s' nicht lesen\n", FileName.c_str());
	}
	line[sizeof(line)-1] = '\0';

	unsigned int AnzZeilen = 1;
	unsigned int AnzWerte = 0;
	unsigned int AnzKnoten = 0;
	while (0 != fgets (line, sizeof(line)-1, fh))
	{
		AnzZeilen ++;

		double X = 0.0;
		double Y = 0.0;
		double Z = 0.0;

		bool Leer = true;
		for (unsigned int i=0; i<strlen(line); i++)
		{
			if (line[i] == ',') line[i] = '.';
			if (line[i] != ' ' && line[i] != '\t' && line[i] != '\n' && line[i] != '\r') Leer = false;
		}

		if (Leer)	continue;

		if (3 != sscanf (line, "%lf%lf%lf", &X, &Y, &Z))
		{
			write_error(2297, "Konnte keine 3 Werte in der Datei '%s' in der Zeile %d lesen\n", FileName.c_str(), AnzZeilen);
		}

		AnzWerte++;

		if (DebugLevel >= 9 || (DebugLevel >= 1 && AnzWerte % 10000 == 0) || (AnzWerte > 0 && AnzWerte % 100000 == 0))
		{
			write_fortschritt("%d Werte gelesen, %d von %d Knoten im Speicher\n", AnzWerte, NodeList->size(), AnzKnoten);
		}

		long XL = (long)(X * 100 + 0.5);
		long YL = (long)(Y * 100 + 0.5);
		int ZI = (long)(Z * 100 + 0.5);

		TInsideTyp InsideTyp = INSIDE;
		if (XyList && XyList->size() > 2) InsideTyp = XyList->IsInsideXYList(XL, YL);
		if (InsideTyp == INSIDE || InsideTyp == ON_LINE)
		{
			AnzKnoten++;

			TNode *Node = new TNode(AnzKnoten, XL, YL, ZI);
			NodeList->Add(Node);
		}
	}
	write_fortschritt("%d Werte gelesen, %d von %d Knoten im Speicher\n", AnzWerte, NodeList->size(), AnzKnoten);

	fclose(fh);

	write_fortschritt("<-Laden des DGM in Datei '%s' als XYZ beendet\n", FileName.c_str());

	return (true);
}

//---------------------------------------------------------------------------
bool LoadDGMSHP(std::string FileName, TNodeList *NodeList, TXYZList *XyList, int DebugLevel)
{
	write_fortschritt("->Laden des DGM in Datei '%s' als SHP gestartet\n", FileName.c_str());

	LoadPRJ(FileName, DebugLevel);

	std::string SHPFileName = ExchangeFileExt(FileName, ".shp");
	SHPHandle hSHP = SHPOpen(SHPFileName.c_str(), "rb");

	if (hSHP == NULL)
	{
		write_error(2201, "Konnte '%s' nicht zum Lesen ffnen\n", SHPFileName.c_str());
	}

	int 		RecordCount = 0;
	TShpType 	ShapeType = SHPT_NULL;
	double		Mins[4];
	double		Maxs[4];

	std::string DBFFileName = ExchangeFileExt(FileName, ".dbf");
	DBFHandle hDBF = DBFOpen(DBFFileName.c_str(), "rb");
	if (hDBF == NULL)
	{
		SHPClose(hSHP);
		write_error(2201, "Konnte '%s' nicht zum Lesen ffnen\n", DBFFileName.c_str());
		return (false);
	}

	SHPGetInfo(hSHP, &RecordCount, &ShapeType, Mins, Maxs);

	if (RecordCount == 0)
	{
		DBFClose(hDBF);
		SHPClose(hSHP);
		write_error(2204, "Es sind keine Objekte in der Shapedatei '%s'\n", SHPFileName.c_str());
	}

	if (ShapeType != SHPT_POINT && ShapeType != SHPT_POINTZ && ShapeType != SHPT_POINTM)
	{
		DBFClose(hDBF);
		SHPClose(hSHP);
		write_error(2213, "Der Typ ('%s') der Shape-Datei '%s' ist nicht POINT, POINTZ oder POINTM.\nNur die genannten Typen sind fr DGM-Daten zulssig.\n", ShapeTypeName(ShapeType), SHPFileName.c_str());
	}

	int DBFRecordCount = DBFGetRecordCount(hDBF);

	if (DBFRecordCount == 0)
	{
		DBFClose(hDBF);
		SHPClose(hSHP);
		write_error(2204, "Es sind keine Objekte in der DBF-Datei '%s'\n", DBFFileName.c_str());
	}

	if (DBFRecordCount != RecordCount)
	{
		DBFClose(hDBF);
		SHPClose(hSHP);
		write_error(2205, "Shape-Datei '%s' und DBF-Datei '%s' enthalten verschieden viele Objekte\n", SHPFileName.c_str(), DBFFileName.c_str());
	}

	int FieldCount = DBFGetFieldCount(hDBF);

	if (ShapeType != SHPT_POINTZ && FieldCount == 0)
	{
		SHPClose(hSHP);
		DBFClose(hDBF);
		write_error(2206, "In der DBF-Datei %s' sind keine Attribute definiert\nEs muss aber mindestens das Attibute 'Z' definiert sein.\n", DBFFileName.c_str());
	}

	std::string ZName = "Z";
	int ZFieldIndex = DBFGetFieldIndex(hDBF, "Z");
	if (ZFieldIndex == -1)
	{
		ZName = "z";
		ZFieldIndex = DBFGetFieldIndex(hDBF, ZName.c_str());
	}
	if (ZFieldIndex == -1)
	{
		ZName = "Topo";
		ZFieldIndex = DBFGetFieldIndex(hDBF, ZName.c_str());
	}
	if (ZFieldIndex == -1)
	{
		ZName = "TOPO";
		ZFieldIndex = DBFGetFieldIndex(hDBF, ZName.c_str());
	}
	if (ZFieldIndex == -1)
	{
		ZName = "topo";
		ZFieldIndex = DBFGetFieldIndex(hDBF, ZName.c_str());
	}
	if (ZFieldIndex == -1)
	{
		ZName = "mNN";
		ZFieldIndex = DBFGetFieldIndex(hDBF, ZName.c_str());
	}
	if (ZFieldIndex == -1)
	{
		ZName = "Hhe";
		ZFieldIndex = DBFGetFieldIndex(hDBF, ZName.c_str());
	}
	if (ZFieldIndex == -1)
	{
		ZName = "Hoehe";
		ZFieldIndex = DBFGetFieldIndex(hDBF, ZName.c_str());
	}
	if (ZFieldIndex == -1)
	{
		ZName = "HOEHE";
		ZFieldIndex = DBFGetFieldIndex(hDBF, ZName.c_str());
	}

	if (ShapeType != SHPT_POINTZ && ZFieldIndex == -1)
	{
		SHPClose(hSHP);
		DBFClose(hDBF);
		write_error(2207, "In der DBF-Datei %s' ist das Attribut 'Z' nicht definiert\nEs muss aber vorhanden sein.\n", DBFFileName.c_str());
	}

	if (ZFieldIndex)
	{
		int Width = 0;
		int Decimals = 0;
		if (DBFGetFieldInfo(hDBF, ZFieldIndex, NULL, &Width, &Decimals) != FTDouble)
		{
			SHPClose(hSHP);
			DBFClose(hDBF);
			write_error(2209, "In der DBF-Datei '%s' ist der Attribut-Typ des Attributes 'Z' nicht Double.\n", DBFFileName.c_str());
		}
	}

	if (ShapeType == SHPT_POINTZ && ZFieldIndex >= 0)
	{
		write_warning(2219, "Es liegen sowohl Hheninformationen an den Punkten (POINTZ-Shape) als auch als Attribut '%s' vor\nEs werden die Informationen des Attributes genommen.\n", ZName.c_str());
	}


	unsigned int AnzWerte = 0;
	unsigned int AnzKnoten = 0;
	for (int i = 0; i<RecordCount; i++)
	{
		SHPObject *psCShape = SHPReadObject(hSHP, i);

		int AnzVert = psCShape->nVertices;

		if (AnzVert <= 0)
		{
			continue;
		}

		for (int j=0; j < AnzVert; j++)
		{
			double X = psCShape->padfX[j];
			double Y = psCShape->padfY[j];
			double Z = psCShape->padfZ[j];

			if (ZFieldIndex >= 0)	Z = DBFReadDoubleAttribute(hDBF, i, ZFieldIndex);

			long XL = (long)(X * 100 + 0.5);
			long YL = (long)(Y * 100 + 0.5);
			int ZI = (long)(Z * 100 + 0.5);

			AnzWerte++;

			if (DebugLevel >= 9 || (DebugLevel >= 1 && AnzWerte % 10000 == 0) || (AnzWerte > 0 && AnzWerte % 100000 == 0))
			{
				write_fortschritt("%d von %d Objekten gelesen, %d von %d Knoten im Speicher\n", AnzWerte, RecordCount, NodeList->size(), AnzKnoten);
			}

			TInsideTyp InsideTyp = INSIDE;
			if (XyList && XyList->size() > 2) InsideTyp = XyList->IsInsideXYList(XL, YL);
			if (InsideTyp == INSIDE || InsideTyp == ON_LINE)
			{
				AnzKnoten++;

				TNode* Node = new TNode(AnzKnoten, XL, YL, ZI);
				NodeList->Add(Node);
			}
		}
		SHPDestroyObject(psCShape);
	}

	write_fortschritt("%d von %d Objekten gelesen, %d von %d Knoten im Speicher\n", AnzWerte, RecordCount, NodeList->size(), AnzKnoten);

	DBFClose(hDBF);
	SHPClose(hSHP);

	write_fortschritt("<-Laden des DGM in Datei '%s' als SHP beendet\n", FileName.c_str());

	return (true);
}

//---------------------------------------------------------------------------
bool LoadDGM2DM(std::string FileName, TNodeList *NodeList, TElementList *ElementList, TXYZList *XyList, int DebugLevel)
{
	write_fortschritt("->Laden des DGM in Datei '%s' als 2DM gestartet\n", FileName.c_str());

	FILE *fh = fopen(FileName.c_str(), "r");
	if (fh == 0)
	{
		write_error(2201, "Konnte '%s' nicht zum Lesen ffnen\n", FileName.c_str());
	}

	char line[1000];

	if (0 == fgets (line, sizeof(line)-1, fh))
	{
		write_error(2298, "Konnte die Kopfzeile in der Datei '%s' nicht lesen\n", FileName.c_str());
	}
	line[sizeof(line)-1] = '\0';

	if (strncmp(line, "MESH2D", 6) != 0)
	{
		write_error(2298, "Die Kopfzeile in der Datei '%s' war nicht korrekt\n", FileName.c_str());
	}

	unsigned int AnzKnoten = 0;
	unsigned int AnzElemente = 0;
	unsigned int AnzZeilen = 1;

	while (NULL != fgets (line, sizeof (line)-1, fh))
	{
		AnzZeilen++;

		if (DebugLevel >= 9 || (DebugLevel >= 1 && AnzZeilen % 10000 == 0) || (AnzZeilen > 0 && AnzZeilen % 100000 == 0))
		{
			write_fortschritt("%d Zeilen, %d Knoten geladen, %d Elemente gezhlt\n", AnzZeilen, NodeList->size(), AnzElemente);
		}

		if (strncmp (line, "E4Q", 3) == 0 || strncmp (line, "E3T", 3) == 0)
		{
			AnzElemente++;
		}
		else if (strncmp (line, "ND", 2) == 0)
		{
			int     Nr;
			double  X;
			double  Y;
			double  Z;

			if (4 != sscanf(line, "ND%d%lf%lf%lf\n", &Nr, &X, &Y, &Z))
			{
				write_error(2297, "Konnte keine 3 Werte in der Datei '%s' in der Zeile %d lesen\n", FileName.c_str(), AnzZeilen);
			}

			long XL = (long)(X * 100 + 0.5);
			long YL = (long)(Y * 100 + 0.5);
			int ZI = (long)(Z * 100 + 0.5);

			TInsideTyp InsideTyp = INSIDE;
			if (XyList && XyList->size() > 2) InsideTyp = XyList->IsInsideXYList(XL, YL);
			if (InsideTyp == INSIDE || InsideTyp == ON_LINE)
			{
				AnzKnoten++;

				TNode *Node = new TNode(Nr, XL, YL, ZI);
				NodeList->Add(Node);
			}
		}
	}

	NodeList->SortByNr();

	if (AnzElemente > 0)
	{
		fseek(fh, 0L, SEEK_SET);

		int AnzZeilen = 0;

		while (ElementList->size() < AnzElemente && NULL != fgets (line, sizeof (line)-1, fh))
		{
			AnzZeilen++;

			if (DebugLevel >= 9 || (DebugLevel >= 1 && AnzZeilen % 10000 == 0) || (AnzZeilen > 0 && AnzZeilen % 100000 == 0))
			{
				write_fortschritt("%d Zeilen, %d Knoten geladen, %d Elemente geladen\n", AnzZeilen, NodeList->size(), ElementList->size());
			}

			int Nr;
			int Node1Nr;
			int Node2Nr;
			int Node3Nr;
			int Node4Nr;
			int Prop;

			if (strncmp(line, "E3T", 3) == 0)
			{
				if (5 != sscanf(line, "E3T%d%d%d%d%d\n", &Nr, &Node1Nr, &Node2Nr, &Node3Nr, &Prop))
				{
					write_error(2297, "Konnte keine 5 Werte in der Datei '%s' in der Zeile %d lesen\n", FileName.c_str(), AnzZeilen);
				}

				TNode *Node1 = NodeList->FindByNr(Node1Nr);
				TNode *Node2 = NodeList->FindByNr(Node2Nr);
				TNode *Node3 = NodeList->FindByNr(Node3Nr);

				if (Node1 && Node2 && Node3)
				{
					TElement *Element = new TElement(Node1, Node2, Node3);
					ElementList->Add(Element);
				}
				continue;
			}

			if (strncmp(line, "E4Q", 3) == 0)
			{
				if (6 != sscanf(line, "E4Q%d%d%d%d%d%d\n", &Nr, &Node1Nr, &Node2Nr, &Node3Nr, &Node4Nr, &Prop))
				{
					write_error(2297, "Konnte keine 6 Werte in der Datei '%s' in der Zeile %d lesen\n", FileName.c_str(), AnzZeilen);
				}

				TNode *Node1 = NodeList->FindByNr(Node1Nr);
				TNode *Node2 = NodeList->FindByNr(Node2Nr);
				TNode *Node3 = NodeList->FindByNr(Node3Nr);
				TNode *Node4 = NodeList->FindByNr(Node4Nr);

				if (Node1 && Node2 && Node3 && Node4)
				{
					TElement *Element = new TElement(Node1, Node2, Node3, Node4);
					ElementList->Add(Element);
				}
				continue;
			}
		}
	}

	fclose(fh);

	write_fortschritt("%d Elemente und %d Knoten geladen\n", ElementList->size(), NodeList->size());
	write_fortschritt("%d Knoten insgesamt\n", AnzKnoten);

	if (XyList && XyList->size() > 2)
	{
		write_fortschritt("->Lschen aussenliegender Kanten gestartet\n");

		int Count = 0;
		for (TElementList::iterator i=ElementList->begin(); i != ElementList->end(); i++)
		{
			TElement *Element = *i;

			TNode* Node1 = Element->Node1;
			TNode* Node2 = Element->Node2;
			TNode* Node3 = Element->Node3;

			long X12 = (Node1->X + Node2->X) / 2;
			long Y12 = (Node1->Y + Node2->Y) / 2;

			if (!XyList->IsInsideXYList(X12, Y12))
			{
				Element->Typ = NO_ELEMENT;
				continue;
			}

			long X23 = (Node2->X + Node3->X) / 2;
			long Y23 = (Node2->Y + Node3->Y) / 2;

			if (!XyList->IsInsideXYList(X23, Y23))
			{
				Element->Typ = NO_ELEMENT;
				continue;
			}

			long X31 = (Node3->X + Node1->X) / 2;
			long Y31 = (Node3->Y + Node1->Y) / 2;

			if (!XyList->IsInsideXYList(X31, Y31))
			{
				Element->Typ = NO_ELEMENT;
				continue;
			}

			Count++;

			if (DebugLevel >= 9 || (DebugLevel >= 1 && Count % 5000 == 0) || (Count % 50000 == 0))
			{
				write_fortschritt("%d von %d Elementen berprft\n", Count, ElementList->size());
			}
		}

		write_fortschritt("<-Lschen aussenliegender Kanten beendet\n");
	}

	write_fortschritt("<-Laden des DGM in Datei '%s' als 2DM beendet\n", FileName.c_str());

	return (true);
}

//---------------------------------------------------------------------
bool LoadProfile(std::string FileName, TProfilList* ProfilList, TGewaesserAchseList* GewaesserAchseList, double VonKmD, double BisKmD, int DebugLevel)
{
	write_fortschritt("->Laden der Profilspuren gestartet\n");

	LoadPRJ(FileName, DebugLevel);

	if (VonKmD > -9999 && BisKmD > -9999)	write_fortschritt("Nur Profile von Station %.4f bis %.4f werden geladen\n", VonKmD, BisKmD);
	else if (VonKmD > -9999)	write_fortschritt("Nur Profile mit einer Station grer als %.4f werden geladen\n", VonKmD);
	else if (BisKmD > -9999)	write_fortschritt("Nur Profile mit einer Station kleiner als %.4f werden geladen\n", BisKmD);

	if (ProfilList == 0)
	{
		dump_error(__FILE__, __LINE__, "ProfilListe ist undefiniert\n");
	}

	std::string SHPFileName = ExchangeFileExt(FileName, ".shp");
	SHPHandle hSHP = SHPOpen(SHPFileName.c_str(), "rb");

	if (hSHP == NULL)
	{
		write_error(2201, "Konnte '%s' nicht zum Lesen ffnen\n", SHPFileName.c_str());
	}

	std::string DBFFileName = ExchangeFileExt(FileName, ".dbf");
	DBFHandle hDBF = DBFOpen(DBFFileName.c_str(), "rb");
	if (hDBF == NULL)
	{
		SHPClose(hSHP);
		write_error(2201, "Konnte '%s' nicht zum Lesen ffnen\n", DBFFileName.c_str());
		return (false);
	}

	int 		SHPRecordCount = 0;
	TShpType	ShapeType = SHPT_NULL;
	double		Mins[4];
	double		Maxs[4];

	SHPGetInfo(hSHP, &SHPRecordCount, &ShapeType, Mins, Maxs);

	if (SHPRecordCount == 0)
	{
		DBFClose(hDBF);
		SHPClose(hSHP);
		write_error(2204, "Es sind keine Objekte in der Shape-Datei '%s'\n", SHPFileName.c_str());
	}

	if (ShapeType != SHPT_ARC && ShapeType != SHPT_ARCZ && ShapeType != SHPT_ARCM)
	{
		DBFClose(hDBF);
		SHPClose(hSHP);
		write_error(2213, "Der Typ ('%s') der Shape-Datei '%s' ist nicht ARC, ARCZ oder ARCM.\nNur die genannten Typen sind fr Profilspuren zulssig.\n", ShapeTypeName(ShapeType), SHPFileName.c_str());
	}

	int DBFRecordCount = DBFGetRecordCount(hDBF);

	if (DBFRecordCount == 0)
	{
		DBFClose(hDBF);
		SHPClose(hSHP);
		write_error(2204, "Es sind keine Objekte in der DBF-Datei '%s'\n", DBFFileName.c_str());
	}

	if (DBFRecordCount != SHPRecordCount)
	{
		DBFClose(hDBF);
		SHPClose(hSHP);
		write_error(2205, "Shape-Datei '%s' und DBF-Datei '%s' enthalten verschieden viele Objekte\n", SHPFileName.c_str(), DBFFileName.c_str());
	}

	int RecordCount = SHPRecordCount;

	int FieldCount = DBFGetFieldCount(hDBF);

	if (FieldCount == 0)
	{
		SHPClose(hSHP);
		DBFClose(hDBF);
		write_error(2206, "In der DBF-Datei %s' sind keine Attribute definiert\nEs mssen aber mindestens die Attibute Z und STATION definiert sein.\n", DBFFileName.c_str());
	}

	std::string StationName = "STATION";
	int StationFieldIndex = DBFGetFieldIndex(hDBF, StationName.c_str());
	if (StationFieldIndex == -1)
	{
		StationName = "Station";
		StationFieldIndex = DBFGetFieldIndex(hDBF, StationName.c_str());
	}
	if (StationFieldIndex == -1)
	{
		StationName = "station";
		StationFieldIndex = DBFGetFieldIndex(hDBF, StationName.c_str());
	}
	if (StationFieldIndex == -1)
	{
		StationName = "KILOMETER";
		StationFieldIndex = DBFGetFieldIndex(hDBF, StationName.c_str());
	}
	if (StationFieldIndex == -1)
	{
		StationName = "Kilometer";
		StationFieldIndex = DBFGetFieldIndex(hDBF, StationName.c_str());
	}
	if (StationFieldIndex == -1)
	{
		StationName = "kilometer";
		StationFieldIndex = DBFGetFieldIndex(hDBF, StationName.c_str());
	}
	if (StationFieldIndex == -1)
	{
		StationName = "KM";
		StationFieldIndex = DBFGetFieldIndex(hDBF, StationName.c_str());
	}
	if (StationFieldIndex == -1)
	{
		StationName = "Km";
		StationFieldIndex = DBFGetFieldIndex(hDBF, StationName.c_str());
	}
	if (StationFieldIndex == -1)
	{
		StationName = "km";
		StationFieldIndex = DBFGetFieldIndex(hDBF, StationName.c_str());
	}

	if (StationFieldIndex == -1)
	{
		SHPClose(hSHP);
		DBFClose(hDBF);
		write_error(2208, "In der DBF-Datei '%s' ist das Attribut 'STATION' nicht definiert\nEs muss aber vorhanden sein.\n", DBFFileName.c_str());
	}

	int Width = 0;
	int Decimals = 0;
	if (DBFGetFieldInfo(hDBF, StationFieldIndex, NULL, &Width, &Decimals) != FTDouble)
	{
		SHPClose(hSHP);
		DBFClose(hDBF);
		write_error(2210, "In der DBF-Datei '%s' ist der Attribut-Typ des Attributes '%s' nicht Fliesskommazahl.\n", DBFFileName.c_str(), StationName.c_str());
	}

	int VonKmI = (int)(VonKmD * 10000.0 + 0.5);
	int BisKmI = (int)(BisKmD * 10000.0 + 0.5);

	int Count = 0;
	for (int i = 0; i<RecordCount; i++)
	{
		double StationD = DBFReadDoubleAttribute(hDBF, i, StationFieldIndex);

		int StationI = (int)(StationD * 10000.0 + 0.5);
		if (StationD < 0)	StationI = (int)(StationD * 10000.0 - 0.5);

		if (StationI >= VonKmI && StationI <= BisKmI)
		{
			if (ProfilList->Find(StationI))
			{
				write_warning(2111, "In der DBF-Datei '%s' kommt das Profil (%.4f) mehrfach vor.\nEs wird nur die erste Definition bercksichtigt.\n", DBFFileName.c_str(), StationI / 10000.0);
				continue;
			}

			SHPObject *psCShape = SHPReadObject(hSHP, i);

			int AnzVert = psCShape->nVertices;

			if (AnzVert <= 0)
			{
				write_warning(2112, "In der DBF-Datei '%s' hat das Profil (%.4f) keine Sttzstellen.\nEs wird ignoriert.\n", DBFFileName.c_str(), StationI / 10000.0);
				continue;
			}

			TProfil* Profil = new TProfil(StationI);

			ProfilList->insert(Profil);

			for (int j=0; j < AnzVert; j++)
			{
				double X = psCShape->padfX[j];
				double Y = psCShape->padfY[j];

				long XL = (long)(X * 100.0 + 0.5);
				long YL = (long)(Y * 100.0 + 0.5);
				Profil->AddPoint(XL, YL);
			}

			SHPDestroyObject(psCShape);
		}

		Count++;

		if (DebugLevel >= 9 || (DebugLevel >= 1 && Count % 10 == 0) || (Count > 0 && Count % 100 == 0))
		{
			write_fortschritt("%d von %d Objekten geladen, %d Profilspuren im Speicher\n", Count, RecordCount, ProfilList->size());
		}
	}

	write_fortschritt("%d von %d Objekten geladen, %d Profilspuren im Speicher\n", Count, RecordCount, ProfilList->size());

	DBFClose(hDBF);
	SHPClose(hSHP);

	write_fortschritt("<-Laden der Profilspuren beendet\n");

	if (DebugLevel >= 9)
	{
		for (TProfilList::iterator i = ProfilList->begin(); i != ProfilList->end(); i++)
		{
			TProfil* Profil = *i;
			write_fortschritt("%d\n", Profil->Station);
		}
	}

	write_fortschritt("->Ausrichten der Profilspuren gestartet\n");

	if (0 == GewaesserAchseList)
	{
		TProfil* VorProfil = 0;
		TProfil* NachProfil = 0;
		Count = 0;
		int SwitchCount = 0;
		TProfilList::iterator i = ProfilList->begin();
		while (i != ProfilList->end())
		{
			Count++;

			if (DebugLevel >= 9 || (DebugLevel >= 1 && Count % 10 == 0) || (Count > 0 && Count % 100 == 0))
			{
				write_fortschritt("%d von %d Profilspuren ausgerichtet, %d gedreht\n", Count, ProfilList->size(), SwitchCount);
			}

			VorProfil = NachProfil;
			NachProfil = *i;

			if (VorProfil == 0)
			{
				i++;
				continue;
			}

			TPoint* VorAnfang = *VorProfil->PointList->begin();
			TPoint* VorEnde   = *VorProfil->PointList->rbegin();

			TPoint* NachAnfang = *NachProfil->PointList->begin();
			TPoint* NachEnde   = *NachProfil->PointList->rbegin();

			long VAx = VorAnfang->X;
			long VAy = VorAnfang->Y;
			long VEx = VorEnde->X;
			long VEy = VorEnde->Y;
			long NAx = NachAnfang->X;
			long NAy = NachAnfang->Y;
			long NEx = NachEnde->X;
			long NEy = NachEnde->Y;

			long ADx = NAx - VAx;
			long ADy = NAy - VAy;
			if (ADx == 0 && ADy == 0)
			{
				write_error(3207, "Profil-Punkteabstand (%.3f %.3f) - (%.3f %.3f) ist zu klein\n", VAx / 100.0, VAy / 100.0, NAx / 100.0, NAy / 100.0);
			}

			long EDx = NEx - VEx;
			long EDy = NEy - VEy;
			if (EDx == 0 && EDy == 0)
			{
				write_error(3207, "Profil-Punkteabstand (%.3f %.3f) - (%.3f %.3f) ist zu klein\n", VEx / 100.0, VEy / 100.0, NEx / 100.0, NEy / 100.0);
			}

			double X = 0.0;
			double Y = 0.0;
			if (Calc2Schnitt(VAx, VAy, NAx, NAy, VEx, VEy, NEx, NEy, &X, &Y))
			{
				// Hier ist ein echter Schnittpunkt gefunden
				// Jetzt muss das NachProfil in seiner Reihenfolge vertauscht werden

				SwitchCount++;

				TProfil* ExChangeProfil = new TProfil(NachProfil->Station);
				for (TPointList::reverse_iterator j = NachProfil->PointList->rbegin(); j != NachProfil->PointList->rend(); j++)
				{
					long X = (long)((*j)->X + 0.5);
					long Y = (long)((*j)->Y + 0.5);

					ExChangeProfil->AddPoint(X, Y);
				}

				TPointList* TempPointList = NachProfil->PointList;
				NachProfil->PointList = ExChangeProfil->PointList;
				ExChangeProfil->PointList = TempPointList;

				delete ExChangeProfil;
			}
			i++;
		}
		write_fortschritt("%d von %d Profilspuren ausgerichtet, %d gedreht\n", Count, ProfilList->size(), SwitchCount);
	}
	else
	{
		Count = 0;
		int SwitchCount = 0;
		TProfilList::iterator i = ProfilList->begin();
		while (i != ProfilList->end())
		{
			Count++;

			if (DebugLevel >= 9 || (DebugLevel >= 1 && Count % 10 == 0) || (Count > 0 && Count % 100 == 0))
			{
				write_fortschritt("%d von %d Profilspuren ausgerichtet, %d gedreht\n", Count, ProfilList->size(), SwitchCount);
			}

			TProfil* Profil = *i;

			bool Found = false;
			for (TGewaesserAchseList::iterator g=GewaesserAchseList->begin(); g != GewaesserAchseList->end(); g++)
			{
				TGewaesserAchse* GewaesserAchse	= *g;

				if (Profil->MaxX < GewaesserAchse->MinX)	continue;
				if (Profil->MinX > GewaesserAchse->MaxX)	continue;
				if (Profil->MaxY < GewaesserAchse->MinY)	continue;
				if (Profil->MinY > GewaesserAchse->MaxY)	continue;

				TPoint*	P0 = 0;
				TPoint*	P1 = 0;
				for (TPointList::iterator p = Profil->PointList->begin(); p != Profil->PointList->end(); p++)
				{
					P0 = P1;
					P1 = *p;

					if (P0 == 0)	continue;

					long P0x = P0->X;
					long P0y = P0->Y;
					long P1x = P1->X;
					long P1y = P1->Y;

					if (P0x	< GewaesserAchse->MinX && P1x <	GewaesserAchse->MinX)	continue;
					if (P0x	> GewaesserAchse->MaxX && P1x >	GewaesserAchse->MaxX)	continue;
					if (P0y	< GewaesserAchse->MinY && P1y <	GewaesserAchse->MinY)	continue;
					if (P0y	> GewaesserAchse->MaxY && P1y >	GewaesserAchse->MaxY)	continue;

					long PDx = P1x - P0x;
					long PDy = P1y - P0y;
					if (PDx	== 0 &&	PDy == 0)
					{
						write_error(3207, "Profil-Punktabstand (%.3f %.3f) - (%.3f %.3f) ist zu	klein\n", P0x /	100.0, P0y / 100.0, P1x	/ 100.0, P1y / 100.0);
					}

					// Es heisst hier s fr spur und nicht g fr gewsserachse

					TPoint*S0 = 0;
					TPoint*S1 = 0;
					for (TPointList::iterator s = GewaesserAchse->PointList->begin(); s != GewaesserAchse->PointList->end(); s++)
					{
						S0 = S1;
						S1 = *s;

						if (S0 == 0)	continue;

						long S0x = S0->X;
						long S0y = S0->Y;
						long S1x = S1->X;
						long S1y = S1->Y;

						long SDx = S1x - S0x;
						long SDy = S1y - S0y;
						if (SDx	== 0 &&	SDy == 0)
						{
							write_error(3208, "Gewsserachsen-Sttzpunkteabstand (%.3f %.3f) - (%.3f %.3f) ist zu klein\n",	S0x / 100.0, S0y / 100.0, S1x /	100.0, S1y / 100.0);
						}

						double X, Y;
						Found = Calc2Schnitt(P0x, P0y, P1x, P1y, S0x, S0y,	S1x, S1y, &X, &Y);

						if (Found)
						{
							// Hier ist ein echter Schnittpunkt gefunden
							// Jetzt prfen wo welcher Punkt liegt

							 if (CrossProduct(X, Y, P1x, P1y, S1x, S1y) < 0.0)
							 {
								if (DebugLevel >= 9)
								{
									write_fortschritt("Profil '%.4f' wurde umgedreht\n", Profil->Station / 10000.0);
								}

								SwitchCount++;

								TProfil* ExChangeProfil = new TProfil(Profil->Station);
								for (TPointList::reverse_iterator j = Profil->PointList->rbegin(); j != Profil->PointList->rend(); j++)
								{
									long X = (long)((*j)->X + 0.5);
									long Y = (long)((*j)->Y + 0.5);

									ExChangeProfil->AddPoint(X, Y);
								}

								TPointList* TempPointList = Profil->PointList;
								Profil->PointList = ExChangeProfil->PointList;
								ExChangeProfil->PointList = TempPointList;

								delete ExChangeProfil;
							}
							break; // Schnittpunkt gefunden
						}
					} // Ende der Iteration ber die Punkte der Achse
					if (Found) break;
				} // Ende der Iteration ber die Punkte eins Profils
				if (Found) break;
			} // Ende der Iteration ber die Achsen
			i++;
		}
		write_fortschritt("%d von %d Profilspuren ausgerichtet, %d gedreht\n", Count, ProfilList->size(), SwitchCount);
	}

	write_fortschritt("<-Ausrichten der Profilspuren beendet\n");


	write_fortschritt("->berprfung der Profilspuren gestartet\n");

	TProfil* VorProfil = 0;
	TProfil* NachProfil = 0;
	Count = 0;
	TProfilList::iterator i = ProfilList->begin();
	while (i != ProfilList->end())
	{
		Count++;

		if (DebugLevel >= 9 || (DebugLevel >= 1 && Count % 10 == 0) || (Count > 0 && Count % 100 == 0))
		{
			write_fortschritt("%d von %d Profilspuren berprft\n", Count, ProfilList->size());
		}

		VorProfil = NachProfil;
		NachProfil = *i;

		if (VorProfil == 0)
		{
			i++;
			continue;
		}


		TPoint* VP1 = 0;
		TPoint* VP2 = 0;
		TPointList::iterator j = VorProfil->PointList->begin();
		while (j != VorProfil->PointList->end())
		{
			VP1 = VP2;
			VP2 = *j;

			if (VP1 == 0)
			{
				j++;
				continue;
			}

			TPoint* NP1 = 0;
			TPoint* NP2 = 0;
			TPointList::iterator k = NachProfil->PointList->begin();
			while (k != NachProfil->PointList->end())
			{
				NP1 = NP2;
				NP2 = *k;

				if (NP1 == 0)
				{
					k++;
					continue;
				}

				double X = 0.0;
				double Y = 0.0;
				if (Calc2Schnitt(VP1->X, VP1->Y, VP2->X, VP2->Y, NP1->X, NP1->Y, NP2->X, NP2->Y, &X, &Y))
                {
                	write_warning(3117, "Die Profile %d und %d schneiden sich bei (%.2f, %.2f)\n", VorProfil->Station / 1000, NachProfil->Station / 1000, X, Y);
                }
                k++;
			}
            j++;
        }
		i++;
	}
	write_fortschritt("%d von %d Profilspuren berprft\n", Count, ProfilList->size());

	write_fortschritt("<-berprfung der Profilspuren beendet\n");

	return (true);
}

//---------------------------------------------------------------------
bool LoadLinien(std::string FileName, TProfilList* SperrenList, TProfilList* GrabenList, TProfilList* Rohre1List, TProfilList* Rohre2List, int DebugLevel)
{
	write_fortschritt("->Laden der Linien Strukturen (Sperren, Grben und Rohre) gestartet\n");

	LoadPRJ(FileName, DebugLevel);

	if (SperrenList == 0 || GrabenList == 0 || Rohre1List == 0)
	{
		dump_error(__FILE__, __LINE__, "Sperrenliste oder Grabenliste oder Rohreliste ist undefiniert\n");
	}

	std::string SHPFileName = ExchangeFileExt(FileName, ".shp");
	SHPHandle hSHP = SHPOpen(SHPFileName.c_str(), "rb");

	if (hSHP == NULL)
	{
		write_error(2201, "Konnte '%s' nicht zum Lesen ffnen\n", SHPFileName.c_str());
	}

	std::string DBFFileName = ExchangeFileExt(FileName, ".dbf");
	DBFHandle hDBF = DBFOpen(DBFFileName.c_str(), "rb");
	if (hDBF == NULL)
	{
		SHPClose(hSHP);
		write_error(2201, "Konnte '%s' nicht zum Lesen ffnen\n", DBFFileName.c_str());
		return (false);
	}

	int 		SHPRecordCount = 0;
	TShpType	ShapeType = SHPT_NULL;
	double		Mins[4];
	double		Maxs[4];

	SHPGetInfo(hSHP, &SHPRecordCount, &ShapeType, Mins, Maxs);

	if (SHPRecordCount == 0)
	{
		DBFClose(hDBF);
		SHPClose(hSHP);

		write_error(2204, "Es sind keine Objekte in der Shape-Datei '%s'\n", SHPFileName.c_str());
	}

	if (ShapeType != SHPT_ARC && ShapeType != SHPT_ARCZ && ShapeType != SHPT_ARCM && ShapeType != SHPT_POLYGON && ShapeType != SHPT_POLYGONZ && ShapeType != SHPT_POLYGONM)
	{
		write_error(2213, "Der Typ ('%s') der Shape-Datei '%s' ist weder ARC, ARCZ, ARCM\nnoch POLYGON, POLYGONZ, POLYHONM.\nNur die genannten Typen sind fr Profilspuren zulssig.\n", ShapeTypeName(ShapeType), SHPFileName.c_str());
	}

	int DBFRecordCount = DBFGetRecordCount(hDBF);

	if (DBFRecordCount == 0)
	{
		DBFClose(hDBF);
		SHPClose(hSHP);

		write_error(2204, "Es sind keine Objekte in der DBF-Datei '%s'\n", DBFFileName.c_str());
	}

	if (DBFRecordCount != SHPRecordCount)
	{
		DBFClose(hDBF);
		SHPClose(hSHP);

		write_error(2205, "Shape-Datei '%s' und DBF-Datei '%s' enthalten verschieden viele Objekte\n", SHPFileName.c_str(), DBFFileName.c_str());
	}

	int RecordCount = SHPRecordCount;

	int FieldCount = DBFGetFieldCount(hDBF);

	if (FieldCount == 0)
	{
		SHPClose(hSHP);
		DBFClose(hDBF);

		write_error(2206, "In der DBF-Datei %s' sind keine Attribute definiert\nEs muss aber mindestens das Attibute TYP definiert sein.\n", DBFFileName.c_str());
	}

	std::string TypName = "TYP";
	int TypFieldIndex = DBFGetFieldIndex(hDBF, TypName.c_str());
	if (TypFieldIndex == -1)
	{
		TypName = "Typ";
		TypFieldIndex = DBFGetFieldIndex(hDBF, TypName.c_str());
	}
	if (TypFieldIndex == -1)
	{
		TypName = "typ";
		TypFieldIndex = DBFGetFieldIndex(hDBF, TypName.c_str());
	}
	if (TypFieldIndex == -1)
	{
		TypName = "TYPE";
		TypFieldIndex = DBFGetFieldIndex(hDBF, TypName.c_str());
	}
	if (TypFieldIndex == -1)
	{
		TypName = "Type";
		TypFieldIndex = DBFGetFieldIndex(hDBF, TypName.c_str());
	}
	if (TypFieldIndex == -1)
	{
		TypName = "type";
		TypFieldIndex = DBFGetFieldIndex(hDBF, TypName.c_str());
	}

	if (TypFieldIndex == -1)
	{
		SHPClose(hSHP);
		DBFClose(hDBF);

		write_error(2207, "In der DBF-Datei %s' ist das Attribut 'Typ' nicht definiert\nEs muss aber vorhanden sein.\n", DBFFileName.c_str());
	}

	int Width = 0;
	int Decimals = 0;
	if (DBFGetFieldInfo(hDBF, TypFieldIndex, NULL, &Width, &Decimals) != FTString)
	{
		SHPClose(hSHP);
		DBFClose(hDBF);

		write_error(2209, "In der DBF-Datei '%s' ist der Attribut-Typ des Attributes '%s' nicht Text.\n", DBFFileName.c_str(), TypName.c_str());
	}

	int Count = 0;
	for (int i=0; i<RecordCount; i++)
	{
		std::string Typ = DBFReadStringAttribute(hDBF, i, TypFieldIndex);

		if (Typ == "SPERRE" || Typ == "Sperre" || Typ == "sperre")
		{
			Typ = "DAMM";
		}
		else if (Typ == "DAMM" || Typ == "Damm" || Typ == "damm")
		{
			Typ = "DAMM";
		}
		else if (Typ == "RINGDEICH" || Typ == "Ringdeich" || Typ == "ringdeich")
		{
			Typ = "DAMM";
		}
		else if (Typ == "GRABEN" || Typ == "Graben" || Typ == "graben")
		{
			Typ = "GRABEN";
		}
		else if (Typ == "BRUCH" || Typ == "Bruch" || Typ == "bruch")
		{
			Typ = "GRABEN";
		}
		else if (Typ == "ROHR" || Typ == "Rohr" || Typ == "rohr" || Typ == "ROHR1" || Typ == "rohr1" || Typ == "Rohr1" || Typ == "ROHR 1" || Typ == "rohr 1" || Typ == "Rohr 1")
		{
			Typ = "ROHR";
		}
		else if (Typ == "GRADIENT" || Typ == "Gradient" || Typ == "gradient" || Typ == "ROHR2" || Typ == "rohr2" || Typ == "Rohr2" || Typ == "ROHR 2" || Typ == "rohr 2" || Typ == "Rohr 2")
		{
			Typ = "GRADIENT";
		}
		else
		{
			write_warning(3112, "Das Attribut 'TYP' in der Datei '%s' darf nur die Werte 'Damm', 'Graben', 'Rohr 1' oder 'Rohr 2' haben.\nDas Linienobjekt '%s' wird ignoriert.\n", DBFFileName.c_str(), Typ.c_str());
			continue;
		}


		SHPObject *psCShape = SHPReadObject(hSHP, i);

		int AnzVert = psCShape->nVertices;

		if (AnzVert <= 0)
		{
			write_warning(2112, "In der SHP-Datei '%s' hat ein Linienobjekt keine Sttzstellen.\nEs wird ignoriert.\n", SHPFileName.c_str());
			continue;
		}


		TProfil* Linie = 0;

		if (Typ == "DAMM")
		{
			Linie = new TProfil(SperrenList->size()+1, NoWsp);
			SperrenList->insert(Linie);
		}
		else if (Typ == "GRABEN")
		{
			Linie = new TProfil(GrabenList->size()+1, NoWsp);
			GrabenList->insert(Linie);
		}
		else if (Typ == "ROHR")
		{
			Linie = new TProfil(Rohre1List->size()+1, NoWsp);
			Rohre1List->insert(Linie);
		}
		else if (Typ == "GRADIENT")
		{
			Linie = new TProfil(Rohre2List->size()+1, NoWsp);
			Rohre2List->insert(Linie);
		}
		else
		{
			dump_error(__FILE__, __LINE__, "Unbekannter Typ '%s' in der Datei '%s'\n", Typ.c_str(), DBFFileName.c_str());
		}


		for (int j=0; j < AnzVert; j++)
		{
			double X = psCShape->padfX[j];
			double Y = psCShape->padfY[j];
			double Z = 0.0;

			long* L1 = (long *)&(psCShape->padfZ[j]);
			long* L2 = L1 +1;

			// Ein Test auf NaN (also einen unbestimmten Knoten)
			if (*L1 == 0x00000000L && *L2 == 0x7FF80000L)	Z = NoZ;
			else											Z = psCShape->padfZ[j];

			if (ShapeType == SHPT_ARC || ShapeType == SHPT_POLYGON || Z == NoZ)
			{
				if (Typ == "DAMM")			Z = DammZ;
				else if (Typ == "GRABEN")	Z = GrabenZ;
				else						Z = 0.0;
			}

			long XL = (long)(X * 100.0 + 0.5);
			long YL = (long)(Y * 100.0 + 0.5);
			int ZL = (int)(Z * 100.0 + 0.5);
			Linie->AddPoint(XL, YL, ZL);
		}

		SHPDestroyObject(psCShape);

		Count++;

		if (DebugLevel >= 9 || (DebugLevel >= 1 && Count % 10 == 0) || (Count > 0 && Count % 100 == 0))
		{
			write_fortschritt("%d von %d Liniensstrukturen geladen.\n%d Sperren, %d Grben, %d Rohre Tpy 1, %d Rohre Tpy 2\n", Count, RecordCount, SperrenList->size(), GrabenList->size(), Rohre1List->size(), Rohre2List->size());
		}
	}
	write_fortschritt("%d von %d Liniensstrukturen geladen.\n%d Sperren, %d Grben, %d Rohre Typ 1, %d Rohre Typ 2\n", Count, RecordCount, SperrenList->size(), GrabenList->size(), Rohre1List->size(), Rohre2List->size());

	DBFClose(hDBF);
	SHPClose(hSHP);

	write_fortschritt("<-Laden der Linien Strukturen (Sperren, Grben und Rohre) beendet\n");

	return (true);
}

//---------------------------------------------------------------------
bool LoadWsp(std::string FileName, TProfilList* ProfilList, std::string Tag, int DebugLevel)
{
	write_fortschritt("->Laden der Wasserstnde gestartet\n");

	std::string Ext = GetFileExt(FileName);

	bool Status = true;

	if (ToUpperCase(Ext) == ".DBF")
	{
		Status = LoadWspDBF(FileName, ProfilList, DebugLevel);
	}
	else if (ToUpperCase(Ext) == ".WST")
	{
		Status = LoadWspWst(FileName, ProfilList, Tag, DebugLevel);
	}
	else if (ToUpperCase(Ext) == ".TXT" || ToUpperCase(Ext) == ".DAT" || ToUpperCase(Ext) == ".ASC")
	{
		Status = LoadWspTxt(FileName, ProfilList, DebugLevel);
	}
	else
	{
		write_error(1227, "Ungltige Dateinamenserweiterung '%s' beim Dateinamen '%s' fr den Parameter -WSP\nErlaubt sind nur WSP, DBF, TXT, DAT oder ASC\n", ToUpperCase(Ext).c_str(), FileName.c_str());
	}

	write_fortschritt("<-Laden der Wasserstnde beendet\n");

	return (Status);
}

//---------------------------------------------------------------------
bool LoadWspTxt(std::string FileName, TProfilList* ProfilList, int DebugLevel)
{
	write_fortschritt("->Laden der Wasserstnde im ASCII-Format gestartet\n");

	if (ProfilList == 0)
	{
		dump_error(__FILE__, __LINE__, "ProfilListe ist undefiniert\n");
	}

	FILE *fh = fopen(FileName.c_str(), "r");
	if (fh == 0)
	{
		write_error(2201, "Konnte '%s' nicht zum Lesen ffnen\n", FileName.c_str());
	}

	int StationCol =0;
	int WspCol =0;

	int AktZeile = 0;

	char line[1000];
	while (fgets (line, sizeof (line)-1, fh))
	{
		AktZeile++;

		line[sizeof(line)-1] = '\0';
		if (strlen(line) > 0) line[strlen(line)-1] = '\0';

		for (unsigned int i=0; i<strlen(line); i++)	if (line[i] == ',')	line[i] = '.';

		char* End = line + strlen(line);

		double		StationD = 0.0;
		double		WspD = 0.0;

		int AktCol = 0;
		int ColOk = 0;
		char* Col =0;
		char* P = line;
		while (P < End)
		{
			P = SkipSpaces(P);

			Col = GetCol(P);
			AktCol++;

			if (AktZeile == 1)
			{
				if (strcmp(Col, "STATION") == 0 || strcmp(Col, "Station") == 0 || strcmp(Col, "station") == 0 || strcmp(Col, "KILOMETER") == 0 || strcmp(Col, "Kilometer") == 0 || strcmp(Col, "kilometer") == 0 || strcmp(Col, "KM") == 0 || strcmp(Col, "Km") == 0 || strcmp(Col, "km") == 0)
				{
					StationCol = AktCol;
					ColOk++;
				}
				if (strcmp(Col, "WSP") == 0 || strcmp(Col, "Wsp") == 0 || strcmp(Col, "wsp") == 0 || strcmp(Col, "WASSERSTAND") == 0 || strcmp(Col, "Wasserstand") == 0 || strcmp(Col, "wasserstand") == 0)
				{
					WspCol = AktCol;
					ColOk++;
				}
			}
			else
			{
				if (AktCol == StationCol)
				{
					if (1 != sscanf(Col, "%lf", &StationD))
					{
						write_error(2216, "Konnte keine Zahl in Spalte %d erkennen\n", AktCol);
					}
					ColOk++;
				}
				if (AktCol == WspCol)
				{
					if (1 != sscanf(Col, "%lf", &WspD))
					{
						write_error(2216, "Konnte keine Zahl in Spalte %d erkennen\n", AktCol);
					}
					ColOk++;
				}
			}
			if (ColOk == 2)	break;

			P = P + strlen(Col) + 1;
		}

		if (ColOk != 2 && AktZeile == 1) write_error(2215, "Nicht genug Informationen in der ersten Zeile\nEs mssen die Feldnamen STATION und WSP vorkommen.\n");

		if (ColOk != 2 && AktZeile > 1) write_error(2215, "Nicht genug Informationen in der Zeile %d.\nEs mssen die Feldnamen STATION und WSP vorkommen.\n", AktZeile);

		if (AktZeile == 1) continue;

		int Station = (int)(StationD * 10000.0 + 0.5);
		if (StationD < 0)	Station = (int)(StationD * 10000.0 - 0.5);

		int Wsp = (int)(WspD * 100.0 + 0.5);
		if (WspD < 0)	Wsp = (int)(WspD * 100.0 - 0.5);

		TProfil* Profil = new TProfil(Station, Wsp);

		unsigned int Anz = ProfilList->size();
		ProfilList->insert(Profil);
		if (ProfilList->size() == Anz)
		{
			write_warning(3119, "Fr den Flusskilometer %.4f wurde der Wasserstand mehrmals angegeben\n", Profil->Station / 10000.0);
			delete Profil;
		}

		if (DebugLevel >= 9 || (DebugLevel >= 1 && AktZeile % 100 == 0) || (AktZeile > 0 && AktZeile % 1000 == 0))
		{
			write_fortschritt("%d Zeilen geladen, %d Wasserstnde im Speicher\n", AktZeile, ProfilList->size());
		}
	}
	write_fortschritt("%d Zeilen geladen, %d Wasserstnde im Speicher\n", AktZeile, ProfilList->size());

	fclose(fh);

	write_fortschritt("<-Laden der Wasserstnde im ASCII-Format beendet\n");

	return (true);
}

//---------------------------------------------------------------------
bool LoadWspWst(std::string FileName, TProfilList* ProfilList, std::string Tag, int DebugLevel)
{
	write_fortschritt("->Laden der Wasserstnde im WST-Format gestartet\n");

	if (ProfilList == 0)
	{
		dump_error(__FILE__, __LINE__, "ProfilListe ist undefiniert\n");
	}

	Tag = Trim(Tag);

	FILE *fh = fopen(FileName.c_str(), "r");
	if (fh == 0)
	{
		write_error(2201, "Konnte '%s' nicht zum Lesen ffnen\n", FileName.c_str());
	}

	int WspCol =0;

	int AktZeile = 0;
	int	AnzCols = 0;

	bool IsOverReadShortBez = false;

	char line[1000];
	if (fgets (line, sizeof (line)-1, fh))
	{
		AktZeile++;

		line[sizeof(line)-1] = '\0';
		if (strlen(line) > 0) line[strlen(line)-1] = '\0';
		if (1 != sscanf(line, "%d", &AnzCols))
		{
			write_error(1, "Konnte in der Datei '%s' in der ersten Zeile die Anzahl der Spalten nicht lesen\n", FileName.c_str());
		}
	}
	else
	{
		write_error(1, "Konnte in der Datei '%s' die erste Zeile nicht lesen\n", FileName.c_str());
	}

	int Col = 0;
	while (fgets (line, sizeof (line)-1, fh))
	{
		AktZeile++;

		line[sizeof(line)-1] = '\0';
		if (strlen(line) > 0 && (line[strlen(line)-1] == '\n' || line[strlen(line)-1] == '\r')) line[strlen(line)-1] = '\0';
		if (strlen(line) > 0 && (line[strlen(line)-1] == '\n' || line[strlen(line)-1] == '\r')) line[strlen(line)-1] = '\0';

		if (strncmp(line, "*!column-bez-text", 17) == 0)
		{
			IsOverReadShortBez = true;
			Col = 1;
			char* Pos = line + 17;

			while (Col <= AnzCols)
			{
				if (*Pos == '\0')
				{
					write_error(2220, "Konnte in Datei '%s' in der Zeile %d nicht die Kennung '%s' finden\n", FileName.c_str(), AktZeile, Tag.c_str());
				}

				bool	IsInString = false;

				while (*Pos && *Pos == ' ')	Pos++;

				if (*Pos == '"')
				{
					Pos++;
					IsInString = true;
				}

				char* Start = Pos;

				if (IsInString == false)
				{
					while (*Pos && *Pos != ' ')	Pos++;
					if (*Pos == ' ')	*Pos++ = '\0';
					if (Trim(Start) == Tag)
					{
						WspCol = Col;
						break;
					}
					else
					{
						Col++;
					}
				}
				else
				{
					char Line2[1000];

					char* Pos2 = Line2;
					while (*Pos && *Pos != '"')
					{
						if (*Pos == '\\')
						{
							Pos++;
							if (*Pos == '\\')		*Pos2++ = '\\';
							else if (*Pos == '"')	*Pos2++ = '"';
							else write_error(2221, "Das Zeichen '\\' leitet in Datei '%s' in Zeile %d eine nicht gltige Steuersequenz ('\\%c' ein)\n", FileName.c_str(), AktZeile, *Pos);
						}
						else
						{
							*Pos2++ = *Pos;
						}
						Pos++;
					}
					*Pos2 = '\0';
					if (Trim(Line2) == Tag)
					{
						WspCol = Col;
						break;
					}
					else
					{
						Col++;
						Pos++;
					}
				}
			}
		}

		if (line[0] == '*')	continue;

		if (WspCol == 0)
		{
			char Cmp[10];
			
			Col = 1;
			char* Pos = line + Col * 9 - 1;
			strncpy(Cmp, Pos, 9);
			Cmp[9] = '\0';

			while (Col <= AnzCols && Trim(Cmp) != Tag)
			{
				Col++;
				Pos = line + Col * 9 - 1;
				strncpy(Cmp, Pos, 9);
				Cmp[9] = '\0';
			}

			if (Col > AnzCols)
			{
				write_error(2220, "Konnte in Datei '%s' in der Zeile %d nicht die Kennung '%s' finden\n", FileName.c_str(), AktZeile, Tag.c_str());
			}

			WspCol = Col;

			continue;
		}

		if (IsOverReadShortBez)
		{
			IsOverReadShortBez = false;
			continue;
		}

		for (unsigned int i=0; i<strlen(line); i++)	if (line[i] == ',')	line[i] = '.';

		double		StationD = 0.0;
		double		WspD = 0.0;

		if (1 != sscanf(line, "%lf", &StationD))
		{
			write_error(1, "Konnte in Datei '%s' in der Zeile %d nicht die Station lesen\n", FileName.c_str(), AktZeile);
		}

		line[Col * 9 + 8] = '\0';
		if (1 != sscanf(line + Col * 9, "%lf", &WspD))
		{
			continue;
		}

		int Station = (int)(StationD * 10000.0 + 0.5);
		if (StationD < 0)	Station = (int)(StationD * 10000.0 - 0.5);

		int Wsp = (int)(WspD * 100.0 + 0.5);
		if (WspD < 0)	Wsp = (int)(WspD * 100.0 - 0.5);

		TProfil* Profil = new TProfil(Station, Wsp);

		unsigned int Anz = ProfilList->size();
		ProfilList->insert(Profil);
		if (ProfilList->size() == Anz)
		{
			write_warning(3119, "Fr den Flusskilometer %.4f wurde der Wasserstand mehrmals angegeben\n", Profil->Station / 10000.0);
			delete Profil;
		}

		if (DebugLevel >= 9 || (DebugLevel >= 1 && AktZeile % 100 == 0) || (AktZeile > 0 && AktZeile % 1000 == 0))
		{
			write_fortschritt("%d Zeilen geladen, %d Wasserstnde im Speicher\n", AktZeile, ProfilList->size());
		}
	}
	write_fortschritt("%d Zeilen geladen, %d Wasserstnde im Speicher\n", AktZeile, ProfilList->size());

	fclose(fh);

	write_fortschritt("<-Laden der Wasserstnde im WST-Format beendet\n");

	return (true);
}

//---------------------------------------------------------------------
bool LoadWspDBF(std::string FileName, TProfilList* ProfilList, int DebugLevel)
{
	write_fortschritt("->Laden der Wasserstnde im DBF-Format gestartet\n");

	if (ProfilList == 0)
	{
		dump_error(__FILE__, __LINE__, "ProfilListe ist undefiniert\n");
	}

	std::string DBFFileName = ExchangeFileExt(FileName, ".dbf");
	DBFHandle hDBF = DBFOpen(DBFFileName.c_str(), "rb");
	if (hDBF == NULL)
	{
		write_error(2201, "Konnte '%s' nicht zum Lesen ffnen\n", DBFFileName.c_str());
		return (false);
	}

	int DBFRecordCount = DBFGetRecordCount(hDBF);

	if (DBFRecordCount == 0)
	{
		DBFClose(hDBF);

		write_error(2204, "Es sind keine Objekte in der DBF-Datei '%s'\n", DBFFileName.c_str());
	}

	int RecordCount = DBFRecordCount;

	int FieldCount = DBFGetFieldCount(hDBF);

	if (FieldCount == 0)
	{
		DBFClose(hDBF);

		write_error(2206, "In der DBF-Datei %s' sind keine Attribute definiert\nEs mssen aber mindestens die Attibute GEW und STATION definiert sein.\n", DBFFileName.c_str());
	}

	std::string StationName = "STATION";
	int StationFieldIndex = DBFGetFieldIndex(hDBF, StationName.c_str());
	if (StationFieldIndex == -1)
	{
		StationName = "Station";
		StationFieldIndex = DBFGetFieldIndex(hDBF, StationName.c_str());
	}
	if (StationFieldIndex == -1)
	{
		StationName = "station";
		StationFieldIndex = DBFGetFieldIndex(hDBF, StationName.c_str());
	}
	if (StationFieldIndex == -1)
	{
		StationName = "KILOMETER";
		StationFieldIndex = DBFGetFieldIndex(hDBF, StationName.c_str());
	}
	if (StationFieldIndex == -1)
	{
		StationName = "Kilometer";
		StationFieldIndex = DBFGetFieldIndex(hDBF, StationName.c_str());
	}
	if (StationFieldIndex == -1)
	{
		StationName = "kilometer";
		StationFieldIndex = DBFGetFieldIndex(hDBF, StationName.c_str());
	}
	if (StationFieldIndex == -1)
	{
		StationName = "KM";
		StationFieldIndex = DBFGetFieldIndex(hDBF, StationName.c_str());
	}
	if (StationFieldIndex == -1)
	{
		StationName = "Km";
		StationFieldIndex = DBFGetFieldIndex(hDBF, StationName.c_str());
	}
	if (StationFieldIndex == -1)
	{
		StationName = "km";
		StationFieldIndex = DBFGetFieldIndex(hDBF, StationName.c_str());
	}

	if (StationFieldIndex == -1)
	{
		DBFClose(hDBF);

		write_error(2208, "In der DBF-Datei '%s' ist das Attribut 'STATION' nicht definiert\nEs muss aber vorhanden sein.\n", DBFFileName.c_str());
	}

	std::string WspName = "WSP";
	int WspFieldIndex = DBFGetFieldIndex(hDBF, WspName.c_str());
	if (WspFieldIndex == -1)
	{
		WspName = "Wsp";
		WspFieldIndex = DBFGetFieldIndex(hDBF, WspName.c_str());
	}
	if (WspFieldIndex == -1)
	{
		WspName = "wsp";
		WspFieldIndex = DBFGetFieldIndex(hDBF, WspName.c_str());
	}
	if (WspFieldIndex == -1)
	{
		WspName = "WASSERSTAND";
		WspFieldIndex = DBFGetFieldIndex(hDBF, WspName.c_str());
	}
	if (WspFieldIndex == -1)
	{
		WspName = "Wasserstand";
		WspFieldIndex = DBFGetFieldIndex(hDBF, WspName.c_str());
	}
	if (WspFieldIndex == -1)
	{
		WspName = "wasserstand";
		WspFieldIndex = DBFGetFieldIndex(hDBF, WspName.c_str());
	}

	if (WspFieldIndex == -1)
	{
		DBFClose(hDBF);

		write_error(2207, "In der DBF-Datei %s' ist das Attribut 'WSP' nicht definiert\nEs muss aber vorhanden sein.\n", DBFFileName.c_str());
	}

	int Width = 0;
	int Decimals = 0;

	if (DBFGetFieldInfo(hDBF, StationFieldIndex, NULL, &Width, &Decimals) != FTDouble)
	{
		DBFClose(hDBF);

		write_error(2210, "In der DBF-Datei '%s' ist der Attribut-Typ des Attributes '%s' nicht Fliesskommazahl.\n", DBFFileName.c_str(), StationName.c_str());
	}

	if (DBFGetFieldInfo(hDBF, WspFieldIndex, NULL, &Width, &Decimals) != FTDouble)
	{
		DBFClose(hDBF);

		write_error(2210, "In der DBF-Datei '%s' ist der Attribut-Typ des Attributes '%s' nicht Fliesskommazahl.\n", DBFFileName.c_str(), WspName.c_str());
	}

	int Count = 0;
	for (int i = 0; i<RecordCount; i++)
	{
		double StationD = DBFReadDoubleAttribute(hDBF, i, StationFieldIndex);
		double WspD = DBFReadDoubleAttribute(hDBF, i, WspFieldIndex);

		int Station = (int)(StationD * 10000.0 + 0.5);
		if (StationD < 0) Station = (int)(StationD * 10000.0 - 0.5);

		int Wsp = (int)(WspD * 100.0 + 0.5);
		if (WspD < 0)	Wsp = (int)(WspD * 100.0 - 0.5);

		if (ProfilList->Find(Station))
		{
			write_warning(2111, "In der DBF-Datei '%s' kommt der Wasserstand (%.4f) mehrfach vor.\nEs wird nur die erste Definition bercksichtigt.\n", DBFFileName.c_str(), Station / 10000.0);
			continue;
		}

		TProfil* Profil = new TProfil(Station, Wsp);

		unsigned int Anz = ProfilList->size();
		ProfilList->insert(Profil);
		if (ProfilList->size() == Anz)
		{
			write_warning(3119, "Fr den Flusskilometer %.4f wurde der Wasserstand mehrmals angegeben\n", Profil->Station / 10000.0);
			delete Profil;
		}

		Count++;

		if (DebugLevel >= 9 || (DebugLevel >= 1 && Count % 100 == 0) || (Count > 0 && Count % 1000 == 0))
		{
			write_fortschritt("%d von %d Objekten gelesen, %d Wasserstnde im Speicher\n", Count, RecordCount, ProfilList->size());
		}
	}

	write_fortschritt("%d von %d Objekten gelesen, %d Wasserstnde im Speicher\n", Count, RecordCount, ProfilList->size());

	DBFClose(hDBF);

	write_fortschritt("<-Laden der Wasserstnde im DBF-Format beendet\n");

	return (true);
}

//---------------------------------------------------------------------
bool LoadAchse(std::string FileName, TGewaesserAchseList* GewaesserAchseList, int DebugLevel)
{
	write_fortschritt("->Laden der Gewsserachsen gestartet\n");

	LoadPRJ(FileName, DebugLevel);

	if (GewaesserAchseList == 0)
	{
		dump_error(__FILE__, __LINE__, "GewsserAchse ist undefiniert\n");
	}

	std::string SHPFileName = ExchangeFileExt(FileName, ".shp");
	SHPHandle hSHP = SHPOpen(SHPFileName.c_str(), "rb");

	if (hSHP == NULL)
	{
		write_error(2201, "Konnte '%s' nicht zum Lesen ffnen\n", SHPFileName.c_str());
	}

	std::string DBFFileName = ExchangeFileExt(FileName, ".dbf");
	DBFHandle hDBF = DBFOpen(DBFFileName.c_str(), "rb");
	if (hDBF == NULL)
	{
		SHPClose(hSHP);
		write_error(2201, "Konnte '%s' nicht zum Lesen ffnen\n", DBFFileName.c_str());
		return (false);
	}

	int 		SHPRecordCount = 0;
	TShpType	ShapeType = SHPT_NULL;
	double		Mins[4];
	double		Maxs[4];

	SHPGetInfo(hSHP, &SHPRecordCount, &ShapeType, Mins, Maxs);

	if (SHPRecordCount == 0)
	{
		DBFClose(hDBF);
		SHPClose(hSHP);

		write_error(2204, "Es sind keine Objekte in der Shape-Datei '%s'\n", SHPFileName.c_str());
	}

	if (ShapeType != SHPT_ARC && ShapeType != SHPT_ARCZ && ShapeType != SHPT_ARCM)
	{
		write_error(2217, "Der Typ ('%s') der Shape-Datei '%s' ist nicht ARC, ARCZ oder ARCM.\nNur die genannten Typen sind fr die Gewsserachse zulssig.\n", ShapeTypeName(ShapeType), SHPFileName.c_str());
	}

	int DBFRecordCount = DBFGetRecordCount(hDBF);

	if (DBFRecordCount == 0)
	{
		DBFClose(hDBF);
		SHPClose(hSHP);

		write_error(2204, "Es sind keine Objekte in der DBF-Datei '%s'\n", DBFFileName.c_str());
	}

	if (DBFRecordCount != SHPRecordCount)
	{
		DBFClose(hDBF);
		SHPClose(hSHP);

		write_error(2205, "Shape-Datei '%s' und DBF-Datei '%s' enthalten verschieden viele Objekte\n", SHPFileName.c_str(), DBFFileName.c_str());
	}

	int RecordCount = SHPRecordCount;

	int Count = 0;
	for (int i = 0; i<RecordCount; i++)
	{
		SHPObject *psCShape = SHPReadObject(hSHP, i);

		int AnzVert = psCShape->nVertices;

		if (AnzVert < 2)
		{
			write_warning(2118, "In der SHP-Datei '%s' hat eine Gewsserachse keine weniger als 2 Sttzstellen.\nSie wird ignoriert.\n", SHPFileName.c_str());
			continue;
		}

		TGewaesserAchse* GewaesserAchse = new TGewaesserAchse();

		GewaesserAchseList->push_back(GewaesserAchse);

		for (int j=0; j < AnzVert; j++)
		{
			double X = psCShape->padfX[j];
			double Y = psCShape->padfY[j];

			long XL = (long)(X * 100.0 + 0.5);
			long YL = (long)(Y * 100.0 + 0.5);
			GewaesserAchse->AddPoint(XL, YL);
		}

		SHPDestroyObject(psCShape);

		Count++;

		if (DebugLevel >= 9 || (DebugLevel >= 1 && Count % 10 == 0) || (Count > 0 && Count % 100 == 0))
		{
			write_fortschritt("%d von %d Objekten gelesen, %d Gewsserachsen im Speicher\n", Count, RecordCount, GewaesserAchseList->size());
		}
	}

	write_fortschritt("%d von %d Objekten gelesen, %d Gewsserachsen im Speicher\n", Count, RecordCount, GewaesserAchseList->size());

	DBFClose(hDBF);
	SHPClose(hSHP);

	write_fortschritt("<-Laden der Gewsserachsen beendet\n");

	write_fortschritt("->Zusammenfhren von Achsensegmenten gestart\n");

	TGewaesserAchseList::iterator FirstAchseIter = GewaesserAchseList->begin();
	int FirstCount = 0;
	TGewaesserAchseList::iterator TestAchseIter = FirstAchseIter;
	int TestCount = FirstCount;
	Count = 0;
	while (FirstAchseIter != GewaesserAchseList->end())
	{
		Count++;

		TestAchseIter++;
		TestCount++;

		if (TestAchseIter == GewaesserAchseList->end())
		{
			FirstAchseIter++;
			FirstCount++;
			TestAchseIter = FirstAchseIter;
			TestCount = FirstCount;
			continue;
		}

		TGewaesserAchse* FirstAchse = *FirstAchseIter;
		TGewaesserAchse* TestAchse = *TestAchseIter;

		if (DebugLevel >= 9 || (DebugLevel >= 8 && Count % 10 == 0) || Count % 100 == 0)
		{
			write_fortschritt("Teste Segment %d (%d) von %d Segmenten mit %d (%d)\n", FirstCount, FirstAchse->PointList->size(), GewaesserAchseList->size(), TestCount, TestAchse->PointList->size());
		}

		// Point First von FirstAchse (PfFirst) und Point Last von Firstachse(PlFirst)
		TPoint* PfFirst = *(FirstAchse->PointList->begin());
		TPoint*PlFirst = *(FirstAchse->PointList->rbegin());

		TPoint*PfTest = *(TestAchse->PointList->begin());
		TPoint*PlTest = *(TestAchse->PointList->rbegin());

		TXYZList* NewXyList = 0;

		double ffDx = PfFirst->X - PfTest->X;
		double ffDy = PfFirst->Y - PfTest->Y;
		double flDx = PfFirst->X - PlTest->X;
		double flDy = PfFirst->Y - PlTest->Y;
		double llDx = PlFirst->X - PlTest->X;
		double llDy = PlFirst->Y - PlTest->Y;
		double lfDx = PlFirst->X - PfTest->X;
		double lfDy = PlFirst->Y - PfTest->Y;

		if (ffDx * ffDx + ffDy * ffDy < 4)
		{
			NewXyList = new TXYZList();

			for (TPointList::reverse_iterator i = TestAchse->PointList->rbegin(); i != TestAchse->PointList->rend(); i++)
			{
				TPoint* Point = *i;
				NewXyList->Add(Point->X, Point->Y);
			}
			for (TPointList::iterator i = FirstAchse->PointList->begin(); i != FirstAchse->PointList->end(); i++)
			{
				TPoint* Point = *i;
				NewXyList->Add(Point->X, Point->Y);
			}
		}
		else if (flDx * flDx + flDy * flDy < 4)
		{
			NewXyList = new TXYZList();

			for (TPointList::iterator i = TestAchse->PointList->begin(); i != TestAchse->PointList->end(); i++)
			{
				TPoint* Point = *i;
				NewXyList->Add(Point->X, Point->Y);
			}
			for (TPointList::iterator i = FirstAchse->PointList->begin(); i != FirstAchse->PointList->end(); i++)
			{
				TPoint* Point = *i;
				NewXyList->Add(Point->X, Point->Y);
			}
		}
		else if (llDx * llDx + llDy * llDy < 4)
		{
			NewXyList = new TXYZList();

			for (TPointList::iterator i = TestAchse->PointList->begin(); i != TestAchse->PointList->end(); i++)
			{
				TPoint* Point = *i;
				NewXyList->Add(Point->X, Point->Y);
			}
			for (TPointList::reverse_iterator i = FirstAchse->PointList->rbegin(); i != FirstAchse->PointList->rend(); i++)
			{
				TPoint* Point = *i;
				NewXyList->Add(Point->X, Point->Y);
			}
		}
		else if (lfDx * lfDx + lfDy * lfDy < 4)
		{
			NewXyList = new TXYZList();

			for (TPointList::reverse_iterator i = TestAchse->PointList->rbegin(); i != TestAchse->PointList->rend(); i++)
			{
				TPoint* Point = *i;
				NewXyList->Add(Point->X, Point->Y);
			}
			for (TPointList::reverse_iterator i = FirstAchse->PointList->rbegin(); i != FirstAchse->PointList->rend(); i++)
			{
				TPoint* Point = *i;
				NewXyList->Add(Point->X, Point->Y);
			}
		}

		if (NewXyList)
		{
			if (DebugLevel >= 9)	write_fortschritt("Segment %d und %d werden verbunden\n", FirstCount, TestCount);

			// Alte Punkte lschen
			for (TPointList::iterator i=FirstAchse->PointList->begin();	i != FirstAchse->PointList->end(); i++)
			{
				TPoint *Point =	*i;

				delete Point;
				Point =	0;
			}
			delete FirstAchse->PointList;
			FirstAchse->PointList = 0;

			if (DebugLevel >= 9) write_fortschritt("Neue XYList (%d)\n", NewXyList->size());

			// Neue Punkte hinzufgen
			for (TXYZList::iterator i = NewXyList->begin(); i != NewXyList->end(); i++)
			{
				TXYZ* Xy = *i;
				FirstAchse->AddPoint(Xy->X, Xy->Y);
			}

			if (DebugLevel >= 9) write_fortschritt("Neue Anzahl (%d)\n", FirstAchse->PointList->size());

			GewaesserAchseList->erase(TestAchseIter);
			delete TestAchse;

			FirstAchseIter = GewaesserAchseList->begin();
			FirstCount = 0;
			TestAchseIter = FirstAchseIter;
			TestCount = FirstCount;

			delete NewXyList;
		}
	}

	write_fortschritt("<-Zusammenfhren von Achsensegmenten beendet\n");

	return (true);
}

//---------------------------------------------------------------------
bool LoadGebiet(std::string FileName, TXYZList *XyList, int DebugLevel)
{
	write_fortschritt("->Laden der hydraulischen Grenze gestartet\n");

	LoadPRJ(FileName, DebugLevel);

	if (XyList == 0)
	{
		dump_error(__FILE__, __LINE__, "XyList ist undefiniert\n");
	}

	std::string SHPFileName = ExchangeFileExt(FileName, ".shp");
	SHPHandle hSHP = SHPOpen(SHPFileName.c_str(), "rb");

	if (hSHP == NULL)
	{
		write_error(2201, "Konnte '%s' nicht zum Lesen ffnen\n", SHPFileName.c_str());
	}

	int 		RecordCount = 0;
	TShpType	ShapeType = SHPT_NULL;
	double		Mins[4];
	double		Maxs[4];

	SHPGetInfo(hSHP, &RecordCount, &ShapeType, Mins, Maxs);

	if (RecordCount == 0)
	{
		SHPClose(hSHP);

		write_error(2204, "Es sind keine Objekte in der Shape-Datei '%s'\n", SHPFileName.c_str());
	}

	if (RecordCount > 1)
	{
		SHPClose(hSHP);

		write_error(2222, "Es ist mehr als ein Objekt in der Shape-Datei '%s'\n", SHPFileName.c_str());
	}

	if (ShapeType != SHPT_POLYGON && ShapeType != SHPT_POLYGONZ && ShapeType != SHPT_POLYGONM)
	{
		write_error(2221, "Der Typ ('%s') der Shape-Datei '%s' ist nicht POLYGON, POLYGONZ oder POYGONM.\nNur die genannten Typen sind fr die hydraulische Grenze zulssig.\n", ShapeTypeName(ShapeType), SHPFileName.c_str());
	}

	SHPObject *psCShape = SHPReadObject(hSHP, 0);

	int AnzPart = psCShape->nParts;

	if (psCShape->panPartType[0] != SHPP_RING)
	{
		write_error(2225, "Der erste Teil des Begrenzungpolygon '%s' ist keine uerer Ring.\n", SHPFileName.c_str());
	}

	int AnzVert = psCShape->nVertices;

	if (AnzPart > 1)
	{
		write_warning(2124, "Das Begrenzungpolygon '%s' besteht aus mehreren Teilen.\nEs wird nur der erste Teil verwendet.\n", SHPFileName.c_str());

		AnzVert = psCShape->panPartStart[1];
	}

	if (AnzVert < 3)
	{
		write_error(2223, "In der SHP-Datei '%s' hat die hydraulische Grenze weniger als 3 Sttzstellen.\n", SHPFileName.c_str());
	}

	for (int j=0; j < AnzVert-1; j++)
	{
		double X = psCShape->padfX[j];
		double Y = psCShape->padfY[j];

		long XL = (long)(X * 100.0 + 0.5);
		long YL = (long)(Y * 100.0 + 0.5);
		XyList->Add(XL, YL);
	}

	SHPDestroyObject(psCShape);

	SHPClose(hSHP);

	write_fortschritt("<-Laden der hydraulischen Grenze beendet\n");

	return (true);
}

//---------------------------------------------------------------------
bool SavePRJ(std::string FileName, int DebugLevel)
{
	if (PrjOut[0] != '\0')
	{
		write_fortschritt("->Speichern einer Projektionsdatei gestartet\n");

		std::string PRJFileName = ExchangeFileExt(FileName, ".prj");
		FILE *ph = fopen(PRJFileName.c_str(), "w");
		if (ph)
		{
			if (EOF == fputs(PrjOut, ph))
			{
				write_fortschritt("Projektionsdatei '%s' konnte nicht geschrieben werden\n", PRJFileName.c_str());
			}
			else if (EOF == fclose(ph))
			{
				write_fortschritt("Projektionsdatei '%s' konnte nicht geschrieben werden\n", PRJFileName.c_str());
			}
		}
		else
		{
			write_fortschritt("Projektionsdatei '%s' konnte nicht zum Schreiben geffnet werden\n", PRJFileName.c_str());
		}
		write_fortschritt("<-Speichern einer Projektionsdatei beendet\n");
	}
	else
	{
		if (DebugLevel >= 3) write_fortschritt("Keine Projektionsinformationen vorhanden\n");
	}

	return (true);
}

//---------------------------------------------------------------------------
bool SaveNet(std::string FileName, TNodeList *NodeList, TElementList *ElementList, int DebugLevel)
{
	write_fortschritt("->Speichern des 2DM-Netzes gestartet\n");

	if (NodeList == 0)		dump_error(__FILE__, __LINE__, "Die Knotenliste ist nicht definiert\n");

	write_fortschritt("->Umnumerierung der Knoten gestartet\n");

	for (unsigned int i=0; i<NodeList->size(); i++)
	{
		if (DebugLevel >= 9 || (DebugLevel >= 1 && i % 5000 == 0) || (i > 0 && i % 50000 == 0))
		{
			write_fortschritt("%d von %d Knoten umnumeriert\n", i, NodeList->size());
		}

		TNode *Node = (*NodeList)[i];

		Node->Nr = i + 1;
	}

	write_fortschritt("%d von %d Knoten umnumeriert\n", NodeList->size(), NodeList->size());

	write_fortschritt("<-Umnumerierung der Knoten beendet\n");

	FILE *fh = fopen (FileName.c_str(), "w");
	if (fh == NULL)
	{
		write_error(4201, "Kann Datei '%s' nicht zum Schreiben ffnen\n", FileName.c_str());
	}

	fprintf(fh, "MESH2D\n");

	if (ElementList && ElementList->size() > 0)
	{
		write_fortschritt("->Speichern der Elemente gestartet\n");

		for (unsigned int i=0; i<ElementList->size(); i++)
		{
			if (DebugLevel >= 9 || (DebugLevel >= 1 && i % 5000 == 0) || (i > 0 && i % 50000 == 0))
			{
				write_fortschritt("%d von %d Elementen gespeichert\n", i, ElementList->size());
			}

			TElement *Element = (*ElementList)[i];

			if (Element->Typ == NO_ELEMENT)
			{
				continue;
			}

			if (Element->Typ != TRI)
			{
				dump_error(__FILE__, __LINE__, "Ein Element hat nicht genau 3 Knoten\n");
			}

			TNode *N1 = Element->Node1;
			TNode *N2 = Element->Node2;
			TNode *N3 = Element->Node3;

			fprintf(fh, "E3T %6d %6d %6d %6d %6d\n", i+1, N1->Nr, N2->Nr, N3->Nr, 1);
		}

		write_fortschritt("%d von %d Elementen gespeichert\n", ElementList->size(), ElementList->size());

		write_fortschritt("<-Speichern der Elemente beendet\n");
	}

	write_fortschritt("->Speichern der Knoten gestartet\n");

	for (unsigned int i=0; i<NodeList->size(); i++)
	{
		if (DebugLevel >= 9 || (DebugLevel >= 1 && i % 5000 == 0) || (i > 0 && i % 50000 == 0))
		{
			write_fortschritt("%d von %d Knoten gespeichert\n", i, NodeList->size());
		}

		TNode *Node = (*NodeList)[i];

		if (Node->Z == NoZ)	continue;

		fprintf(fh, "ND %6d %15.9e %15.9e %15.9e\n", Node->Nr, Node->X / 100.0, Node->Y / 100.0, Node->Z / 100.0);
	}

	write_fortschritt("%d von %d Knoten gespeichert\n", NodeList->size(), NodeList->size());

	write_fortschritt("<-Speichern der Knoten beendet\n");

	write_fortschritt("->Speichern der Modell-Parameter gestartet\n");

	fprintf(fh, "BEGPARAMDEF\n");
	fprintf(fh, "GM  HYDRO_AS-2D\n");
	fprintf(fh, "SI  1\n");
	fprintf(fh, "DY  1\n");
	fprintf(fh, "TD  500  20000\n");
	fprintf(fh, "KEY  \"hydro_as-2d\"\n");
	fprintf(fh, "PD  \"Rechenlauf fortsetzen ?\"  0  0\n");
	fprintf(fh, "PD  \"Zeitintervall SMS [s]\"  2  5000\n");
	fprintf(fh, "PD  \"Hmin [m]\"  2  0.01\n");
	fprintf(fh, "PD  \"VELMAX [m/s]\"  2  15\n");
	fprintf(fh, "PD  \"Amin\"  2  1e-015\n");
	fprintf(fh, "PD  \"CMUVISC\"  2  0.6\n");
	fprintf(fh, "NUME  2\n");
	fprintf(fh, "BD 0  \"KUK / Druckabfluss\"  1  1 \"KUK [m  NN]\"\n");
	fprintf(fh, "BCDISP  0  1  1  0  0  0 0\n");
	fprintf(fh, "BD 0  \"Pegelpunkt\"  2  0\n");
	fprintf(fh, "BCDISP  0  2  1  0  0  0 0\n");
	fprintf(fh, "BD 1  \"Zulauf\"  1  2 \"Abfluss [m^3/s]\" \"Richtung CCW \"\n");
	fprintf(fh, "BCDISP  1  1  1  0  255  0 0\n");
	fprintf(fh, "BD 1  \"Auslauf / Ie\"  2  1 \"EL_Geflle [Promille]\"\n");
	fprintf(fh, "BCDISP  1  2  1  0  0  0 255\n");
	fprintf(fh, "BD 1  \"Auslauf / W-Q\"  3  11 \"dQ [m^3/s]\" \"WSP_0 [m  NN]\" \"WSP_1 [m  NN]\" \"WSP_2 [m  NN]\" \"WSP_3 [m  NN]\" \"WSP_4 [m  NN]\" \"WSP_5 [m  NN]\" \"WSP_6 [m  NN]\" \"WSP_7 [m  NN]\" \"WSP_8 [m  NN]\" \"WSP_9 [m  NN]\"\n");
	fprintf(fh, "BCDISP  1  3  1  0  0  255 255\n");
	fprintf(fh, "BD 1  \"Durchlass / Rechteck\"  4  6 \"Abflusskoeffizient\" \"Breite [m]\" \"Hhe [m]\" \"Sohlhhe z1 [m  NN]\" \"Sohlhhe z2 [m  NN]\" \"Formel: 0=Poleni / 1=DU_BUAT\"\n");
	fprintf(fh, "BCDISP  1  4  1  0  255  255 0\n");
	fprintf(fh, "BD 1  \"Durchlass / kreisfrmig\"  5  5 \"Abflusskoeffizient\" \"Durchmesser [m]\" \"Sohlhhe z1 [m  NN]\" \"Sohlhhe z2 [m  NN]\" \"Formel: 0=Poleni / 1=DU_BUAT\"\n");
	fprintf(fh, "BCDISP  1  5  1  0  128  255 0\n");
	fprintf(fh, "BD 1  \"Wehrberfall\"  6  4 \"berfallbeiwert\" \"Segmentbreite [m]\" \"Kronenhhe [m  NN]\" \"Formel: 0=Poleni / 1=DU_BUAT\"\n");
	fprintf(fh, "BCDISP  1  6  1  0  255  0 255\n");
	fprintf(fh, "BD 1  \"Kontrollquerschnitt\"  7  0\n");
	fprintf(fh, "BCDISP  1  7  1  0  0  0 0\n");
	fprintf(fh, "MD  2 \"Strickler\" \"Wirbelviskositt\"\n");
	fprintf(fh, "ENDPARAMDEF\n");
	fprintf(fh, "BEG2DMBC\n");

	fprintf(fh, "END2DMBC\n");

	write_fortschritt("<-Speichern der Modell-Parameter beendet\n");

	fclose (fh);

	write_fortschritt("<-Speichern des 2DM-Netzes beendet\n");

	return (true);
}

//---------------------------------------------------------------------------
bool SaveSol(std::string FileName, TNodeList *NodeList, int DebugLevel)
{
	write_fortschritt("->Speichern der SOL-Wasserstnde gestartet\n");

	if (NodeList == 0)		dump_error(__FILE__, __LINE__, "Die Knotenliste ist nicht definiert\n");

	FILE *fh = fopen (FileName.c_str(), "w");
	if (fh == NULL)
	{
		write_error(4201, "Kann Datei '%s' nicht zum Schreiben ffnen\n", FileName.c_str());
	}

	fprintf(fh, "SCALAR\n");
	fprintf(fh, "ND %d\n", NodeList->size());
	fprintf(fh, "ST 0\n");
	fprintf(fh, "TS 0\n");

	for (unsigned int i=0; i<NodeList->size(); i++)
	{
		if (DebugLevel >= 9 || (DebugLevel >= 1 && i % 5000 == 0) || (i > 0 && i % 50000 == 0))
		{
			write_fortschritt("%d von %d Wasserstnde gespeichert\n", i, NodeList->size());
		}

		TNode *Node = (*NodeList)[i];

		fprintf(fh, "%15.9f\n", Node->Wsp / 100.0);
	}

	write_fortschritt("%d von %d Wasserstnde gespeichert\n", NodeList->size(), NodeList->size());

	fclose (fh);

	write_fortschritt("<-Speichern der SOL-Wasserstnde beendet\n");

	return (true);
}

//---------------------------------------------------------------------------
bool SaveProfile(std::string FileName, TProfilList *ProfilList, int DebugLevel)
{
	write_fortschritt("->Speichern der Profile gestartet\n");

	SavePRJ(FileName, DebugLevel);

	DBFHandle DBFHandle = DBFCreate(FileName.c_str());
	if(DBFHandle == NULL )
	{
		write_error(4201, "Kann Datei '%s' nicht zum Schreiben ffnen\n", FileName.c_str());
	}

	if (DBFAddField(DBFHandle, "STATION", FTDouble, 10, 4) == -1)
	{
		DBFClose(DBFHandle);
		write_error(4204, "Kann bei Datei '%s' das Attribut 'STATION' nicht erzeugen\n", FileName.c_str());
	}

	if (DBFAddField(DBFHandle, "WSP", FTDouble, 10, 2) == -1)
	{
		DBFClose(DBFHandle);
		write_error(4204, "Kann bei Datei '%s' das Attribut 'WSP' nicht erzeugen\n", FileName.c_str());
	}

	SHPHandle SHPHandle = SHPCreate(FileName.c_str(), SHPT_ARC);
	SHPSetMaxRecords(SHPHandle, ProfilList->size());

	int AnzProfil = 0;
	for (TProfilList::iterator i = ProfilList->begin(); i != ProfilList->end(); i++)
	{
		TProfil* Profil = *i;

		double		Station = Profil->Station / 10000.0;
		double		Wsp = Profil->Wsp / 100.0;

		double	*x = (double *)malloc(sizeof(double) * (Profil->PointList->size() + 1));
		double 	*y = (double *)malloc(sizeof(double) * (Profil->PointList->size() + 1));
		double 	*z = (double *)malloc(sizeof(double) * (Profil->PointList->size() + 1));

		int AnzPoint = 0;
		for (TPointList::iterator j = Profil->PointList->begin(); j != Profil->PointList->end(); j++)
		{
			TPoint* Point = *j;

			double X = Point->X / 100.0;
			double Y = Point->Y / 100.0;

			x[AnzPoint] = X;
			y[AnzPoint] = Y;
			z[AnzPoint] = 0.0;

			AnzPoint++;
		}

		SHPObject *psShape = SHPCreateSimpleObject(SHPT_ARC, AnzPoint, x, y, z);
		SHPWriteObject(SHPHandle, -1, psShape);
		SHPDestroyObject(psShape);

		DBFWriteDoubleAttribute(DBFHandle, AnzProfil, 0, Station);
		DBFWriteDoubleAttribute(DBFHandle, AnzProfil, 1, Wsp);

		free(x);
		free(y);
		free(z);

		AnzProfil++;

		if (DebugLevel >= 9 || (DebugLevel >= 1 && AnzProfil % 100 == 0) || (AnzProfil > 0 && AnzProfil % 1000 == 0))
		{
			write_fortschritt("%d von %d Profile gespeichert\n", AnzProfil, ProfilList->size());
		}
	}
	write_fortschritt("%d von %d Profile gespeichert\n", AnzProfil, ProfilList->size());

	DBFClose(DBFHandle);
	SHPClose(SHPHandle);

	write_fortschritt("<-Speichern der Profile beendet\n");

	return (true);
}

//---------------------------------------------------------------------------
bool SavePolygon(std::string FileName, TXYZList *XyList, int DebugLevel)
{
	write_fortschritt("->Speichern des Begrenzungspolygons gestartet\n");

	SavePRJ(FileName, DebugLevel);

	DBFHandle DBFHandle = DBFCreate(FileName.c_str());
	if(DBFHandle == NULL )
	{
		write_error(4201, "Kann Datei '%s' nicht zum Schreiben ffnen\n", FileName.c_str());
	}

	if (DBFAddField(DBFHandle, "ID", FTInteger, 10, 0) == -1)
	{
		DBFClose(DBFHandle);
		write_error(4204, "Kann bei Datei '%s' das Attribut 'ID' nicht erzeugen\n", FileName.c_str());
	}

	SHPHandle SHPHandle = SHPCreate(FileName.c_str(), SHPT_POLYGON);

	double	*x = (double *)malloc(sizeof(double) * (XyList->size()+1));
	double 	*y = (double *)malloc(sizeof(double) * (XyList->size()+1));
	double 	*z = (double *)malloc(sizeof(double) * (XyList->size()+1));

	int AnzPoint = 0;
	for (TXYZList::iterator i = XyList->begin(); i != XyList->end(); i++)
	{
		x[AnzPoint] = (*i)->X / 100.0;
		y[AnzPoint] = (*i)->Y / 100.0;
		z[AnzPoint] = 0.0;

		AnzPoint++;

		if (DebugLevel >= 9 || (DebugLevel >= 1 && AnzPoint % 1000 == 0) || (AnzPoint > 0 && AnzPoint % 10000 == 0))
		{
			write_fortschritt("%d von %d Sttzstellen gespeichert\n", AnzPoint, XyList->size()+1);
		}
	}

	TXYZList::iterator i = XyList->begin();
	x[AnzPoint] = (*i)->X / 100.0;
	y[AnzPoint] = (*i)->Y / 100.0;
	z[AnzPoint] = 0.0;

	AnzPoint++;

	SHPObject *psShape = SHPCreateSimpleObject(SHPT_POLYGON, AnzPoint, x, y, z);
	SHPWriteObject(SHPHandle, -1, psShape);
	SHPDestroyObject(psShape);

	DBFWriteIntegerAttribute(DBFHandle, 0, 0, 0);

	free(x);
	free(y);
	free(z);

	DBFClose(DBFHandle);
	SHPClose(SHPHandle);

	write_fortschritt("%d von %d Sttzstellen gespeichert\n", AnzPoint, XyList->size()+1);

	write_fortschritt("<-Speichern des Begrenzungspolygons beendet\n");

	return (true);
}

//---------------------------------------------------------------------------
void CreatePolygon(SHPObject* psObject, int nParts, int* panPartStart, int* panPartType, int nVertices, double* padfX, double* padfY)
{
	psObject->nSHPType = SHPT_POLYGON;
	psObject->nShapeId = -1;

	psObject->nParts = nParts;

	psObject->panPartStart = panPartStart;
	psObject->panPartType = panPartType;

	psObject->padfX = padfX;
	psObject->padfY = padfY;
	psObject->padfZ = 0;
	psObject->padfM = 0;

	/* -------------------------------------------------------------------- */
	/*      Compute the extents.                                            */
	/* -------------------------------------------------------------------- */
	psObject->nVertices = nVertices;

	/* -------------------------------------------------------------------- */
	/*      Build extents for this object.                                  */
	/* -------------------------------------------------------------------- */
	psObject->dfZMin = psObject->dfZMax = 0;
	psObject->dfMMin = psObject->dfMMax = 0;

	psObject->dfXMin = psObject->dfXMax = psObject->padfX[0];
	psObject->dfYMin = psObject->dfYMax = psObject->padfY[0];

	for (int i = 1; i < psObject->nVertices; i++)
	{
		if (psObject->dfXMin > psObject->padfX[i])	psObject->dfXMin = psObject->padfX[i];
		if (psObject->dfYMin > psObject->padfY[i])	psObject->dfYMin = psObject->padfY[i];

		if (psObject->dfXMax < psObject->padfX[i])	psObject->dfXMax = psObject->padfX[i];
		if (psObject->dfYMax < psObject->padfY[i])	psObject->dfYMax = psObject->padfY[i];
	}
}

//---------------------------------------------------------------------------
bool SavePolygone(std::string FileName, TErgebnisPolygonList *PolygonList, int DebugLevel)
{
	write_fortschritt("->Speichern der Ergebnispolygone gestartet\n");

	SavePRJ(FileName, DebugLevel);

	DBFHandle DBFHandle = DBFCreate(FileName.c_str());
	if(DBFHandle == NULL )
	{
		write_error(4201, "Kann Datei '%s' nicht zum Schreiben ffnen\n", FileName.c_str());
	}

	if (DBFAddField(DBFHandle, "DIFF", FTDouble, 12, 4) == -1)
	{
		DBFClose(DBFHandle);
		write_error(4204, "Kann bei Datei '%s' das Attribut 'DIFF' nicht erzeugen\n", FileName.c_str());
	}

	write_fortschritt("->Lschen der zu kleinen Ringe gestartet\n");

	int	Count = 0;
	int DelCount = 0;
	long double MinArea = -1;

	for (TErgebnisPolygonList::iterator j = PolygonList->begin(); j != PolygonList->end(); j++)
	{
		Count++;

		if (DebugLevel >= 9 || (DebugLevel >= 1 && Count % 10000 == 0) || (Count > 0 && Count % 100000 == 0))
		{
			write_fortschritt("%d von %d Polygonen, %d Polygone gelscht\n", Count, PolygonList->size(), DelCount);
		}

		TErgebnisPolygon* EP1 = *j;

		long double Area = fabs(EP1->Area());
		if (MinArea < 0 || MinArea > Area)	MinArea = Area;

		// 10000 entspricht 100 cm * 100 cm
		if (Area < 10000.0)
		{
			DelCount++;

			EP1->Diff = NotValid;
		}
	}
	write_fortschritt("%d Ringe gelscht\n", DelCount);
	write_fortschritt("kleinste Ringflche: %.3f\n", (double)MinArea / 10000.0);

	write_fortschritt("<-Lschen der zu kleinen Ringe beendet\n");

	SHPHandle SHPHandle = SHPCreate(FileName.c_str(), SHPT_POLYGON);

	int		*ps = 0;
	int		*pt = 0;
	double	*x = 0;
	double 	*y = 0;

	size_t OldPartSize = 0;
	size_t OldPointSize = 0;

	Count = 0;
	TErgebnisPolygonList::iterator i = PolygonList->begin();
	while (i != PolygonList->end())
	{
		TErgebnisPolygon* ErgebnisPolygon = *i;

        if (ErgebnisPolygon->Diff == NotValid)
        {
			// Dieses Polygon wird nicht mehr gebraucht
			i++;
	        continue;
        }

		double AktDiff = ErgebnisPolygon->Diff;

		size_t PointSize = 0;
		int	AnzPolyParts = 0;
		int	AnzPolyPoints = 0;
		for (TErgebnisPolygonList::iterator j = PolygonList->begin(); j != PolygonList->end(); j++)
		{
			TErgebnisPolygon* EP1 = *j;

	        if (EP1->Diff == NotValid)
            {
				// Dieses Polygon wird nicht mehr gebraucht
	            continue;
            }

			if (EP1->Diff != AktDiff)
			{
				// Andere Diffs werden im Moment nicht bearbeitet
				continue;
			}

			TXYZList::iterator First = EP1->begin();
			TXYZ *Xy = *First;
			long X = Xy->X;
			long Y = Xy->Y;

			// Jetzt wird getestet, ob der erste Punkt innerhalb liegt
			bool IsInnerRing = false;
			for (TErgebnisPolygonList::iterator k = PolygonList->begin(); k != PolygonList->end(); k++)
			{
				TErgebnisPolygon* EP2 = *k;

		        if (EP2->Diff == NotValid)
    		    {
					// Dieses Polygon wird nicht mehr gebraucht
			       	continue;
	    	    }

				if (EP2 == EP1)
				{
					// Nur mit anderen Polygonen vergleichen
					continue;
				}

				if (EP2->Diff != AktDiff)
				{
					// Andere Diffs werden im Moment nicht bearbeitet
					continue;
				}

				if (EP2->IsInsideXYList(X, Y))
				{
					IsInnerRing = true;
					break;
				}
			}

			// Dieses Polygon ist ein neuer Ring
			AnzPolyParts++;

			size_t PartSize = sizeof(int) * AnzPolyParts;
			if (PartSize > OldPartSize)
			{
				OldPartSize = PartSize * 2;
				ps = SfIntRealloc(ps, OldPartSize);
				pt = SfIntRealloc(pt, OldPartSize);
			}

			// Alle Punkte und der erste nochmal am Ende
			PointSize = PointSize + sizeof(double) * (EP1->size() + 1);
			if (PointSize > OldPointSize)
			{
				OldPointSize = PointSize * 2;

				x = SfDoubleRealloc(x, OldPointSize);
				y = SfDoubleRealloc(y, OldPointSize);
			}

			ps[AnzPolyParts-1] = AnzPolyPoints;

			if (IsInnerRing)	pt[AnzPolyParts-1] = SHPP_INNERRING;
			else				pt[AnzPolyParts-1] = SHPP_RING;

			for (TErgebnisPolygon::iterator m = EP1->begin(); m != EP1->end(); m++)
			{
                x[AnzPolyPoints] = (*m)->X / 100.0;
				y[AnzPolyPoints] = (*m)->Y / 100.0;

				if (DebugLevel == 9)
				{
					write_fortschritt("%.2f %.2f\n", x[AnzPolyPoints], y[AnzPolyPoints]);
				}

				AnzPolyPoints++;
			}

			TXYZList::iterator m = EP1->begin();
			x[AnzPolyPoints] = (*m)->X / 100.0;
			y[AnzPolyPoints] = (*m)->Y / 100.0;

			if (DebugLevel == 9)
			{
				write_fortschritt("%.2f %.2f\n", x[AnzPolyPoints], y[AnzPolyPoints]);
			}

			AnzPolyPoints++;
		}

		write_fortschritt("%d. Ergebnispolygon (Wassertiefe %.2f, %d Ringe, %d Sttzstellen)\n", Count+1, AktDiff / 100.0, AnzPolyParts, AnzPolyPoints);

		if (AnzPolyParts > 0 && AnzPolyPoints > 0)
		{
			DBFWriteDoubleAttribute(DBFHandle, Count, 0, AktDiff / 100.0);

			SHPObject *psShape = SfSHPObjectRealloc(0, 1);

			CreatePolygon(psShape, AnzPolyParts, ps, pt, AnzPolyPoints, x, y);

			SHPRewindObject(SHPHandle, psShape);
			SHPWritePolygon(SHPHandle, psShape);

			SfSHPObjectFree (psShape);

			Count++;
		}


		for (TErgebnisPolygonList::iterator j = PolygonList->begin(); j != PolygonList->end(); j++)
		{
			TErgebnisPolygon* EP1 = *j;

			if (EP1->Diff == AktDiff)
			{
				EP1->Diff = NotValid;
			}
		}

		i = PolygonList->begin();
	}

	SfIntFree(ps);
	SfIntFree(pt);
	SfDoubleFree(x);
	SfDoubleFree(y);

	DBFClose(DBFHandle);
	SHPClose(SHPHandle);

	write_fortschritt("<-Speichern der Ergebnispolygone beendet\n");

	return (true);
}

//---------------------------------------------------------------------------
int OutElement(int AnzPolygone, TNode* Node1, TNode* Node2, TNode* Node3, int TopoDiff, DBFHandle PolygonDBFHandle, SHPHandle PolygonSHPHandle)
{
	double Wsp = 0.0;
	double Depth = 0.0;
	double Topo = 0.0;

	long Z1 = Node1->Z;
	long Z2 = Node2->Z;
	long Z3 = Node3->Z;

	long W1 = Node1->Wsp;
	long W2 = Node2->Wsp;
	long W3 = Node3->Wsp;

	if (W1 < Z1 + TopoDiff)	W1 = NoWsp;
	if (W2 < Z2 + TopoDiff)	W2 = NoWsp;
	if (W3 < Z3 + TopoDiff)	W3 = NoWsp;

	if (W1 == NoWsp && W2 == NoWsp && W3 == NoWsp)	Wsp = NoWsp;
	else if (W1 == NoWsp && W2 == NoWsp)			Wsp = W3 / 100.0;
	else if (W1 == NoWsp && W3 == NoWsp)			Wsp = W2 / 100.0;
	else if (W2 == NoWsp && W3 == NoWsp)			Wsp = W1 / 100.0;
	else if (W1 == NoWsp)							Wsp = (W2 + W3) / 200.0;
	else if (W2 == NoWsp)							Wsp = (W1 + W3) / 200.0;
	else if (W3 == NoWsp)							Wsp = (W1 + W2) / 200.0;
	else											Wsp = (W1 + W2 + W3) / 300.0;

	Topo = (Z1 + Z2 + Z3) / 300.0;

	if (Wsp == NoWsp)	Depth = 0.0;
	else				Depth = Wsp - Topo;

	if (Depth < TopoDiff)	Depth = TopoDiff;

	double X1 = Node1->X / 100.0;
	double Y1 = Node1->Y / 100.0;
	double X2 = Node2->X / 100.0;
	double Y2 = Node2->Y / 100.0;
	double X3 = Node3->X / 100.0;
	double Y3 = Node3->Y / 100.0;

	double DoubleArea = X1 * Y2 - X2 * Y1 + X2 * Y3 - X3 * Y2 + X3 * Y1 - X1 * Y3;

	if (DoubleArea < 0.0)
	{
		SHPWriteElement(PolygonSHPHandle, X1, Y1, X2, Y2, X3, Y3);
	}
	else
	{
		SHPWriteElement(PolygonSHPHandle, X1, Y1, X3, Y3, X2, Y2);
	}

	DBFWriteDoubleAttribute(PolygonDBFHandle, AnzPolygone, 0, Topo);
	DBFWriteDoubleAttribute(PolygonDBFHandle, AnzPolygone, 1, Wsp);
	DBFWriteDoubleAttribute(PolygonDBFHandle, AnzPolygone, 2, Depth);

	AnzPolygone++;

	return (AnzPolygone);
}

//---------------------------------------------------------------------------
bool SaveElements(std::string FileName, TElementList *ElementList, int TopoDiff, int DebugLevel)
{
	write_fortschritt("->Speichern der Elemente gestartet\n");

	SavePRJ(FileName, DebugLevel);

	DBFHandle DBFHandle = DBFCreate(FileName.c_str());
	if(DBFHandle == NULL )
	{
		write_error(4201, "Kann Datei '%s' nicht zum Schreiben ffnen\n", FileName.c_str());
	}

	if (DBFAddField(DBFHandle, "Topo", FTDouble, 10, 2) == -1)
	{
		DBFClose(DBFHandle);
		write_error(4204, "Kann bei Datei '%s' das Attribut 'Topo' nicht erzeugen\n", FileName.c_str());
	}

	if (DBFAddField(DBFHandle, "Wsp", FTDouble, 10, 2) == -1)
	{
		DBFClose(DBFHandle);
		write_error(4204, "Kann bei Datei '%s' das Attribut 'Wsp' nicht erzeugen\n", FileName.c_str());
	}

	if (DBFAddField(DBFHandle, "Depth", FTDouble, 10, 2) == -1)
	{
		DBFClose(DBFHandle);
		write_error(4204, "Kann bei Datei '%s' das Attribut 'Depth' nicht erzeugen\n", FileName.c_str());
	}

	SHPHandle SHPHandle = SHPCreate(FileName.c_str(), SHPT_POLYGON);
	SHPSetMaxRecords(SHPHandle, ElementList->size()+1);

	int AnzPolygone = 0;
	for (TElementList::iterator i=ElementList->begin(); i != ElementList->end(); i++)
	{
		if (DebugLevel >= 9 || (DebugLevel >= 1 && AnzPolygone % 5000 == 0) || (AnzPolygone > 0 && AnzPolygone % 50000 == 0))
		{
			write_fortschritt("%d von %d Elementen gespeichert\n", AnzPolygone, ElementList->size());
		}

		TElement *Element = *i;

		if (Element->Typ == TRI)
		{
			TNode *Node1 = Element->Node1;
			TNode *Node2 = Element->Node2;
			TNode *Node3 = Element->Node3;

			AnzPolygone = OutElement(AnzPolygone, Node1, Node2, Node3, TopoDiff, DBFHandle, SHPHandle);
		}
        else if (Element->Typ == QUAD)
		{
			TNode *Node1 = Element->Node1;
			TNode *Node2 = Element->Node2;
			TNode *Node3 = Element->Node3;
			TNode *Node4 = Element->Node4;

			AnzPolygone = OutElement(AnzPolygone, Node1, Node2, Node3, TopoDiff, DBFHandle, SHPHandle);
			AnzPolygone = OutElement(AnzPolygone, Node1, Node3, Node4, TopoDiff, DBFHandle, SHPHandle);
        }
	}

	SHPClose(SHPHandle);
	DBFClose(DBFHandle);

	write_fortschritt("%d von %d Elementen gespeichert\n", AnzPolygone, ElementList->size());

	write_fortschritt("<-Speichern der Elemente beendet\n");

	return (true);
}

//---------------------------------------------------------------------------
bool SaveNodes(std::string FileName, TNodeList *NodeList, int DebugLevel)
{
	write_fortschritt("->Speichern der Knoten gestartet\n");

	SavePRJ(FileName, DebugLevel);

	DBFHandle DBFHandle = DBFCreate(FileName.c_str());
	if(DBFHandle == NULL )
	{
		write_error(4201, "Kann Datei '%s' nicht zum Schreiben ffnen\n", FileName.c_str());
	}

	if (DBFAddField(DBFHandle, "Topo", FTDouble, 10, 2) == -1)
	{
		DBFClose(DBFHandle);
		write_error(4204, "Kann bei Datei '%s' das Attribut 'Topo' nicht erzeugen\n", FileName.c_str());
	}

	if (DBFAddField(DBFHandle, "Wsp", FTDouble, 10, 2) == -1)
	{
		DBFClose(DBFHandle);
		write_error(4204, "Kann bei Datei '%s' das Attribut 'Wsp' nicht erzeugen\n", FileName.c_str());
	}

	if (DBFAddField(DBFHandle, "Depth", FTDouble, 10, 2) == -1)
	{
		DBFClose(DBFHandle);
		write_error(4204, "Kann bei Datei '%s' das Attribut 'Depth' nicht erzeugen\n", FileName.c_str());
	}

	SHPHandle SHPHandle = SHPCreate(FileName.c_str(), SHPT_POINT);
	SHPSetMaxRecords(SHPHandle, NodeList->size());

	int AnzNodes = NodeList->size();
	int Count = 0;
	for (TNodeList::iterator i=NodeList->begin(); i != NodeList->end(); i++)
	{
		if (DebugLevel >= 9 || (DebugLevel >= 1 && Count % 5000 == 0) || (Count > 0 && Count % 50000 == 0))
		{
			write_fortschritt("%d von %d Knoten gespeichert\n", Count, AnzNodes);
		}

		TNode *Node = *i;

		if (Node->Z != NoZ)
		{
			double X = Node->X / 100.0;
			double Y = Node->Y / 100.0;

			SHPWritePoint(SHPHandle, X, Y);

//			SHPObject *psShape = SHPCreateSimpleObject(SHPT_POINT, 1, &X, &Y, &Z);
//			SHPWriteObject(SHPHandle, -1, psShape);
//			SHPDestroyObject(psShape);

			DBFWriteDoubleAttribute(DBFHandle, Count, 0, Node->Z / 100.0);

			if (Node->Wsp == NoWsp)
			{
				DBFWriteDoubleAttribute(DBFHandle, Count, 1, 0.00);
				DBFWriteDoubleAttribute(DBFHandle, Count, 2, 0.00);
			}
			else
			{
				DBFWriteDoubleAttribute(DBFHandle, Count, 1, Node->Wsp / 100.0);
				DBFWriteDoubleAttribute(DBFHandle, Count, 2, Node->Wsp / 100.0 - Node->Z / 100.0);
			}

			Count++;
		}
		else
		{
			AnzNodes--;
		}
	}

	SHPClose(SHPHandle);
	DBFClose(DBFHandle);

	write_fortschritt("%d von %d Knoten gespeichert\n", Count, AnzNodes);

	write_fortschritt("<-Speichern der Knoten beendet\n");

	return (true);
}

//---------------------------------------------------------------------------
bool SaveEdges(std::string FileName, TEdgeList *EdgeList, int DebugLevel)
{
	write_fortschritt("->Speichern der Kanten gestartet\n");

	SavePRJ(FileName, DebugLevel);

	DBFHandle DBFHandle = DBFCreate(FileName.c_str());
	if(DBFHandle == NULL )
	{
		write_error(4201, "Kann Datei '%s' nicht zum Schreiben ffnen\n", FileName.c_str());
	}

	if (DBFAddField(DBFHandle, "Nr1", FTInteger, 10, 0) == -1)
	{
		DBFClose(DBFHandle);
		write_error(4204, "Kann bei Datei '%s' das Attribut 'Nr1' nicht erzeugen\n", FileName.c_str());
	}

	if (DBFAddField(DBFHandle, "Nr2", FTInteger, 10, 0) == -1)
	{
		DBFClose(DBFHandle);
		write_error(4204, "Kann bei Datei '%s' das Attribut 'Nr2' nicht erzeugen\n", FileName.c_str());
	}

	if (DBFAddField(DBFHandle, "Aussen", FTInteger, 1, 0) == -1)
	{
		DBFClose(DBFHandle);
		write_error(4204, "Kann bei Datei '%s' das Attribut 'Aussen' nicht erzeugen\n", FileName.c_str());
	}

	SHPHandle SHPHandle = SHPCreate(FileName.c_str(), SHPT_ARC);

	int Count = 0;
	for (TEdgeVector::iterator i = EdgeList->EdgeVector.begin(); i != EdgeList->EdgeVector.end(); i++)
	{
		if (DebugLevel >= 9 || (DebugLevel >= 1 && Count % 5000 == 0) || (Count > 0 && Count % 50000 == 0))
		{
			write_fortschritt("%d von %d Kanten gespeichert\n", Count, EdgeList->EdgeVector.size());
		}

		TEdge* Edge = *i;

		TNode* Node1 = Edge->Node1;
		TNode* Node2 = Edge->Node2;

		int Nr1 = Node1->Nr;
		int Nr2 = Node2->Nr;

		double	x[2];
		double 	y[2];
		double 	z[2];

		x[0] = Node1->X / 100.0;
		y[0] = Node1->Y / 100.0;
		x[1] = Node2->X / 100.0;
		y[1] = Node2->Y / 100.0;
		z[0] = 0.0;
		z[1] = 0.0;

		SHPObject *psShape = SHPCreateSimpleObject(SHPT_ARC, 2, x, y, z);
		SHPWriteObject(SHPHandle, -1, psShape);
		SHPDestroyObject(psShape);

		DBFWriteIntegerAttribute(DBFHandle, Count, 0, Nr1);
		DBFWriteIntegerAttribute(DBFHandle, Count, 1, Nr2);
		if (Edge->IsBoundary)	DBFWriteIntegerAttribute(DBFHandle, Count, 2, 1);
		else					DBFWriteIntegerAttribute(DBFHandle, Count, 2, 0);

		Count++;
	}

	DBFClose(DBFHandle);
	SHPClose(SHPHandle);

	write_fortschritt("%d von %d Kanten gespeichert\n", Count, EdgeList->EdgeVector.size());

	write_fortschritt("<-Speichern der Kanten beendet\n");

	return (true);
}

//---------------------------------------------------------------------------
bool SaveXml(std::string FileName, TParameter *Parameter)
{
	write_fortschritt("->Speichern der Meta-Daten gestartet\n");

	FileName = FileName + ".xml";

	FILE *fh = fopen (FileName.c_str(), "w");
	if (fh == NULL)
	{
		write_error(4201, "Kann Datei '%s' nicht zum Schreiben ffnen\n", FileName.c_str());
	}











	long Secs, Mins, Hours, Days, Months, Years;
	GetTimeAndDate(&Secs, &Mins, &Hours, &Days, &Months, &Years);

	fprintf(fh, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
	fprintf(fh, "<iso19115full:MD_Metadata xmlns:iso19115full=\"http://schemas.opengis.net/iso19115full\" xmlns:iso19119=\"http://schemas.opengis.net/iso19119\" xmlns:csw=\"http://www.opengis.net/cat/csw\" xmlns:smXML=\"http://metadata.dgiwg.org/smXML\" xmlns:ows=\"http://www.opengis.net/ows\" xmlns:dc=\"http://www.purl.org/dc/elements/1.1/\" xmlns:terms=\"http://www.purl.org/dc/terms/\" xmlns:gml=\"http://www.opengis.net/gml\" xmlns:xlink=\"http://www.w3.org/1999/xlink\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\">\n");
	fprintf(fh, "    <iso19115full:language>\n");
	fprintf(fh, "        <smXML:CharacterString>de</smXML:CharacterString>\n");
	fprintf(fh, "    </iso19115full:language>\n");
	fprintf(fh, "    <iso19115full:characterSet>\n");
	fprintf(fh, "        <smXML:MD_CharacterSetCode codeList=\"http://metadata.dgiwg.org/codelistRegistry?MD_CharacterSetCode\" codeListValue=\"utf8\"/>\n");
	fprintf(fh, "    </iso19115full:characterSet>\n");
	fprintf(fh, "    <iso19115full:dateStamp>\n");
	fprintf(fh, "        <smXML:Date>%04ld-%02ld-%02ld</smXML:Date>\n", Years, Months, Days);
	fprintf(fh, "    </iso19115full:dateStamp>\n");
	fprintf(fh, "    <iso19115full:identificationInfo>\n");
	fprintf(fh, "        <smXML:MD_DataIdentification>\n");
	fprintf(fh, "            <smXML:citation>\n");
	fprintf(fh, "                <smXML:CI_Citation>\n");
	fprintf(fh, "                    <smXML:title>\n");
	fprintf(fh, "                        <smXML:CharacterString>FLYS-2 WSPLGEN</smXML:CharacterString>\n");
	fprintf(fh, "                    </smXML:title>\n");
	fprintf(fh, "                    <smXML:date>\n");
	fprintf(fh, "                        <smXML:CI_Date>\n");
	fprintf(fh, "                            <smXML:date>\n");
	fprintf(fh, "                                <smXML:Date>%04ld-%02ld-%02ld</smXML:Date>\n", Years, Months, Days);
	fprintf(fh, "                            </smXML:date>\n");
	fprintf(fh, "                            <smXML:dateType>\n");
	fprintf(fh, "                                <smXML:CI_DateTypeCode codeList=\"http://metadata.dgiwg.org/codelistRegistry?CI_DateTypeCode\" codeListValue=\"creation\"/>\n");
	fprintf(fh, "                            </smXML:dateType>\n");
	fprintf(fh, "                        </smXML:CI_Date>\n");
	fprintf(fh, "                    </smXML:date>\n");
	fprintf(fh, "                </smXML:CI_Citation>\n");
	fprintf(fh, "            </smXML:citation>\n");
	fprintf(fh, "            <smXML:abstract>\n");
	fprintf(fh, "                <smXML:CharacterString>%s</smXML:CharacterString>\n", ToUTF8("Linien gleicher Differenzen zwischen Wasserstand und Gelndehhen"));
	fprintf(fh, "            </smXML:abstract>\n");
	fprintf(fh, "            <smXML:language>\n");
	fprintf(fh, "                <smXML:CharacterString>de</smXML:CharacterString>\n");
	fprintf(fh, "            </smXML:language>\n");
	fprintf(fh, "        </smXML:MD_DataIdentification>\n");
	fprintf(fh, "    </iso19115full:identificationInfo>\n");
	fprintf(fh, "    <iso19115full:dataQualityInfo>\n");
	fprintf(fh, "        <smXML:DQ_DataQuality>\n");
	fprintf(fh, "            <smXML:LI_Lineage>\n");
	fprintf(fh, "                <smXML:statement>\n");
	fprintf(fh, "                    <smXML:CharacterString>\n");

	char line[1000];

	if (Parameter->IsSetPar)	sprintf(line, "Parameterdatei(\"%s\")", Parameter->FileNamePar.c_str());
	else						line[0] = '\0';
	strcpy(line, ToUTF8(line));
	fprintf(fh, "                        %s\n", line);

	if (Parameter->IsSetVonKm && Parameter->IsSetBisKm)	sprintf(line, "GewsserStrecke(Von(%.2f) Bis(%.2f))", Parameter->VonKm, Parameter->BisKm);
	else if (Parameter->IsSetVonKm)						sprintf(line, "GewsserStrecke(Von(%.2f) Bis(-))", Parameter->VonKm);
	else if (Parameter->IsSetBisKm)						sprintf(line, "GewsserStrecke(Von(-) Bis(%.2f))", Parameter->BisKm);
	else 												line[0] = '\0';
	strcpy(line, ToUTF8(line));
	fprintf(fh, "                        %s\n", line);

	sprintf(line, "Delta(Von(%.2f) Bis(%.2f) Diff(%.2f))", Parameter->Von / 100.0, Parameter->Bis / 100.0, Parameter->Diff / 100.0);
	strcpy(line, ToUTF8(line));
	fprintf(fh, "                        %s\n", line);

	if (Parameter->IsSetLin) 	sprintf(line, "Szenario");
	else if (Parameter->Sperre) sprintf(line, "Aktuelles-berschwemmungsgebiet");
	else 						sprintf(line, "Potentielles-berschwemmungsgebiet");
	strcpy(line, ToUTF8(line));
	fprintf(fh, "                        %s\n", line);

	sprintf(line, "DigitalesGelndeModell(\"%s\")", Parameter->FileNameDgm.c_str());
	strcpy(line, ToUTF8(line));
	fprintf(fh, "                        %s\n", line);

	sprintf(line, "Gewsserquerprofilspuren(\"%s\")", Parameter->FileNamePro.c_str());
	strcpy(line, ToUTF8(line));
	fprintf(fh, "                        %s\n", line);

	if (Parameter->IsSetWspTag)	sprintf(line, "WspTag(\"%s\")", Parameter->WspTag.c_str());
	else						line[0] = '\0';
	strcpy(line, ToUTF8(line));
	fprintf(fh, "                        %s\n", line);

	sprintf(line, "Wasserstnde(\"%s\")", Parameter->FileNameWsp.c_str());
	strcpy(line, ToUTF8(line));
	fprintf(fh, "                        %s\n", line);

	for (TStringList::iterator i = Parameter->FileNamesLin.begin(); i != Parameter->FileNamesLin.end(); i++)
	{
		std::string FileNameLin = *i;

		if (Parameter->IsSetLin)	sprintf(line, "Linienobjekte(\"%s\")", FileNameLin.c_str());
		else						line[0] = '\0';
		strcpy(line, ToUTF8(line));
		fprintf(fh, "                        %s\n", line);
	}

	if (Parameter->IsSetAchse)	sprintf(line, "Gewsserachse(\"%s\")", Parameter->FileNameAchse.c_str());
	else						line[0] = '\0';
	strcpy(line, ToUTF8(line));
	fprintf(fh, "                        %s\n", line);


	fprintf(fh, "                    </smXML:CharacterString>\n");
	fprintf(fh, "                </smXML:statement>\n");
	fprintf(fh, "            </smXML:LI_Lineage>\n");
	fprintf(fh, "        </smXML:DQ_DataQuality>\n");
	fprintf(fh, "    </iso19115full:dataQualityInfo>\n");
	fprintf(fh, "</iso19115full:MD_Metadata>\n");

	fclose (fh);

	write_fortschritt("<-Speichern der Meta-Daten beendet\n");

	return (true);
}

