//
// $Id: parameter.cpp 101 2008-07-10 09:20:12Z mrchip $
//
// Copyright (C) 2005 STADT-LAND-FLUSS INGENIEURDIENSTE GmbH
//
// Authors:
// Ulrich Kiel <u.kiel@S-L-F.de>
//
// This program is free software under the GPL (>=v2)
// Read the file COPYING coming with WSPLGEN for details.
//

//---------------------------------------------------------------------------
#include <string.h>
#include "tools.h"
#include "file.h"
#include "parameter.h"

//---------------------------------------------------------------------------
TParameter::TParameter(int argc, char **argv)
{
	write_fortschritt("->Auswertung der Kommandozeilen-Parameter gestartet\n");

	IsSetDebug = false;
	DebugLevel = 0;

	IsSetPar = false;
	FileNamePar = "";

	IsSetDelta = false;
	IsSetVon = false;
	Von = 0;
	IsSetBis = false;
	Bis = 0;
	IsSetDiff = false;
	Diff = 0;

	IsSetGel = false;
	Sperre = true;

	IsSetDgm = false;
	FileNameDgm = "";

	IsSetPro = false;
	FileNamePro = "";

	IsSetWsp = false;
	FileNameWsp = "";

	IsSetWspTag = false;
	WspTag = "";

	IsSetLin = false;


	IsSetAchse = false;
	FileNameAchse = "";

	IsSetGebiet = false;
	FileNameGebiet = "";

	IsSetStrecke = false;
	IsSetVonKm = false;
	IsSetBisKm = false;
	VonKm = -9999.9;
	BisKm = +9999.9;

	IsSetDist = false;
	ProfilDistance = 999999;

	IsSetSaveTri = false;
	IsSetSave2DM = false;
	IsSetSaveSol = false;

	IsSetAusgabe = false;
	FileNameAusgabe = "WSPLGEN.SHP";

	if (argc == 0 || argv == 0)
	{
		write_error(1201, "Es wurden keine Parameter bergeben.\nWenigstens die Parameter -DGM=<DGM-Datei>, -PRO=<Profil-Datei>, -ACHSE=<Achse-Datei> und -WSP=<Wasserstands-Datei>\nmssen angegeben werden.\n");
	}

	std::vector<std::string>	ParList;

	int AnzPar = 0;
	for (int i=1, j=0; i<argc || j<AnzPar; i++)
	{
		std::string Parameter = "";

		if (i < argc)
		{
			Parameter = argv[i];
		}
		else
		{
			Parameter = ParList[j];
			j++;
		}

		Parameter = Trim(Parameter);

		if (Parameter.size() == 0)
		{
			// Leerzeilen werden berlesen
		}
		else if (Parameter.substr(0, 1) == "'")
		{
			write_fortschritt("%s\n", Parameter.c_str()+1);
		}
		else if (Parameter.substr(0, 1) == "#" || Parameter.substr(0, 1) == "/" || Parameter.substr(0, 1) == ";")
		{
			// Echter Kommentar
        }
		else if (Parameter.substr(0, 7) == "-OUTPUT" || Parameter.substr(0, 7) == "-output")
		{
			///////////////////////////////////////////////
			// DEBUG
			///////////////////////////////////////////////
			write_fortschritt("Parameter -OUTPUT erkannt.\n");

			if (IsSetDebug)
			{
				write_warning(1106, "Der Parameter -OUTPUT wurde mehrfach angegeben.\nEs wird der hchste Outputlevel verwendet\n");
			}

			IsSetDebug = true;

			if (Parameter.substr(7, 1) != "=")
			{
				write_warning(1105, "Es wurden keine Argumente fr -OUTPUT bergeben.\nOutputlevel wird 0 (Kaum Informationen).\n");
				DebugLevel = 0;
                continue;
			}

			int NewDebugLevel;
			if (1 == sscanf(Parameter.substr(8).c_str(), "%d", &NewDebugLevel))
			{
				if (NewDebugLevel > DebugLevel) DebugLevel = NewDebugLevel;
			}

			write_fortschritt("OutputLevel: '%d'\n", DebugLevel);

			if (DebugLevel == 0) write_fortschritt("Es werden keine zustzlichen Informationen ausgegeben\n");
			if (DebugLevel >= 1) write_fortschritt("Es werden kurze Zwischeninformationen und mehr Fortschrittanzeigen ausgegeben\n");
			if (DebugLevel >= 2) write_fortschritt("Es werden zustzlich mehr Zwischeninformationen ausgegeben\n");
			if (DebugLevel >= 3) write_fortschritt("Es werden zustzlich deutlich mehr Zwischeninformationen ausgegeben\n");
			if (DebugLevel >= 4) write_fortschritt("Es werden zustzlich mehr Fortschrittanzeigen und die wichtigsten Testausgabedateien ausgegeben\n");
			if (DebugLevel >= 5) write_fortschritt("Es werden zustzlich alle wichtigen Testausgabedateien ausgegeben\n");
			if (DebugLevel >= 6) write_fortschritt("Es werden zustzlich wichtigen Testausgabedateien ausgegeben\n");
			if (DebugLevel >= 7) write_fortschritt("Es werden zustzlich alle Informationen ausgegeben\n");
			if (DebugLevel >= 8) write_fortschritt("Es werden zustzlich alle Tests durchgefhrt\n");
			if (DebugLevel == 9) write_fortschritt("Es werden alle Informationen und Testausgabedateien erzeugt\n");
		}
		else if (Parameter.substr(0, 4) == "-PAR" || Parameter.substr(0, 4) == "-par")
		{
			///////////////////////////////////////////////
			// PAR
			///////////////////////////////////////////////
			if (IsSetDebug) write_fortschritt("Parameter -PAR erkannt.\n");

			if (IsSetPar)
			{
				write_warning(1106, "Der Parameter -PAR wurde mehrfach angegeben.\n");
				continue;
			}

			if (Parameter.substr(4, 1) != "=")
			{
				write_error(1207, "Es wurden kein Dateiname fr -PAR bergeben.\nDas Format lautet korrekt -PAR=<Dateiname>.\n");
			}

			FileNamePar = Parameter.substr(5);

			if (FileNamePar[0] == '"') 						FileNamePar = FileNamePar.substr(1);
			if (FileNamePar[FileNamePar.length()-1] == '"') FileNamePar = FileNamePar.substr(0, FileNamePar.length()-1);

			IsSetPar = true;

			write_fortschritt("Parameterdatei: '%s'\n", FileNamePar.c_str());

			LoadPar(FileNamePar, &ParList);
			AnzPar = ParList.size();
		}
		else if (Parameter.substr(0, 8) == "-STRECKE" || Parameter.substr(0, 8) == "-strecke")
		{
			///////////////////////////////////////////////
			// STRECKE
			///////////////////////////////////////////////
			if (IsSetDebug) write_fortschritt("Parameter -STRECKE erkannt.\n");

			if (IsSetStrecke)
			{
				write_warning(1106, "Der Parameter -STRECKE wurde mehrfach angegeben.\n");
				continue;
			}

			if (Parameter.substr(8, 1) != "=")
			{
				write_warning(1226, "Es wurden keine Argumente fr -STRECKE bergeben.\nDas Format lautet korrekt -STRECKE=<Von>,<Bis>.\n");
			}

			double D = 0.0;
			if (1 == sscanf(Parameter.substr(9).c_str(), "%lf", &D))
			{
				IsSetVonKm = true;
				VonKm = D;
			}

			unsigned int FirstKommaPos = Parameter.find(",");

			if (FirstKommaPos < Parameter.size())
			{
				if (1 == sscanf(Parameter.substr(FirstKommaPos + 1).c_str(), "%lf", &D))
				{
					IsSetBisKm = true;
					BisKm = D;
				}
			}

			IsSetStrecke = true;

			if (IsSetVonKm && IsSetBisKm)	write_fortschritt("Strecke: Von = %.3f km, Bis = %.3f km\n", VonKm, BisKm);
			else if (IsSetVonKm)			write_fortschritt("Strecke: Von = %.3f km, Bis = <default>\n", VonKm);
			else if (IsSetBisKm)			write_fortschritt("Strecke: Von = <default>, Bis = %.3f km\n", BisKm);
			else							write_fortschritt("Strecke: Von = <default>, Bis = <default>\n");
		}
		else if (Parameter.substr(0, 5) == "-DIST" || Parameter.substr(0, 5) == "-dist")
		{
			///////////////////////////////////////////////
			// DIST
			///////////////////////////////////////////////
			if (IsSetDebug) write_fortschritt("Parameter -DIST erkannt.\n");

			if (IsSetDist)
			{
				write_warning(1106, "Der Parameter -DIST wurde mehrfach angegeben.\n");
				continue;
			}

			if (Parameter.substr(5, 1) != "=")
			{
				write_warning(1226, "Es wurden keine Argumente fr -DIST bergeben.\nDas Format lautet korrekt -DIST=<Abstand>.\n");
			}

			double D = 0.0;
			if (1 != sscanf(Parameter.substr(6).c_str(), "%lf", &D))
			{
				write_error(1229, "Das Argumente fr -DIST konnte nicht verarbeitet werden.\nDas Format lautet korrekt -DIST=<Abstand>.\n");
			}

			IsSetDist = true;
			ProfilDistance = D * 10000 * 100; // Umrechnung von km in Meter und weiter in cm

			write_fortschritt("Gewnschter Abstand der Profile: %.3f km\n", D);
		}
		else if (Parameter.substr(0, 8) == "-SAVETRI" || Parameter.substr(0, 8) == "-savetri")
		{
			///////////////////////////////////////////////
			// SAVETRI
			///////////////////////////////////////////////
			if (IsSetDebug) write_fortschritt("Parameter -SAVETRI erkannt.\n");

			if (IsSetSaveTri)
			{
				write_warning(1106, "Der Parameter -SAVETRI wurde mehrfach angegeben.\n");
				continue;
			}

			IsSetSaveTri = true;

			write_fortschritt("Triangulation wird gespeichert\n");
		}
		else if (Parameter.substr(0, 8) == "-SAVE2DM" || Parameter.substr(0, 8) == "-save2dm")
		{
			///////////////////////////////////////////////
			// SAVE2DM
			///////////////////////////////////////////////
			if (IsSetDebug) write_fortschritt("Parameter -SAVE2DM erkannt.\n");

			if (IsSetSave2DM)
			{
				write_warning(1106, "Der Parameter -SAVE2DM wurde mehrfach angegeben.\n");
				continue;
			}

			IsSetSave2DM = true;

			write_fortschritt("2DM-Netz wird gespeichert\n");
		}
		else if (Parameter.substr(0, 8) == "-SAVESOL" || Parameter.substr(0, 8) == "-savesol")
		{
			///////////////////////////////////////////////
			// SAVESOL
			///////////////////////////////////////////////
			if (IsSetDebug) write_fortschritt("Parameter -SAVESOL erkannt.\n");

			if (IsSetSaveSol)
			{
				write_warning(1106, "Der Parameter -SAVESOL wurde mehrfach angegeben.\n");
				continue;
			}

			IsSetSaveSol = true;

			write_fortschritt("Wasserstnde werden als SOL-Datei gespeichert\n");
		}
		else if (Parameter.substr(0, 6) == "-DELTA" || Parameter.substr(0, 6) == "-delta")
		{
			///////////////////////////////////////////////
			// DELTA
			///////////////////////////////////////////////
			if (IsSetDebug) write_fortschritt("Parameter -DELTA erkannt.\n");

			if (IsSetDelta)
			{
				write_warning(1106, "Der Parameter -DELTA wurde mehrfach angegeben.\n");
				continue;
			}

			if (Parameter.substr(6, 1) != "=")
			{
				write_warning(1108, "Es wurden keine Argumente fr -DELTA bergeben.\nDas Format lautet korrekt -DELTA=<Von>,<Bis>,<Diff>.\n");
                continue;
			}

			unsigned int FirstKommaPos = Parameter.find(",");

			if (FirstKommaPos >= Parameter.size())
			{
				write_warning(1108, "Es wurden nicht genug Argumente fr den Parameter -DELTA bergeben.\nDas Format lautet korrekt -DELTA=<Von>,<Bis>,<Diff>.\n");
                continue;
			}

			unsigned int SecondKommaPos = Parameter.find(",", FirstKommaPos + 1);

			if (SecondKommaPos >= Parameter.size())
			{
				write_error(1209, "Es wurden nicht genau 2 Kommata beim Parameter -DELTA verwendet.\nDas Format lautet korrekt -DELTA=<Von>,<Bis>,<Diff>.\n");
			}

			double D = 0.0;
			if (1 == sscanf(Parameter.substr(7).c_str(), "%lf", &D))
			{
				IsSetVon = true;
				Von = (int)(D * 100.0 + 0.5);
				if (D < 0.0) Von = (int)(D * 100.0 - 0.5);
			}

			if (1 == sscanf(Parameter.substr(FirstKommaPos + 1).c_str(), "%lf", &D))
			{
				IsSetBis = true;
				Bis = (int)(D * 100.0 + 0.5);
				if (D < 0.0) Bis = (int)(D * 100.0 - 0.5);
			}

			if (1 == sscanf(Parameter.substr(SecondKommaPos + 1).c_str(), "%lf", &D))
			{
				IsSetDiff = true;
				Diff = (int)(D * 100.0 + 0.5);
			}

			IsSetDelta = true;

			if (IsSetVon && IsSetBis && IsSetDiff)	write_fortschritt("Delta: Von = %.2f Bis = %.2f Diff = %.2f\n", Von / 100.0, Bis / 100.0, Diff / 100.0);
			else if (IsSetVon && IsSetBis)			write_fortschritt("Delta: Von = %.2f Bis = %.2f Diff = <default>\n", Von / 100.0, Bis / 100.0);
			else if (IsSetVon && IsSetDiff)			write_fortschritt("Delta: Von = %.2f Bis = <default> Diff = %.2f\n", Von / 100.0, Diff / 100.0);
			else if (IsSetBis && IsSetBis)			write_fortschritt("Delta: Von = <default> Bis = %.2f Diff = %.2f\n", Bis / 100.0, Diff / 100.0);
			else if (IsSetVon)						write_fortschritt("Delta: Von = %.2f Bis = <default> Diff = <default>\n", Von / 100.0);
			else if (IsSetBis)						write_fortschritt("Delta: Von = <default> Bis = %.2f Diff = <default>\n", Bis / 100.0);
			else if (IsSetDiff)						write_fortschritt("Delta: Von = <default> Bis = <default> Diff = %.2f\n", Diff / 100.0);
			else									write_fortschritt("Delta: Von = <default> Bis = <default> Diff = <default>\n");
		}
		else if (Parameter.substr(0, 4) == "-GEL" || Parameter.substr(0, 4) == "-gel")
		{
			///////////////////////////////////////////////
			// GEL
			///////////////////////////////////////////////
			if (IsSetDebug) write_fortschritt("Parameter -GEL erkannt.\n");

			if (IsSetGel)
			{
				write_warning(1106, "Der Parameter -GEL wurde mehrfach angegeben.\n");
				continue;
			}

			IsSetGel = true;

			if (Parameter.substr(4, 1) != "=")
			{
				write_warning(1105, "Es wurden keine Argumente fr -GEL bergeben.\nDas Format lautet korrekt -GEL=<SPERRE|NOSPERRE>.\nGelnde wirkt als Sperre.\n");
				Sperre = true;
                continue;
			}

			if (Parameter.substr(5, 6) == "SPERRE" || Parameter.substr(5, 6) == "sperre")
			{
				Sperre = true;
			}
			else if (Parameter.substr(5, 8) == "NOSPERRE" || Parameter.substr(5, 8) == "nosperre")
			{
				Sperre = false;
			}
			else
			{
				write_error(1214, "Es wurden weder SPERRE noch NOSPERRE mit -GEL verwendet.\nDas Format lautet korrekt -GEL=<SPERRE|NOSPERRE>.\n");
			}

			if (Sperre)	write_fortschritt("Gelnde wirkt als Sperre\n");
			else		write_fortschritt("Gelnde wirkt nicht als Sperre\n");
		}
		else if (Parameter.substr(0, 4) == "-DGM" || Parameter.substr(0, 4) == "-dgm")
		{
			if (IsSetDebug) write_fortschritt("Parameter -DGM erkannt.\n");

			if (IsSetDgm)
			{
				write_warning(1106, "Der Parameter -DGM wurde mehrfach angegeben.\n");
				continue;
			}

			if (Parameter.substr(4, 1) != "=")
			{
				write_error(1207, "Es wurden kein Dateiname fr -DGM bergeben.\nDas Format lautet korrekt -DGM=<Dateiname>.\n");
			}

			FileNameDgm = Parameter.substr(5);

			if (FileNameDgm[0] == '"') 						FileNameDgm = FileNameDgm.substr(1);
			if (FileNameDgm[FileNameDgm.length()-1] == '"') FileNameDgm = FileNameDgm.substr(0, FileNameDgm.length()-1);

			IsSetDgm = true;

			write_fortschritt("DGM-Datei: '%s'\n", FileNameDgm.c_str());
		}
		else if (Parameter.substr(0, 4) == "-PRO" || Parameter.substr(0, 4) == "-pro")
		{
			if (IsSetDebug) write_fortschritt("Parameter -PRO erkannt.\n");

			if (IsSetPro)
			{
				write_warning(1106, "Der Parameter -PRO wurde mehrfach angegeben.\n");
				continue;
			}

			if (Parameter.substr(4, 1) != "=")
			{
				write_error(1207, "Es wurden kein Dateiname fr -PRO bergeben.\nDas Format lautet korrekt -PRO=<Dateiname>.\n");
			}

			FileNamePro = Parameter.substr(5);

			if (FileNamePro[0] == '"') 						FileNamePro = FileNamePro.substr(1);
			if (FileNamePro[FileNamePro.length()-1] == '"') FileNamePro = FileNamePro.substr(0, FileNamePro.length()-1);

			IsSetPro = true;

			write_fortschritt("Profillagen-Datei: '%s'\n", FileNamePro.c_str());
		}
		else if (Parameter.substr(0, 7) == "-WSPTAG" || Parameter.substr(0, 7) == "-wsptag")
		{
			if (IsSetDebug) write_fortschritt("Parameter -WSPTAG erkannt.\n");

			if (IsSetWspTag)
			{
				write_warning(1106, "Der Parameter -WSPTAG wurde mehrfach angegeben.\n");
				continue;
			}

			if (Parameter.substr(7, 1) != "=")
			{
				write_error(1207, "Es wurden keine Zeichenkette fr -WSPTAG bergeben.\nDas Format lautet korrekt -WSPTAG=<Bezeichnung>.\n");
			}

			WspTag = Parameter.substr(8);

			if (WspTag[0] == '"') 	WspTag = WspTag.substr(1);
			if (WspTag[WspTag.length()-1] == '"') WspTag = WspTag.substr(0, WspTag.length()-1);

			IsSetWspTag = true;

			write_fortschritt("WSP-Bezeichner: '%s'\n", WspTag.c_str());

			continue;
		}
		else if (Parameter.substr(0, 4) == "-WSP" || Parameter.substr(0, 4) == "-wsp")
		{
			if (IsSetDebug) write_fortschritt("Parameter -WSP erkannt.\n");

			if (IsSetWsp)
			{
				write_warning(1106, "Der Parameter -WSP wurde mehrfach angegeben.\n");
				continue;
			}

			if (Parameter.substr(4, 1) != "=")
			{
				write_error(1207, "Es wurden kein Dateiname fr -WSP bergeben.\nDas Format lautet korrekt -WSP=<Dateiname>.\n");
			}

			FileNameWsp = Parameter.substr(5);

			if (FileNameWsp[0] == '"') 						FileNameWsp = FileNameWsp.substr(1);
			if (FileNameWsp[FileNameWsp.length()-1] == '"') FileNameWsp = FileNameWsp.substr(0, FileNameWsp.length()-1);

			IsSetWsp = true;

			write_fortschritt("WSP-Datei: '%s'\n", FileNameWsp.c_str());

			continue;
		}
		else if (Parameter.substr(0, 4) == "-LIN" || Parameter.substr(0, 4) == "-LIN")
		{
			if (IsSetDebug) write_fortschritt("Parameter -LIN erkannt.\n");

			if (Parameter.substr(4, 1) != "=")
			{
				write_error(1207, "Es wurden kein Dateiname fr -LIN bergeben.\nDas Format lautet korrekt -LIN=<Dateiname>.\n");
			}

			std::string FileNameLin = Parameter.substr(5);

			if (FileNameLin[0] == '"') 						FileNameLin = FileNameLin.substr(1);
			if (FileNameLin[FileNameLin.length()-1] == '"') FileNameLin = FileNameLin.substr(0, FileNameLin.length()-1);

			FileNamesLin.push_back(FileNameLin);

			IsSetLin = true;

			write_fortschritt("Bruch- und Sperr-Datei: '%s'\n", FileNameLin.c_str());
		}
		else if (Parameter.substr(0, 6) == "-ACHSE" || Parameter.substr(0, 6) == "-achse")
		{
			if (IsSetDebug) write_fortschritt("Parameter -ACHSE erkannt.\n");

			if (IsSetAchse)
			{
				write_warning(1106, "Der Parameter -ACHSE wurde mehrfach angegeben.\n");
				continue;
			}

			if (Parameter.substr(6, 1) != "=")
			{
				write_error(1207, "Es wurden kein Dateiname fr -ACHSE bergeben.\nDas Format lautet korrekt -ACHSE=<Dateiname>.\n");
			}

			FileNameAchse = Parameter.substr(7);

			if (FileNameAchse[0] == '"') 						FileNameAchse = FileNameAchse.substr(1);
			if (FileNameAchse[FileNameAchse.length()-1] == '"') FileNameAchse = FileNameAchse.substr(0, FileNameAchse.length()-1);

			IsSetAchse = true;

			write_fortschritt("Gewsserachse: '%s'\n", FileNameAchse.c_str());

			continue;
		}
		else if (Parameter.substr(0, 7) == "-GEBIET" || Parameter.substr(0, 7) == "-gebiet")
		{
			if (IsSetDebug) write_fortschritt("Parameter -GEBIET erkannt.\n");

			if (IsSetGebiet)
			{
				write_warning(1106, "Der Parameter -GEBIET wurde mehrfach angegeben.\n");
				continue;
			}

			if (Parameter.substr(7, 1) != "=")
			{
				write_error(1207, "Es wurden kein Dateiname fr -GEBIET bergeben.\nDas Format lautet korrekt -GEBIET=<Dateiname>.\n");
			}

			FileNameGebiet = Parameter.substr(8);

			if (FileNameGebiet[0] == '"') 						 	FileNameGebiet = FileNameGebiet.substr(1);
			if (FileNameGebiet[FileNameGebiet.length()-1] == '"') 	FileNameGebiet = FileNameGebiet.substr(0, FileNameGebiet.length()-1);

			IsSetGebiet = true;

			write_fortschritt("Hydraulische Grenze: '%s'\n", FileNameGebiet.c_str());

			continue;
		}
		else if (Parameter.substr(0, 8) == "-AUSGABE" || Parameter.substr(0, 8) == "-ausgabe")
		{
			if (IsSetDebug) write_fortschritt("Parameter -AUSGABE erkannt.\n");

			if (IsSetAusgabe)
			{
				write_warning(1106, "Der Parameter -AUSGABE wurde mehrfach angegeben.\n");
				continue;
			}

			if (Parameter.substr(8, 1) != "=")
			{
				write_error(1207, "Es wurden kein Dateiname fr -AUSGABE bergeben.\nDas Format lautet korrekt -AUSGABE=<Dateiname>.\n");
			}

			FileNameAusgabe = Parameter.substr(9);

			if (FileNameAusgabe[0] == '"') 							FileNameAusgabe = FileNameAusgabe.substr(1);
			if (FileNameAusgabe[FileNameAusgabe.length()-1] == '"') FileNameAusgabe = FileNameAusgabe.substr(0, FileNameAusgabe.length()-1);

			IsSetAusgabe = true;

			write_fortschritt("Ausgabe-Datei: '%s'\n", FileNameAusgabe.c_str());
		}
		else if (Parameter.substr(0, 1) == "$")
		{
			// Geheimer Parameter zum Umleiten der Ausgaben
			// Siehe WSPLGEN - main am Anfang !!!!
		}
		else
		{
			write_warning(1124, "Eine unbekannte Zeichenkette '%s' wurde als Parameter bergeben.\nDie Zeichenkette wird ignoriert.\n", Parameter.c_str());
		}
	}

	if (false == IsSetDgm)
	{
		write_error(1202, "Der Parameter -DGM wurde nicht angegeben.\nWenigstens die Parameter\n-DGM=<DGM-Datei>, -PRO=<Profil-Datei>, -ACHSE=<Achsen-Datei> und -WSP=<Wasserstands-Datei>\nmssen angegeben werden.\n");
	}

	if (false == IsSetPro)
	{
		write_error(1203, "Der Parameter -PRO wurde nicht angegeben.\nWenigstens die Parameter\n-DGM=<DGM-Datei>, -PRO=<Profil-Datei>, -ACHSE=<Achsen-Datei> und -WSP=<Wasserstands-Datei>\nmssen angegeben werden.\n");
	}

	if (false == IsSetWsp)
	{
		write_error(1204, "Der Parameter -WSP wurde nicht angegeben.\nWenigstens die Parameter\n-DGM=<DGM-Datei>, -PRO=<Profil-Datei>, -ACHSE=<Achsen-Datei> und -WSP=<Wasserstands-Datei>\nmssen angegeben werden.\n");
	}

	if (false == IsSetAchse)
	{
		write_error(1230, "Der Parameter -ACHSE wurde nicht angegeben.\nWenigstens die Parameter\n-DGM=<DGM-Datei>, -PRO=<Profil-Datei>, -ACHSE=<Achsen-Datei> und -WSP=<Wasserstands-Datei>\nmssen angegeben werden.\n");
	}

	if (true == IsSetLin && false == Sperre)
	{
		write_warning(1115, "-GEL=NOSPERRE wurde zusammen mit -LIN=<Dateiname> verwendet.\nDies liefert keine sinnvollen Ergebnisse.\nEs wird daher -GEL=SPERRE angenommen.\n");
		Sperre = true;
	}

	if (IsSetVon && IsSetBis && Von > Bis)
	{
		write_error(1210, "Das Argument <Von> ist grer als das Argument <Bis>.\nDie Reihenfolge der Argumente wurde wahrscheinlich nicht beachtet.\nZuerst kommt <Von>, dann <Bis>, und dann <Diff>\n");
	}

	if (IsSetDiff && Diff < 0)
	{
		write_error(1211, "Das Argument <Diff> ist kleiner als Null.\n");
	}

	if (IsSetVon && IsSetBis && IsSetDiff && Bis - Von > 0.0 && Diff <= 0.0)
	{
		write_error(1212, "Das Argument <Diff> darf nur dann Null sein, wenn <Von> und <Bis> gleich sind.\n");
	}

	if (IsSetVon && IsSetBis && IsSetDiff && Bis - Von < Diff)
	{
		write_warning(1113, "Das Argument <Diff> ist grer als die Differenz zwischen <Von> und <Bis>.\nDies ist nicht sinnvoll.\nEs wird nur eine Abstufung mit dem Wert <Von> generiert.\n");
	}

	if (false == IsSetAusgabe)
	{
		write_warning(1117, "Der Parameter -AUSGABE wurde nicht angegeben.\nEs wird als Ausgabedateiname WSPLGEN.SHP benutzt.\n");
	}

	if (false == IsSetGel)
	{
		write_warning(1118, "Der Parameter -GEL wurde nicht angegeben.\nEs wird -GEL=SPERRE angenommen.\n");
	}

	if (false == IsSetVon)
	{
		write_warning(1119, "Der Parameter -DELTA oder sein Argument <Von> wurden nicht angegeben.\nEs wird <Von> als Null angenommen.\n");
	}

	if (false == IsSetBis)
	{
		write_warning(1120, "Der Parameter -DELTA oder sein Argument <Bis> wurden nicht angegeben.\nEs wird <Bis> als der maximale Wasserstand angenommen.\n");
	}

	if (false == IsSetDiff)
	{
		write_warning(1121, "Der Parameter -DELTA oder sein Argument <Diff> wurden nicht angegeben.\nEs wird <Diff> als (<Bis> minus <Von>) geteilt durch 10 angenommen.\n");
	}

	write_fortschritt("<-Auswertung der Kommandozeilen-Parameter beendet\n");
}


