//
// $Id: qhull.cpp 101 2008-07-10 09:20:12Z mrchip $
//
// Copyright (C) 2005 STADT-LAND-FLUSS INGENIEURDIENSTE GmbH
//
// Authors:
// Ulrich Kiel <u.kiel@S-L-F.de>
//
// This program is free software under the GPL (>=v2)
// Read the file COPYING coming with WSPLGEN for details.
//

#ifdef __BORLANDC__
#pragma hdrstop
#endif

//---------------------------------------------------------------------
#include "tools.h"
#include "qhull.h"

//---------------------------------------------------------------------
double DistanceBetween(TXYZ* SXy, TXYZ* EXy, TXYZ* Xy)
{
	double X1 = SXy->X;
	double Y1 = SXy->Y;
	double X2 = EXy->X;
	double Y2 = EXy->Y;
	double Xp = Xy->X;
	double Yp = Xy->Y;

	if (fabs(X2 - X1) < 0.0001 && fabs(Y2 - Y1) < 0.0001)
	{
		write_error(3221, "Die Punkte liegen zu dicht zusammen");
	}
	else if (fabs(X2 - X1) < 0.0001)
	{
		double D = Xp - X1;
		if (Y2 > Y1)	D = -D;
		return (D);
	}
	else if (fabs(Y2 - Y1) < 0.0001)
	{
		double D = Yp - Y1;
		if (X2 < X1)	D = -D;
		return (D);
	}

	// Die Gerade waagerecht um den Punkt X1,Y1 drehen
	// und den Punkt Xp,Yp, mitnehmen

	double alpha = atan((Y2 - Y1) / (X2 - X1));

	if (X2 - X1 < 0.0)  alpha = alpha + M_PI;
	if (alpha < 0.0)    alpha = alpha + 2 * M_PI;
	if (alpha >= 2 * M_PI)   alpha = alpha - 2 * M_PI;

	alpha = -alpha;

	// Verschiebung damit X1,Y1 Drehurprung ist
	{
		X2 = X2 - X1;
		Y2 = Y2 - Y1;
		double Xd = X2 * cos(alpha) - Y2 * sin(alpha);
		double Yd = X2 * sin(alpha) + Y2 * cos(alpha);
		X2 = Xd + X1;
		Y2 = Yd + Y1;
	}

	{
		Xp = Xp - X1;
		Yp = Yp - Y1;
		double Xd = Xp * cos(alpha) - Yp * sin(alpha);
		double Yd = Xp * sin(alpha) + Yp * cos(alpha);
		Xp = Xd + X1;
		Yp = Yd + Y1;
	}

	double D = Yp - Y1;

	return (D);
}

//---------------------------------------------------------------------
TXYZ* FindPointBetween(TXYZList *XYList, TXYZ* SXy, TXYZ* EXy)
{
	double MaxD = 0.0001;
	TXYZ* MaxXy = 0;

	for (TXYZList::iterator i = XYList->begin(); i != XYList->end(); i++)
	{
		TXYZ* Xy = *i;

		double D = DistanceBetween(SXy, EXy, Xy);

		if (D > MaxD)
		{
			MaxD = D;
			MaxXy = Xy;
		}
	}

	return (MaxXy);
}

//---------------------------------------------------------------------
void ExpandQHull(TXYZList *XYList, TXYZList *QHull, TXYZ* SXy, TXYZ* EXy)
{
	TXYZ* Xy = FindPointBetween(XYList, SXy, EXy);

	if (Xy == 0)	return;

	ExpandQHull(XYList, QHull, SXy, Xy);
	QHull->Add(Xy);
	ExpandQHull(XYList, QHull, Xy, EXy);
}

//---------------------------------------------------------------------
void QHull(TXYZList *XYList, TXYZList* QHull, int DebugLevel)
{
	write_fortschritt("->Erzeugen des Begrenzungspolygons gestartet\n");

	if (XYList->size() < 3)
	{
		write_error(3222, "Es sind in der Punktliste nur %d Sttzstellen vorhanden.\nEs werden aber mindestens 3 Knoten bentigt\n", XYList->size());
	}

	// Den linkesten und rechtesten Punkt suchen
	// Die gehren beide zur konvexen Hlle

	TXYZ* LXy = 0;
	TXYZ* RXy = 0;
	for (TXYZList::iterator i = XYList->begin(); i != XYList->end(); i++)
	{
		TXYZ* Xy = *i;

		if (LXy == 0 || Xy->X < LXy->X)	LXy = Xy;
		if (RXy == 0 || Xy->X > RXy->X)	RXy = Xy;
	}

	QHull->Add(LXy);
    if (DebugLevel > 0)	write_fortschritt("Oberen Teil der Umhllung bilden.\n");
	ExpandQHull(XYList, QHull, LXy, RXy);
	QHull->Add(RXy);
    if (DebugLevel > 0)	write_fortschritt("Unteren Teil der Umhllung bilden.\n");
	ExpandQHull(XYList, QHull, RXy, LXy);

	write_fortschritt("<-Erzeugen des Begrenzungspolygons beendet\n");
}
