// $Id: shape.cpp 92 2007-09-02 20:13:20Z mrchip $
//
// Copyright (C) 2005 STADT-LAND-FLUSS INGENIEURDIENSTE GmbH
//
// Authors:
// Ulrich Kiel <u.kiel@S-L-F.de>
//
// Diese Datei beruht auf der Shapelib von Frank Warmerdam
//
// Es wurden die beiden Hauptdateien shpopen.c und dbfopen.c zusammengefgt und
// an die Strktur von WSPLGEN angepasst
//
// Einige 'assert' wuurden durch 'dump_error' Aufrufe ersetzt
//
// Die original Shapelib wurde als ZIP-Archive den Quellen beigefgt.
//
// Die beiden Funktionen SHPWritePoint und SHPWriteElement wurden der Einfachheit
// halber eingefhrt.
// 
// Hier der original Kopf der Datei shpopen.c

/******************************************************************************
 * $Id: shape.cpp 92 2007-09-02 20:13:20Z mrchip $
 *
 * Project:  Shapelib
 * Purpose:  Implementation of core Shapefile read/write functions.
 * Author:   Frank Warmerdam, warmerdam@pobox.com
 *
 ******************************************************************************
 * Copyright (c) 1999, 2001, Frank Warmerdam
 *
 * This software is available under the following "MIT Style" license,
 * or at the option of the licensee under the LGPL (see LICENSE.LGPL).  This
 * option is discussed in more detail in shapelib.html.
 *
 * --
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 ******************************************************************************/


#ifdef __BORLANDC__
#pragma hdrstop
#endif

//---------------------------------------------------------------------------

#include <stdio.h>
#include <math.h>

#include "shape.h"

#include "tools.h"

#define ByteCopy(a, b, c)	memcpy(b, a, c)
#define	TRUE				1
#define	FALSE				0

static int 	bBigEndian;

static FILE*	fpSHX = 0;
static FILE*	fpSHP = 0;

/************************************************************************/
/*                          SHPWriteHeader()                            */
/*                                                                      */
/*      Write out a header for the .shp and .shx files as well as the	*/
/*		contents of the index (.shx) file.								*/
/************************************************************************/
static void SHPWriteHeader(SHPHandle psSHP)
{
	uchar   abyHeader[100];
	int32	i32;
	double	dValue;
	int32	*panSHX;

	/* -------------------------------------------------------------------- */
	/*      Prepare header block for .shp file.                             */
	/* -------------------------------------------------------------------- */

	for (int i = 0; i < 100; i++)	abyHeader[i] = 0;

	/* magic cookie */
	abyHeader[2] = 0x27;
	abyHeader[3] = 0x0a;

	/* file size */
	i32 = psSHP->nFileSize / 2;
	ByteCopy(&i32, abyHeader+24, 4);
	if(!bBigEndian) Swap4Bytes(abyHeader+24);

	i32 = 1000;								/* version */
	ByteCopy( &i32, abyHeader+28, 4);
	if( bBigEndian ) Swap4Bytes(abyHeader+28);

	i32 = psSHP->nShapeType;				/* shape type */
	ByteCopy( &i32, abyHeader+32, 4);
	if( bBigEndian ) Swap4Bytes(abyHeader+32);

	dValue = psSHP->adBoundsMin[0];			/* set bounds */
	ByteCopy( &dValue, abyHeader+36, 8);
	if( bBigEndian ) Swap8Bytes(abyHeader+36);

	dValue = psSHP->adBoundsMin[1];
    ByteCopy( &dValue, abyHeader+44, 8);
    if( bBigEndian ) Swap8Bytes(abyHeader+44);

    dValue = psSHP->adBoundsMax[0];
    ByteCopy( &dValue, abyHeader+52, 8);
    if( bBigEndian ) Swap8Bytes(abyHeader+52);

    dValue = psSHP->adBoundsMax[1];
	ByteCopy( &dValue, abyHeader+60, 8);
    if( bBigEndian ) Swap8Bytes(abyHeader+60);

    dValue = psSHP->adBoundsMin[2];			/* z */
    ByteCopy( &dValue, abyHeader+68, 8);
    if( bBigEndian ) Swap8Bytes(abyHeader+68);

	dValue = psSHP->adBoundsMax[2];
    ByteCopy( &dValue, abyHeader+76, 8);
    if( bBigEndian ) Swap8Bytes(abyHeader+76);

    dValue = psSHP->adBoundsMin[3];			/* m */
    ByteCopy( &dValue, abyHeader+84, 8);
    if( bBigEndian ) Swap8Bytes(abyHeader+84);

    dValue = psSHP->adBoundsMax[3];
    ByteCopy( &dValue, abyHeader+92, 8);
    if( bBigEndian ) Swap8Bytes(abyHeader+92);

	/* -------------------------------------------------------------------- */
	/*      Write .shp file header.                                         */
	/* -------------------------------------------------------------------- */
	fseek (fpSHP, 0, 0);
	fwrite (abyHeader, 100, 1, fpSHP);

	/* -------------------------------------------------------------------- */
	/*      Prepare, and write .shx file header.                            */
	/* -------------------------------------------------------------------- */
    i32 = (psSHP->nRecords * 2 * sizeof(int32) + 100) / 2;   /* file size */
    ByteCopy (&i32, abyHeader+24, 4);
    if (!bBigEndian) Swap4Bytes(abyHeader+24);
    
    fseek (fpSHX, 0, 0);
    fwrite (abyHeader, 100, 1, fpSHX);

	/* -------------------------------------------------------------------- */
	/*      Write out the .shx contents.                                    */
	/* -------------------------------------------------------------------- */
	panSHX = SfIntRealloc(0, 2 * psSHP->nRecords);

	for (int i = 0; i < psSHP->nRecords; i++)
    {
		panSHX[i*2  ] = psSHP->panRecOffset[i] / 2;
		panSHX[i*2+1] = psSHP->panRecSize[i] / 2;
		if (!bBigEndian) Swap4Bytes(panSHX + i * 2);
		if (!bBigEndian) Swap4Bytes(panSHX + i * 2 + 1);
    }

	fwrite(panSHX, sizeof(int32) * 2, psSHP->nRecords, fpSHX);

	SfIntFree (panSHX);
}

/************************************************************************/
/*                              SHPOpen()                               */
/*                                                                      */
/*      Open the .shp and .shx files based on the basename of the       */
/*      files or either file name.                                      */
/************************************************************************/
   
SHPHandle SHPAPI_CALL SHPOpen(const char * pszLayer, const char * pszAccess)
{
	char		*pszFullname, *pszBasename;
	SHPHandle	psSHP;
    
    uchar		*pabyBuf;
    double		dValue;
    
/* -------------------------------------------------------------------- */
/*      Ensure the access string is one of the legal ones.  We          */
/*      ensure the result string indicates binary to avoid common       */
/*      problems on Windows.                                            */
/* -------------------------------------------------------------------- */
	if(strcmp(pszAccess,"rb+") == 0 || strcmp(pszAccess,"r+b") == 0 || strcmp(pszAccess,"r+") == 0)	pszAccess = "r+b";
	else                                                                                            pszAccess = "rb";

/* -------------------------------------------------------------------- */
/*	Establish the byte order on this machine.			*/
/* -------------------------------------------------------------------- */
	int Testint = 1;
	if( *((uchar *) &Testint) == 1 )	bBigEndian = FALSE;
	else                        		bBigEndian = TRUE;

/* -------------------------------------------------------------------- */
/*	Initialize the info structure.					*/
/* -------------------------------------------------------------------- */
	psSHP = SfSHPInfoRealloc(0, 1);

	psSHP->bUpdated = FALSE;

	/* -------------------------------------------------------------------- */
	/*	Compute the base (layer) name.  If there is any extension	*/
	/*	on the passed in filename we will strip it off.			*/
	/* -------------------------------------------------------------------- */
	pszBasename = (char*)SfByteRealloc(0, strlen(pszLayer)+5);
	strcpy( pszBasename, pszLayer );
	int i = strlen(pszBasename)-1;
	while(i > 0 && pszBasename[i] != '.' && pszBasename[i] != '/' && pszBasename[i] != '\\')
	{
		i--;
	}

	if( pszBasename[i] == '.' ) pszBasename[i] = '\0';

	/* -------------------------------------------------------------------- */
	/*	Open the .shp and .shx files.  Note that files pulled from			*/
	/*	a PC to Unix with upper case filenames won't work!					*/
	/* -------------------------------------------------------------------- */
	pszFullname = (char*)SfByteRealloc(0, strlen(pszBasename) + 5);
	sprintf( pszFullname, "%s.shp", pszBasename );
	fpSHP = fopen(pszFullname, pszAccess );
	if( fpSHP == NULL )
	{
		sprintf( pszFullname, "%s.SHP", pszBasename );
		fpSHP = fopen(pszFullname, pszAccess );
	}

	if( fpSHP == NULL )
	{
		SfSHPInfoFree(psSHP);
		SfByteFree((uchar*)pszBasename);
		SfByteFree((uchar*)pszFullname);
		return( NULL );
	}

	sprintf( pszFullname, "%s.shx", pszBasename );
	fpSHX = fopen(pszFullname, pszAccess );
	if( fpSHX == NULL )
	{
		sprintf( pszFullname, "%s.SHX", pszBasename );
		fpSHX = fopen(pszFullname, pszAccess );
	}

	if( fpSHX == NULL )
    {
		fclose(fpSHP);
		fpSHP = 0;
		SfSHPInfoFree(psSHP);
		SfByteFree((uchar*)pszBasename);
		SfByteFree((uchar*)pszFullname);
		return(NULL);
    }

	SfByteFree((uchar*)pszFullname);
	SfByteFree((uchar*)pszBasename);

	/* -------------------------------------------------------------------- */
	/*  Read the file size from the SHP file.				*/
	/* -------------------------------------------------------------------- */
	pabyBuf = SfByteRealloc(0, 100);

	fread(pabyBuf, 100, 1, fpSHP);

	psSHP->nFileSize = (pabyBuf[24] * 256 * 256 * 256 + pabyBuf[25] * 256 * 256 + pabyBuf[26] * 256 + pabyBuf[27]) * 2;

	/* -------------------------------------------------------------------- */
	/*  Read SHX file Header info                                           */
	/* -------------------------------------------------------------------- */
	fread(pabyBuf, 100, 1, fpSHX);

	if( pabyBuf[0] != 0 || pabyBuf[1] != 0 || pabyBuf[2] != 0x27 || (pabyBuf[3] != 0x0a && pabyBuf[3] != 0x0d))
	{
		fclose(fpSHP);
		fpSHP = 0;
		fclose(fpSHX);
		fpSHX = 0;
		SfSHPInfoFree(psSHP);
		return(NULL);
	}

    psSHP->nRecords = pabyBuf[27] + pabyBuf[26] * 256 + pabyBuf[25] * 256 * 256 + pabyBuf[24] * 256 * 256 * 256;
	psSHP->nRecords = (psSHP->nRecords * 2 - 100) / 8;

	psSHP->nShapeType = pabyBuf[32];

	if( psSHP->nRecords < 0 || psSHP->nRecords > 256000000 )
    {
        /* this header appears to be corrupt.  Give up. */
		fclose(fpSHP);
		fpSHP = 0;
		fclose(fpSHX);
		fpSHX = 0;
		SfSHPInfoFree(psSHP);
		return(NULL);
	}

	/* -------------------------------------------------------------------- */
	/*      Read the bounds.                                                */
	/* -------------------------------------------------------------------- */
    if( bBigEndian ) Swap8Bytes(pabyBuf+36 );
    memcpy( &dValue, pabyBuf+36, 8 );
    psSHP->adBoundsMin[0] = dValue;

    if( bBigEndian ) Swap8Bytes(pabyBuf+44 );
    memcpy( &dValue, pabyBuf+44, 8 );
    psSHP->adBoundsMin[1] = dValue;

    if( bBigEndian ) Swap8Bytes(pabyBuf+52 );
    memcpy( &dValue, pabyBuf+52, 8 );
    psSHP->adBoundsMax[0] = dValue;

    if( bBigEndian ) Swap8Bytes(pabyBuf+60 );
    memcpy( &dValue, pabyBuf+60, 8 );
    psSHP->adBoundsMax[1] = dValue;

    if( bBigEndian ) Swap8Bytes(pabyBuf+68 );		/* z */
    memcpy( &dValue, pabyBuf+68, 8 );
    psSHP->adBoundsMin[2] = dValue;
    
	if( bBigEndian ) Swap8Bytes(pabyBuf+76 );
    memcpy( &dValue, pabyBuf+76, 8 );
    psSHP->adBoundsMax[2] = dValue;
    
	if( bBigEndian ) Swap8Bytes(pabyBuf+84 );		/* m */
    memcpy( &dValue, pabyBuf+84, 8 );
    psSHP->adBoundsMin[3] = dValue;

    if( bBigEndian ) Swap8Bytes(pabyBuf+92 );
    memcpy( &dValue, pabyBuf+92, 8 );
    psSHP->adBoundsMax[3] = dValue;

	SfByteFree (pabyBuf);

	/* -------------------------------------------------------------------- */
	/*	Read the .shx file to get the offsets to each record in 			*/
	/*	the .shp file.														*/
	/* -------------------------------------------------------------------- */
	psSHP->nMaxRecords = psSHP->nRecords;

	if (psSHP->nMaxRecords > 0)
	{
		psSHP->panRecOffset = SfIntRealloc(0, psSHP->nMaxRecords);
		psSHP->panRecSize =   SfIntRealloc(0, psSHP->nMaxRecords);
		pabyBuf = SfByteRealloc(0, 8 * psSHP->nMaxRecords);

		fread (pabyBuf, 8, psSHP->nRecords, fpSHX);

    	for(int i = 0; i < psSHP->nRecords; i++ )
    	{
			int32		nOffset, nLength;

			memcpy( &nOffset, pabyBuf + i * 8, 4 );
			if( !bBigEndian ) Swap4Bytes(&nOffset );

			memcpy( &nLength, pabyBuf + i * 8 + 4, 4 );
			if( !bBigEndian ) Swap4Bytes(&nLength );

			psSHP->panRecOffset[i] = nOffset*2;
			psSHP->panRecSize[i] = nLength*2;
	    }

		SfByteFree (pabyBuf);
    }

    return( psSHP );
}

/************************************************************************/
/*                              SHPClose()                              */
/*								       	*/
/*	Close the .shp and .shx files.					*/
/************************************************************************/

void SHPAPI_CALL SHPClose(SHPHandle psSHP)
{
	/* -------------------------------------------------------------------- */
	/*	Update the header if we have modified anything.						*/
	/* -------------------------------------------------------------------- */
	if (psSHP->bUpdated)
    {
		SHPWriteHeader (psSHP);
    }

	/* -------------------------------------------------------------------- */
	/*      Free all resources, and close files.                            */
	/* -------------------------------------------------------------------- */

	fclose (fpSHX);
	fpSHX = 0;
	fclose (fpSHP);
	fpSHP = 0;

	SfIntFree (psSHP->panRecOffset);
	SfIntFree (psSHP->panRecSize);

	if (psSHP->pabyRec != NULL)
	{
		SfByteFree (psSHP->pabyRec);
	}

	SfSHPInfoFree (psSHP);
}

/************************************************************************/
/*                             SHPGetInfo()                             */
/*                                                                      */
/*      Fetch general information about the shape file.                 */
/************************************************************************/

void SHPAPI_CALL SHPGetInfo(SHPHandle psSHP, int * pnEntities, TShpType *pnShapeType, double * padfMinBound, double * padfMaxBound)

{
    int		i;
    
	if (pnEntities != NULL) *pnEntities = psSHP->nRecords;

	if (pnShapeType != NULL)
	{
		switch(psSHP->nShapeType)
		{
			case 0: *pnShapeType = SHPT_NULL; break;
			case 1: *pnShapeType = SHPT_POINT; break;
			case 3: *pnShapeType = SHPT_ARC; break;
			case 5: *pnShapeType = SHPT_POLYGON; break;
			case 8: *pnShapeType = SHPT_MULTIPOINT; break;
			case 11: *pnShapeType = SHPT_POINTZ; break;
			case 13: *pnShapeType = SHPT_ARCZ; break;
			case 15: *pnShapeType = SHPT_POLYGONZ; break;
			case 18: *pnShapeType = SHPT_MULTIPOINTZ; break;
			case 21: *pnShapeType = SHPT_POINTM; break;
			case 23: *pnShapeType = SHPT_ARCM; break;
			case 28: *pnShapeType = SHPT_POLYGONM; break;
			case 31: *pnShapeType = SHPT_MULTIPATCH; break;
			default: *pnShapeType = SHPT_NULL;
		}
	}

	for (i = 0; i < 4; i++)
	{
		if( padfMinBound != NULL ) padfMinBound[i] = psSHP->adBoundsMin[i];
		if( padfMaxBound != NULL ) padfMaxBound[i] = psSHP->adBoundsMax[i];
    }
}

/************************************************************************/
/*                             SHPCreate()                              */
/*                                                                      */
/*      Create a new shape file and return a handle to the open         */
/*      shape file with read/write access.                              */
/************************************************************************/
SHPHandle SHPAPI_CALL SHPCreate( const char * pszLayer, int nShapeType )
{
	char	*pszBasename, *pszFullname;
	uchar   abyHeader[100];
	int32	i32;
	double	dValue;

/* -------------------------------------------------------------------- */
/*      Establish the byte order on this system.                        */
/* -------------------------------------------------------------------- */
	int testint = 1;
	if( *((uchar *) &testint) == 1 )	bBigEndian = FALSE;
	else                        		bBigEndian = TRUE;

/* -------------------------------------------------------------------- */
/*	Compute the base (layer) name.  If there is any extension			*/
/*	on the passed in filename we will strip it off.						*/
/* -------------------------------------------------------------------- */
	pszBasename = (char*)SfByteRealloc(0, strlen(pszLayer)+5);
	strcpy (pszBasename, pszLayer);
	int i = strlen(pszBasename) - 1;
	while(i > 0 && pszBasename[i] != '.' && pszBasename[i] != '/' && pszBasename[i] != '\\')
	{
		i--;
	}

	if( pszBasename[i] == '.' ) pszBasename[i] = '\0';

	/* -------------------------------------------------------------------- */
	/*      Open the two files so we can write their headers.               */
	/* -------------------------------------------------------------------- */
	pszFullname = (char*)SfByteRealloc(0, strlen(pszBasename) + 5);
	sprintf( pszFullname, "%s.shp", pszBasename );
    fpSHP = fopen(pszFullname, "wb" );
	if(fpSHP == NULL) return(NULL);

	sprintf( pszFullname, "%s.shx", pszBasename );
	fpSHX = fopen(pszFullname, "wb" );
	if(fpSHX == NULL) return(NULL);

	SfByteFree((uchar*)pszFullname);
	SfByteFree((uchar*)pszBasename);

/* -------------------------------------------------------------------- */
/*      Prepare header block for .shp file.                             */
/* -------------------------------------------------------------------- */
	for(int i = 0; i < 100; i++) abyHeader[i] = 0;

    abyHeader[2] = 0x27;				/* magic cookie */
    abyHeader[3] = 0x0a;

    i32 = 50;						/* file size */
    ByteCopy( &i32, abyHeader+24, 4 );
	if(!bBigEndian) Swap4Bytes(abyHeader+24 );
    
	i32 = 1000;						/* version */
    ByteCopy( &i32, abyHeader+28, 4 );
	if(bBigEndian) Swap4Bytes(abyHeader+28 );
    
    i32 = nShapeType;					/* shape type */
    ByteCopy( &i32, abyHeader+32, 4 );
	if(bBigEndian) Swap4Bytes(abyHeader+32 );

    dValue = 0.0;					/* set bounds */
    ByteCopy( &dValue, abyHeader+36, 8 );
    ByteCopy( &dValue, abyHeader+44, 8 );
    ByteCopy( &dValue, abyHeader+52, 8 );
    ByteCopy( &dValue, abyHeader+60, 8 );

	/* -------------------------------------------------------------------- */
	/*      Write .shp file header.                                         */
	/* -------------------------------------------------------------------- */
    fwrite( abyHeader, 100, 1, fpSHP );

	/* -------------------------------------------------------------------- */
	/*      Prepare, and write .shx file header.                            */
	/* -------------------------------------------------------------------- */
    i32 = 50;						/* file size */
    ByteCopy( &i32, abyHeader+24, 4 );
	if(!bBigEndian) Swap4Bytes(abyHeader+24);
    
	fwrite(abyHeader, 100, 1, fpSHX);

	/* -------------------------------------------------------------------- */
	/*      Close the files, and then open them as regular existing files.  */
	/* -------------------------------------------------------------------- */
	fclose( fpSHP );
	fpSHP = 0;
	fclose( fpSHX );
	fpSHX = 0;

    return(SHPOpen(pszLayer, "r+b"));
}

/************************************************************************/
/*                           _SHPSetBounds()                            */
/*                                                                      */
/*      Compute a bounds rectangle for a shape, and set it into the     */
/*      indicated location in the record.                               */
/************************************************************************/

static void	_SHPSetBounds( uchar * pabyRec, SHPObject * psShape )
{
    ByteCopy( &(psShape->dfXMin), pabyRec +  0, 8 );
    ByteCopy( &(psShape->dfYMin), pabyRec +  8, 8 );
    ByteCopy( &(psShape->dfXMax), pabyRec + 16, 8 );
    ByteCopy( &(psShape->dfYMax), pabyRec + 24, 8 );

    if( bBigEndian )
    {
        Swap8Bytes(pabyRec + 0 );
        Swap8Bytes(pabyRec + 8 );
        Swap8Bytes(pabyRec + 16 );
        Swap8Bytes(pabyRec + 24 );
    }
}

/************************************************************************/
/*                         SHPComputeExtents()                          */
/*                                                                      */
/*      Recompute the extents of a shape.  Automatically done by        */
/*      SHPCreateObject().                                              */
/************************************************************************/

void SHPAPI_CALL SHPComputeExtents(SHPObject * psObject)
{
	/* -------------------------------------------------------------------- */
	/*      Build extents for this object.                                  */
	/* -------------------------------------------------------------------- */
	if (psObject->nVertices > 0)
	{
		psObject->dfXMin = psObject->dfXMax = psObject->padfX[0];
		psObject->dfYMin = psObject->dfYMax = psObject->padfY[0];
		psObject->dfZMin = psObject->dfZMax = psObject->padfZ[0];
		psObject->dfMMin = psObject->dfMMax = psObject->padfM[0];
	}

	for (int i = 0; i < psObject->nVertices; i++)
	{
		if (psObject->dfXMin > psObject->padfX[i])	psObject->dfXMin = psObject->padfX[i];
		if (psObject->dfYMin > psObject->padfY[i])	psObject->dfYMin = psObject->padfY[i];
		if (psObject->dfZMin > psObject->padfZ[i])	psObject->dfZMin = psObject->padfZ[i];
		if (psObject->dfMMin > psObject->padfM[i])	psObject->dfMMin = psObject->padfM[i];

		if (psObject->dfXMax < psObject->padfX[i])	psObject->dfXMax = psObject->padfX[i];
		if (psObject->dfYMax < psObject->padfY[i])	psObject->dfYMax = psObject->padfY[i];
		if (psObject->dfZMax < psObject->padfZ[i])	psObject->dfZMax = psObject->padfZ[i];
		if (psObject->dfMMax < psObject->padfM[i])	psObject->dfMMax = psObject->padfM[i];
	}
}

/************************************************************************/
/*                          SHPCreateObject()                           */
/*                                                                      */
/*      Create a shape object.  It should be freed with                 */
/*      SHPDestroyObject().                                             */
/************************************************************************/
SHPObject SHPAPI_CALL1(*) SHPCreateObject( int nSHPType, int nShapeId, int nParts, int * panPartStart, int * panPartType, int nVertices, double * padfX, double * padfY, double * padfZ, double * padfM )
{
	SHPObject	*psObject;
	int			i, bHasM, bHasZ;

	psObject = SfSHPObjectRealloc(0, 1);
	psObject->nSHPType = nSHPType;
	psObject->nShapeId = nShapeId;

	/* -------------------------------------------------------------------- */
	/*	Establish whether this shape type has M, and Z values.		*/
	/* -------------------------------------------------------------------- */
	if (nSHPType == SHPT_ARCM || nSHPType == SHPT_POINTM || nSHPType == SHPT_POLYGONM || nSHPType == SHPT_MULTIPOINTM)
	{
		bHasM = TRUE;
		bHasZ = FALSE;
	}
	else if(nSHPType == SHPT_ARCZ || nSHPType == SHPT_POINTZ || nSHPType == SHPT_POLYGONZ || nSHPType == SHPT_MULTIPOINTZ || nSHPType == SHPT_MULTIPATCH)
	{
		bHasM = TRUE;
		bHasZ = TRUE;
	}
	else
	{
		bHasM = FALSE;
		bHasZ = FALSE;
	}

	/* -------------------------------------------------------------------- */
	/*      Capture parts.  Note that part type is optional, and            */
	/*      defaults to ring.                                               */
	/* -------------------------------------------------------------------- */
	if (nSHPType == SHPT_ARC || nSHPType == SHPT_POLYGON || nSHPType == SHPT_ARCM || nSHPType == SHPT_POLYGONM || nSHPType == SHPT_ARCZ || nSHPType == SHPT_POLYGONZ || nSHPType == SHPT_MULTIPATCH)
	{
		if (nParts > 1)	psObject->nParts = nParts;
		else			psObject->nParts = 1;

		psObject->panPartStart = SfIntRealloc(0, psObject->nParts);
		psObject->panPartType = SfIntRealloc(0, psObject->nParts);
		psObject->panPartStart[0] = 0;
		psObject->panPartType[0] = SHPP_RING;

		for( i = 0; i < nParts; i++ )
		{
			psObject->panPartStart[i] = panPartStart[i];
			if( panPartType != NULL )	psObject->panPartType[i] = panPartType[i];
			else                        psObject->panPartType[i] = SHPP_RING;
		}
	}

	/* -------------------------------------------------------------------- */
	/*      Capture vertices.  Note that Z and M are optional, but X and    */
	/*      Y are not.                                                      */
	/* -------------------------------------------------------------------- */
	if (nVertices > 0)
	{
		psObject->padfX = SfDoubleRealloc(0, nVertices);
		psObject->padfY = SfDoubleRealloc(0, nVertices);
		psObject->padfZ = SfDoubleRealloc(0, nVertices);
		psObject->padfM = SfDoubleRealloc(0, nVertices);

		for (int i = 0; i < nVertices; i++)
		{
			psObject->padfX[i] = padfX[i];
			psObject->padfY[i] = padfY[i];
			if (padfZ != NULL && bHasZ)	psObject->padfZ[i] = padfZ[i];
			if (padfM != NULL && bHasM)	psObject->padfM[i] = padfM[i];
		}
	}

	/* -------------------------------------------------------------------- */
	/*      Compute the extents.                                            */
	/* -------------------------------------------------------------------- */
	psObject->nVertices = nVertices;
	SHPComputeExtents(psObject);

    return( psObject );
}

/************************************************************************/
/*                       SHPCreateSimpleObject()                        */
/*                                                                      */
/*      Create a simple (common) shape object.  Destroy with            */
/*      SHPDestroyObject().                                             */
/************************************************************************/
SHPObject SHPAPI_CALL1(*) SHPCreateSimpleObject( int nSHPType, int nVertices, double * padfX, double * padfY, double * padfZ)
{
	return(SHPCreateObject(nSHPType, -1, 0, NULL, NULL, nVertices, padfX, padfY, padfZ, NULL));
}
                                  
/************************************************************************/
/*                           SHPSetMaxRecords()                         */
/*                                                                      */
/*      Sets nMaxRecords und reallocs enough Memory                     */
/************************************************************************/
void SHPAPI_CALL SHPSetMaxRecords(SHPHandle psSHP, int nNewMaxRecords)
{
	if (nNewMaxRecords > psSHP->nMaxRecords)
	{
		psSHP->nMaxRecords = nNewMaxRecords;
		psSHP->panRecOffset = SfIntRealloc(psSHP->panRecOffset, psSHP->nMaxRecords);
		psSHP->panRecSize = SfIntRealloc(psSHP->panRecSize, psSHP->nMaxRecords);
	}
}

/************************************************************************/
/*                           SHPWritePoint()                            */
/*                                                                      */
/*      Write out the vertices of a new structure.  Note that it is     */
/*      only possible to write vertices at the end of the file.         */
/************************************************************************/
int SHPAPI_CALL SHPWritePoint(SHPHandle psSHP, double X, double Y)
{
	int		nRecordOffset, nRecordSize;
	uchar	pabyRec[28];
	int32	i32;

	psSHP->bUpdated = TRUE;

	/* -------------------------------------------------------------------- */
	/*      Ensure that shape object matches the type of the file it is     */
	/*      being written to.                                               */
	/* -------------------------------------------------------------------- */
	assert(SHPT_POINT == psSHP->nShapeType);

	/* -------------------------------------------------------------------- */
	/*      Add the new entity to the in memory index.                      */
	/* -------------------------------------------------------------------- */
	if (psSHP->nRecords + 1 > psSHP->nMaxRecords)
	{
		psSHP->nMaxRecords =(int)(psSHP->nMaxRecords * 1.3 + 100);
		psSHP->panRecOffset = SfIntRealloc(psSHP->panRecOffset, psSHP->nMaxRecords);
		psSHP->panRecSize = SfIntRealloc(psSHP->panRecSize, psSHP->nMaxRecords);
	}

	nRecordSize = 28;

	int nShapeId = psSHP->nRecords++;

	psSHP->panRecOffset[nShapeId] = nRecordOffset = psSHP->nFileSize;
	psSHP->panRecSize[nShapeId] = nRecordSize - 8;
	psSHP->nFileSize += nRecordSize;

	/* -------------------------------------------------------------------- */
	/*      Set the record number, record size, shape type, X, Y.           */
	/* -------------------------------------------------------------------- */
	i32 = nShapeId+1;						/* record # */
	if(!bBigEndian) Swap4Bytes(&i32);
	ByteCopy(&i32, pabyRec + 0, 4);

	i32 = (nRecordSize-8)/2;				/* record size */
	if(!bBigEndian) Swap4Bytes(&i32);
	ByteCopy(&i32, pabyRec + 4, 4);

	i32 = SHPT_POINT;						/* shape type */
	if(bBigEndian) Swap4Bytes(&i32);
	ByteCopy( &i32, pabyRec + 8, 4);

	ByteCopy(&X, pabyRec + 12, 8);
	if (bBigEndian) Swap8Bytes(pabyRec + 12);

	ByteCopy(&Y, pabyRec + 20, 8);
	if (bBigEndian) Swap8Bytes(pabyRec + 20);

	/* -------------------------------------------------------------------- */
	/*      Write out record.                                               */
	/* -------------------------------------------------------------------- */
	if( fseek(fpSHP, nRecordOffset, 0) != 0 || fwrite(pabyRec, nRecordSize, 1, fpSHP) < 1 )
	{
		dump_error(__FILE__, __LINE__, "Error in fseek() or fwrite().\n");
		return (-1);
	}

	/* -------------------------------------------------------------------- */
	/*	Expand file wide bounds based on this shape.			*/
	/* -------------------------------------------------------------------- */
	if (psSHP->adBoundsMin[0] == 0.0 && psSHP->adBoundsMax[0] == 0.0 && psSHP->adBoundsMin[1] == 0.0 && psSHP->adBoundsMax[1] == 0.0)
	{
		psSHP->adBoundsMin[0] = psSHP->adBoundsMax[0] = X;
		psSHP->adBoundsMin[1] = psSHP->adBoundsMax[1] = Y;
		psSHP->adBoundsMin[2] = psSHP->adBoundsMax[2] = 0.0;
		psSHP->adBoundsMin[3] = psSHP->adBoundsMax[3] = 0.0;
	}

	if (psSHP->adBoundsMin[0] > X)		psSHP->adBoundsMin[0] = X;
	if (psSHP->adBoundsMin[1] > Y)		psSHP->adBoundsMin[1] = Y;
	if (psSHP->adBoundsMin[2] > 0.0)	psSHP->adBoundsMin[2] = 0.0;
	if (psSHP->adBoundsMin[3] > 0.0)	psSHP->adBoundsMin[3] = 0.0;

	if (psSHP->adBoundsMax[0] < X)		psSHP->adBoundsMax[0] = X;
	if (psSHP->adBoundsMax[1] < Y)		psSHP->adBoundsMax[1] = Y;
	if (psSHP->adBoundsMax[2] < 0.0)	psSHP->adBoundsMax[2] = 0.0;
	if (psSHP->adBoundsMax[3] < 0.0)	psSHP->adBoundsMax[3] = 0.0;


	return(nShapeId);
}

/************************************************************************/
/*                           SHPWriteElement()                          */
/*                                                                      */
/*      Write out the 3 vertices of a new element.  Note that it is     */
/*      only possible to write vertices at the end of the file.         */
/************************************************************************/
int SHPAPI_CALL SHPWriteElement(SHPHandle psSHP, double X1, double Y1, double X2, double Y2, double X3, double Y3)
{
	int		nRecordOffset, nRecordSize;
	uchar	pabyRec[120];
    int32	i32;

	psSHP->bUpdated = TRUE;

	/* -------------------------------------------------------------------- */
	/*      Ensure that shape object matches the type of the file it is     */
	/*      being written to.                                               */
	/* -------------------------------------------------------------------- */
	assert(SHPT_POLYGON == psSHP->nShapeType);

	/* -------------------------------------------------------------------- */
	/*      Add the new entity to the in memory index.                      */
	/* -------------------------------------------------------------------- */
	if (psSHP->nRecords + 1 > psSHP->nMaxRecords)
	{
		psSHP->nMaxRecords =(int)(psSHP->nMaxRecords * 1.3 + 100);
		psSHP->panRecOffset = SfIntRealloc(psSHP->panRecOffset, psSHP->nMaxRecords);
		psSHP->panRecSize = SfIntRealloc(psSHP->panRecSize, psSHP->nMaxRecords);
	}

	/* -------------------------------------------------------------------- */
	/*  Extract vertices for a Polygon or Arc.								*/
	/* -------------------------------------------------------------------- */

	double XMin = X1;
	double YMin = Y1;
	double XMax = X1;
	double YMax = Y1;

	if (XMin > X2)	XMin = X2;
	if (XMin > X3)	XMin = X3;
	if (YMin > Y2)	YMin = Y2;
	if (YMin > Y3)	YMin = Y3;

	if (XMax < X2)	XMax = X2;
	if (XMax < X3)	XMax = X3;
	if (YMax < Y2)	YMax = Y2;
	if (YMax < Y3)	YMax = Y3;

	ByteCopy (&XMin, pabyRec + 12, 8);        	// 12 - 19
	ByteCopy (&YMin, pabyRec + 20, 8);         	// 20 - 27
	ByteCopy (&XMax, pabyRec + 28, 8);         	// 28 - 35
	ByteCopy (&YMax, pabyRec + 36, 8);        	// 36 - 43

	if (bBigEndian)
	{
		Swap8Bytes (pabyRec + 12);
		Swap8Bytes (pabyRec + 20);
		Swap8Bytes (pabyRec + 28);
		Swap8Bytes (pabyRec + 36);
	}

	int32 nParts = 1;
	int32 nPoints = 4;

	ByteCopy (&nParts, pabyRec + 44, 4);		// 44 - 47
	ByteCopy (&nPoints, pabyRec + 48, 4);		// 48 - 51
	if (bBigEndian) Swap4Bytes(pabyRec + 44);
	if (bBigEndian) Swap4Bytes(pabyRec + 48);

	i32 = 0;
	ByteCopy (&i32, pabyRec + 52, 4);			// 52 - 56	PanStart
	if (bBigEndian) Swap4Bytes(pabyRec + 52);

	double X = 0.0;
	double Y = 0.0;

	X = X1;
	Y = Y1;
	ByteCopy (&X, pabyRec + 56, 8);			// 56 - 63
	ByteCopy (&Y, pabyRec + 64, 8);			// 64 - 72
	if (bBigEndian)	Swap8Bytes(pabyRec + 56);
	if (bBigEndian)	Swap8Bytes(pabyRec + 64);

	X = X2;
	Y = Y2;
	ByteCopy(&X, pabyRec + 72, 8);				// 72 - 79
	ByteCopy(&Y, pabyRec + 80, 8);				// 80 - 87
	if (bBigEndian)	Swap8Bytes(pabyRec + 72);
	if (bBigEndian)	Swap8Bytes(pabyRec + 80);

	X = X3;
	Y = Y3;
	ByteCopy (&X, pabyRec + 88, 8);				// 88 - 95
	ByteCopy (&Y, pabyRec + 96, 8);				// 96 - 103
	if (bBigEndian)	Swap8Bytes(pabyRec + 88);
	if (bBigEndian)	Swap8Bytes(pabyRec + 96);

	X = X1;
	Y = Y1;
	ByteCopy (&X, pabyRec + 104, 8);			// 104 - 111
	ByteCopy (&Y, pabyRec + 112, 8);			// 112 - 119
	if (bBigEndian)	Swap8Bytes(pabyRec + 104);
	if (bBigEndian)	Swap8Bytes(pabyRec + 112);

	nRecordSize = 120;


	/* -------------------------------------------------------------------- */
	/*      Establish where we are going to put this record. If we are      */
	/*      rewriting and existing record, and it will fit, then put it     */
	/*      back where the original came from.  Otherwise write at the end. */
	/* -------------------------------------------------------------------- */

	int nShapeId = psSHP->nRecords++;

	psSHP->panRecOffset[nShapeId] = nRecordOffset = psSHP->nFileSize;
	psSHP->panRecSize[nShapeId] = nRecordSize-8;
	psSHP->nFileSize += nRecordSize;

	/* -------------------------------------------------------------------- */
	/*      Set the shape type, record number, and record size.             */
	/* -------------------------------------------------------------------- */
	i32 = nShapeId+1;						/* record # */
	ByteCopy (&i32, pabyRec + 0, 4);			// 0 - 3
	if (!bBigEndian) Swap4Bytes (pabyRec + 0);

	i32 = (nRecordSize-8)/2;				/* record size */
	ByteCopy (&i32, pabyRec + 4, 4);		// 4 - 7
	if (!bBigEndian) Swap4Bytes (pabyRec + 4);

	i32 = SHPT_POLYGON;						/* shape type */
	ByteCopy (&i32, pabyRec + 8, 4);		// 8 - 11
	if (bBigEndian) Swap4Bytes (pabyRec + 8);

	/* -------------------------------------------------------------------- */
	/*      Write out record.                                               */
	/* -------------------------------------------------------------------- */
	if (fseek(fpSHP, nRecordOffset, 0) != 0 || fwrite(pabyRec, nRecordSize, 1, fpSHP) < 1)
    {
		dump_error(__FILE__, __LINE__, "Error in fseek() or fwrite().\n");
		return (-1);
    }

	/* -------------------------------------------------------------------- */
	/*	Expand file wide bounds based on this shape.						*/
	/* -------------------------------------------------------------------- */
	if (psSHP->adBoundsMin[0] == 0.0 && psSHP->adBoundsMax[0] == 0.0 && psSHP->adBoundsMin[1] == 0.0 && psSHP->adBoundsMax[1] == 0.0)
	{
		psSHP->adBoundsMin[0] = XMin;
		psSHP->adBoundsMin[1] = YMin;
		psSHP->adBoundsMax[0] = XMax;
		psSHP->adBoundsMax[1] = YMax;
		psSHP->adBoundsMin[2] = psSHP->adBoundsMax[2] = 0.0;
		psSHP->adBoundsMin[3] = psSHP->adBoundsMax[3] = 0.0;
	}

	if (psSHP->adBoundsMin[0] > XMin)	psSHP->adBoundsMin[0] = XMin;
	if (psSHP->adBoundsMin[1] > YMin)	psSHP->adBoundsMin[1] = YMin;

	if (psSHP->adBoundsMax[0] < XMax)	psSHP->adBoundsMax[0] = XMax;
	if (psSHP->adBoundsMax[1] < YMax)	psSHP->adBoundsMax[1] = YMax;

	return(nShapeId);
}


/************************************************************************/
/*                           SHPWritePolygon()                          */
/*                                                                      */
/*      Write out the vertices of a new structure.  Note that it is     */
/*      only possible to write vertices at the end of the file.         */
/************************************************************************/
int SHPAPI_CALL SHPWritePolygon(SHPHandle psSHP, SHPObject* psObject)
{
	psSHP->bUpdated = TRUE;

	/* -------------------------------------------------------------------- */
	/*      Ensure that shape object matches the type of the file it is     */
	/*      being written to.                                               */
	/* -------------------------------------------------------------------- */
	assert(psObject->nSHPType == psSHP->nShapeType || psObject->nSHPType == SHPT_NULL);

	int nShapeId = -1;

	/* -------------------------------------------------------------------- */
	/*      Add the new entity to the in memory index.                      */
	/* -------------------------------------------------------------------- */
	if (nShapeId == -1 && psSHP->nRecords + 1 > psSHP->nMaxRecords)
	{
		psSHP->nMaxRecords =(int)(psSHP->nMaxRecords * 1.3 + 100);
		psSHP->panRecOffset = SfIntRealloc(psSHP->panRecOffset, psSHP->nMaxRecords);
		psSHP->panRecSize = SfIntRealloc(psSHP->panRecSize, psSHP->nMaxRecords);
	}

	/* -------------------------------------------------------------------- */
	/*      Initialize record.                                              */
	/* -------------------------------------------------------------------- */
	uchar* pabyRec = SfByteRealloc(0, psObject->nVertices * 4 * sizeof(double) + psObject->nParts * 8 + 128);

	/* -------------------------------------------------------------------- */
	/*  Extract vertices for a Polygon or Arc.								*/
	/* -------------------------------------------------------------------- */

	int32 nPoints = psObject->nVertices;
	int32 nParts = psObject->nParts;

	_SHPSetBounds( pabyRec + 12, psObject );

	if( bBigEndian ) Swap4Bytes(&nPoints );
	if( bBigEndian ) Swap4Bytes(&nParts );

	ByteCopy( &nParts, pabyRec + 36 + 8, 4 );
	ByteCopy( &nPoints, pabyRec + 40 + 8, 4 );

	int nRecordSize = 52;

	// Write part start positions.
	ByteCopy( psObject->panPartStart, pabyRec + 44 + 8, 4 * psObject->nParts );
	for(int i = 0; i < psObject->nParts; i++ )
	{
		if( bBigEndian ) Swap4Bytes(pabyRec + 44 + 8 + 4*i );
		nRecordSize += 4;
	}

	// Write the (x,y) vertex values.
	for(int i = 0; i < psObject->nVertices; i++ )
	{
		ByteCopy( psObject->padfX + i, pabyRec + nRecordSize, 8 );
		ByteCopy( psObject->padfY + i, pabyRec + nRecordSize + 8, 8 );

		if (bBigEndian)	Swap8Bytes(pabyRec + nRecordSize );

		if (bBigEndian)	Swap8Bytes(pabyRec + nRecordSize + 8 );

		nRecordSize = nRecordSize + 2 * 8;
	}

	/* -------------------------------------------------------------------- */
	/*      Establish where we are going to put this record. If we are      */
	/*      rewriting and existing record, and it will fit, then put it     */
	/*      back where the original came from.  Otherwise write at the end. */
	/* -------------------------------------------------------------------- */
	nShapeId = psSHP->nRecords++;

	int nRecordOffset = psSHP->nFileSize;
	psSHP->panRecOffset[nShapeId] = nRecordOffset;
	psSHP->panRecSize[nShapeId] = nRecordSize-8;
	psSHP->nFileSize += nRecordSize;

	/* -------------------------------------------------------------------- */
	/*      Set the shape type, record number, and record size.             */
	/* -------------------------------------------------------------------- */
	int32	i32 = 0;

	i32 = nShapeId+1;						/* record # */
	if( !bBigEndian ) Swap4Bytes(&i32 );
	ByteCopy( &i32, pabyRec, 4 );

	i32 = (nRecordSize-8)/2;				/* record size */
	if( !bBigEndian ) Swap4Bytes(&i32 );
	ByteCopy( &i32, pabyRec + 4, 4 );

	i32 = psObject->nSHPType;				/* shape type */
	if( bBigEndian ) Swap4Bytes(&i32 );
	ByteCopy( &i32, pabyRec + 8, 4 );

	/* -------------------------------------------------------------------- */
	/*      Write out record.                                               */
	/* -------------------------------------------------------------------- */
	if( fseek(fpSHP, nRecordOffset, 0) != 0 || fwrite(pabyRec, nRecordSize, 1, fpSHP) < 1 )
	{
		SfByteFree(pabyRec);
		dump_error(__FILE__, __LINE__, "Error in fseek() or fwrite().\n");
		return (-1);
	}

	SfByteFree(pabyRec);

	/* -------------------------------------------------------------------- */
	/*	Expand file wide bounds based on this shape.						*/
	/* -------------------------------------------------------------------- */
	if (psSHP->adBoundsMin[0] == 0.0 && psSHP->adBoundsMax[0] == 0.0 && psSHP->adBoundsMin[1] == 0.0 && psSHP->adBoundsMax[1] == 0.0)
	{
		psSHP->adBoundsMin[0] = psSHP->adBoundsMax[0] = psObject->padfX[0];
		psSHP->adBoundsMin[1] = psSHP->adBoundsMax[1] = psObject->padfY[0];
		psSHP->adBoundsMin[2] = psSHP->adBoundsMax[2] = 0;
		psSHP->adBoundsMin[3] = psSHP->adBoundsMax[3] = 0;
	}

	for (int i = 0; i < psObject->nVertices; i++ )
	{
		if (psSHP->adBoundsMin[0] > psObject->padfX[i])	psSHP->adBoundsMin[0] = psObject->padfX[i];
		if (psSHP->adBoundsMin[1] > psObject->padfY[i])	psSHP->adBoundsMin[1] = psObject->padfY[i];

		if (psSHP->adBoundsMax[0] < psObject->padfX[i])	psSHP->adBoundsMax[0] = psObject->padfX[i];
		if (psSHP->adBoundsMax[1] < psObject->padfY[i])	psSHP->adBoundsMax[1] = psObject->padfY[i];
	}

	return(nShapeId);
}


/************************************************************************/
/*                           SHPWriteObject()                           */
/*                                                                      */
/*      Write out the vertices of a new structure.  Note that it is     */
/*      only possible to write vertices at the end of the file.         */
/************************************************************************/
int SHPAPI_CALL SHPWriteObject(SHPHandle psSHP, int nShapeId, SHPObject* psObject)
{
	int		nRecordOffset, i, nRecordSize;
	uchar	*pabyRec;
	int32	i32;

	psSHP->bUpdated = TRUE;

	/* -------------------------------------------------------------------- */
	/*      Ensure that shape object matches the type of the file it is     */
	/*      being written to.                                               */
	/* -------------------------------------------------------------------- */
	assert(psObject->nSHPType == psSHP->nShapeType || psObject->nSHPType == SHPT_NULL);

	/* -------------------------------------------------------------------- */
	/*      Ensure that -1 is used for appends.  Either blow an             */
	/*      assertion, or if they are disabled, set the shapeid to -1       */
	/*      for appends.                                                    */
	/* -------------------------------------------------------------------- */
	assert(nShapeId == -1 || (nShapeId >= 0 && nShapeId < psSHP->nRecords));

	if (nShapeId != -1 && nShapeId >= psSHP->nRecords)	nShapeId = -1;

	/* -------------------------------------------------------------------- */
	/*      Add the new entity to the in memory index.                      */
	/* -------------------------------------------------------------------- */
	if (nShapeId == -1 && psSHP->nRecords + 1 > psSHP->nMaxRecords)
	{
		psSHP->nMaxRecords =(int)(psSHP->nMaxRecords * 1.3 + 100);
		psSHP->panRecOffset = SfIntRealloc(psSHP->panRecOffset, psSHP->nMaxRecords);
		psSHP->panRecSize = SfIntRealloc(psSHP->panRecSize, psSHP->nMaxRecords);
	}

	/* -------------------------------------------------------------------- */
	/*      Initialize record.                                              */
	/* -------------------------------------------------------------------- */
	pabyRec = SfByteRealloc(0, psObject->nVertices * 4 * sizeof(double) + psObject->nParts * 8 + 128);

	if(psObject->nSHPType == SHPT_POLYGON || psObject->nSHPType == SHPT_POLYGONZ || psObject->nSHPType == SHPT_POLYGONM || psObject->nSHPType == SHPT_ARC || psObject->nSHPType == SHPT_ARCZ || psObject->nSHPType == SHPT_ARCM || psObject->nSHPType == SHPT_MULTIPATCH)
	{
		/* -------------------------------------------------------------------- */
		/*  Extract vertices for a Polygon or Arc.								*/
		/* -------------------------------------------------------------------- */

		int32		nPoints, nParts;
		int    		i;

		nPoints = psObject->nVertices;
		nParts = psObject->nParts;

		_SHPSetBounds( pabyRec + 12, psObject );

		if( bBigEndian ) Swap4Bytes(&nPoints );
		if( bBigEndian ) Swap4Bytes(&nParts );

		ByteCopy( &nParts, pabyRec + 36 + 8, 4 );
		ByteCopy( &nPoints, pabyRec + 40 + 8, 4 );

		nRecordSize = 52;

		// Write part start positions.
		ByteCopy( psObject->panPartStart, pabyRec + 44 + 8, 4 * psObject->nParts );
		for( i = 0; i < psObject->nParts; i++ )
		{
			if( bBigEndian ) Swap4Bytes(pabyRec + 44 + 8 + 4*i );
			nRecordSize += 4;
		}

		// Write multipatch part types if needed.
		if( psObject->nSHPType == SHPT_MULTIPATCH )
		{
			memcpy( pabyRec + nRecordSize, psObject->panPartType, 4*psObject->nParts );
			for( i = 0; i < psObject->nParts; i++ )
            {
                if( bBigEndian ) Swap4Bytes(pabyRec + nRecordSize );
                nRecordSize += 4;
            }
		}

		// Write the (x,y) vertex values.
		for( i = 0; i < psObject->nVertices; i++ )
		{
			ByteCopy( psObject->padfX + i, pabyRec + nRecordSize, 8 );
			ByteCopy( psObject->padfY + i, pabyRec + nRecordSize + 8, 8 );

			if (bBigEndian)	Swap8Bytes(pabyRec + nRecordSize );

			if (bBigEndian)	Swap8Bytes(pabyRec + nRecordSize + 8 );

			nRecordSize = nRecordSize + 2 * 8;
		}

		// Write the Z coordinates (if any).
		if( psObject->nSHPType == SHPT_POLYGONZ
			|| psObject->nSHPType == SHPT_ARCZ
			|| psObject->nSHPType == SHPT_MULTIPATCH )
		{
			ByteCopy( &(psObject->dfZMin), pabyRec + nRecordSize, 8 );
			if( bBigEndian ) Swap8Bytes(pabyRec + nRecordSize );
			nRecordSize += 8;

			ByteCopy( &(psObject->dfZMax), pabyRec + nRecordSize, 8 );
			if( bBigEndian ) Swap8Bytes(pabyRec + nRecordSize );
			nRecordSize += 8;

			for( i = 0; i < psObject->nVertices; i++ )
			{
				ByteCopy( psObject->padfZ + i, pabyRec + nRecordSize, 8 );
                if( bBigEndian ) Swap8Bytes(pabyRec + nRecordSize );
                nRecordSize += 8;
			}
        }

		/*
         * Write the M values, if any.
		*/
        if( psObject->nSHPType == SHPT_POLYGONM
            || psObject->nSHPType == SHPT_ARCM
#ifndef DISABLE_MULTIPATCH_MEASURE            
            || psObject->nSHPType == SHPT_MULTIPATCH
#endif            
            || psObject->nSHPType == SHPT_POLYGONZ
            || psObject->nSHPType == SHPT_ARCZ )
        {
            ByteCopy( &(psObject->dfMMin), pabyRec + nRecordSize, 8 );
            if( bBigEndian ) Swap8Bytes(pabyRec + nRecordSize );
            nRecordSize += 8;

            ByteCopy( &(psObject->dfMMax), pabyRec + nRecordSize, 8 );
            if( bBigEndian ) Swap8Bytes(pabyRec + nRecordSize );
			nRecordSize += 8;

            for( i = 0; i < psObject->nVertices; i++ )
            {
                ByteCopy( psObject->padfM + i, pabyRec + nRecordSize, 8 );
                if( bBigEndian ) Swap8Bytes(pabyRec + nRecordSize );
                nRecordSize += 8;
            }
        }
    }
    else if( psObject->nSHPType == SHPT_MULTIPOINT || psObject->nSHPType == SHPT_MULTIPOINTZ || psObject->nSHPType == SHPT_MULTIPOINTM )
    {
		/* -------------------------------------------------------------------- */
		/*  Extract vertices for a MultiPoint.									*/
		/* -------------------------------------------------------------------- */

		int32		nPoints;

		nPoints = psObject->nVertices;

        _SHPSetBounds( pabyRec + 12, psObject );

		if( bBigEndian ) Swap4Bytes(&nPoints );
		ByteCopy( &nPoints, pabyRec + 44, 4 );
	
		for( i = 0; i < psObject->nVertices; i++ )
		{
			ByteCopy( psObject->padfX + i, pabyRec + 48 + i*16, 8 );
	    	ByteCopy( psObject->padfY + i, pabyRec + 48 + i*16 + 8, 8 );

	    	if( bBigEndian ) Swap8Bytes(pabyRec + 48 + i*16 );
	    	if( bBigEndian ) Swap8Bytes(pabyRec + 48 + i*16 + 8 );
		}

		nRecordSize = 48 + 16 * psObject->nVertices;

        if( psObject->nSHPType == SHPT_MULTIPOINTZ )
        {
            ByteCopy( &(psObject->dfZMin), pabyRec + nRecordSize, 8 );
            if( bBigEndian ) Swap8Bytes(pabyRec + nRecordSize );
            nRecordSize += 8;

            ByteCopy( &(psObject->dfZMax), pabyRec + nRecordSize, 8 );
            if( bBigEndian ) Swap8Bytes(pabyRec + nRecordSize );
            nRecordSize += 8;
            
            for( i = 0; i < psObject->nVertices; i++ )
            {
                ByteCopy( psObject->padfZ + i, pabyRec + nRecordSize, 8 );
                if( bBigEndian ) Swap8Bytes(pabyRec + nRecordSize );
                nRecordSize += 8;
			}
        }

        if( psObject->nSHPType == SHPT_MULTIPOINTZ
            || psObject->nSHPType == SHPT_MULTIPOINTM )
		{
            ByteCopy( &(psObject->dfMMin), pabyRec + nRecordSize, 8 );
            if( bBigEndian ) Swap8Bytes(pabyRec + nRecordSize );
            nRecordSize += 8;

            ByteCopy( &(psObject->dfMMax), pabyRec + nRecordSize, 8 );
            if( bBigEndian ) Swap8Bytes(pabyRec + nRecordSize );
            nRecordSize += 8;
            
            for( i = 0; i < psObject->nVertices; i++ )
            {
                ByteCopy( psObject->padfM + i, pabyRec + nRecordSize, 8 );
				if( bBigEndian ) Swap8Bytes(pabyRec + nRecordSize );
                nRecordSize += 8;
            }
        }
    }
	else if (psObject->nSHPType == SHPT_POINT || psObject->nSHPType == SHPT_POINTZ || psObject->nSHPType == SHPT_POINTM)
	{
		/* -------------------------------------------------------------------- */
		/*      Write point.							*/
		/* -------------------------------------------------------------------- */

		ByteCopy( psObject->padfX, pabyRec + 12, 8 );
		if (bBigEndian) Swap8Bytes(pabyRec + 12);

		ByteCopy( psObject->padfY, pabyRec + 20, 8 );
		if (bBigEndian) Swap8Bytes(pabyRec + 20);

		nRecordSize = 28;

		if (psObject->nSHPType == SHPT_POINTZ)
		{
			ByteCopy(psObject->padfZ, pabyRec + nRecordSize, 8);
			if(bBigEndian) Swap8Bytes(pabyRec + nRecordSize);
			nRecordSize += 8;
        }
        
		if( psObject->nSHPType == SHPT_POINTZ || psObject->nSHPType == SHPT_POINTM )
        {
            ByteCopy(psObject->padfM, pabyRec + nRecordSize, 8);
			if( bBigEndian ) Swap8Bytes(pabyRec + nRecordSize);
            nRecordSize += 8;
        }
    }
    else if( psObject->nSHPType == SHPT_NULL )
	{
		/* -------------------------------------------------------------------- */
		/*      Not much to do for null geometries.                             */
		/* -------------------------------------------------------------------- */
        nRecordSize = 12;
    }
    else
    {
        /* unknown type */
		dump_error(__FILE__, __LINE__, "Unbekannter Shape-Typ %d\n", psObject->nSHPType);
        return (0);
	}

	/* -------------------------------------------------------------------- */
	/*      Establish where we are going to put this record. If we are      */
	/*      rewriting and existing record, and it will fit, then put it     */
	/*      back where the original came from.  Otherwise write at the end. */
	/* -------------------------------------------------------------------- */
    if( nShapeId == -1 || psSHP->panRecSize[nShapeId] < nRecordSize-8 )
    {
		if (nShapeId == -1) nShapeId = psSHP->nRecords++;

        psSHP->panRecOffset[nShapeId] = nRecordOffset = psSHP->nFileSize;
        psSHP->panRecSize[nShapeId] = nRecordSize-8;
        psSHP->nFileSize += nRecordSize;
    }
    else
    {
        nRecordOffset = psSHP->panRecOffset[nShapeId];
    }

	/* -------------------------------------------------------------------- */
	/*      Set the shape type, record number, and record size.             */
	/* -------------------------------------------------------------------- */
	i32 = nShapeId+1;					/* record # */
    if( !bBigEndian ) Swap4Bytes(&i32 );
    ByteCopy( &i32, pabyRec, 4 );

    i32 = (nRecordSize-8)/2;				/* record size */
    if( !bBigEndian ) Swap4Bytes(&i32 );
    ByteCopy( &i32, pabyRec + 4, 4 );

    i32 = psObject->nSHPType;				/* shape type */
    if( bBigEndian ) Swap4Bytes(&i32 );
    ByteCopy( &i32, pabyRec + 8, 4 );

	/* -------------------------------------------------------------------- */
	/*      Write out record.                                               */
	/* -------------------------------------------------------------------- */
	if( fseek(fpSHP, nRecordOffset, 0) != 0 || fwrite(pabyRec, nRecordSize, 1, fpSHP) < 1 )
    {
		SfByteFree(pabyRec);
		dump_error(__FILE__, __LINE__, "Error in fseek() or fwrite().\n");
		return (-1);
    }
    
	SfByteFree(pabyRec);

	/* -------------------------------------------------------------------- */
	/*	Expand file wide bounds based on this shape.			*/
	/* -------------------------------------------------------------------- */
    if (psSHP->adBoundsMin[0] == 0.0 && psSHP->adBoundsMax[0] == 0.0 && psSHP->adBoundsMin[1] == 0.0 && psSHP->adBoundsMax[1] == 0.0 && psObject->nSHPType != SHPT_NULL)
    {
        psSHP->adBoundsMin[0] = psSHP->adBoundsMax[0] = psObject->padfX[0];
        psSHP->adBoundsMin[1] = psSHP->adBoundsMax[1] = psObject->padfY[0];
		psSHP->adBoundsMin[2] = psSHP->adBoundsMax[2] = psObject->padfZ[0];
        psSHP->adBoundsMin[3] = psSHP->adBoundsMax[3] = psObject->padfM[0];
    }

	for (int i = 0; i < psObject->nVertices; i++ )
	{
		if (psSHP->adBoundsMin[0] > psObject->padfX[i])	psSHP->adBoundsMin[0] = psObject->padfX[i];
		if (psSHP->adBoundsMin[1] > psObject->padfY[i])	psSHP->adBoundsMin[1] = psObject->padfY[i];
		if (psSHP->adBoundsMin[2] > psObject->padfZ[i])	psSHP->adBoundsMin[2] = psObject->padfZ[i];
		if (psSHP->adBoundsMin[3] > psObject->padfM[i])	psSHP->adBoundsMin[3] = psObject->padfM[i];

		if (psSHP->adBoundsMax[0] < psObject->padfX[i])	psSHP->adBoundsMax[0] = psObject->padfX[i];
		if (psSHP->adBoundsMax[1] < psObject->padfY[i])	psSHP->adBoundsMax[1] = psObject->padfY[i];
		if (psSHP->adBoundsMax[2] < psObject->padfZ[i])	psSHP->adBoundsMax[2] = psObject->padfZ[i];
		if (psSHP->adBoundsMax[3] < psObject->padfM[i])	psSHP->adBoundsMax[3] = psObject->padfM[i];
	}

	return(nShapeId);
}

/************************************************************************/
/*                          SHPReadObject()                             */
/*                                                                      */
/*      Read the vertices, parts, and other non-attribute information	*/
/*	for one shape.							*/
/************************************************************************/

SHPObject SHPAPI_CALL1(*) SHPReadObject( SHPHandle psSHP, int hEntity )
{
	SHPObject		*psShape;

/* -------------------------------------------------------------------- */
/*      Validate the record/entity number.                              */
/* -------------------------------------------------------------------- */
	if( hEntity < 0 || hEntity >= psSHP->nRecords )	return( NULL );

/* -------------------------------------------------------------------- */
/*      Ensure our record buffer is large enough.                       */
/* -------------------------------------------------------------------- */
	if( psSHP->panRecSize[hEntity]+8 > psSHP->nBufSize )
	{
		psSHP->nBufSize = psSHP->panRecSize[hEntity]+8;
		psSHP->pabyRec = SfByteRealloc(psSHP->pabyRec, psSHP->nBufSize);
    }

/* -------------------------------------------------------------------- */
/*      Read the record.                                                */
/* -------------------------------------------------------------------- */
    fseek( fpSHP, psSHP->panRecOffset[hEntity], 0 );
	fread( psSHP->pabyRec, psSHP->panRecSize[hEntity]+8, 1, fpSHP );

/* -------------------------------------------------------------------- */
/*	Allocate and minimally initialize the object.			*/
/* -------------------------------------------------------------------- */
	psShape = SfSHPObjectRealloc(0, 1);
    psShape->nShapeId = hEntity;

    memcpy( &psShape->nSHPType, psSHP->pabyRec + 8, 4 );
    if( bBigEndian ) Swap4Bytes(&(psShape->nSHPType) );

/* ==================================================================== */
/*  Extract vertices for a Polygon or Arc.				*/
/* ==================================================================== */
    if( psShape->nSHPType == SHPT_POLYGON || psShape->nSHPType == SHPT_ARC
        || psShape->nSHPType == SHPT_POLYGONZ
        || psShape->nSHPType == SHPT_POLYGONM
        || psShape->nSHPType == SHPT_ARCZ
        || psShape->nSHPType == SHPT_ARCM
        || psShape->nSHPType == SHPT_MULTIPATCH )
    {
	int32		nPoints, nParts;
	int    		i, nOffset;

/* -------------------------------------------------------------------- */
/*	Get the X/Y bounds.						*/
/* -------------------------------------------------------------------- */
        memcpy( &(psShape->dfXMin), psSHP->pabyRec + 8 +  4, 8 );
        memcpy( &(psShape->dfYMin), psSHP->pabyRec + 8 + 12, 8 );
        memcpy( &(psShape->dfXMax), psSHP->pabyRec + 8 + 20, 8 );
        memcpy( &(psShape->dfYMax), psSHP->pabyRec + 8 + 28, 8 );

	if( bBigEndian ) Swap8Bytes(&(psShape->dfXMin) );
	if( bBigEndian ) Swap8Bytes(&(psShape->dfYMin) );
	if( bBigEndian ) Swap8Bytes(&(psShape->dfXMax) );
	if( bBigEndian ) Swap8Bytes(&(psShape->dfYMax) );

/* -------------------------------------------------------------------- */
/*      Extract part/point count, and build vertex and part arrays      */
/*      to proper size.                                                 */
/* -------------------------------------------------------------------- */
	memcpy( &nPoints, psSHP->pabyRec + 40 + 8, 4 );
	memcpy( &nParts, psSHP->pabyRec + 36 + 8, 4 );

	if( bBigEndian ) Swap4Bytes(&nPoints );
	if( bBigEndian ) Swap4Bytes(&nParts );

	psShape->nVertices = nPoints;
	psShape->padfX = SfDoubleRealloc(0, nPoints);
	psShape->padfY = SfDoubleRealloc(0, nPoints);
	psShape->padfZ = SfDoubleRealloc(0, nPoints);
	psShape->padfM = SfDoubleRealloc(0, nPoints);

	psShape->nParts = nParts;
	psShape->panPartStart = SfIntRealloc(0, nParts);
	psShape->panPartType = SfIntRealloc(0, nParts);

    for( i = 0; i < nParts; i++ ) psShape->panPartType[i] = SHPP_RING;

/* -------------------------------------------------------------------- */
/*      Copy out the part array from the record.                        */
/* -------------------------------------------------------------------- */
	memcpy( psShape->panPartStart, psSHP->pabyRec + 44 + 8, 4 * nParts );
	for( i = 0; i < nParts; i++ )
	{
	    if( bBigEndian ) Swap4Bytes(psShape->panPartStart+i );
	}

	nOffset = 44 + 8 + 4*nParts;

/* -------------------------------------------------------------------- */
/*      If this is a multipatch, we will also have parts types.         */
/* -------------------------------------------------------------------- */
        if( psShape->nSHPType == SHPT_MULTIPATCH )
        {
            memcpy( psShape->panPartType, psSHP->pabyRec + nOffset, 4*nParts );
            for( i = 0; i < nParts; i++ )
            {
                if( bBigEndian ) Swap4Bytes(psShape->panPartType+i );
            }

            nOffset += 4*nParts;
        }
        
/* -------------------------------------------------------------------- */
/*      Copy out the vertices from the record.                          */
/* -------------------------------------------------------------------- */
	for( i = 0; i < nPoints; i++ )
	{
	    memcpy(psShape->padfX + i,
		   psSHP->pabyRec + nOffset + i * 16,
		   8 );

	    memcpy(psShape->padfY + i,
		   psSHP->pabyRec + nOffset + i * 16 + 8,
		   8 );

	    if( bBigEndian ) Swap8Bytes(psShape->padfX + i );
	    if( bBigEndian ) Swap8Bytes(psShape->padfY + i );
	}

        nOffset += 16*nPoints;
        
/* -------------------------------------------------------------------- */
/*      If we have a Z coordinate, collect that now.                    */
/* -------------------------------------------------------------------- */
        if( psShape->nSHPType == SHPT_POLYGONZ
            || psShape->nSHPType == SHPT_ARCZ
            || psShape->nSHPType == SHPT_MULTIPATCH )
        {
            memcpy( &(psShape->dfZMin), psSHP->pabyRec + nOffset, 8 );
            memcpy( &(psShape->dfZMax), psSHP->pabyRec + nOffset + 8, 8 );
            
            if( bBigEndian ) Swap8Bytes(&(psShape->dfZMin) );
            if( bBigEndian ) Swap8Bytes(&(psShape->dfZMax) );
            
            for( i = 0; i < nPoints; i++ )
            {
                memcpy( psShape->padfZ + i,
                        psSHP->pabyRec + nOffset + 16 + i*8, 8 );
                if( bBigEndian ) Swap8Bytes(psShape->padfZ + i );
            }

            nOffset += 16 + 8*nPoints;
        }

/* -------------------------------------------------------------------- */
/*      If we have a M measure value, then read it now.  We assume      */
/*      that the measure can be present for any shape if the size is    */
/*      big enough, but really it will only occur for the Z shapes      */
/*      (options), and the M shapes.                                    */
/* -------------------------------------------------------------------- */
        if( psSHP->panRecSize[hEntity]+8 >= nOffset + 16 + 8*nPoints )
        {
            memcpy( &(psShape->dfMMin), psSHP->pabyRec + nOffset, 8 );
            memcpy( &(psShape->dfMMax), psSHP->pabyRec + nOffset + 8, 8 );
            
            if( bBigEndian ) Swap8Bytes(&(psShape->dfMMin) );
            if( bBigEndian ) Swap8Bytes(&(psShape->dfMMax) );
            
            for( i = 0; i < nPoints; i++ )
            {
                memcpy( psShape->padfM + i,
                        psSHP->pabyRec + nOffset + 16 + i*8, 8 );
                if( bBigEndian ) Swap8Bytes(psShape->padfM + i );
            }
        }
        
    }

/* ==================================================================== */
/*  Extract vertices for a MultiPoint.					*/
/* ==================================================================== */
    else if( psShape->nSHPType == SHPT_MULTIPOINT
             || psShape->nSHPType == SHPT_MULTIPOINTM
             || psShape->nSHPType == SHPT_MULTIPOINTZ )
    {
	int32		nPoints;
	int    		i, nOffset;

	memcpy( &nPoints, psSHP->pabyRec + 44, 4 );
	if( bBigEndian ) Swap4Bytes(&nPoints );

	psShape->nVertices = nPoints;
	psShape->padfX = SfDoubleRealloc(0, nPoints);
	psShape->padfY = SfDoubleRealloc(0, nPoints);
	psShape->padfZ = SfDoubleRealloc(0, nPoints);
	psShape->padfM = SfDoubleRealloc(0, nPoints);

	for( i = 0; i < nPoints; i++ )
	{
		memcpy(psShape->padfX+i, psSHP->pabyRec + 48 + 16 * i, 8 );
		memcpy(psShape->padfY+i, psSHP->pabyRec + 48 + 16 * i + 8, 8 );

		if( bBigEndian ) Swap8Bytes(psShape->padfX + i );
		if( bBigEndian ) Swap8Bytes(psShape->padfY + i );
	}

    nOffset = 48 + 16*nPoints;
        
/* -------------------------------------------------------------------- */
/*	Get the X/Y bounds.						*/
/* -------------------------------------------------------------------- */
        memcpy( &(psShape->dfXMin), psSHP->pabyRec + 8 +  4, 8 );
        memcpy( &(psShape->dfYMin), psSHP->pabyRec + 8 + 12, 8 );
        memcpy( &(psShape->dfXMax), psSHP->pabyRec + 8 + 20, 8 );
        memcpy( &(psShape->dfYMax), psSHP->pabyRec + 8 + 28, 8 );

	if( bBigEndian ) Swap8Bytes(&(psShape->dfXMin) );
	if( bBigEndian ) Swap8Bytes(&(psShape->dfYMin) );
	if( bBigEndian ) Swap8Bytes(&(psShape->dfXMax) );
	if( bBigEndian ) Swap8Bytes(&(psShape->dfYMax) );

/* -------------------------------------------------------------------- */
/*      If we have a Z coordinate, collect that now.                    */
/* -------------------------------------------------------------------- */
        if( psShape->nSHPType == SHPT_MULTIPOINTZ )
        {
            memcpy( &(psShape->dfZMin), psSHP->pabyRec + nOffset, 8 );
            memcpy( &(psShape->dfZMax), psSHP->pabyRec + nOffset + 8, 8 );
            
            if( bBigEndian ) Swap8Bytes(&(psShape->dfZMin) );
            if( bBigEndian ) Swap8Bytes(&(psShape->dfZMax) );
            
            for( i = 0; i < nPoints; i++ )
            {
                memcpy( psShape->padfZ + i,
                        psSHP->pabyRec + nOffset + 16 + i*8, 8 );
                if( bBigEndian ) Swap8Bytes(psShape->padfZ + i );
            }

            nOffset += 16 + 8*nPoints;
        }

/* -------------------------------------------------------------------- */
/*      If we have a M measure value, then read it now.  We assume      */
/*      that the measure can be present for any shape if the size is    */
/*      big enough, but really it will only occur for the Z shapes      */
/*      (options), and the M shapes.                                    */
/* -------------------------------------------------------------------- */
        if( psSHP->panRecSize[hEntity]+8 >= nOffset + 16 + 8*nPoints )
        {
            memcpy( &(psShape->dfMMin), psSHP->pabyRec + nOffset, 8 );
            memcpy( &(psShape->dfMMax), psSHP->pabyRec + nOffset + 8, 8 );
            
            if( bBigEndian ) Swap8Bytes(&(psShape->dfMMin) );
            if( bBigEndian ) Swap8Bytes(&(psShape->dfMMax) );
            
            for( i = 0; i < nPoints; i++ )
            {
                memcpy( psShape->padfM + i,
                        psSHP->pabyRec + nOffset + 16 + i*8, 8 );
                if( bBigEndian ) Swap8Bytes(psShape->padfM + i );
            }
        }
    }

/* ==================================================================== */
/*      Extract vertices for a point.                                   */
/* ==================================================================== */
    else if( psShape->nSHPType == SHPT_POINT
             || psShape->nSHPType == SHPT_POINTM
             || psShape->nSHPType == SHPT_POINTZ )
    {
        int	nOffset;
        
	psShape->nVertices = 1;
	psShape->padfX = SfDoubleRealloc(0, 1);
	psShape->padfY = SfDoubleRealloc(0, 1);
	psShape->padfZ = SfDoubleRealloc(0, 1);
	psShape->padfM = SfDoubleRealloc(0, 1);

	memcpy( psShape->padfX, psSHP->pabyRec + 12, 8 );
	memcpy( psShape->padfY, psSHP->pabyRec + 20, 8 );

	if( bBigEndian ) Swap8Bytes(psShape->padfX );
	if( bBigEndian ) Swap8Bytes(psShape->padfY );

    nOffset = 20 + 8;
        
/* -------------------------------------------------------------------- */
/*      If we have a Z coordinate, collect that now.                    */
/* -------------------------------------------------------------------- */
        if( psShape->nSHPType == SHPT_POINTZ )
        {
            memcpy( psShape->padfZ, psSHP->pabyRec + nOffset, 8 );
        
            if( bBigEndian ) Swap8Bytes(psShape->padfZ );
            
            nOffset += 8;
        }

/* -------------------------------------------------------------------- */
/*      If we have a M measure value, then read it now.  We assume      */
/*      that the measure can be present for any shape if the size is    */
/*      big enough, but really it will only occur for the Z shapes      */
/*      (options), and the M shapes.                                    */
/* -------------------------------------------------------------------- */
        if( psSHP->panRecSize[hEntity]+8 >= nOffset + 8 )
        {
            memcpy( psShape->padfM, psSHP->pabyRec + nOffset, 8 );
        
            if( bBigEndian ) Swap8Bytes(psShape->padfM );
        }

/* -------------------------------------------------------------------- */
/*      Since no extents are supplied in the record, we will apply      */
/*      them from the single vertex.                                    */
/* -------------------------------------------------------------------- */
        psShape->dfXMin = psShape->dfXMax = psShape->padfX[0];
        psShape->dfYMin = psShape->dfYMax = psShape->padfY[0];
        psShape->dfZMin = psShape->dfZMax = psShape->padfZ[0];
        psShape->dfMMin = psShape->dfMMax = psShape->padfM[0];
    }

    return( psShape );
}

/************************************************************************/
/*                            SHPTypeName()                             */
/************************************************************************/

const char SHPAPI_CALL1(*)
SHPTypeName( int nSHPType )

{
    switch( nSHPType )
    {
      case SHPT_NULL:
        return "NullShape";

      case SHPT_POINT:
        return "Point";

      case SHPT_ARC:
        return "Arc";

      case SHPT_POLYGON:
        return "Polygon";

      case SHPT_MULTIPOINT:
        return "MultiPoint";
        
      case SHPT_POINTZ:
        return "PointZ";

      case SHPT_ARCZ:
        return "ArcZ";

      case SHPT_POLYGONZ:
        return "PolygonZ";

      case SHPT_MULTIPOINTZ:
        return "MultiPointZ";
        
      case SHPT_POINTM:
        return "PointM";

      case SHPT_ARCM:
        return "ArcM";

      case SHPT_POLYGONM:
        return "PolygonM";

      case SHPT_MULTIPOINTM:
        return "MultiPointM";

      case SHPT_MULTIPATCH:
        return "MultiPatch";

      default:
        return "UnknownShapeType";
    }
}

/************************************************************************/
/*                          SHPPartTypeName()                           */
/************************************************************************/

const char SHPAPI_CALL1(*)
SHPPartTypeName( int nPartType )

{
    switch( nPartType )
    {
      case SHPP_TRISTRIP:
        return "TriangleStrip";
        
      case SHPP_TRIFAN:
        return "TriangleFan";

      case SHPP_OUTERRING:
        return "OuterRing";

      case SHPP_INNERRING:
        return "InnerRing";

      case SHPP_FIRSTRING:
        return "FirstRing";

      case SHPP_RING:
        return "Ring";

      default:
        return "UnknownPartType";
    }
}

/************************************************************************/
/*                          SHPDestroyObject()                          */
/************************************************************************/

void SHPAPI_CALL SHPDestroyObject( SHPObject * psShape )
{
	if( psShape == NULL ) return;
    
	if( psShape->padfX != NULL )		SfDoubleFree( psShape->padfX );
	if( psShape->padfY != NULL )        SfDoubleFree( psShape->padfY );
	if( psShape->padfZ != NULL )        SfDoubleFree( psShape->padfZ );
	if( psShape->padfM != NULL )        SfDoubleFree( psShape->padfM );

	if( psShape->panPartStart != NULL )        SfIntFree( psShape->panPartStart );
	if( psShape->panPartType != NULL )         SfIntFree( psShape->panPartType );

	SfSHPObjectFree( psShape );
}

/************************************************************************/
/*                          SHPRewindObject()                           */
/*                                                                      */
/*      Reset the winding of polygon objects to adhere to the           */
/*      specification.                                                  */
/************************************************************************/
int SHPAPI_CALL SHPRewindObject(SHPHandle, SHPObject * psObject)
{
	int  iOpRing, bAltered = 0;

	/* -------------------------------------------------------------------- */
	/*      Do nothing if this is not a polygon object.                     */
	/* -------------------------------------------------------------------- */
	if (psObject->nSHPType != SHPT_POLYGON && psObject->nSHPType != SHPT_POLYGONZ && psObject->nSHPType != SHPT_POLYGONM)
	{
		return(0);
	}

	/* -------------------------------------------------------------------- */
	/*      Process each of the rings.                                      */
	/* -------------------------------------------------------------------- */
	for (iOpRing = 0; iOpRing < psObject->nParts; iOpRing++)
	{
		int      iVert, nVertCount, nVertStart, iCheckRing;
		double   dfSum;

/* -------------------------------------------------------------------- */
/*      Determine if this ring is an inner ring or an outer ring        */
/*      relative to all the other rings.  For now we assume the         */
/*      first ring is outer and all others are inner, but eventually    */
/*      we need to fix this to handle multiple island polygons and      */
/*      unordered sets of rings.                                        */
/* -------------------------------------------------------------------- */
		double dfTestX = psObject->padfX[psObject->panPartStart[iOpRing]];
		double dfTestY = psObject->padfY[psObject->panPartStart[iOpRing]];

		int bInner = FALSE;
        for( iCheckRing = 0; iCheckRing < psObject->nParts; iCheckRing++ )
        {
			if (iCheckRing == iOpRing)	continue;
            
            nVertStart = psObject->panPartStart[iCheckRing];

			if( iCheckRing == psObject->nParts-1 )
			{
				nVertCount = psObject->nVertices - psObject->panPartStart[iCheckRing];
			}
			else
			{
				nVertCount = psObject->panPartStart[iCheckRing+1] - psObject->panPartStart[iCheckRing];
			}

			for (int iEdge = 0; iEdge < nVertCount; iEdge++)
            {
                int iNext;

                if( iEdge < nVertCount-1 )
					iNext = iEdge+1;
                else
                    iNext = 0;

                if( (psObject->padfY[iEdge+nVertStart] < dfTestY 
                     && psObject->padfY[iNext+nVertStart] >= dfTestY)
                    || (psObject->padfY[iNext+nVertStart] < dfTestY 
                        && psObject->padfY[iEdge+nVertStart] >= dfTestY) )
                {
                    if( psObject->padfX[iEdge+nVertStart] 
                        + (dfTestY - psObject->padfY[iEdge+nVertStart])
                           / (psObject->padfY[iNext+nVertStart]
                              - psObject->padfY[iEdge+nVertStart])
                           * (psObject->padfX[iNext+nVertStart]
                              - psObject->padfX[iEdge+nVertStart]) < dfTestX )
                        bInner = !bInner;
                }
            }
        }

/* -------------------------------------------------------------------- */
/*      Determine the current order of this ring so we will know if     */
/*      it has to be reversed.                                          */
/* -------------------------------------------------------------------- */
		nVertStart = psObject->panPartStart[iOpRing];

        if( iOpRing == psObject->nParts-1 )
            nVertCount = psObject->nVertices - psObject->panPartStart[iOpRing];
        else
            nVertCount = psObject->panPartStart[iOpRing+1] 
                - psObject->panPartStart[iOpRing];

        dfSum = 0.0;
        for( iVert = nVertStart; iVert < nVertStart+nVertCount-1; iVert++ )
        {
			dfSum += psObject->padfX[iVert] * psObject->padfY[iVert+1]
                - psObject->padfY[iVert] * psObject->padfX[iVert+1];
        }

        dfSum += psObject->padfX[iVert] * psObject->padfY[nVertStart]
               - psObject->padfY[iVert] * psObject->padfX[nVertStart];

/* -------------------------------------------------------------------- */
/*      Reverse if necessary.                                           */
/* -------------------------------------------------------------------- */
        if( (dfSum < 0.0 && bInner) || (dfSum > 0.0 && !bInner) )
        {
            int   i;

			bAltered++;
            for( i = 0; i < nVertCount/2; i++ )
            {
                double dfSaved;

                /* Swap X */
                dfSaved = psObject->padfX[nVertStart+i];
                psObject->padfX[nVertStart+i] = 
					psObject->padfX[nVertStart+nVertCount-i-1];
				psObject->padfX[nVertStart+nVertCount-i-1] = dfSaved;

				/* Swap Y */
				dfSaved = psObject->padfY[nVertStart+i];
				psObject->padfY[nVertStart+i] =
					psObject->padfY[nVertStart+nVertCount-i-1];
				psObject->padfY[nVertStart+nVertCount-i-1] = dfSaved;

				/* Swap Z */
				if( psObject->padfZ )
                {
                    dfSaved = psObject->padfZ[nVertStart+i];
                    psObject->padfZ[nVertStart+i] = 
                        psObject->padfZ[nVertStart+nVertCount-i-1];
                    psObject->padfZ[nVertStart+nVertCount-i-1] = dfSaved;
                }

				/* Swap M */
				if( psObject->padfM )
				{
					dfSaved = psObject->padfM[nVertStart+i];
					psObject->padfM[nVertStart+i] =
						psObject->padfM[nVertStart+nVertCount-i-1];
					psObject->padfM[nVertStart+nVertCount-i-1] = dfSaved;
				}
			}
		}
	}

	return bAltered;
}


static int	nStringFieldLen = 0;
static char * pszStringField = NULL;


/************************************************************************/
/*                           DBFWriteHeader()                           */
/*                                                                      */
/*      This is called to write out the file header, and field          */
/*      descriptions before writing any actual data records.  This      */
/*      also computes all the DBFDataSet field offset/size/decimals     */
/*      and so forth values.                                            */
/************************************************************************/
static void DBFWriteHeader(DBFHandle psDBF)

{
	uchar	abyHeader[XBASE_FLDHDR_SZ];
	int		i;

	if (!psDBF->bNoHeader)	return;

    psDBF->bNoHeader = FALSE;

	/* -------------------------------------------------------------------- */
	/*	Initialize the file header information.								*/
	/* -------------------------------------------------------------------- */
	for (i = 0; i < XBASE_FLDHDR_SZ; i++)	abyHeader[i] = 0;

    abyHeader[0] = 0x03;		/* memo field? - just copying 	*/

	/* date updated on close, record count preset at zero */

	abyHeader[8] = (uchar)(psDBF->nHeaderLength % 256);
	abyHeader[9] = (uchar)(psDBF->nHeaderLength / 256);

	abyHeader[10] = (uchar)(psDBF->nRecordLength % 256);
	abyHeader[11] = (uchar)(psDBF->nRecordLength / 256);

	/* -------------------------------------------------------------------- */
	/*      Write the initial 32 byte file header, and all the field        */
	/*      descriptions.                                     				*/
	/* -------------------------------------------------------------------- */
    fseek( psDBF->fp, 0, 0 );
    fwrite( abyHeader, XBASE_FLDHDR_SZ, 1, psDBF->fp );
    fwrite( psDBF->pszHeader, XBASE_FLDHDR_SZ, psDBF->nFields, psDBF->fp );

	/* -------------------------------------------------------------------- */
	/*      Write out the newline character if there is room for it.        */
	/* -------------------------------------------------------------------- */
    if( psDBF->nHeaderLength > 32*psDBF->nFields + 32 )
    {
        char	cNewline;

        cNewline = 0x0d;
        fwrite( &cNewline, 1, 1, psDBF->fp );
    }
}

/************************************************************************/
/*                           DBFFlushRecord()                           */
/*                                                                      */
/*      Write out the current record if there is one.                   */
/************************************************************************/

static void DBFFlushRecord( DBFHandle psDBF )

{
    int		nRecordOffset;

    if( psDBF->bCurrentRecordModified && psDBF->nCurrentRecord > -1 )
    {
	psDBF->bCurrentRecordModified = FALSE;

	nRecordOffset = psDBF->nRecordLength * psDBF->nCurrentRecord 
	                                             + psDBF->nHeaderLength;

	fseek( psDBF->fp, nRecordOffset, 0 );
	fwrite( psDBF->pszCurrentRecord, psDBF->nRecordLength, 1, psDBF->fp );
    }
}

/************************************************************************/
/*                              DBFOpen()                               */
/*                                                                      */
/*      Open a .dbf file.                                               */
/************************************************************************/
   
DBFHandle SHPAPI_CALL
DBFOpen( const char * pszFilename, const char * pszAccess )

{
    DBFHandle		psDBF;
    unsigned char		*pabyBuf;
    int			nFields, nHeadLen, nRecLen, iField, i;
    char		*pszBasename, *pszFullname;

/* -------------------------------------------------------------------- */
/*      We only allow the access strings "rb" and "r+".                  */
/* -------------------------------------------------------------------- */
    if( strcmp(pszAccess,"r") != 0 && strcmp(pszAccess,"r+") != 0 
        && strcmp(pszAccess,"rb") != 0 && strcmp(pszAccess,"rb+") != 0
        && strcmp(pszAccess,"r+b") != 0 )
        return( NULL );

    if( strcmp(pszAccess,"r") == 0 )
        pszAccess = "rb";
 
    if( strcmp(pszAccess,"r+") == 0 )
        pszAccess = "rb+";

/* -------------------------------------------------------------------- */
/*	Compute the base (layer) name.  If there is any extension	*/
/*	on the passed in filename we will strip it off.			*/
/* -------------------------------------------------------------------- */
	pszBasename = (char *) SfByteRealloc(0, strlen(pszFilename)+5);
	strcpy( pszBasename, pszFilename );
    for( i = strlen(pszBasename)-1; 
	 i > 0 && pszBasename[i] != '.' && pszBasename[i] != '/'
	       && pszBasename[i] != '\\';
	 i-- ) {}

    if( pszBasename[i] == '.' )
        pszBasename[i] = '\0';

	pszFullname = (char *) SfByteRealloc(0, strlen(pszBasename) + 5);
	sprintf( pszFullname, "%s.dbf", pszBasename );

	psDBF = (DBFHandle) SfDBFInfoRealloc(0, 1);
	psDBF->fp = fopen( pszFullname, pszAccess );

	if( psDBF->fp == NULL )
	{
        sprintf( pszFullname, "%s.DBF", pszBasename );
        psDBF->fp = fopen(pszFullname, pszAccess );
    }
    
	SfByteFree( (uchar*)pszBasename );
	SfByteFree( (uchar*)pszFullname );
    
	if( psDBF->fp == NULL )
    {
		SfDBFInfoFree( psDBF );
		return( NULL );
	}

	psDBF->bNoHeader = FALSE;
    psDBF->nCurrentRecord = -1;
    psDBF->bCurrentRecordModified = FALSE;

/* -------------------------------------------------------------------- */
/*  Read Table Header info                                              */
/* -------------------------------------------------------------------- */
	pabyBuf = SfByteRealloc(0, 500);
	if( fread( pabyBuf, 32, 1, psDBF->fp ) != 1 )
	{
		fclose( psDBF->fp );
		SfByteFree( pabyBuf );
		SfDBFInfoFree( psDBF );
        return NULL;
    }

	psDBF->nRecords = pabyBuf[4] + pabyBuf[5]*256 + pabyBuf[6]*256*256 + pabyBuf[7]*256*256*256;

	psDBF->nHeaderLength = nHeadLen = pabyBuf[8] + pabyBuf[9]*256;
    psDBF->nRecordLength = nRecLen = pabyBuf[10] + pabyBuf[11]*256;
    
    psDBF->nFields = nFields = (nHeadLen - 32) / 32;

	psDBF->pszCurrentRecord = (char *) SfByteRealloc(0, nRecLen);

/* -------------------------------------------------------------------- */
/*  Read in Field Definitions                                           */
/* -------------------------------------------------------------------- */
    
	pabyBuf = SfByteRealloc(pabyBuf, nHeadLen);
    psDBF->pszHeader = (char *) pabyBuf;

    fseek( psDBF->fp, 32, 0 );
    if( fread( pabyBuf, nHeadLen-32, 1, psDBF->fp ) != 1 )
    {
		fclose( psDBF->fp );
		SfByteFree( pabyBuf );
		SfDBFInfoFree( psDBF );
        return NULL;
    }

	psDBF->panFieldOffset = SfIntRealloc(0, nFields);
	psDBF->panFieldSize = SfIntRealloc(0, nFields);
	psDBF->panFieldDecimals = SfIntRealloc(0, nFields);
	psDBF->pachFieldType = (char *)SfByteRealloc(0, nFields);

	for( iField = 0; iField < nFields; iField++ )
	{
		unsigned char		*pabyFInfo;

		pabyFInfo = pabyBuf+iField*32;

		if( pabyFInfo[11] == 'N' || pabyFInfo[11] == 'F' )
		{
			psDBF->panFieldSize[iField] = pabyFInfo[16];
			psDBF->panFieldDecimals[iField] = pabyFInfo[17];
		}
		else
		{
			psDBF->panFieldSize[iField] = pabyFInfo[16] + pabyFInfo[17]*256;
			psDBF->panFieldDecimals[iField] = 0;
		}

		psDBF->pachFieldType[iField] = (char) pabyFInfo[11];

		if( iField == 0 )	psDBF->panFieldOffset[iField] = 1;
		else             	psDBF->panFieldOffset[iField] = psDBF->panFieldOffset[iField-1] + psDBF->panFieldSize[iField-1];
	}

	return( psDBF );
}

/************************************************************************/
/*                              DBFClose()                              */
/************************************************************************/

void SHPAPI_CALL DBFClose(DBFHandle psDBF)
{
/* -------------------------------------------------------------------- */
/*      Write out header if not already written.                        */
/* -------------------------------------------------------------------- */
	if( psDBF->bNoHeader )
		DBFWriteHeader( psDBF );

    DBFFlushRecord( psDBF );

/* -------------------------------------------------------------------- */
/*      Update last access date, and number of records if we have	*/
/*	write access.                					*/
/* -------------------------------------------------------------------- */
    if( psDBF->bUpdated )
    {
	unsigned char		abyFileHeader[32];

	fseek( psDBF->fp, 0, 0 );
	fread( abyFileHeader, 32, 1, psDBF->fp );

	abyFileHeader[1] = 95;			/* YY */
	abyFileHeader[2] = 7;			/* MM */
	abyFileHeader[3] = 26;			/* DD */

	abyFileHeader[4] = (uchar)((psDBF->nRecords) % 256);
	abyFileHeader[5] = (uchar)((psDBF->nRecords / 256) % 256);
	abyFileHeader[6] = (uchar)((psDBF->nRecords / (256*256)) % 256);
	abyFileHeader[7] = (uchar)((psDBF->nRecords / (256*256*256)) % 256);

	fseek( psDBF->fp, 0, 0 );
	fwrite( abyFileHeader, 32, 1, psDBF->fp );
    }

/* -------------------------------------------------------------------- */
/*      Close, and free resources.                                      */
/* -------------------------------------------------------------------- */
    fclose( psDBF->fp );

    if( psDBF->panFieldOffset != NULL )
    {
		SfIntFree( psDBF->panFieldOffset );
		SfIntFree( psDBF->panFieldSize );
		SfIntFree( psDBF->panFieldDecimals );
		SfByteFree( (uchar*)psDBF->pachFieldType );
    }

	SfByteFree( (uchar*)psDBF->pszHeader );
	SfByteFree( (uchar*)psDBF->pszCurrentRecord );

	SfDBFInfoFree( psDBF );

    if( pszStringField != NULL )
	{
		SfByteFree( (uchar*)pszStringField );
        pszStringField = NULL;
        nStringFieldLen = 0;
    }
}

/************************************************************************/
/*                             DBFCreate()                              */
/*                                                                      */
/*      Create a new .dbf file.                                         */
/************************************************************************/

DBFHandle SHPAPI_CALL DBFCreate( const char * pszFilename )
{
	DBFHandle	psDBF;
	char	*pszFullname, *pszBasename;
	int		i;

/* -------------------------------------------------------------------- */
/*	Compute the base (layer) name.  If there is any extension	*/
/*	on the passed in filename we will strip it off.			*/
/* -------------------------------------------------------------------- */
	pszBasename = (char*)SfByteRealloc(0, strlen(pszFilename)+5);
	strcpy( pszBasename, pszFilename );
	for( i = strlen(pszBasename)-1; i > 0 && pszBasename[i] != '.' && pszBasename[i] != '/' && pszBasename[i] != '\\'; i-- ) {}

	if( pszBasename[i] == '.' ) pszBasename[i] = '\0';

	pszFullname = (char *) SfByteRealloc(0, strlen(pszBasename) + 5);
	sprintf( pszFullname, "%s.dbf", pszBasename );
	SfByteFree( (uchar*)pszBasename );

/* -------------------------------------------------------------------- */
/*      Create the file.                                                */
/* -------------------------------------------------------------------- */
	FILE* fp = fopen (pszFullname, "wb");
	if (fp == NULL) return(NULL);

	fputc(0, fp);
	fclose(fp);

	fp = fopen(pszFullname, "rb+");
	if (fp == NULL) return(NULL);

	SfByteFree((uchar*)pszFullname);

/* -------------------------------------------------------------------- */
/*	Create the info structure.					*/
/* -------------------------------------------------------------------- */
	psDBF = SfDBFInfoRealloc(0, 1);

	psDBF->fp = fp;
    psDBF->nRecords = 0;
    psDBF->nFields = 0;
    psDBF->nRecordLength = 1;
    psDBF->nHeaderLength = 33;
    
    psDBF->panFieldOffset = NULL;
    psDBF->panFieldSize = NULL;
    psDBF->panFieldDecimals = NULL;
    psDBF->pachFieldType = NULL;
    psDBF->pszHeader = NULL;

    psDBF->nCurrentRecord = -1;
    psDBF->bCurrentRecordModified = FALSE;
    psDBF->pszCurrentRecord = NULL;

    psDBF->bNoHeader = TRUE;

    return( psDBF );
}

/************************************************************************/
/*                            DBFAddField()                             */
/*                                                                      */
/*      Add a field to a newly created .dbf file before any records     */
/*      are written.                                                    */
/************************************************************************/

int SHPAPI_CALL DBFAddField(DBFHandle psDBF, const char *pszFieldName, DBFFieldType eType, int nWidth, int nDecimals )
{
    char	*pszFInfo;
	int		i;

/* -------------------------------------------------------------------- */
/*      Do some checking to ensure we can add records to this file.     */
/* -------------------------------------------------------------------- */
	if( psDBF->nRecords > 0 ) return( -1 );

	if( !psDBF->bNoHeader ) return( -1 );

	if( eType != FTDouble && nDecimals != 0 ) return( -1 );

	if( nWidth < 1 ) return -1;

/* -------------------------------------------------------------------- */
/*      SfRealloc all the arrays larger to hold the additional field      */
/*      information.                                                    */
/* -------------------------------------------------------------------- */
    psDBF->nFields++;

	psDBF->panFieldOffset = SfIntRealloc(psDBF->panFieldOffset, psDBF->nFields);
	psDBF->panFieldSize = SfIntRealloc(psDBF->panFieldSize, psDBF->nFields);
	psDBF->panFieldDecimals = SfIntRealloc(psDBF->panFieldDecimals, psDBF->nFields);
	psDBF->pachFieldType = (char *)SfByteRealloc((uchar*)psDBF->pachFieldType, psDBF->nFields);

/* -------------------------------------------------------------------- */
/*      Assign the new field information fields.                        */
/* -------------------------------------------------------------------- */
    psDBF->panFieldOffset[psDBF->nFields-1] = psDBF->nRecordLength;
    psDBF->nRecordLength += nWidth;
    psDBF->panFieldSize[psDBF->nFields-1] = nWidth;
    psDBF->panFieldDecimals[psDBF->nFields-1] = nDecimals;

	if( eType == FTLogical )        psDBF->pachFieldType[psDBF->nFields-1] = 'L';
	else if( eType == FTString )    psDBF->pachFieldType[psDBF->nFields-1] = 'C';
    else                            psDBF->pachFieldType[psDBF->nFields-1] = 'N';

/* -------------------------------------------------------------------- */
/*      Extend the required header information.                         */
/* -------------------------------------------------------------------- */
	psDBF->nHeaderLength += 32;
    psDBF->bUpdated = FALSE;

	psDBF->pszHeader = (char *)SfByteRealloc((uchar*)psDBF->pszHeader,psDBF->nFields*32);

    pszFInfo = psDBF->pszHeader + 32 * (psDBF->nFields-1);

	for( i = 0; i < 32; i++ )  pszFInfo[i] = '\0';

	if( (int) strlen(pszFieldName) < 10 )          strncpy( pszFInfo, pszFieldName, strlen(pszFieldName));
	else                                           strncpy( pszFInfo, pszFieldName, 10);

    pszFInfo[11] = psDBF->pachFieldType[psDBF->nFields-1];

	if (eType == FTString)
	{
		pszFInfo[16] = (uchar)(nWidth % 256);
		pszFInfo[17] = (uchar)(nWidth / 256);
	}
	else
	{
		pszFInfo[16] = (uchar)(nWidth);
        pszFInfo[17] = (uchar)(nDecimals);
    }
    
	/* -------------------------------------------------------------------- */
	/*      Make the current record buffer appropriately larger.            */
	/* -------------------------------------------------------------------- */
	psDBF->pszCurrentRecord = (char *) SfByteRealloc((uchar*)psDBF->pszCurrentRecord, psDBF->nRecordLength);

    return( psDBF->nFields-1 );
}

/************************************************************************/
/*                          DBFReadAttribute()                          */
/*                                                                      */
/*      Read one of the attribute fields of a record.                   */
/************************************************************************/
static void *DBFReadAttribute(DBFHandle psDBF, int hEntity, int iField, char chReqType)
{
	int		nRecordOffset;
	uchar	*pabyRec;
	void	*pReturnField = NULL;

	static double dDoubleField;

/* -------------------------------------------------------------------- */
/*      Verify selection.                                               */
/* -------------------------------------------------------------------- */
	if (hEntity < 0 || hEntity >= psDBF->nRecords) return(NULL);
	if (iField < 0 || iField >= psDBF->nFields) return(NULL);

/* -------------------------------------------------------------------- */
/*	Have we read the record?					*/
/* -------------------------------------------------------------------- */
    if( psDBF->nCurrentRecord != hEntity )
	{
		DBFFlushRecord(psDBF);

		nRecordOffset = psDBF->nRecordLength * hEntity + psDBF->nHeaderLength;

		if(fseek( psDBF->fp, nRecordOffset, 0 ) != 0)
		{
			fprintf( stderr, "fseek(%d) failed on DBF file.\n", nRecordOffset );
			return NULL;
		}

		if (fread( psDBF->pszCurrentRecord, psDBF->nRecordLength, 1, psDBF->fp ) != 1)
		{
			fprintf (stderr, "fread(%d) failed on DBF file.\n", psDBF->nRecordLength);
			return NULL;
		}

		psDBF->nCurrentRecord = hEntity;
	}

	pabyRec = (unsigned char *) psDBF->pszCurrentRecord;

	/* -------------------------------------------------------------------- */
	/*	Ensure our field buffer is large enough to hold this buffer.		*/
	/* -------------------------------------------------------------------- */
	if (psDBF->panFieldSize[iField]+1 > nStringFieldLen)
	{
		nStringFieldLen = psDBF->panFieldSize[iField]*2 + 10;
		pszStringField = (char *) SfByteRealloc((uchar*)pszStringField,nStringFieldLen);
	}

	/* -------------------------------------------------------------------- */
	/*	Extract the requested field.										*/
	/* -------------------------------------------------------------------- */
	strncpy(pszStringField, ((const char *) pabyRec) + psDBF->panFieldOffset[iField], psDBF->panFieldSize[iField]);
	pszStringField[psDBF->panFieldSize[iField]] = '\0';

	pReturnField = pszStringField;

	/* -------------------------------------------------------------------- */
	/*      Decode the field.                                               */
	/* -------------------------------------------------------------------- */
	if (chReqType == 'N')
	{
		dDoubleField = atof(pszStringField);
		pReturnField = &dDoubleField;
	}

	/* -------------------------------------------------------------------- */
	/*      Should we trim white space off the string attribute value?      */
	/* -------------------------------------------------------------------- */
#ifdef TRIM_DBF_WHITESPACE
	else
	{
		char	*pchSrc, *pchDst;

		pchDst = pchSrc = pszStringField;
		while (*pchSrc == ' ') pchSrc++;

		while (*pchSrc != '\0') *(pchDst++) = *(pchSrc++);
		*pchDst = '\0';

		while (pchDst != pszStringField && *(--pchDst) == ' ') *pchDst = '\0';
	}
#endif

	return (pReturnField);
}

/************************************************************************/
/*                        DBFReadIntAttribute()                         */
/*                                                                      */
/*      Read an integer attribute.                                      */
/************************************************************************/

int SHPAPI_CALL DBFReadIntegerAttribute(DBFHandle psDBF, int iRecord, int iField)
{
	double	*pdValue;

	pdValue = (double *) DBFReadAttribute(psDBF, iRecord, iField, 'N');

	if (pdValue == NULL) return 0;
	else				 return((int)*pdValue);
}

/************************************************************************/
/*                        DBFReadDoubleAttribute()                      */
/*                                                                      */
/*      Read a double attribute.                                        */
/************************************************************************/
double SHPAPI_CALL DBFReadDoubleAttribute(DBFHandle psDBF, int iRecord, int iField)
{
	double	*pdValue;

	pdValue = (double *)DBFReadAttribute(psDBF, iRecord, iField, 'N');

	if (pdValue == NULL) return(0.0);
	else                 return(*pdValue);
}

/************************************************************************/
/*                        DBFReadStringAttribute()                      */
/*                                                                      */
/*      Read a string attribute.                                        */
/************************************************************************/

const char SHPAPI_CALL1(*)
DBFReadStringAttribute( DBFHandle psDBF, int iRecord, int iField )

{
	return( (const char *) DBFReadAttribute( psDBF, iRecord, iField, 'C' ) );
}

/************************************************************************/
/*                        DBFReadLogicalAttribute()                     */
/*                                                                      */
/*      Read a logical attribute.                                       */
/************************************************************************/

const char SHPAPI_CALL1(*)
DBFReadLogicalAttribute( DBFHandle psDBF, int iRecord, int iField )

{
    return( (const char *) DBFReadAttribute( psDBF, iRecord, iField, 'L' ) );
}

/************************************************************************/
/*                         DBFIsAttributeNULL()                         */
/*                                                                      */
/*      Return TRUE if value for field is NULL.                         */
/*                                                                      */
/*      Contributed by Jim Matthews.                                    */
/************************************************************************/

int SHPAPI_CALL
DBFIsAttributeNULL( DBFHandle psDBF, int iRecord, int iField )

{
    const char	*pszValue;

    pszValue = DBFReadStringAttribute( psDBF, iRecord, iField );

	switch(psDBF->pachFieldType[iField])
    {
      case 'N':
      case 'F':
        /* NULL numeric fields have value "****************" */
		return pszValue[0] == '*';

	  case 'D':
        /* NULL date fields have value "00000000" */
        return strncmp(pszValue,"00000000",8) == 0;

      case 'L':
        /* NULL boolean fields have value "?" */ 
        return pszValue[0] == '?';

      default:
        /* empty string fields are considered NULL */
        return strlen(pszValue) == 0;
    }
}

/************************************************************************/
/*                          DBFGetFieldCount()                          */
/*                                                                      */
/*      Return the number of fields in this table.                      */
/************************************************************************/

int SHPAPI_CALL
DBFGetFieldCount( DBFHandle psDBF )

{
    return( psDBF->nFields );
}

/************************************************************************/
/*                         DBFGetRecordCount()                          */
/*                                                                      */
/*      Return the number of records in this table.                     */
/************************************************************************/

int SHPAPI_CALL
DBFGetRecordCount( DBFHandle psDBF )

{
    return( psDBF->nRecords );
}

/************************************************************************/
/*                          DBFGetFieldInfo()                           */
/*                                                                      */
/*      Return any requested information about the field.               */
/************************************************************************/

DBFFieldType SHPAPI_CALL
DBFGetFieldInfo( DBFHandle psDBF, int iField, char * pszFieldName,
                 int * pnWidth, int * pnDecimals )

{
    if( iField < 0 || iField >= psDBF->nFields )
        return( FTInvalid );

    if( pnWidth != NULL )
        *pnWidth = psDBF->panFieldSize[iField];

    if( pnDecimals != NULL )
        *pnDecimals = psDBF->panFieldDecimals[iField];

    if( pszFieldName != NULL )
    {
	int	i;

	strncpy( pszFieldName, (char *) psDBF->pszHeader+iField*32, 11 );
	pszFieldName[11] = '\0';
	for( i = 10; i > 0 && pszFieldName[i] == ' '; i-- )
	    pszFieldName[i] = '\0';
    }

    if ( psDBF->pachFieldType[iField] == 'L' )
	return( FTLogical);

    else if( psDBF->pachFieldType[iField] == 'N' 
             || psDBF->pachFieldType[iField] == 'F'
             || psDBF->pachFieldType[iField] == 'D' )
    {
	if( psDBF->panFieldDecimals[iField] > 0 )
	    return( FTDouble );
	else
	    return( FTInteger );
    }
    else
    {
	return( FTString );
    }
}

/************************************************************************/
/*                         DBFWriteAttribute()                          */
/*									*/
/*	Write an attribute record to the file.				*/
/************************************************************************/

static int DBFWriteAttribute(DBFHandle psDBF, int hEntity, int iField, void * pValue )

{
    int	       	nRecordOffset, i, j, nRetResult = TRUE;
	unsigned char	*pabyRec;
    char	szSField[400], szFormat[20];

/* -------------------------------------------------------------------- */
/*	Is this a valid record?						*/
/* -------------------------------------------------------------------- */
    if( hEntity < 0 || hEntity > psDBF->nRecords ) return( FALSE );

	if( psDBF->bNoHeader ) DBFWriteHeader(psDBF);

/* -------------------------------------------------------------------- */
/*      Is this a brand new record?                                     */
/* -------------------------------------------------------------------- */
    if( hEntity == psDBF->nRecords )
    {
		DBFFlushRecord( psDBF );

		psDBF->nRecords++;
		for( i = 0; i < psDBF->nRecordLength; i++ ) psDBF->pszCurrentRecord[i] = ' ';

		psDBF->nCurrentRecord = hEntity;
    }

/* -------------------------------------------------------------------- */
/*      Is this an existing record, but different than the last one     */
/*      we accessed?                                                    */
/* -------------------------------------------------------------------- */
    if( psDBF->nCurrentRecord != hEntity )
	{
		DBFFlushRecord( psDBF );

		nRecordOffset = psDBF->nRecordLength * hEntity + psDBF->nHeaderLength;

		fseek( psDBF->fp, nRecordOffset, 0 );
		fread( psDBF->pszCurrentRecord, psDBF->nRecordLength, 1, psDBF->fp );

		psDBF->nCurrentRecord = hEntity;
    }

    pabyRec = (unsigned char *) psDBF->pszCurrentRecord;

    psDBF->bCurrentRecordModified = TRUE;
    psDBF->bUpdated = TRUE;

/* -------------------------------------------------------------------- */
/*      Translate NULL value to valid DBF file representation.          */
/*                                                                      */
/*      Contributed by Jim Matthews.                                    */
/* -------------------------------------------------------------------- */
    if( pValue == NULL )
    {
        switch(psDBF->pachFieldType[iField])
        {
          case 'N':
          case 'F':
	    /* NULL numeric fields have value "****************" */
			memset( (char *) (pabyRec+psDBF->panFieldOffset[iField]), '*', psDBF->panFieldSize[iField] );
			break;

		  case 'D':
		/* NULL date fields have value "00000000" */
			memset( (char *) (pabyRec+psDBF->panFieldOffset[iField]), '0', psDBF->panFieldSize[iField] );
            break;

          case 'L':
	    /* NULL boolean fields have value "?" */ 
			memset( (char *) (pabyRec+psDBF->panFieldOffset[iField]), '?', psDBF->panFieldSize[iField] );
            break;

          default:
            /* empty string fields are considered NULL */
			memset( (char *) (pabyRec+psDBF->panFieldOffset[iField]), '\0', psDBF->panFieldSize[iField] );
            break;
        }
        return TRUE;
    }

/* -------------------------------------------------------------------- */
/*      Assign all the record fields.                                   */
/* -------------------------------------------------------------------- */
	switch(psDBF->pachFieldType[iField])
	{
		case 'D':
		case 'N':
		case 'F':
			if( psDBF->panFieldDecimals[iField] == 0 )
			{
				int		nWidth = psDBF->panFieldSize[iField];

				if( (int)sizeof(szSField)-2 < nWidth )	nWidth = sizeof(szSField)-2;

				sprintf(szFormat, "%%%dd", nWidth );
				sprintf(szSField, szFormat, (int) *((double *) pValue) );
				if( (int)strlen(szSField) > psDBF->panFieldSize[iField] )
				{
					szSField[psDBF->panFieldSize[iField]] = '\0';
					nRetResult = FALSE;
				}

				strncpy((char *) (pabyRec+psDBF->panFieldOffset[iField]), szSField, strlen(szSField) );
			}
			else
			{
				int		nWidth = psDBF->panFieldSize[iField];

				if( (int)sizeof(szSField)-2 < nWidth )				nWidth = sizeof(szSField)-2;

				sprintf( szFormat, "%%%d.%df",			 nWidth, psDBF->panFieldDecimals[iField] );
				sprintf(szSField, szFormat, *((double *) pValue) );
				if( (int) strlen(szSField) > psDBF->panFieldSize[iField] )
				{
					szSField[psDBF->panFieldSize[iField]] = '\0';
					nRetResult = FALSE;
				}
				strncpy((char *) (pabyRec+psDBF->panFieldOffset[iField]),			szSField, strlen(szSField) );
			}
			break;
		case 'L':
			if (psDBF->panFieldSize[iField] >= 1  && (*(char*)pValue == 'F' || *(char*)pValue == 'T'))
			*(pabyRec+psDBF->panFieldOffset[iField]) = *(char*)pValue;
			break;
		default:
			if( (int) strlen((char *) pValue) > psDBF->panFieldSize[iField] )
			{
				j = psDBF->panFieldSize[iField];
				nRetResult = FALSE;
			}
			else
			{
				memset( pabyRec+psDBF->panFieldOffset[iField], ' ',	psDBF->panFieldSize[iField] );
				j = strlen((char *) pValue);
			}

			strncpy((char *) (pabyRec+psDBF->panFieldOffset[iField]),		(char *) pValue, j );
			break;
	}

	return (nRetResult);
}

/************************************************************************/
/*                     DBFWriteAttributeDirectly()                      */
/*                                                                      */
/*      Write an attribute record to the file, but without any          */
/*      reformatting based on type.  The provided buffer is written     */
/*      as is to the field position in the record.                      */
/************************************************************************/

int DBFWriteAttributeDirectly(DBFHandle psDBF, int hEntity, int iField, void * pValue )
{
	int	       	nRecordOffset, i, j;
    unsigned char	*pabyRec;

/* -------------------------------------------------------------------- */
/*	Is this a valid record?						*/
/* -------------------------------------------------------------------- */
    if( hEntity < 0 || hEntity > psDBF->nRecords )
        return( FALSE );

    if( psDBF->bNoHeader )
        DBFWriteHeader(psDBF);

/* -------------------------------------------------------------------- */
/*      Is this a brand new record?                                     */
/* -------------------------------------------------------------------- */
    if( hEntity == psDBF->nRecords )
    {
	DBFFlushRecord( psDBF );

	psDBF->nRecords++;
	for( i = 0; i < psDBF->nRecordLength; i++ )
	    psDBF->pszCurrentRecord[i] = ' ';

	psDBF->nCurrentRecord = hEntity;
    }

/* -------------------------------------------------------------------- */
/*      Is this an existing record, but different than the last one     */
/*      we accessed?                                                    */
/* -------------------------------------------------------------------- */
    if( psDBF->nCurrentRecord != hEntity )
    {
	DBFFlushRecord( psDBF );

	nRecordOffset = psDBF->nRecordLength * hEntity + psDBF->nHeaderLength;

	fseek( psDBF->fp, nRecordOffset, 0 );
	fread( psDBF->pszCurrentRecord, psDBF->nRecordLength, 1, psDBF->fp );

	psDBF->nCurrentRecord = hEntity;
    }

    pabyRec = (unsigned char *) psDBF->pszCurrentRecord;

/* -------------------------------------------------------------------- */
/*      Assign all the record fields.                                   */
/* -------------------------------------------------------------------- */
	if( (int)strlen((char *) pValue) > psDBF->panFieldSize[iField] )
	{
		j = psDBF->panFieldSize[iField];
	}
    else
    {
        memset( pabyRec+psDBF->panFieldOffset[iField], ' ',
                psDBF->panFieldSize[iField] );
        j = strlen((char *) pValue);
    }

    strncpy((char *) (pabyRec+psDBF->panFieldOffset[iField]),
            (char *) pValue, j );

    psDBF->bCurrentRecordModified = TRUE;
	psDBF->bUpdated = TRUE;

	return( TRUE );
}

/************************************************************************/
/*                      DBFWriteDoubleAttribute()                       */
/*                                                                      */
/*      Write a double attribute.                                       */
/************************************************************************/

int SHPAPI_CALL DBFWriteDoubleAttribute( DBFHandle psDBF, int iRecord, int iField, double dValue )
{
	return( DBFWriteAttribute( psDBF, iRecord, iField, (void *) &dValue ) );
}

/************************************************************************/
/*                      DBFWriteIntegerAttribute()                      */
/*                                                                      */
/*      Write a integer attribute.                                      */
/************************************************************************/

int SHPAPI_CALL DBFWriteIntegerAttribute( DBFHandle psDBF, int iRecord, int iField, int nValue )
{
	double	dValue = nValue;

    return( DBFWriteAttribute( psDBF, iRecord, iField, (void *) &dValue ) );
}

/************************************************************************/
/*                      DBFWriteStringAttribute()                       */
/*                                                                      */
/*      Write a string attribute.                                       */
/************************************************************************/

int SHPAPI_CALL DBFWriteStringAttribute (DBFHandle psDBF, int iRecord, int iField, const char *pszValue)
{
	return(DBFWriteAttribute(psDBF, iRecord, iField, (void *)pszValue));
}

/************************************************************************/
/*                      DBFWriteNULLAttribute()                         */
/*                                                                      */
/*      Write a string attribute.                                       */
/************************************************************************/

int SHPAPI_CALL DBFWriteNULLAttribute( DBFHandle psDBF, int iRecord, int iField )
{
    return( DBFWriteAttribute( psDBF, iRecord, iField, NULL ) );
}

/************************************************************************/
/*                      DBFWriteLogicalAttribute()                      */
/*                                                                      */
/*      Write a logical attribute.                                      */
/************************************************************************/

int SHPAPI_CALL DBFWriteLogicalAttribute( DBFHandle psDBF, int iRecord, int iField, const char lValue)
{
    return( DBFWriteAttribute( psDBF, iRecord, iField, (void *) (&lValue) ) );
}

/************************************************************************/
/*                         DBFWriteTuple()                              */
/*									*/
/*	Write an attribute record to the file.				*/
/************************************************************************/

int SHPAPI_CALL DBFWriteTuple(DBFHandle psDBF, int hEntity, void * pRawTuple )
{
    int	       	nRecordOffset, i;
    unsigned char	*pabyRec;

/* -------------------------------------------------------------------- */
/*	Is this a valid record?						*/
/* -------------------------------------------------------------------- */
    if( hEntity < 0 || hEntity > psDBF->nRecords )
        return( FALSE );

	if( psDBF->bNoHeader ) DBFWriteHeader(psDBF);

/* -------------------------------------------------------------------- */
/*      Is this a brand new record?                                     */
/* -------------------------------------------------------------------- */
	if( hEntity == psDBF->nRecords )
    {
		DBFFlushRecord( psDBF );

		psDBF->nRecords++;
		for( i = 0; i < psDBF->nRecordLength; i++ )
			psDBF->pszCurrentRecord[i] = ' ';

		psDBF->nCurrentRecord = hEntity;
    }

/* -------------------------------------------------------------------- */
/*      Is this an existing record, but different than the last one     */
/*      we accessed?                                                    */
/* -------------------------------------------------------------------- */
    if( psDBF->nCurrentRecord != hEntity )
    {
		DBFFlushRecord( psDBF );

		nRecordOffset = psDBF->nRecordLength * hEntity + psDBF->nHeaderLength;

		fseek( psDBF->fp, nRecordOffset, 0 );
		fread( psDBF->pszCurrentRecord, psDBF->nRecordLength, 1, psDBF->fp );

		psDBF->nCurrentRecord = hEntity;
    }

    pabyRec = (unsigned char *) psDBF->pszCurrentRecord;

    memcpy ( pabyRec, pRawTuple,  psDBF->nRecordLength );

	psDBF->bCurrentRecordModified = TRUE;
    psDBF->bUpdated = TRUE;

    return( TRUE );
}

/************************************************************************/
/*                          DBFReadTuple()                              */
/*                                                                      */
/*      Read one of the attribute fields of a record.                   */
/************************************************************************/

const char SHPAPI_CALL1(*) DBFReadTuple(DBFHandle psDBF, int hEntity )
{
    int	       	nRecordOffset;
    unsigned char	*pabyRec;
    static char	*pReturnTuple = NULL;

    static int	nTupleLen = 0;

/* -------------------------------------------------------------------- */
/*	Have we read the record?					*/
/* -------------------------------------------------------------------- */
    if( hEntity < 0 || hEntity >= psDBF->nRecords )
        return( NULL );

    if( psDBF->nCurrentRecord != hEntity )
    {
		DBFFlushRecord( psDBF );

		nRecordOffset = psDBF->nRecordLength * hEntity + psDBF->nHeaderLength;

		fseek( psDBF->fp, nRecordOffset, 0 );
		fread( psDBF->pszCurrentRecord, psDBF->nRecordLength, 1, psDBF->fp );

		psDBF->nCurrentRecord = hEntity;
    }

    pabyRec = (unsigned char *) psDBF->pszCurrentRecord;

	if ( nTupleLen < psDBF->nRecordLength)
	{
		nTupleLen = psDBF->nRecordLength;
		pReturnTuple = (char *) SfByteRealloc((uchar*)pReturnTuple, psDBF->nRecordLength);
    }
    
    memcpy ( pReturnTuple, pabyRec, psDBF->nRecordLength );
        
    return( pReturnTuple );
}

/************************************************************************/
/*                          DBFCloneEmpty()                              */
/*                                                                      */
/*      Read one of the attribute fields of a record.                   */
/************************************************************************/

DBFHandle SHPAPI_CALL DBFCloneEmpty(DBFHandle psDBF, const char * pszFilename )
{
	DBFHandle	newDBF;

	newDBF = DBFCreate ( pszFilename );
	if ( newDBF == NULL ) return ( NULL );

	newDBF->pszHeader = (char *) SfByteRealloc (0, 32 * psDBF->nFields );
	memcpy ( newDBF->pszHeader, psDBF->pszHeader, 32 * psDBF->nFields );

	newDBF->nFields = psDBF->nFields;
	newDBF->nRecordLength = psDBF->nRecordLength;
	newDBF->nHeaderLength = 32 * (psDBF->nFields+1);

	newDBF->panFieldOffset = SfIntRealloc (0, psDBF->nFields );
	memcpy ( newDBF->panFieldOffset, psDBF->panFieldOffset, sizeof(int) * psDBF->nFields );

	newDBF->panFieldSize = SfIntRealloc (0, psDBF->nFields );
	memcpy ( newDBF->panFieldSize, psDBF->panFieldSize, sizeof(int) * psDBF->nFields );

	newDBF->panFieldDecimals = SfIntRealloc (0, psDBF->nFields );
	memcpy ( newDBF->panFieldDecimals, psDBF->panFieldDecimals, sizeof(int) * psDBF->nFields );

	newDBF->pachFieldType = (char *) SfByteRealloc (0,  sizeof(int) * psDBF->nFields );
	memcpy ( newDBF->pachFieldType, psDBF->pachFieldType, sizeof(int) * psDBF->nFields );

	newDBF->bNoHeader = TRUE;
	newDBF->bUpdated = TRUE;

	DBFWriteHeader ( newDBF );
	DBFClose ( newDBF );

	newDBF = DBFOpen ( pszFilename, "rb+" );

	return ( newDBF );
}

/************************************************************************/
/*                       DBFGetNativeFieldType()                        */
/*                                                                      */
/*      Return the DBase field type for the specified field.            */
/*                                                                      */
/*      Value can be one of: 'C' (String), 'D' (Date), 'F' (Float),     */
/*                           'N' (Numeric, with or without decimal),    */
/*                           'L' (Logical),                             */
/*                           'M' (Memo: 10 digits .DBT block ptr)       */
/************************************************************************/

char SHPAPI_CALL DBFGetNativeFieldType( DBFHandle psDBF, int iField )
{
	if( iField >=0 && iField < psDBF->nFields )	return psDBF->pachFieldType[iField];
	return  ' ';
}

/************************************************************************/
/*                            str_to_upper()                            */
/************************************************************************/
static void str_to_upper (char *string)
{
	int len = strlen (string);

	short i = -1;
	while (++i < len)
	{
		if (isalpha(string[i]) && islower(string[i]))
		{
			string[i] = (char)toupper((int)string[i]);
		}
	}
}

/************************************************************************/
/*                          DBFGetFieldIndex()                          */
/*                                                                      */
/*      Get the index number for a field in a .dbf file.                */
/*                                                                      */
/*      Contributed by Jim Matthews.                                    */
/************************************************************************/
int SHPAPI_CALL DBFGetFieldIndex(DBFHandle psDBF, const char *pszFieldName)
{
    char          name[12], name1[12], name2[12];
	int           i;

    strncpy(name1, pszFieldName,11);
	name1[11] = '\0';
    str_to_upper(name1);

    for( i = 0; i < DBFGetFieldCount(psDBF); i++ )
	{
        DBFGetFieldInfo( psDBF, i, name, NULL, NULL );
        strncpy(name2,name,11);
        str_to_upper(name2);

		if(!strncmp(name1,name2,10))
			return(i);
	}
	return(-1);
}

/************************************************************************/
/*                          Test_WritePoints()                          */
/*                                                                      */
/*      Write a small point file.                                       */
/************************************************************************/
static void Test_WritePoints(int nSHPType, const char *pszFilename)

{
	SHPHandle	hSHPHandle;
    SHPObject	*psShape;
	double		x, y, z, m;

	hSHPHandle = SHPCreate(pszFilename, nSHPType);

	x = 3500001.0;
	y = 5800002.0;
	z = 3.0;
	m = 4.0;
	psShape = SHPCreateObject(nSHPType, -1, 0, NULL, NULL, 1, &x, &y, &z, &m);
	SHPWriteObject(hSHPHandle, -1, psShape);
	SHPDestroyObject(psShape);

	x = 3500010.0;
	y = 5800020.0;
	z = 30.0;
	m = 40.0;
	psShape = SHPCreateObject(nSHPType, -1, 0, NULL, NULL, 1, &x, &y, &z, &m);
	SHPWriteObject(hSHPHandle, -1, psShape);
	SHPDestroyObject(psShape);

	SHPClose(hSHPHandle);
}

/************************************************************************/
/*                       Test_WriteMultiPoints()                        */
/*                                                                      */
/*      Write a small multipoint file.                                  */
/************************************************************************/
static void Test_WriteMultiPoints( int nSHPType, const char *pszFilename )
{
    SHPHandle	hSHPHandle;
    SHPObject	*psShape;
    double	x[4], y[4], z[4], m[4];
    int		i, iShape;

    hSHPHandle = SHPCreate( pszFilename, nSHPType );

	for( iShape = 0; iShape < 3; iShape++ )
	{
        for( i = 0; i < 4; i++ )
		{
            x[i] = iShape * 10 + i + 1.15;
            y[i] = iShape * 10 + i + 2.25;
            z[i] = iShape * 10 + i + 3.35;
            m[i] = iShape * 10 + i + 4.45;
        }
        
        psShape = SHPCreateObject( nSHPType, -1, 0, NULL, NULL,
                                   4, x, y, z, m );
        SHPWriteObject( hSHPHandle, -1, psShape );
        SHPDestroyObject( psShape );
    }    

    SHPClose( hSHPHandle );
}

/************************************************************************/
/*                         Test_WriteArcPoly()                          */
/*                                                                      */
/*      Write a small arc or polygon file.                              */
/************************************************************************/
static void Test_WriteArcPoly(int nSHPType, const char *pszFilename)
{
    SHPHandle	hSHPHandle;
    SHPObject	*psShape;
	double		x[100], y[100], z[100], m[100];
	int			anPartStart[100];
	int			anPartType[100], *panPartType;

	hSHPHandle = SHPCreate(pszFilename, nSHPType);

	if (nSHPType == SHPT_MULTIPATCH)	panPartType = anPartType;
	else                                panPartType = NULL;

	for (int iShape = 0; iShape < 3; iShape++)
	{
        x[0] = 1.0;
		y[0] = 1.0 + iShape * 3;
		x[1] = 2.0;
		y[1] = 1.0 + iShape * 3;
		x[2] = 2.0;
		y[2] = 2.0 + iShape * 3;
		x[3] = 1.0;
		y[3] = 2.0 + iShape * 3;
		x[4] = 1.0;
		y[4] = 1.0 + iShape * 3;

		for (int i = 0; i < 5; i++)
        {
            z[i] = iShape * 10 + i + 3.35;
			m[i] = iShape * 10 + i + 4.45;
        }
        
		psShape = SHPCreateObject(nSHPType, -1, 0, NULL, NULL, 5, x, y, z, m);
		SHPWriteObject(hSHPHandle, -1, psShape);
		SHPDestroyObject(psShape);
	}

	/* -------------------------------------------------------------------- */
	/*      Do a multi part polygon (shape).  We close it, and have two     */
	/*      inner rings.                                                    */
	/* -------------------------------------------------------------------- */

	// Eine lange Liste ALLER Knoten der drei Polygone
	x[0] = 0.0;
    y[0] = 0.0;
    x[1] = 0;
    y[1] = 100;
    x[2] = 100;
    y[2] = 100;
	x[3] = 100;
    y[3] = 0;
    x[4] = 0;
    y[4] = 0;

    x[5] = 10;
    y[5] = 20;
    x[6] = 30;
	y[6] = 20;
    x[7] = 30;
    y[7] = 40;
    x[8] = 10;
    y[8] = 40;
	x[9] = 10;
    y[9] = 20;

    x[10] = 60;
    y[10] = 20;
    x[11] = 90;
    y[11] = 20;
    x[12] = 90;
    y[12] = 40;
    x[13] = 60;
    y[13] = 40;
    x[14] = 60;
    y[14] = 20;

	for(int i = 0; i < 15; i++ )
    {
        z[i] = i;
		m[i] = i * 2;
    }

	// Die Startpunkte der drei Polygone
	anPartStart[0] = 0;
	anPartStart[1] = 5;
	anPartStart[2] = 10;

	// Die Typen der drei Polygone
	anPartType[0] = SHPP_RING;
    anPartType[1] = SHPP_INNERRING;
    anPartType[2] = SHPP_INNERRING;
    
	psShape = SHPCreateObject (nSHPType, -1, 3, anPartStart, panPartType, 15, x, y, z, m);
	SHPWriteObject(hSHPHandle, -1, psShape);
	SHPDestroyObject(psShape);

	SHPClose(hSHPHandle);
}

void	TestCreateDBF(char *FileName, int Count)
{
	/* -------------------------------------------------------------------- */
	/*	Create the database.						*/
	/* -------------------------------------------------------------------- */
	DBFHandle hDBF = DBFCreate(FileName);
	if( hDBF == NULL )
	{
		printf( "DBFCreate failed.\n");
		exit( 2 );
	}

	if( DBFAddField(hDBF, "ID", FTString, 2, 0) == -1)
	{
		printf( "DBFAddField failed.\n");
		exit( 4 );
	}

	DBFClose( hDBF );

	/* -------------------------------------------------------------------- */
	/*	Open the database.													*/
	/* -------------------------------------------------------------------- */
	hDBF = DBFOpen(FileName, "r+b" );
	if( hDBF == NULL )
	{
		printf( "DBFOpen failed.\n");
		exit( 2 );
	}

	/* -------------------------------------------------------------------- */
	/*	Do we have the correct number of arguments?							*/
	/* -------------------------------------------------------------------- */
	if( DBFGetFieldCount(hDBF) != 1)
	{
		printf( "Got fields, but require\n");
		exit( 3 );
	}

	/* -------------------------------------------------------------------- */
	/*	Loop assigning the new records values.								*/
	/* -------------------------------------------------------------------- */
	for (int i=0; i < Count; i++)
	{
		/* -------------------------------------------------------------------- */
		/*	Loop assigning the new field values.								*/
		/* -------------------------------------------------------------------- */
		for (int j = 0; j < DBFGetFieldCount(hDBF); j++ )
		{
			if( DBFGetFieldInfo(hDBF, j, NULL, NULL, NULL) == FTString)
			{
				DBFWriteStringAttribute(hDBF, i, j, "id");
			}
			else
			{
				DBFWriteDoubleAttribute(hDBF, i, j, 9182);
			}
		}
	}

	/* -------------------------------------------------------------------- */
	/*      Close and cleanup.                                              */
	/* -------------------------------------------------------------------- */
	DBFClose( hDBF );
}

void	TestShapeFileLib(void)
{
	// Test_WritePoints( SHPT_NULL, "ShapeTest\\test0.shp" );
	// TestCreateDBF("ShapeTest\\test0.dbf", 2);

	Test_WritePoints( SHPT_POINT, "ShapeTest\\test1.shp" );
	Test_WritePoints( SHPT_POINTZ, "ShapeTest\\test2.shp" );
	Test_WritePoints( SHPT_POINTM, "ShapeTest\\test3.shp" );

	Test_WriteMultiPoints( SHPT_MULTIPOINT, "ShapeTest\\test4.shp" );
	Test_WriteMultiPoints( SHPT_MULTIPOINTZ, "ShapeTest\\test5.shp" );
	Test_WriteMultiPoints( SHPT_MULTIPOINTM, "ShapeTest\\test6.shp" );

	Test_WriteArcPoly( SHPT_ARC, "ShapeTest\\test7.shp" );
	Test_WriteArcPoly( SHPT_ARCZ, "ShapeTest\\test8.shp" );
	Test_WriteArcPoly( SHPT_ARCM, "ShapeTest\\test9.shp" );

	Test_WriteArcPoly( SHPT_POLYGON, "ShapeTest\\test10.shp" );
	Test_WriteArcPoly( SHPT_POLYGONZ, "ShapeTest\\test11.shp" );
	Test_WriteArcPoly( SHPT_POLYGONM, "ShapeTest\\test12.shp" );
	Test_WriteArcPoly( SHPT_MULTIPATCH, "ShapeTest\\test13.shp" );

	TestCreateDBF("ShapeTest\\test1.dbf", 2);
	TestCreateDBF("ShapeTest\\test2.dbf", 2);
	TestCreateDBF("ShapeTest\\test3.dbf", 2);
	TestCreateDBF("ShapeTest\\test4.dbf", 3);
	TestCreateDBF("ShapeTest\\test5.dbf", 3);
	TestCreateDBF("ShapeTest\\test6.dbf", 3);
	TestCreateDBF("ShapeTest\\test7.dbf", 4);
	TestCreateDBF("ShapeTest\\test8.dbf", 4);
	TestCreateDBF("ShapeTest\\test9.dbf", 4);
	TestCreateDBF("ShapeTest\\test10.dbf", 4);
	TestCreateDBF("ShapeTest\\test11.dbf", 4);
	TestCreateDBF("ShapeTest\\test12.dbf", 4);
	TestCreateDBF("ShapeTest\\test13.dbf", 4);
}





