//
// $Id: test_tools.cpp 101 2008-07-10 09:20:12Z mrchip $
//
// Copyright (C) 2005 STADT-LAND-FLUSS INGENIEURDIENSTE GmbH
//
// Authors:
// Ulrich Kiel <u.kiel@S-L-F.de>
//
// This program is free software under the GPL (>=v2)
// Read the file COPYING coming with WSPLGEN for details.
//

#ifdef __BORLANDC__
#pragma hdrstop
#endif

//----------------------------------------------------------------------------
#ifdef __BORLANDC__
#define DGMXYZFILE		"test_daten\\test_buildedgelist.xyz"
#define DGM2DMFILE		"test_daten\\dgm2.2dm"
#define PROFILFILE		"test_daten\\profile2.shp"
#define ACHSEFILE		"test_daten\\achse2.shp"
#define WSPFILE			"test_daten\\profile2.dbf"
#define LINIENFILE		"test_daten\\linien2.shp"
#else
#define DGMXYZFILE		"test_daten/test_buildedgelist.xyz"
#define DGM2DMFILE		"test_daten/dgm2.2dm"
#define PROFILFILE		"test_daten/profile2.shp"
#define ACHSEFILE		"test_daten/achse2.shp"
#define WSPFILE			"test_daten/profile2.dbf"
#define LINIENFILE		"test_daten/linien2.shp"
#endif

//---------------------------------------------------------------------------
#include <stdio.h>

#include "file.h"
#include "tri.h"

#include "tools.h"

void test_tools(void)
{
	////////////////////////////////////////
	// Test der Tools
	////////////////////////////////////////

	printf ("*************************************************************************\n");
	printf ("Test Tools\n");

	////////////////////////////////////////
	// Test von TransferWsp
	////////////////////////////////////////

	{
		printf ("Test TransferWsp:                                                    ");

		TNodeList *NodeList = new TNodeList();
		TNodeList *NassNodeList = new TNodeList();
		TElementList *ElementList = new TElementList();
		LoadDGM(DGM2DMFILE, NodeList, ElementList, 0, 8);

		TProfilList *ProfilList = new TProfilList();
		LoadProfile(PROFILFILE, ProfilList, 0, -9999.9, 9999.9, 8);

		TGewaesserAchseList *AchsenList = new TGewaesserAchseList();
		LoadAchse(ACHSEFILE, AchsenList, 8);

		TProfilList *WspList = new TProfilList();
		LoadWsp(WSPFILE, WspList, "", 8);

		InterpolateWsp(ProfilList, WspList, 8);
		delete WspList;
		WspList = 0;

		TProfilList *SperrenList = new TProfilList();
		TProfilList *GrabenList = new TProfilList();
		TProfilList *RohreList = new TProfilList();
		TProfilList *GradientenList = new TProfilList();
		LoadLinien(LINIENFILE, SperrenList, GrabenList, RohreList, GradientenList, 8);

		TEdgeList* EdgeList = new TEdgeList;
		BuildEdgeList(NodeList, ElementList, EdgeList, 8);
		// Den alten Knoten Kanten Index lschen
		NodeList->ClearEdgeIndex();
		// Einen neuen Knoten Kanten Index aufbauen
		EdgeList->BuildEdgeIndex();

		// Jetzt die Wasserstnde der Profile bertragen
		TransferWsp(ProfilList, NodeList, AchsenList, SperrenList, NassNodeList, true, 8);


		int Count = 0;
		for (TNodeList::iterator i=NodeList->begin(); i != NodeList->end(); i++)
        {
			TNode* Node = *i;

            if (Node->Wsp != NoWsp)	Count++;
        }

		if (Count != 20)	printf("Failed\n");
		else	   	  		printf("Pass\n");

        delete GradientenList;
        delete RohreList;
        delete GrabenList;
        delete SperrenList;
		delete ProfilList;

		delete AchsenList;
		delete EdgeList;
        delete ElementList;
        NassNodeList->clear();
        delete NassNodeList;
        delete NodeList;
	}

	////////////////////////////////////////
	// Test von ToUTF8
	////////////////////////////////////////

	{
		printf ("Test ToUTF8:                                                         ");

		char Test1[15];
		char Test2[15];

		sprintf(Test1, "%c%c%c%c%c%c%c%c%c%c%c%c%c%c", 0xc3, 0xa4, 0xc3, 0xb6, 0xc3, 0xbc, 0xc3, 0x84, 0xc3, 0x96, 0xc3, 0x9c, 0xc3, 0x9f);

		strncpy(Test2, ToUTF8(""), 15);

		if (strncmp(Test2, Test1, 14) != 0)	printf("Failed\n");
		else						  		printf("Pass\n");
	}

	////////////////////////////////////////
	// Test von Trim
	////////////////////////////////////////

	{
		printf ("Test Trim:                                                           ");

		if (Trim("") != "")							printf("Failed\n");
		else if (Trim(" ") != "")					printf("Failed\n");
		else if (Trim("   ") != "")					printf("Failed\n");
		else if (Trim("a b") != "a b")				printf("Failed\n");
		else if (Trim(" a b ") != "a b")			printf("Failed\n");
		else if (Trim("  a b  ") != "a b")			printf("Failed\n");
		else if (Trim("   a b") != "a b")			printf("Failed\n");
		else if (Trim("a b  ") != "a b")			printf("Failed\n");
		else										printf("Pass\n");
	}

	////////////////////////////////////////
	// Test von Calc2Schnitt
	////////////////////////////////////////

	{
		printf ("Test Calc2Schnitt:                                                   ");

		TNode N1(1, 1, 1, 1);
		TNode N2(2, 4, 4, 1);
		TNode N3(3, 1, 4, 1);
		TNode N4(4, 4, 1, 1);

		TNode N5(5, 350000000, 580000000, 100);
		TNode N6(5, 350000100, 580000200, 100);
		TNode N7(7, 350000060, 580000060, 100);
		TNode N8(8, 350000060, 580000250, 100);

		TNode N9 ( 9, 350000000, 580000000, 100);
		TNode N10(10, 350000110, 580000200, 100);
		TNode N11(11, 349999390, 579998780, 100);
		TNode N12(12, 350000890, 580001750, 100);

		double X1, Y1;
		bool OK1 = Calc2Schnitt(N1.X, N1.Y, N2.X, N2.Y, N3.X, N3.Y, N4.X, N4.Y, &X1, &Y1);

		double X2, Y2;
		bool OK2 = Calc2Schnitt(N1.X, N1.Y, N3.X, N3.Y, N2.X, N2.Y, N4.X, N4.Y, &X2, &Y2);

		double X3, Y3;
		bool OK3 = Calc2Schnitt(N1.X, N1.Y, N4.X, N4.Y, N2.X, N2.Y, N3.X, N3.Y, &X3, &Y3);

		double X4, Y4;
		bool OK4 = Calc2Schnitt(N5.X, N5.Y, N6.X, N6.Y, N7.X, N7.Y, N8.X, N8.Y, &X4, &Y4);

		double X5, Y5;
		bool OK5 = Calc2Schnitt(N7.X, N7.Y, N8.X, N8.Y, N5.X, N5.Y, N6.X, N6.Y, &X5, &Y5);

		double X6, Y6;
		bool OK6 = Calc2Schnitt(N9.X, N9.Y, N10.X, N10.Y, N11.X, N11.Y, N12.X, N12.Y, &X6, &Y6);

		double X7, Y7;
		bool OK7 = Calc2Schnitt(N11.X, N11.Y, N12.X, N12.Y, N9.X, N9.Y, N10.X, N10.Y, &X7, &Y7);

		if (false == OK1)							printf("Failed\n");
		else if (fabs(X1 - 2.5) > 0.01)				printf("Failed\n");
		else if (fabs(Y1 - 2.5) > 0.01)				printf("Failed\n");
		else if (true == OK2)						printf("Failed\n");
		else if (true == OK3)						printf("Failed\n");
		else if (false == OK4)						printf("Failed\n");
		else if (fabs(X4 - 350000060) > 0.01)		printf("Failed\n");
		else if (fabs(Y4 - 580000120) > 0.01)		printf("Failed\n");
		else if (false == OK5)						printf("Failed\n");
		else if (fabs(X5 - 350000060) > 0.01)		printf("Failed\n");
		else if (fabs(Y5 - 580000120) > 0.01)		printf("Failed\n");
		else if (false == OK6)						printf("Failed\n");
		else if (fabs(X6 - 350000075.39) > 0.01)	printf("Failed\n");
		else if (fabs(Y6 - 580000137.08) > 0.01)	printf("Failed\n");
		else if (false == OK7)						printf("Failed\n");
		else if (fabs(X7 - 350000075.39) > 0.01)	printf("Failed\n");
		else if (fabs(Y7 - 580000137.08) > 0.01)	printf("Failed\n");
		else										printf("Pass\n");
	}

	////////////////////////////////////////
	// Test von CircumCircle
	////////////////////////////////////////

	{
		printf ("Test CircumCircle:                                                   ");

		TNode N1(1, 1, 1, 1);
		TNode N2(2, 3, 4, 1);
		TNode N3(3, 3, 6, 1);
		TNode N4(4, 3, 7, 1);
		TNode N5(5, 350083339, 587315140, 100);
		TNode N6(6, 350083554, 587315565, 100);
		TNode N7(7, 350080924, 587315903, 100);
		TNode N8(8, 350081510, 587317062, 100);

		long double CX1, CY1, R1;
		bool OK1 = CircumCircle(N1.X, N1.Y, N2.X, N2.Y, N3.X, N3.Y, &CX1, &CY1, &R1);

		long double CX2, CY2, R2;
		bool OK2 = CircumCircle(N2.X, N2.Y, N3.X, N3.Y, N4.X, N4.Y, &CX2, &CY2, &R2);

		long double CX3, CY3, R3;
		bool OK3 = CircumCircle(N5.X, N5.Y, N6.X, N6.Y, N7.X, N7.Y, &CX3, &CY3, &R3);

		long double CX4, CY4, R4;
		bool OK4 = CircumCircle(N5.X, N5.Y, N6.X, N6.Y, N8.X, N8.Y, &CX4, &CY4, &R4);

		if (false == OK1)							printf("Failed\n");
		else if (fabs(sqrt(R1) - 4.8541) > 0.001)	printf("Failed\n");
		else if (fabs(CX1 - (-1.75)) > 0.001)		printf("Failed\n");
		else if (fabs(CY1 - 5) > 0.001)				printf("Failed\n");
		else if (true == OK2)						printf("Failed\n");
		else if (false == OK3)						printf("Failed\n");
		else if (fabs(sqrt(R3) - 1343.476) > 0.01)	printf("Failed\n");
		else if (fabs(CX3 - 350082266.676) > 0.01)	printf("Failed\n");
		else if (fabs(CY3 - 587315949.352) > 0.01)	printf("Failed\n");
		else if (false == OK4)						printf("Failed\n");
		else if (fabs(sqrt(R4) - 1344.582) > 0.01)	printf("Failed\n");
		else if (fabs(CX4 - 350082265.674) > 0.01)	printf("Failed\n");
		else if (fabs(CY4 - 587315949.859) > 0.01)	printf("Failed\n");
		else										printf("Pass\n");
	}

	////////////////////////////////////////
	// Test der Zeichenkettenhilfsfunktionen
	////////////////////////////////////////

	{
		printf ("Test ToUpperCase:                                                    ");

		std::string Lower = "aAbBcCdDeEfFgGhHiIjJkKlLmMnNoOpPqQrRsStTuUvVwWxXyYzZaabbccddeeff";
		std::string Upper = ToUpperCase(Lower);

		if (Upper != "AABBCCDDEEFFGGHHIIJJKKLLMMNNOOPPQQRRSSTTUUVVWWXXYYZZAABBCCDDEEFF")	printf("Failed\n");
		else																				printf("Pass\n");
	}

	{
		printf ("Test GetFileExt:                                                     ");

		std::string FileName1 = "c:\\hallo\\test.adf";
		std::string Ext1 = GetFileExt(FileName1);

		std::string FileName2 = "c:\\hallo\\test";
		std::string Ext2 = GetFileExt(FileName2);

		std::string FileName3 = "c:\\hallo\\test.txt.SHP";
		std::string Ext3 = GetFileExt(FileName3);

		if (Ext1 != ".adf" || Ext2 != "" || Ext3 != ".SHP")	printf("Failed\n");
		else			 									printf("Pass\n");
	}

	{
		printf ("Test ExchangeFileExt:                                                ");

		std::string FileName = "c:\\hallo\\test.adf";
		std::string NewFileName = ExchangeFileExt(FileName, ".dat");

		if (NewFileName != "c:\\hallo\\test.dat")	printf("Failed\n");
		else			 							printf("Pass\n");
	}

	{
		printf ("Test GetFilePath:                                                    ");

		std::string FileName1 = "c:\\hallo\\test.adf";
		std::string Path1 = GetFilePath(FileName1);

		std::string FileName2 = "c:/test/einwenigunix.txt";
		std::string Path2 = GetFilePath(FileName2);

		std::string FileName3 = "c:test/nichtganzkorrekt.txt";
		std::string Path3 = GetFilePath(FileName3);

		std::string FileName4 = "c:nochwenigerkorrekt.txt";
		std::string Path4 = GetFilePath(FileName4);

		std::string FileName5 = "c:\\";
		std::string Path5 = GetFilePath(FileName5);

		std::string FileName6 = "test.txt";
		std::string Path6 = GetFilePath(FileName6);

		if (Path1 != "c:\\hallo\\")		printf("Failed\n");
		else if (Path2 != "c:/test/")	printf("Failed\n");
		else if (Path3 != "c:test/")	printf("Failed\n");
		else if (Path4 != "c:")			printf("Failed\n");
		else if (Path5 != "c:\\")		printf("Failed\n");
		else if (Path6 != "")			printf("Failed\n");
		else			 			 	printf("Pass\n");
	}

	////////////////////////////////////////
	// Test der Funktion IsInside
	////////////////////////////////////////

	{
		printf ("Test IsInside:                                                       ");

		TXYZList *XyList = new TXYZList();

		int Alpha1 = IsInside(XyList, 1.0, 1.0);

		XyList->push_back(new TXYZ(2, 3));

		int Alpha2 = IsInside(XyList, 2.0, 3.0);
		int Alpha3 = IsInside(XyList, 1.0, 1.0);

		XyList->push_back(new TXYZ(2, 0));

		int Alpha4 = IsInside(XyList, 2.0, 1.0);

		XyList->push_back(new TXYZ(0, 0));
		XyList->push_back(new TXYZ(0, 2));

		int Alpha5 = IsInside(XyList, 1.0, 1.0);
		int Alpha6 = IsInside(XyList, 3.0, 3.0);
		int Alpha7 = IsInside(XyList, 2.0, 2.0);

		if (Alpha1 != 0)		printf("Failed\n");
		else if (Alpha2 != -1)	printf("Failed\n");
		else if (Alpha3 != 0)	printf("Failed\n");
		else if (Alpha4 != -1)	printf("Failed\n");
		else if (Alpha5 != -4)	printf("Failed\n");
		else if (Alpha6 != 0)	printf("Failed\n");
		else if (Alpha7 != -1)	printf("Failed\n");
		else					printf("Pass\n");

		delete XyList;
	}

	////////////////////////////////////////
	// Test von BuildEdgeList
	////////////////////////////////////////

	{
		printf ("Test BuildEdgeList:                                                  ");

		TNodeList *NodeList = new TNodeList();
		TElementList *ElementList = new TElementList();
		TEdgeList* EdgeList = new TEdgeList;

		LoadDGM(DGMXYZFILE, NodeList, ElementList, 0, 8);

		Triangulate(NodeList, ElementList, 0, 8);

		BuildEdgeList(NodeList, ElementList, EdgeList, 8);

		bool WrongOrder = false;
		for (TEdgeVector::iterator i=EdgeList->EdgeVector.begin(); i!=EdgeList->EdgeVector.end(); i++)
		{
			TEdge* Edge = *i;
			int Node1Nr = Edge->Node1->Nr;
			int Node2Nr = Edge->Node2->Nr;

			if (Node1Nr >= Node2Nr)
			{
				WrongOrder = true;
			}
		}

		if (EdgeList->EdgeVector.size() != 29)	printf("Failed\n");
		else if (WrongOrder)					printf("Failed\n");
		else									printf("Pass\n");

        delete EdgeList;
        delete ElementList;
        delete NodeList;
	}

	printf ("*************************************************************************\n");
}

