//
// $Id: tri.cpp 104 2008-07-21 20:39:54Z mrchip $
//
// Copyright (C) 2005 STADT-LAND-FLUSS INGENIEURDIENSTE GmbH
//
// Authors:
// Ulrich Kiel <u.kiel@S-L-F.de>
//
// This program is free software under the GPL (>=v2)
// Read the file COPYING coming with WSPLGEN for details.
//

#ifdef __BORLANDC__
#pragma hdrstop
#endif

//---------------------------------------------------------------------
#include <sys/timeb.h>

#include "tools.h"
#include "tri.h"
#include "file.h"

//---------------------------------------------------------------------
void FindLeftMostNeighbours(TNodeList* NodeList, TNode* *Node1, TNode* *Node2)
{
	write_fortschritt("->Suchen der Startkante gestartet\n");

	double Min1 = 0.0;
	double Min2 = 0.0;

	for (unsigned int i = 0; i < NodeList->size(); i++)
	{
		TNode *Node = (*NodeList)[i];

		if (*Node1 == 0)
		{
			Min1 = Node->X;
			*Node1 = Node;
		}
		else if	(*Node2 == 0)
		{
			if (Min1 > Node->X)
			{
				Min2 = Min1;
				*Node2 = *Node1;

				Min1 = Node->X;
				*Node1 = Node;
			}
			else
			{
				Min2 = Node->X;
				*Node2 = Node;
			}
		}
		else if (Min1 > Node->X)
		{
			Min2 = Min1;
			*Node2 = *Node1;

			Min1 = Node->X;
			*Node1 = Node;
		}
		else if (Min2 > Node->X)
		{
			Min2 = Node->X;
			*Node2 = Node;
		}
	}
	if (*Node1 && *Node2)	write_fortschritt("Kante: %d - %d als Startkante\n", (*Node1)->Nr, (*Node2)->Nr);

	write_fortschritt("<-Suchen der Startkante beendet\n");
}

//---------------------------------------------------------------------
unsigned int CompleteFacet(TNodeList* NodeList, TElementList* ElementList, TEdgeList *EdgeList, unsigned int AktEdgeIndex, int DebugLevel)
{
	TEdge* AktEdge = EdgeList->EdgeVector[AktEdgeIndex++];

	TNode *Node1 = AktEdge->Node2;
	TNode *Node2 = AktEdge->Node1;

	// Zuerst testen, ob es Verbindungen unter Wchter-Knoten waren

	if (AktEdgeIndex > 1 && Node1->Z == NoZ && Node2->Z == NoZ)	return (AktEdgeIndex);

	if (EdgeList->FindWithIndex(Node1, Node2)) return (AktEdgeIndex);

	long double cx = 0;
	long double cy = 0;
	long double cr = 0;

	long X1 = Node1->X;
	long Y1 = Node1->Y;
	long X2 = Node2->X;
	long Y2 = Node2->Y;

	double Dx = X2 - X1;
	double Dy = Y2 - Y1;

	TNodeList* FoundNodeList = 0;

	double X = (X1 + X2) / 2;
	double Y = (Y1 + Y2) / 2;

	int MaxWeite = 256;
	int TestWeite = 2;
	int FoundAnz = 0;
	do
	{
		double SX = X - TestWeite * Dy;
		double SY = Y + TestWeite * Dx;

		CircumCircle(Node1->X, Node1->Y, Node2->X, Node2->Y, SX, SY, &cx, &cy, &cr);

		FoundNodeList = NodeList->FindAllByXY((double)cx, (double)cy, (double)(sqrt(cr) + 0.001));

		FoundAnz = 0;
		TNodeList::iterator i = FoundNodeList->begin();
		while (i != FoundNodeList->end())
		{
			TNode* Node = *i;

			Node->Wsp = 1;

			if (Node->Nr == Node1->Nr)
			{
				Node->Wsp = 0;
			}
			else if (Node->Nr == Node2->Nr)
			{
				Node->Wsp = 0;
			}
			else if (CrossProduct(Node1->X, Node1->Y, Node2->X, Node2->Y, Node->X, Node->Y) < 0.0000001)
			{
				Node->Wsp = 0;
			}
			else
			{
				FoundAnz++;
			}
			i++;
		}

		TestWeite = TestWeite * 2;

	} while (FoundAnz <= 0 && TestWeite <= MaxWeite);

/*
	// Dann versuche ich es noch mal mit allen
	if (FoundNodeList->size() == 0)
	{
		for (TNodeList::iterator i = NodeList->begin(); i != NodeList->end(); i++)
		{
			TNode* Node = *i;
			if (Node->Nr == Node1->Nr || Node->Nr == Node2->Nr) continue;
			FoundNodeList->push_back(Node);
		}
	}
*/

	TNodeList::iterator i = FoundNodeList->begin();
	TNode* NewNode = 0;
	while (i != FoundNodeList->end() && (*i)->Wsp < 1)	i++;
	if (i != FoundNodeList->end())	NewNode = *i++;

	if (NewNode)
	{
		CircumCircle(Node1->X, Node1->Y, Node2->X, Node2->Y, NewNode->X, NewNode->Y, &cx, &cy, &cr);

		while (i != FoundNodeList->end())
		{
			TNode* Node = *i++;

			// Dieser Knoten durfte es nicht sein
			// siehe oben (Wsp = 0)
			if (Node->Wsp < 1)	continue;

			long X = Node->X;
			long Y = Node->Y;

			double Dx = (double)cx - X;
			double Dy = (double)cy - Y;

			// Liegt er im Umkreis ?
			if (Dx * Dx + Dy * Dy < cr - 0.01)
			{
				// Er ist deutlich dichter dran, also ist er besser
				NewNode = Node;
				CircumCircle(Node1->X, Node1->Y, Node2->X, Node2->Y, NewNode->X, NewNode->Y, &cx, &cy, &cr);
				continue;
			}

			if (Dx * Dx + Dy * Dy > cr + 0.01)
			{
				// Er ist deutlich weiter weg, also brauchen wir ihn nicht weiter zu beachten
				continue;
			}

			// Die beiden Knoten sind gleich oder wenigstens fast gleich gut
			// Dann hilft nur eine andere Art der Unterscheidung

			long double CP = CrossProduct(Node1->X, Node1->Y, NewNode->X, NewNode->Y, Node->X, Node->Y);

			if (CP < 0.0000001)
			{
				// die Diagonalen sind Node1 - Node und NewNode - Node2

				// Gab es die Diagonale Node1 - Node ?
				TEdge* Edge = EdgeList->FindWithIndex(Node1, Node);

				if (Edge)
				{
					// Dann mssen wir Node auch nehmen
					NewNode = Node;
					CircumCircle(Node1->X, Node1->Y, Node2->X, Node2->Y, NewNode->X, NewNode->Y, &cx, &cy, &cr);
					continue;
				}

				// Gab es schon die Diagonale NewNode - Node2 ?
				Edge = EdgeList->FindWithIndex(NewNode, Node2);

				// Dann mssen wir NewNode auch nehmen
				if (Edge)
				{
					continue;
				}
			}
			else
			{
				// die Diagonalen sind Node - Node2 und Node1 - NewNode

				// Gab es die Diagonale Node - Node2 ?
				TEdge* Edge = EdgeList->FindWithIndex(Node, Node2);

				if (Edge)
				{
					// Dann mssen wir Node auch nehmen
					NewNode = Node;
					CircumCircle(Node1->X, Node1->Y, Node2->X, Node2->Y, NewNode->X, NewNode->Y, &cx, &cy, &cr);
					continue;
				}

				// Gab es schon die Diagonale Node1 - NewNode ?
				Edge = EdgeList->FindWithIndex(Node1, NewNode);

				// Dann mssen wir NewNode auch nehmen
				if (Edge)
				{
					continue;
				}
			}

			if (Dx * Dx + Dy * Dy < cr)
			{
				// Er ist dichter dran, also ist er besser
				// Rundungsfehler spielen keine Rolle mehr
				NewNode = Node;
				CircumCircle(Node1->X, Node1->Y, Node2->X, Node2->Y, NewNode->X, NewNode->Y, &cx, &cy, &cr);
				continue;
			}
		}
	}

	if (NewNode)
	{
		if (DebugLevel >= 8)
		{
			if (EdgeList->FindWithIndex(Node2, NewNode))
			{
				dump_error(__FILE__, __LINE__, "Kante %d in Bearbeitung.\nDie Kante %d - %d wurde bereits erzeugt\n", AktEdgeIndex-1, Node2->Nr, NewNode->Nr);
			}
			if (EdgeList->FindWithIndex(NewNode, Node1))
			{
				dump_error(__FILE__, __LINE__, "Kante %d in Bearbeitung.\nDie Kante %d - %d wurde bereits erzeugt\n", AktEdgeIndex-1, NewNode->Nr, Node1->Nr);
			}
		}

		if (AktEdgeIndex == 1)
		{
			EdgeList->AddWithIndex(Node1, Node2);
		}
		EdgeList->AddWithIndex(Node2, NewNode);
		EdgeList->AddWithIndex(NewNode, Node1);

		TElement *Element = 0;
		Element = new TElement(Node1, Node2, NewNode);
		ElementList->Add(Element);
	}
	else
	{
		AktEdge->IsBoundary = true;
	}

	return (AktEdgeIndex);
}

//---------------------------------------------------------------------
bool	Triangulate(TNodeList *NodeList, TElementList *ElementList, TXYZList* BegrenzungsPolygon, int DebugLevel)
{
	write_fortschritt("->Triangulierung der Knoten gestartet\n");

	if (NodeList->size() < 3)
	{
		write_error(1234, "Es sind im DGM nur %d Knoten vorhanden.\nEs werden aber mindestens 3 Knoten bentigt\n", NodeList->size());
	}

	// Noch ein paar Wchter-Punkte einfgen

	NodeList->Scale();

	long MinX = NodeList->MinX;
	long MaxX = NodeList->MaxX;
	long MinY = NodeList->MinY;
	long MaxY = NodeList->MaxY;

	long Dx = MaxX - MinX;
	long Dy = MaxY - MinY;

	Dx = Dx / 2;
	Dy = Dy / 2;

	if (Dx <= 10)	Dx = 10;
	if (Dy <= 10)	Dy = 10;


	TNode* NewNode1 = new TNode(++NodeList->MaxNr, MinX - Dx, MinY - Dy, NoZ);
	TNode* NewNode2 = new TNode(++NodeList->MaxNr, MinX - Dx, MaxY + Dy, NoZ);
	TNode* NewNode3 = new TNode(++NodeList->MaxNr, MaxX + Dx, MaxY + Dy, NoZ);
	TNode* NewNode4 = new TNode(++NodeList->MaxNr, MaxX + Dx, MinY - Dy, NoZ);

	NodeList->Add(NewNode1);
	NodeList->Add(NewNode2);
	NodeList->Add(NewNode3);
	NodeList->Add(NewNode4);

	// Jetzt noch die Strukturen vorbereiten

	NodeList->SortByXY();
	NodeList->ClearEdgeIndex();

	// Erst mal genug Speicher besorgen
    ElementList->reserve((size_t)(NodeList->size() * 2.2));
	write_fortschritt("Kapazitt der Elementliste wurde auf %d erhht\n", ElementList->capacity());

	TEdgeList *EdgeList = new TEdgeList;

	// Erst mal genug Speicher besorgen
    EdgeList->EdgeVector.reserve((size_t)(NodeList->size() * 2.2 * 2.2));
	write_fortschritt("Kapazitt der Kantenliste wurde auf %d erhht\n", EdgeList->EdgeVector.capacity());

	TEdge* NewEdge = new TEdge(NewNode1, NewNode2);
	EdgeList->Add(NewEdge);

	int TimeDiff = 5000;
	if (DebugLevel >= 2)	TimeDiff = 1000;

	struct timeb Now;
	ftime (&Now);

	int StartMSec = Now.time * 1000 + Now.millitm;
	int NextMSec = StartMSec + TimeDiff;

	unsigned int AktEdgeIndex = 0;
	while (AktEdgeIndex < EdgeList->EdgeVector.size())
	{
		AktEdgeIndex = CompleteFacet(NodeList, ElementList, EdgeList, AktEdgeIndex, DebugLevel);

/*
if (AktEdgeIndex >= 8)
{
//	SaveElements("tmp_elements.shp", ElementList, 0.0, DebugLevel);
	SaveNet("tmp.2dm", NodeList, ElementList, DebugLevel);
}
*/

		ftime (&Now);
		int NowMSec = Now.time * 1000 + Now.millitm;

		if (NowMSec > NextMSec)
		{
			NextMSec = NowMSec + TimeDiff;

			int DiffMSec = NowMSec - StartMSec;

			write_fortschritt("%d von ca. %d Elementen erzeugt\n", ElementList->size(), NodeList->size() * 2);

			if (ElementList->size() > 0)
			{
				unsigned int EleBisher = ElementList->size();
				unsigned int EleGesamt = NodeList->size() * 2;
				double TimeBisher = DiffMSec;
				double MitTimeBisher = TimeBisher / EleBisher;
				double MitTimeNoch = 0.2;

				double TimePerElement = (MitTimeBisher *  EleBisher + MitTimeNoch * EleGesamt - MitTimeNoch * EleBisher) / EleGesamt;
				double RestMSec = (EleGesamt - EleBisher) * TimePerElement;

				write_fortschritt("Dauer bisher: %.0f sec -> Dauer noch ca. %.0f sec\n", DiffMSec / 1000.0, RestMSec / 1000.0);
			}
		}
	}

	write_fortschritt("->Freigeben des Speichers der Kanten gestartet\n");
	delete EdgeList;
	EdgeList = 0;
	write_fortschritt("<-Freigeben des Speichers der Kanten beendet\n");

	NodeList->ReScale();

// SaveNet("_tmp.2dm", NodeList, ElementList, DebugLevel);


	write_fortschritt("->Lschen der berschssigen Elemente gestartet\n");

	int Count = 0;
	for (TElementList::iterator i=ElementList->begin(); i != ElementList->end(); i++)
	{
		TElement *Element = *i;

		if (Element->Node1->Z == NoZ || Element->Node2->Z == NoZ || Element->Node3->Z == NoZ)
		{
			Element->Typ = NO_ELEMENT;
			continue;
		}

		if (BegrenzungsPolygon && BegrenzungsPolygon->size() > 0)
		{
			TNode* Node1 = Element->Node1;
			TNode* Node2 = Element->Node2;
			TNode* Node3 = Element->Node3;

			long X12 = (Node1->X + Node2->X) / 2;
			long Y12 = (Node1->Y + Node2->Y) / 2;

			if (!BegrenzungsPolygon->IsInsideXYList(X12, Y12))
			{
				Element->Typ = NO_ELEMENT;
				continue;
			}

			long X23 = (Node2->X + Node3->X) / 2;
			long Y23 = (Node2->Y + Node3->Y) / 2;

			if (!BegrenzungsPolygon->IsInsideXYList(X23, Y23))
			{
				Element->Typ = NO_ELEMENT;
				continue;
			}

			long X31 = (Node3->X + Node1->X) / 2;
			long Y31 = (Node3->Y + Node1->Y) / 2;

			if (!BegrenzungsPolygon->IsInsideXYList(X31, Y31))
			{
				Element->Typ = NO_ELEMENT;
				continue;
			}
		}

		Count++;

		if (DebugLevel >= 9 || (DebugLevel >= 1 && Count % 5000 == 0) || (Count % 50000 == 0))
		{
			write_fortschritt("%d von %d Elementen berprft\n", Count, ElementList->size());
		}
	}

	write_fortschritt("<-Lschen der berschssigen Elemente beendet\n");

	write_fortschritt("<-Triangulierung der Knoten beendet\n");

	return (true);
}

