//
// $Id:	xy.cpp 64 2006-09-25 20:54:55Z mrchip $
//
// Copyright (C) 2005 STADT-LAND-FLUSS INGENIEURDIENSTE	GmbH
//
// Authors:
// Ulrich Kiel <u.kiel@S-L-F.de>
//
// This	program	is free	software under the GPL (>=v2)
// Read	the file COPYING coming	with WSPLGEN for details.
//

#ifdef __BORLANDC__
#pragma	hdrstop
#endif

//---------------------------------------------------------------------------
#include <algorithm>
#include <math.h>
#include "tools.h"

#include "file.h"
#include "xy.h"

//---------------------------------------------------------------------------
static const double		MinNodeDistanceKonst = 1;

//---------------------------------------------------------------------
// TXY
//---------------------------------------------------------------------

//---------------------------------------------------------------------
TXYZ::TXYZ(long x, long y, int z)
{
	X = x;
	Y = y;
	Z = z;
}

//---------------------------------------------------------------------
TXYZ::TXYZ(TXYZ *xy)
{
	X = xy->X;
	Y = xy->Y;
	Z = xy->Z;
}

//---------------------------------------------------------------------
// TNode
//---------------------------------------------------------------------

//---------------------------------------------------------------------
TNode::TNode(int nr, long x, long y, int z, int	wsp) : TXYZ (x, y, z)
{
	Nr = nr;
	Z = z;
	Wsp = wsp;
	EdgeIndex = 0;
}

//---------------------------------------------------------------------
TNode::TNode(TNode *node) : TXYZ (node->X, node->Y, node->Z)
{
	Nr = node->Nr;
	Wsp = node->Wsp;
	EdgeIndex = 0;
}

//---------------------------------------------------------------------
TNode::~TNode(void)
{
	ClearEdgeIndex();
}

//---------------------------------------------------------------------
void TNode::ClearEdgeIndex(void)
{
	TEdgeIndex *AktEdgeIndex = EdgeIndex;
	while (AktEdgeIndex)
	{
		TEdgeIndex *NextEdgeIndex = AktEdgeIndex->NextEdgeIndex;
		delete AktEdgeIndex;
		AktEdgeIndex = NextEdgeIndex;
	}
	EdgeIndex = 0;
}

//---------------------------------------------------------------------
void TNode::EraseEdgeIndex(TEdge *Edge)
{
/*
	TEdgeIndex *AktEdgeIndex1 = EdgeIndex;
	while (AktEdgeIndex1)
	{
		TEdgeIndex *NextEdgeIndex = AktEdgeIndex1->NextEdgeIndex;
		TEdge* E = AktEdgeIndex1->Edge;
		write_fortschritt("Knoten %d: Kante %d - %d\n",	Nr, E->Node1->Nr, E->Node2->Nr);
		AktEdgeIndex1 =	NextEdgeIndex;
	}

	write_fortschritt("\nKante %d -	%d soll	gelscht werden\n", Edge->Node1->Nr, Edge->Node2->Nr);
*/


	bool Found = false;
	TEdgeIndex *AktEdgeIndex = EdgeIndex;
	TEdgeIndex *PrevEdgeIndex = EdgeIndex;
	while (AktEdgeIndex)
	{
		TEdgeIndex *NextEdgeIndex = AktEdgeIndex->NextEdgeIndex;

		if (AktEdgeIndex->Edge == Edge)
		{
			Found =	true;

			if (AktEdgeIndex == EdgeIndex)	EdgeIndex = NextEdgeIndex;
			else							PrevEdgeIndex->NextEdgeIndex = NextEdgeIndex;

			delete AktEdgeIndex;
			AktEdgeIndex = 0;

			break;
		}

		PrevEdgeIndex =	AktEdgeIndex;
		AktEdgeIndex = NextEdgeIndex;
	}

/*
	TEdgeIndex *AktEdgeIndex2 = EdgeIndex;
	while (AktEdgeIndex2)
	{
		TEdgeIndex *NextEdgeIndex = AktEdgeIndex2->NextEdgeIndex;
		TEdge* E = AktEdgeIndex2->Edge;
		write_fortschritt("Knoten %d: Kante %d - %d\n",	Nr, E->Node1->Nr, E->Node2->Nr);
		AktEdgeIndex2 =	NextEdgeIndex;
	}
	write_fortschritt("\n");
*/

	if (false == Found)
	{
		dump_error(__FILE__, __LINE__, "Konnte zu lschende Kante nicht	finden\n");
	}
}

//---------------------------------------------------------------------
struct TNodeSortByNr
{
	 bool operator()(TNode*	const &Node1, TNode* const &Node2)
	 {
		  return(Node1->Nr < Node2->Nr);
	 }
};

//---------------------------------------------------------------------
// TPoint
//---------------------------------------------------------------------

//---------------------------------------------------------------------
TPoint::TPoint(long x, long y, int z, int meter) : TXYZ (x, y, z)
{
	Meter =	meter;
}

//---------------------------------------------------------------------
TPoint::TPoint(TPoint *point) :	TXYZ (point->X, point->Y, point->Z)
{
	Meter =	point->Meter;
}


//---------------------------------------------------------------------
// TElement
//---------------------------------------------------------------------

//---------------------------------------------------------------------
TElement::TElement(TNode* node1, TNode*	node2, TNode* node3)
{
	Typ = TRI;

	Node1 =	node1;
	Node2 =	node2;
	Node3 =	node3;
}

//---------------------------------------------------------------------
TElement::TElement(TNode* node1, TNode*	node2, TNode* node3, TNode* node4)
{
	Typ = QUAD;

	Node1 =	node1;
	Node2 =	node2;
	Node3 =	node3;
	Node4 =	node4;
}

//---------------------------------------------------------------------
TElement::TElement(TElement *element)
{
	Typ = element->Typ;

	Node1 =	element->Node1;
	Node2 =	element->Node2;
	Node3 =	element->Node3;
	Node4 =	element->Node4;
}

//---------------------------------------------------------------------
// TXYZList
//---------------------------------------------------------------------

//---------------------------------------------------------------------------
TXYZList::TXYZList(void)
{
	QuadTree = 0;
	MinX = 0;
	MaxX = 0;
	MinY = 0;
	MaxY = 0;
}

//---------------------------------------------------------------------------
TXYZList::~TXYZList(void)
{
	if (QuadTree)	delete QuadTree;
	QuadTree = 0;

	Clear();
}

//---------------------------------------------------------------------
void TXYZList::Clear(void)
{
	for (unsigned int i=0; i<size(); i++)
	{
		TXYZ *Xy = (*this)[i];
		delete Xy;
		Xy = 0;
	}
	clear();

	MinX = 0;
	MaxX = 0;
	MinY = 0;
	MaxY = 0;
}

//---------------------------------------------------------------------
TXYZList* TXYZList::Copy(void)
{
	TXYZList	*NewXyList = new TXYZList;

	for (unsigned int i=0; i<size(); i++)
	{
		TXYZ *OldXy = (*this)[i];
		TXYZ *NewXy = new TXYZ(OldXy);
		NewXyList->push_back(NewXy);
	}
	NewXyList->MinX	= MinX;
	NewXyList->MaxX	= MaxX;
	NewXyList->MinY	= MinY;
	NewXyList->MaxY	= MaxY;

	return(NewXyList);
}

//---------------------------------------------------------------------
void TXYZList::SortByXY(void)
{
	write_fortschritt("->Sortieren der Punkte gestartet\n");

	if (QuadTree)	delete QuadTree;
	QuadTree = new TQuadTree<TXYZList, TXYZ>();

	for (iterator i	= begin(); i !=	end(); i++)
	{
		TXYZ* Xy	= *i;
		QuadTree->Insert(MinX, MaxX, MinY, MaxY, Xy);
	}

	write_fortschritt("<-Sortieren der Punkte beendet\n");
}

//---------------------------------------------------------------------
TXYZ* TXYZList::FindByXY(long X, long Y)
{
	if (size() <= 0) return	(0);

	TXYZ* Xy = QuadTree->Search(X, Y, MinX, MaxX, MinY, MaxY);

	if (Xy)
	{
/*
		if (Xy->X == X && Xy->Y	== Y)
		{
			return (Xy);
		}
		else
		{
			dump_error(__FILE__, __LINE__, "Falschen Knoten	gefunden\n");
		}
*/

		return(Xy);
	}

/*
	for (unsigned int i=0; i<size(); i++)
	{
		TXY *Xy	= (* this)[i];

		if (Xy->X >= X - Eps &&	Xy->X <= X + Eps && Xy->Y >= Y - Eps &&	Xy->Y <= Y + Eps)
		{
			dump_error(__FILE__, __LINE__, "Existierenden Knoten nicht gefunden\n");
		}
	}
*/

	return (0);
}

//---------------------------------------------------------------------
TInsideTyp TXYZList::IsInsideXYList(long	X, long	Y)
{
	if (X <	MinX ||	X > MaxX || Y <	MinY ||	Y > MaxY)	return (NOT_INSIDE);

	int alpha = IsInside(this, X, Y);

	if (alpha == 0)		return (NOT_INSIDE);	// not inside
	if (alpha == 4)		return (INSIDE);		// inside
	if (alpha == -4)	return (INSIDE);		// inside
	if (alpha == -1)	return (ON_LINE);		// on line
	if (alpha == 2)	    return (ON_LINE);		// on line twisted
	if (alpha == -2)    return (ON_LINE);		// on line twisted

	if (alpha == 8)		return (INSIDE);		// double inside
	if (alpha == -8)	return (INSIDE);		// double inside

	if (alpha == 16)	return (INSIDE);		// triple inside
	if (alpha == -16)	return (INSIDE);		// triple inside

	if (alpha % 4 == 0)	return (INSIDE);		// some	inside

	dump_error(__FILE__, __LINE__, "Alpha (%d) ungltig\n",	alpha);

	return (UNDEFINED);
}

//---------------------------------------------------------------------
void TXYZList::Add(TXYZ *Xyz)
{
	push_back(Xyz);

	if (size() == 1)
	{
		MinX = Xyz->X;
		MaxX = Xyz->X;
		MinY = Xyz->Y;
		MaxY = Xyz->Y;
	}
	else
	{
		if (MinX > Xyz->X) MinX = Xyz->X;
		if (MaxX < Xyz->X) MaxX = Xyz->X;
		if (MinY > Xyz->Y) MinY = Xyz->Y;
		if (MaxY < Xyz->Y) MaxY = Xyz->Y;
	}
}

//---------------------------------------------------------------------
void TXYZList::Add(long X, long Y, int Z)
{
	TXYZ *Xyz = new TXYZ(X, Y, Z);
	Add(Xyz);
}

//---------------------------------------------------------------------
TXYZList* TXYZList::Cut(TXYZList* ProfilPolygon, TProfil* CutProfil)
{
	write_fortschritt("->Beschneiden des Begrenzungspolygon gestartet\n");

	write_fortschritt("Beschneidung mit Profil %.4f\n", CutProfil->Station / 10000.0);

	// Zuerst (wenn notwendig )den ersten Punkt am Ende hinzufgen
	TXYZ* XyFirst = *begin();
	TXYZ* XyLast = *rbegin();

	if (XyFirst->X != XyLast->X || XyFirst->Y != XyLast->Y)
	{
		Add(XyFirst->X, XyFirst->Y);
	}

	// Jetzt das CutProfil verlngern, damit es wirklich zwei Schnittpunkte gibt
	TXYZList* CutXyList = new TXYZList();

	long X = 0;
	long Y = 0;

	TPoint*	FPoint = *CutProfil->PointList->begin();
	TPoint*	LPoint = *CutProfil->PointList->rbegin();

	CutProfil->GetXY(-999999, &X, &Y);
	CutXyList->Add(X, Y);
	for	(TPointList::iterator i=CutProfil->PointList->begin(); i != CutProfil->PointList->end(); i++)
	{
		TPoint*	Point =	*i;

		CutXyList->Add(Point->X, Point->Y);
	}
	CutProfil->GetXY(999999, &X, &Y);
	CutXyList->Add(X, Y);

	int 	After0 = -1;
	long	X0 = FPoint->X;
	long	Y0 = FPoint->Y;
	long double	SX0 = 0;
	long double	SY0 = 0;
	long double	D00 = 0.0;
	long double	D01 = 0.0;

	int 	After1 = -1;
	long	X1 = LPoint->X;
	long	Y1 = LPoint->Y;
	long double  SX1 = 0;
	long double  SY1 = 0;
	long double	D10 = 0.0;
	long double	D11 = 0.0;

	TXYZ* Xy0 = 0;
	TXYZ* Xy1 = 0;
	int Index = 0;
	for (TXYZList::iterator i=begin(); i	!= end(); i++, Index++)
	{
		Xy0 = Xy1;
		Xy1 = *i;

		if (Xy0	== 0)
		{
			continue;
		}

		if (Xy0->X == Xy1->X && Xy0->Y == Xy1->Y)
		{
			continue;
		}

		TXYZ* CutXy0 = 0;
		TXYZ* CutXy1 = 0;
		bool Schnitt = false;
		double SX = 0;
		double SY = 0;
		for	(TXYZList::iterator j=CutXyList->begin(); j != CutXyList->end(); j++)
		{
			CutXy0 = CutXy1;
			CutXy1 = *j;

			if (CutXy0 == 0)	continue;

			Schnitt	= Calc2Schnitt(Xy0->X, Xy0->Y, Xy1->X, Xy1->Y, CutXy0->X, CutXy0->Y, CutXy1->X,	CutXy1->Y, &SX,	&SY);

			if (Schnitt)
			{
				long double D0 = (SX - X0) * (SX - X0) + (SY - Y0) * (SY - Y0);
				long double D1 = (SX - X1) * (SX - X1) + (SY - Y1) * (SY - Y1);

				if (After0 == -1 && After1 == -1)
				{
					if (D0 < D1)
					{
						After0 = Index;
						SX0 = SX;
						SY0 = SY;
						D00 = D0;
						D01 = D1;
					}
					else
					{
						After1 = Index;
						SX1 = SX;
						SY1 = SY;
						D10 = D0;
						D11 = D1;
					}
					break;
				}
				else if (After1 == -1) // -> Also After0 != -1
				{
					if (SX == SX0 && SY == SY0)	continue;

					if (D0 + D01 < D00 + D1)
					{
						After1 = After0;
						SX1 = SX1;
						SY1 = SY1;
						D10 = D00;
						D11 = D01;

						After0 = Index;
						SX0 = SX;
						SY0 = SY;
						D00 = D0;
						D01 = D1;
					}
					else
					{
						After1 = Index;
						SX1 = SX;
						SY1 = SY;
						D10 = D0;
						D11 = D1;
					}
					break;
				}
				else if (After0 == -1) // -> Also After1 != -1
				{
					if (SX == SX1 && SY == SY1)	continue;

					if (D0 + D11 < D10 + D1)
					{
						After0 = Index;
						SX0 = SX;
						SY0 = SY;
						D00 = D0;
						D01 = D1;
					}
					else
					{
						After0 = After0;
						SX0 = SX1;
						SY0 = SY1;
						D00 = D10;
						D01 = D11;

						After1 = Index;
						SX1 = SX;
						SY1 = SY;
						D10 = D0;
						D11 = D1;
					}
					break;
				}
				else
				{
					if (SX == SX0 && SY == SY0)	continue;
					if (SX == SX1 && SY == SY1)	continue;

					if (D0 + D01 < D00 + D11 && D0 + D01 < D00 + D1 && D0 + D01 < D0 + D11 && D0 + D01 < D10 + D1)
					{
						After1 = After0;
						SX1 = SX0;
						SY1 = SY0;
						D10 = D00;
						D11 = D01;

						After0 = Index;
						SX0 = SX;
						SY0 = SY;
						D00 = D0;
						D01 = D1;
					}
					else if (D0 + D11 < D00 + D11 && D0 + D11 < D00 + D1 && D0 + D11 < D0 + D01 && D0 + D11 < D10 + D1)
					{
						After0 = Index;
						SX0 = SX;
						SY0 = SY;
						D00 = D0;
						D01 = D1;
					}
					else if (D00 + D1 < D00 + D11 && D00 + D1 < D0 + D01 && D00 + D1 < D0 + D11 && D00 + D1 < D10 + D1)
					{
						After1 = Index;
						SX1 = SX;
						SY1 = SY;
						D10 = D0;
						D11 = D1;
					}
					else if (D10 + D1 < D00 + D11 && D10 + D1 < D00 + D1 && D10 + D1 < D0 + D11 && D10 + D1 < D0 + D01)
					{
						After0 = After1;
						SX0 = SX1;
						SY0 = SY1;
						D00 = D10;
						D01 = D11;

						After1 = Index;
						SX1 = SX;
						SY1 = SY;
						D10 = D0;
						D11 = D1;
					}
					break;
				}
			}
		}
	}

	if (After0 == -1 && After1 == -1)
	{
		TXYZList* SideZero = new TXYZList();

		TXYZList::iterator iter=begin();
		unsigned int i = 0;
		while (i < size())
		{
			TXYZ *Xy = *iter++;

			SideZero->Add(Xy->X, Xy->Y);

			i++;
		}

		write_fortschritt("<-Beschneiden des Begrenzungspolygon beendet\n");

		return (SideZero);
	}
	else if (After0 == -1 || After1 == -1)
	{
		dump_error(__FILE__, __LINE__, "Es wurden nur ein Schnittpunkt gefunden.\n");

		return (0);
	}

	// Damit die Reihenfolge stimmt evtl. vertauschen
	if (After0 > After1)
	{
		int TempAfter = After0;
		long double TempSX = SX0;
		long double TempSY = SY0;

		After0 = After1;
		SX0 = SX1;
		SY0 = SY1;

		After1 = TempAfter;
		SX1 = TempSX;
		SY1 = TempSY;
	}

	TXYZList* SideZero = new TXYZList();
	TXYZList* SideOne = new TXYZList();

	TXYZList::iterator iter=begin();
	unsigned int i = 0;
	while (i < (unsigned int)After0)
	{
		TXYZ *Xy = *iter++;

		SideZero->Add(Xy->X, Xy->Y);

		i++;
	}

	long LSX0 = (long)(SX0 + 0.5);
	long LSY0 = (long)(SY0 + 0.5);
	SideZero->Add(LSX0, LSY0);

	SideOne->Add(LSX0, LSY0);

	while (i < (unsigned int)After1)
	{
		TXYZ *Xy = *iter++;

		SideOne->Add(Xy->X, Xy->Y);

		i++;
	}

	long LSX1 = (long)(SX1 + 0.5);
	long LSY1 = (long)(SY1 + 0.5);
	SideOne->Add(LSX1, LSY1);

	SideZero->Add(LSX1, LSY1);

	while (i < size())
	{
		TXYZ *Xy = *iter++;

		SideZero->Add(Xy->X, Xy->Y);

		i++;
	}

/*
	write_fortschritt("\n");

	int Count = 0;
	for	(TXYZList::iterator i=SideZero->begin();	i != SideZero->end(); i++)
	{
		TXY* Xy	= *i;

		write_fortschritt("%d %d %d\n",	++Count, Xy->X,	Xy->Y);
	}
	write_fortschritt("\n");

	Count = 0;
	for	(TXYZList::iterator i=SideOne->begin(); i != SideOne->end(); i++)
	{
		TXY* Xy	= *i;

		write_fortschritt("%d %d %d\n",	++Count, Xy->X,	Xy->Y);
	}
	write_fortschritt("\n");
*/

	// Nun haben wir 2 Polygone
	// Als nchstes	wird getestet, welches das richtige ist.


	// Falls das ProfilPolygon grer als das Begrenzungspolygon ist,
	// werden hier die Punkte innerhalb	gezhlt
	int IsInsideZeroCount =	0;
	for	(TXYZList::iterator i=SideZero->begin(); i != SideZero->end(); i++)
	{
		TXYZ* Xy = *i;

		if (ProfilPolygon->IsInsideXYList(Xy->X, Xy->Y))	IsInsideZeroCount++;
	}

	int IsInsideOneCount = 0;
	for	(TXYZList::iterator i=SideOne->begin(); i != SideOne->end(); i++)
	{
		TXYZ* Xy = *i;

		if (ProfilPolygon->IsInsideXYList(Xy->X, Xy->Y))	IsInsideOneCount++;
	}

	// Falls das ProfilPolygon kleiner als das Begrenzungspolygon ist,
	// werden hier die Punkte innerhalb	gezhlt
	for	(TXYZList::iterator i=ProfilPolygon->begin(); i != ProfilPolygon->end();	i++)
	{
		TXYZ* Xy = *i;

		if (SideZero->IsInsideXYList(Xy->X, Xy->Y))	IsInsideZeroCount++;
		if (SideOne->IsInsideXYList(Xy->X, Xy->Y))	IsInsideOneCount++;
	}

	delete CutXyList;

	write_fortschritt("<-Beschneiden des Begrenzungspolygon beendet\n");

	if (IsInsideZeroCount > IsInsideOneCount)
	{
		delete SideOne;
		return (SideZero);
	}
	else
	{
		delete SideZero;
		return (SideOne);
	}
}

//---------------------------------------------------------------------
// TNodeList
//---------------------------------------------------------------------

//---------------------------------------------------------------------------
TNodeList::TNodeList(void)
{
	QuadTree = 0;
	SearchNodeList = 0;
	MaxNr =	-1;
	MinX = 0;
	MaxX = 0;
	MinY = 0;
	MaxY = 0;
	DiffX =	0;
	DiffY =	0;
}

//---------------------------------------------------------------------------
TNodeList::~TNodeList(void)
{
	Clear();
}

//---------------------------------------------------------------------
void TNodeList::Clear(void)
{
	write_fortschritt("->Lschen des Suchbaumes der Knoten gestartet\n");
	if (QuadTree)	delete QuadTree;
	QuadTree = 0;
	write_fortschritt("<-Lschen des Suchbaumes der Knoten beendet\n");

	if (SearchNodeList)
	{
		SearchNodeList->clear();
		delete SearchNodeList;
		SearchNodeList = 0;
	}

	if (size() > 0)
	{
		int Count = 0;
		for (TNodeList::iterator i = begin(); i	!= end(); i++)
		{
			TNode *Node = *i;
			delete Node;
			Node = 0;

			Count++;

			if (Count % 100000 == 0)	write_fortschritt("%d von %d Objekte freigegeben\n", Count, size());
		}
		write_fortschritt("%d Objekte freigegeben\n", Count);
	}

	clear();

	MaxNr =	-1;

	// Die Minima und Maxima werden falsch belegt, damit Add erkennen kann, ob
	// die Maxima und Minima schon definiert wurden
	MinX = 1;
	MaxX = 0;
	MinY = 1;
	MaxY = 0;
	DiffX =	0;
	DiffY =	0;
}

//---------------------------------------------------------------------
TNodeList* TNodeList::Copy(void)
{
	TNodeList *NewNodeList = new TNodeList;

	for (unsigned int i=0; i<size(); i++)
	{
		TNode *OldNode = (*this)[i];
		TNode *NewNode = new TNode(OldNode);
		NewNodeList->push_back(NewNode);
	}

	NewNodeList->MaxNr = MaxNr;
	NewNodeList->MinX = MinX;
	NewNodeList->MaxX = MaxX;
	NewNodeList->MinY = MinY;
	NewNodeList->MaxY = MaxY;
	NewNodeList->DiffX = DiffX;
	NewNodeList->DiffY = DiffY;

	return(NewNodeList);
}

//---------------------------------------------------------------------
void TNodeList::CreateQuadTree(long minX, long minY, long maxX, long maxY)
{
	QuadTree = new TQuadTree<TNodeList, TNode>();
	MinX = minX;
	MinY = minY;
	MaxX = maxX;
	MaxY = maxY;
}

//---------------------------------------------------------------------
void TNodeList::Add(TNode *Node)
{
	if (capacity() < size()	+ 1)
	{
		reserve((size_t)(capacity() * 1.1 + 100000));
		write_fortschritt("Kapazitt der Knotenliste wurde auf %d erhht\n", capacity());
	}

	push_back(Node);

	if (MinX > MaxX || MinY > MaxY)
	{
		MaxNr =	Node->Nr;
		MinX = Node->X;
		MaxX = Node->X;
		MinY = Node->Y;
		MaxY = Node->Y;
	}
	else
	{
		if (MaxNr < Node->Nr) MaxNr = Node->Nr;
		if (MinX > Node->X) MinX = Node->X;
		if (MaxX < Node->X) MaxX = Node->X;
		if (MinY > Node->Y) MinY = Node->Y;
		if (MaxY < Node->Y) MaxY = Node->Y;
	}
}

//---------------------------------------------------------------------
bool TNodeList::AddWithQuadTree(TNode *Node)
{
	if (Node->X < MinX || Node->X >	MaxX ||	Node->Y	< MinY || Node->Y > MaxY)
	{
		dump_error(__FILE__, __LINE__, "AddWithQuadTree	funktioniert nicht mit Knoten die ausserhalb legen\nKnoten %d %d %d\nBounding Box %d %d	%d %d\n", Node->Nr, Node->X, Node->Y, MinX, MaxX, MinY,	MaxY);
	}

	// Es werden nur neue Knoten aufgenommen
	// Doppelte werden so vermieden
	if (QuadTree->Insert(MinX, MaxX, MinY, MaxY, Node))
	{
		Add(Node);
		return (true);
	}
	else
	{
		return (false);
	}
}

//---------------------------------------------------------------------
bool TNodeList::AddUnique(TNode *node)
{
	// Es werden nur neue Knoten aufgenommen
	// Doppelte werden so vermieden

	bool Found = false;
	for (iterator i = begin(); i != end(); i++)
	{
		TNode* Node = *i;

		if (Node->Nr == node->Nr)
		{
			Found = true;
			break;
		}
	}

	if (false == Found)
	{
		Add(node);
		return (true);
	}
	else
	{
		return (false);
	}
}

//---------------------------------------------------------------------
void TNodeList::SortByXY(void)
{
	write_fortschritt("->Sortieren der Knoten gestartet\n");

	write_fortschritt("->Lschen und Neuaufbauen des Suchbaumes der Knoten gestartet\n");
	if (QuadTree)	delete QuadTree;
	QuadTree = new TQuadTree<TNodeList, TNode>();
	write_fortschritt("<-Lschen und Neuaufbauen des Suchbaumes der Knoten beendet\n");

	for (iterator i	= begin(); i !=	end(); i++)
	{
		TNode* Node = *i;

		QuadTree->Insert(MinX, MaxX, MinY, MaxY, Node);
	}

	write_fortschritt("<-Sortieren der Knoten beendet\n");
}

//---------------------------------------------------------------------
void TNodeList::SortByNr(void)
{
	write_fortschritt("->Sortieren der Knoten gestartet\n");

	std::sort(this->begin(), this->end(), TNodeSortByNr());

	write_fortschritt("<-Sortieren der Knoten beendet\n");
}

//---------------------------------------------------------------------
TNode* TNodeList::FindByNr(int nr)
{
	if (size() <= 0) return	(0);

	int Under = 0;
	int Over = size()-1;
	while (Under <=	Over)
	{
		int Index = (Over + Under) / 2;

		TNode *Node = (*this)[Index];

		if (nr < Node->Nr)
		{
			Over = Index - 1;
		}
		else if	(nr > Node->Nr)
		{
			Under =	Index +	1;
		}
		else
		{
			return(Node);
		}
	}

	return (0);
}

//---------------------------------------------------------------------
TNode* TNodeList::FindByXY(long	X, long	Y)
{
	if (size() <= 0) return	(0);

	TNode* Node = QuadTree->Search(X, Y, MinX, MaxX, MinY, MaxY);

	if (Node)
	{
/*
		if (Node->X == X && Node->Y == Y)
		{
			return (Node);
		}
		else
		{
			dump_error(__FILE__, __LINE__, "Falschen Knoten	gefunden\n");
		}
*/

		return (Node);
	}

/*
	for (unsigned int i=0; i<size(); i++)
	{
		TNode *Node = (* this)[i];

		if (Node->X >= X - Eps && Node->X <= X + Eps &&	Node->Y	>= Y - Eps && Node->Y <= Y + Eps)
		{
			dump_error(__FILE__, __LINE__, "Existierenden Knoten (%.4f,%.4f	-> %.4f,%.4f) nicht gefunden\n", X, Y, Node->X,	Node->Y);
		}
	}
*/

	return (0);
}

//---------------------------------------------------------------------
TNodeList* TNodeList::FindAllByXY(double X, double Y, double R)
{
	// Wichtig ist das kleingeschriebene clear
	// In dieser SearchNodelist sind Knoten	drin, die nur geliehen sind.
	// Die Knoten selbst sollen nicht gelscht werden

	if (SearchNodeList)	SearchNodeList->clear();
	else				SearchNodeList = new TNodeList;

	if (size() <= 0) return(SearchNodeList);

	QuadTree->SearchAll(X, Y, R, SearchNodeList, MinX, MaxX, MinY, MaxY);

	return (SearchNodeList);
}

//---------------------------------------------------------------------
bool TNodeList::Interpolate(TProfilList	*SperrenList, long X, long Y, double R,	int *ZWert, int	*WspWert)
{
	TNodeList *SearchNodeList = FindAllByXY(X, Y, R);

	TNode* TLNode =	0;
	double TLD = 0.0;

	TNode* BLNode =	0;
	double BLD = 0.0;

	TNode* TRNode =	0;
	double TRD = 0.0;

	TNode* BRNode =	0;
	double BRD = 0.0;

	for (TNodeList::iterator i = SearchNodeList->begin(); i	!= SearchNodeList->end(); i++)
	{
		TNode* Node = *i;

		long NDx = X - Node->X;
		long NDy = Y - Node->Y;
		if (NDx	== 0 &&	NDy == 0)
		{
			*ZWert = Node->Z;
			*WspWert = Node->Wsp;

			return (true);
		}

		double DummyX =	0.0;
		double DummyY =	0.0;
		double DummyZ =	0.0;
		double DummyMeter = 0.0;

		if (SperrenList	&& SperrenList->size() > 0 && CalcProfilSchnitt(Node->X, Node->Y, X, Y,	SperrenList, &DummyX, &DummyY, &DummyZ, &DummyMeter)) continue;

		if (Node->X >= X && Node->Y > Y)
		{
			if (0 == TRNode)
			{
				TRNode = Node;
				long Dx	= TRNode->X - X;
				long Dy	= TRNode->Y - Y;

				TRD = Dx*Dx + Dy*Dy;
				continue;
			}
			else
			{
				long Dx	= Node->X - X;
				long Dy	= Node->Y - Y;

				double D = Dx*Dx + Dy*Dy;

				if (D <	TRD)
				{
					TRNode = Node;
					TRD = D;
				}
			}
			continue;
		}

		if (Node->X < X	&& Node->Y >= Y)
		{
			if (0 == TLNode)
			{
				TLNode = Node;
				long Dx	= TLNode->X - X;
				long Dy	= TLNode->Y - Y;

				TLD = Dx*Dx + Dy*Dy;
				continue;
			}
			else
			{
				long Dx	= Node->X - X;
				long Dy	= Node->Y - Y;

				double D = Dx*Dx + Dy*Dy;

				if (D <	TLD)
				{
					TLNode = Node;
					TLD = D;
				}
			}
			continue;
		}

		if (Node->X <= X && Node->Y < Y)
		{
			if (0 == BLNode)
			{
				BLNode = Node;
				long Dx	= BLNode->X - X;
				long Dy	= BLNode->Y - Y;

				BLD = Dx*Dx + Dy*Dy;
				continue;
			}
			else
			{
				long Dx	= Node->X - X;
				long Dy	= Node->Y - Y;

				double D = Dx*Dx + Dy*Dy;

				if (D <	BLD)
				{
					BLNode = Node;
					BLD = D;
				}
			}
			continue;
		}

		if (Node->X > X	&& Node->Y <= Y)
		{
			if (0 == BRNode)
			{
				BRNode = Node;
				long Dx	= BRNode->X - X;
				long Dy	= BRNode->Y - Y;

				BRD = Dx*Dx + Dy*Dy;
				continue;
			}
			else
			{
				long Dx	= Node->X - X;
				long Dy	= Node->Y - Y;

				double D = Dx*Dx + Dy*Dy;

				if (D <	BRD)
				{
					BRNode = Node;
					BRD = D;
				}
			}
			continue;
		}
	}

	int ZAnzahl = 0;
	double ZSumme =	0.0;
	int WspAnzahl =	0;
	double WspSumme	= 0.0;
	if (TLNode != 0)
	{
		TLD = TLD * TLD;
		ZSumme = ZSumme	+ 1 / TLD;
		ZAnzahl++;

		if (TLNode->Wsp	!= NoWsp)
		{
			WspSumme = WspSumme + 1	/ TLD;
			WspAnzahl++;
		}
	}
	if (TRNode != 0)
	{
		TRD = TRD * TRD;
		ZSumme = ZSumme	+ 1 / TRD;
		ZAnzahl++;

		if (TRNode->Wsp	!= NoWsp)
		{
			WspSumme = WspSumme + 1	/ TRD;
			WspAnzahl++;
		}
	}
	if (BLNode != 0)
	{
		BLD = BLD * BLD;
		ZSumme = ZSumme	+ 1 / BLD;
		ZAnzahl++;

		if (BLNode->Wsp	!= NoWsp)
		{
			WspSumme = WspSumme + 1	/ BLD;
			WspAnzahl++;
		}
	}
	if (BRNode != 0)
	{
		BRD = BRD * BRD;
		ZSumme = ZSumme	+ 1 / BRD;
		ZAnzahl++;

		if (BRNode->Wsp	!= NoWsp)
		{
			WspSumme = WspSumme + 1	/ BRD;
			WspAnzahl++;
		}
	}

	if (WspAnzahl >	0)
	{
		int Wsp	= 0;

		if (TLNode != 0	&& TLNode->Wsp != NoWsp) Wsp = Wsp + (int)(1 / TLD / WspSumme *	TLNode->Wsp + 0.5);
		if (TRNode != 0	&& TRNode->Wsp != NoWsp) Wsp = Wsp + (int)(1 / TRD / WspSumme *	TRNode->Wsp + 0.5);
		if (BLNode != 0	&& BLNode->Wsp != NoWsp) Wsp = Wsp + (int)(1 / BLD / WspSumme *	BLNode->Wsp + 0.5);
		if (BRNode != 0	&& BRNode->Wsp != NoWsp) Wsp = Wsp + (int)(1 / BRD / WspSumme *	BRNode->Wsp + 0.5);

		*WspWert = Wsp;
	}
	else
	{
		*WspWert = NoWsp;
	}

	if (ZAnzahl > 0)
	{
		int Z =	0;

		if (TLNode != 0) Z = Z + (int)(1 / TLD / ZSumme	* TLNode->Z + 0.5);
		if (TRNode != 0) Z = Z + (int)(1 / TRD / ZSumme	* TRNode->Z + 0.5);
		if (BLNode != 0) Z = Z + (int)(1 / BLD / ZSumme	* BLNode->Z + 0.5);
		if (BRNode != 0) Z = Z + (int)(1 / BRD / ZSumme	* BRNode->Z + 0.5);

		*ZWert = Z;

		return (true);
	}

	return (false);
}

//---------------------------------------------------------------------
void TNodeList::ClearEdgeIndex(void)
{
	if (size() > 0)
	{
		write_fortschritt("->Lschen des Knoten/Kanten-Index gestartet\n");

		int Count = 0;
		for (iterator i	= begin(); i !=	end(); i++)
		{
			TNode *Node = *i;
			Node->ClearEdgeIndex();

			Count++;

			if (Count % 100000 == 0)	write_fortschritt("%d von %d Objekte freigegeben\n", Count, size());
		}

		write_fortschritt("<-Lschen des Knoten/Kanten-Index beendet\n");
	}
}

//---------------------------------------------------------------------
int TNodeList::GetMaxDepth(int DebugLevel)
{
	write_fortschritt("->Ermittlung	der maximalen Wassertiefe gestartet\n");

	int MaxDepth = 0;
	int Count = 0;
	for (iterator i	= begin(); i !=	end(); i++)
	{
		TNode* Node = *i;

		if (Node->Wsp == NoWsp)	continue;

		int Depth = Node->Wsp -	Node->Z;

		if (Depth < 0)
		{
			dump_error(__FILE__, __LINE__, "Der Wasserstand	ist negativ\n");
		}

		if (Count == 0 || Depth	> MaxDepth)	MaxDepth = Depth;

		Count++;

		if (DebugLevel >= 9 || (DebugLevel >= 1	&& Count % 10000 == 0) || (Count > 0 &&	Count %	100000 == 0))
		{
			write_fortschritt("Knoten %d von %d untersucht.\n", Count, size());
		}
	}
	write_fortschritt("Knoten %d von %d untersucht.\n", Count, size());

	write_fortschritt("Maximale Wassertiefe: %.3f.\n", MaxDepth / 100.0);

	write_fortschritt("<-Ermittlung	der maximalen Wassertiefe beendet\n");

	return (MaxDepth);
}

//---------------------------------------------------------------------
int TNodeList::GetMaxWsp(int DebugLevel)
{
	write_fortschritt("->Ermittlung	des maximalen Wasserstandes gestartet\n");

	int MaxWsp = 0;
	int Count = 0;
	for (iterator i	= begin(); i !=	end(); i++)
	{
		TNode* Node = *i;

		if (Node->Wsp == NoWsp)	continue;

		int Wsp	= Node->Wsp;

		if (Wsp	< 0.0)
		{
			dump_error(__FILE__, __LINE__, "Der Wasserstand	ist negativ\n");
		}

		if (Count == 0 || Wsp >	MaxWsp)	MaxWsp = Wsp;

		Count++;

		if (DebugLevel >= 9 || (DebugLevel >= 1	&& Count % 10000 == 0) || (Count > 0 &&	Count %	100000 == 0))
		{
			write_fortschritt("Knoten %d von %d untersucht.\n", Count, size());
		}
	}
	write_fortschritt("Knoten %d von %d untersucht.\n", Count, size());

	write_fortschritt("Maximaler Wasserstand: %.3f.\n", MaxWsp / 100.0);

	write_fortschritt("<-Ermittlung	des maximalen Wasserstandes beendet\n");

	return (MaxWsp);
}

//---------------------------------------------------------------------
int TNodeList::GetMaxZ(int DebugLevel)
{
	write_fortschritt("->Ermittlung	der maximale Gelndehhe gestartet\n");

	int MaxZ = 0;
	int Count = 0;
	for (iterator i	= begin(); i !=	end(); i++)
	{
		TNode* Node = *i;

		int Z =	Node->Z;

		if (Count == 0 || Z > MaxZ)	MaxZ = Z;

		Count++;

		if (DebugLevel >= 9 || (DebugLevel >= 1	&& Count % 10000 == 0) || (Count > 0 &&	Count %	100000 == 0))
		{
			write_fortschritt("Knoten %d von %d untersucht.\n", Count, size());
		}
	}
	write_fortschritt("Knoten %d von %d untersucht.\n", Count, size());

	write_fortschritt("Maximale Gelndehhe: %.3f.\n", MaxZ	/ 100.0);

	write_fortschritt("<-Ermittlung	der maximale Gelndehhe beendet\n");

	return (MaxZ);
}

//---------------------------------------------------------------------
void TNodeList::Scale(void)
{
	write_fortschritt("->Verschiebung des Koordinatensystems gestartet\n");

	DiffX =	MinX;
	DiffY =	MinY;

	MinX = MinX - DiffX;
	MaxX = MaxX - DiffX;
	MinY = MinY - DiffY;
	MaxY = MaxY - DiffY;

	for (iterator i = begin(); i != end(); i++)
	{
		TNode* Node = *i;

		Node->X	= Node->X - DiffX;
		Node->Y	= Node->Y - DiffY;
	}

	write_fortschritt("<-Verschiebung des Koordinatensystems beendet\n");
}

//---------------------------------------------------------------------
void TNodeList::ReScale(void)
{
	write_fortschritt("->Zurckverschiebung des Koordinatensystems gestartet\n");

	for (iterator i	= begin(); i !=	end(); i++)
	{
		TNode* Node = *i;

		Node->X	= Node->X + DiffX;
		Node->Y	= Node->Y + DiffY;
	}

	MinX = MinX + DiffX;
	MaxX = MaxX + DiffX;
	MinY = MinY + DiffY;
	MaxY = MaxY + DiffY;

	DiffX =	0;
	DiffY =	0;

	write_fortschritt("<-Zurckverschiebung des Koordinatensystems beendet\n");
}

//---------------------------------------------------------------------
// TElementList
//---------------------------------------------------------------------

//---------------------------------------------------------------------------
TElementList::~TElementList(void)
{
	Clear();
}

//---------------------------------------------------------------------
void TElementList::Clear(void)
{
	if (size() > 0)
	{
		int Count = 0;
		for (iterator i=begin(); i != end(); i++)
		{
			TElement *Element = *i;
			delete Element;
			Element	= 0;

			Count++;

			if (Count % 100000 == 0)	write_fortschritt("%d von %d Objekte freigegeben\n", Count, size());
		}
		write_fortschritt("%d Objekte freigegeben\n", Count);
	}
	clear();
}

//---------------------------------------------------------------------
void TElementList::Add(TElement* Element)
{
	if (capacity() < size()	+ 1)
	{
		reserve((size_t)(capacity() * 1.1 + 200000));
		write_fortschritt("Kapazitt der Elementliste wurde auf %d erhht\n", capacity());
	}

	push_back(Element);
}

//---------------------------------------------------------------------------
int TElementList::HelpChangeElements(TNodeList *NodeList, TProfilList *LinienList, int DebugLevel)
{
	int MaxNodeNr =	NodeList->MaxNr;

	int ChangeCount	= 0;
	int Count = 0;
	for (unsigned int i = 0; i < size(); i++)
	{
		TElement* Element = (*this)[i];

		Count++;

		if (DebugLevel >= 9 || (DebugLevel >= 1	&& Count % 5000	== 0) || (Count	% 50000	== 0))
		{
			write_fortschritt("%d von %d Elementen, %d Elemente angepasst\n", Count, size(), ChangeCount);
		}

		// Hier	werden die zum Lschen markierten Elemente ignoriert.
		if (Element->Typ == NO_ELEMENT)	continue;


		// Hier	werden die neuen  Elemente wieder auf einen korrekten Typ gesetzt und dann ignoriert.
		if (Element->Typ == NEWTRI)
		{
			Element->Typ = TRI;
			continue;
		}

		if (Element->Typ == QUAD)
		{
			dump_error(__FILE__, __LINE__, "Viereckselemente werden	im Moment noch nicht untersttzt");
		}

		TNode* Node1 = Element->Node1;
		TNode* Node2 = Element->Node2;
		TNode* Node3 = Element->Node3;

		TElement* NewElement = 0;

		double dummyMeter = 0.0;

		if (Node1->X ==	Node2->X && Node1->Y ==	Node2->Y)
		{
			Element->Typ = NO_ELEMENT;
			continue;
		}

		if (Node2->X ==	Node3->X && Node2->Y ==	Node3->Y)
		{
			Element->Typ = NO_ELEMENT;
			continue;
		}

		if (Node3->X ==	Node1->X && Node3->Y ==	Node1->Y)
		{
			Element->Typ = NO_ELEMENT;
			continue;
		}

		double XD12 = 0.0;
		double YD12 = 0.0;
		double ZD12 = 0.0;
		double XD23 = 0.0;
		double YD23 = 0.0;
		double ZD23 = 0.0;
		double XD31 = 0.0;
		double YD31 = 0.0;
		double ZD31 = 0.0;

		long X12 = 0;
		long Y12 = 0;
		int Z12 = 0;
		long X23 = 0;
		long Y23 = 0;
		int Z23 = 0;
		long X31 = 0;
		long Y31 = 0;
		int Z31 = 0;

		TNode* Node12 =	0;
		TNode* Node23 =	0;
		TNode* Node31 =	0;


		bool Schnitt12 = CalcProfilSchnitt(Node1->X, Node1->Y, Node2->X, Node2->Y, LinienList, &XD12, &YD12, &ZD12, &dummyMeter);
		bool Schnitt23 = CalcProfilSchnitt(Node2->X, Node2->Y, Node3->X, Node3->Y, LinienList, &XD23, &YD23, &ZD23, &dummyMeter);
		bool Schnitt31 = CalcProfilSchnitt(Node3->X, Node3->Y, Node1->X, Node1->Y, LinienList, &XD31, &YD31, &ZD31, &dummyMeter);


		if (Schnitt12)
		{
			X12 = (long)(XD12 + 0.5);
			Y12 = (long)(YD12 + 0.5);
			Z12 = (int)(ZD12 + 0.5);

			TNodeList* FoundNodeList = NodeList->FindAllByXY(X12, Y12, 4);

			if (FoundNodeList->size() > 0)
			{
				for (TNodeList::iterator j=FoundNodeList->begin(); j !=	FoundNodeList->end(); j++)
				{
					TNode* Node = *j;

					if (Node1->X ==	Node->X	&& Node1->Y == Node->Y)
					{
						Node->Z	= Z12;
						Schnitt12 = false;
					}
					else if	(Node2->X == Node->X &&	Node2->Y == Node->Y)
					{
						Node->Z	= Z12;
						Schnitt12 = false;
					}
					else if	(Node3->X == Node->X &&	Node3->Y == Node->Y)
					{
						Node->Z	= Z12;
						Schnitt12 = false;
					}

					double D = (Node->X - X12) * (Node->X -	X12) + (Node->Y	- Y12) * (Node->Y - Y12);

					if (Node12 == 0	|| D < (Node12->X - X12) * (Node12->X -	X12) + (Node12->Y - Y12) * (Node12->Y -	Y12))
					{
						Node12 = Node;
					}
				}
			}

			if (0 == Node12)
			{
				Node12 = new TNode(++MaxNodeNr,	X12, Y12, Z12);
				if (false == NodeList->AddWithQuadTree(Node12))
				{
					dump_error(__FILE__, __LINE__, "Der Knoten (%.2f, %.2f) wurde nicht gefunden und konnte nicht eingefgt werden\n", X12 / 100.0, Y12 / 100.0);
				}
			}
		}

		if (Schnitt23)
		{
			X23 = (long)(XD23 + 0.5);
			Y23 = (long)(YD23 + 0.5);
			Z23 = (int)(ZD23 + 0.5);

			TNodeList* FoundNodeList = NodeList->FindAllByXY(X23, Y23, 4);

			if (FoundNodeList->size() > 0)
			{
				for (TNodeList::iterator j=FoundNodeList->begin(); j !=	FoundNodeList->end(); j++)
				{
					TNode* Node = *j;

					if (Node1->X ==	Node->X	&& Node1->Y == Node->Y)
					{
						Node->Z	= Z23;
						Schnitt23 = false;
					}
					else if	(Node2->X == Node->X &&	Node2->Y == Node->Y)
					{
						Node->Z	= Z23;
						Schnitt23 = false;
					}
					else if	(Node3->X == Node->X &&	Node3->Y == Node->Y)
					{
						Node->Z	= Z23;
						Schnitt23 = false;
					}

					double D = (Node->X - X23) * (Node->X -	X23) + (Node->Y	- Y23) * (Node->Y - Y23);

					if (Node23 == 0	|| D < (Node23->X - X23) * (Node23->X -	X23) + (Node23->Y - Y23) * (Node23->Y -	Y23))
					{
						Node23 = Node;
					}
				}
			}

			if (0 == Node23)
			{
				Node23 = new TNode(++MaxNodeNr,	X23, Y23, Z23);
				if(false == NodeList->AddWithQuadTree(Node23))
				{
					dump_error(__FILE__, __LINE__, "Der Knoten (%.2f, %.2f)	wurde nicht gefunden und konnte	nicht eingefgt	werden\n", X23 / 100.0,	Y23 / 100.0);
				}
			}
		}

		if (Schnitt31)
		{
			X31 = (long)(XD31 + 0.5);
			Y31 = (long)(YD31 + 0.5);
			Z31 = (int)(ZD31 + 0.5);

			TNodeList* FoundNodeList = NodeList->FindAllByXY(X31, Y31, 4);

			if (FoundNodeList->size() > 0)
			{
				for (TNodeList::iterator j=FoundNodeList->begin(); j !=	FoundNodeList->end(); j++)
				{
					TNode* Node = *j;

					if (Node1->X ==	Node->X	&& Node1->Y == Node->Y)
					{
						Node->Z	= Z31;
						Schnitt31 = false;
					}
					else if	(Node2->X == Node->X &&	Node2->Y == Node->Y)
					{
						Node->Z	= Z31;
						Schnitt31 = false;
					}
					else if	(Node3->X == Node->X &&	Node3->Y == Node->Y)
					{
						Node->Z	= Z31;
						Schnitt31 = false;
					}

					double D = (Node->X - X31) * (Node->X -	X31) + (Node->Y	- Y31) * (Node->Y - Y31);

					if (Node31 == 0	|| D < (Node31->X - X31) * (Node31->X -	X31) + (Node31->Y - Y31) * (Node31->Y -	Y31))
					{
						Node31 = Node;
					}
				}
			}

			if (0 == Node31)
			{
				Node31 = new TNode(++MaxNodeNr,	X31, Y31, Z31);
				if(false == NodeList->AddWithQuadTree(Node31))
				{
					dump_error(__FILE__, __LINE__, "Der Knoten (%.2f, %.2f)	wurde nicht gefunden und konnte	nicht eingefgt	werden\n", X31 / 100.0,	Y31 / 100.0);
				}
			}
		}

		if (Schnitt12 && Schnitt23 && Schnitt31)
		{
			NewElement = new TElement(Node1, Node12, Node31);
			NewElement->Typ	= NEWTRI;
			push_back(NewElement);

			NewElement = new TElement(Node12, Node2, Node23);
			NewElement->Typ	= NEWTRI;
			push_back(NewElement);

			NewElement = new TElement(Node12, Node23, Node31);
			NewElement->Typ	= NEWTRI;
			push_back(NewElement);

			NewElement = new TElement(Node23, Node3, Node31);
			NewElement->Typ	= NEWTRI;
			push_back(NewElement);

			if (DebugLevel >= 9) write_fortschritt("Element	%d : %d	- %d - %d wird gendert\n", Count, Node1->Nr, Node2->Nr, Node3->Nr);

			Element->Typ = NO_ELEMENT;
			ChangeCount++;
		}
		else if	(Schnitt12 && Schnitt23)
		{
			NewElement = new TElement(Node1, Node12, Node3);
			NewElement->Typ	= NEWTRI;
			push_back(NewElement);

			NewElement = new TElement(Node12, Node2, Node23);
			NewElement->Typ	= NEWTRI;
			push_back(NewElement);

			NewElement = new TElement(Node12, Node23, Node3);
			NewElement->Typ	= NEWTRI;
			push_back(NewElement);

			if (DebugLevel >= 9) write_fortschritt("Element	%d : %d	- %d - %d wird gendert\n", Count, Node1->Nr, Node2->Nr, Node3->Nr);

			Element->Typ = NO_ELEMENT;
			ChangeCount++;
		}
		else if	(Schnitt12 && Schnitt31)
		{
			NewElement = new TElement(Node1, Node12, Node31);
			NewElement->Typ	= NEWTRI;
			push_back(NewElement);

			NewElement = new TElement(Node12, Node2, Node31);
			NewElement->Typ	= NEWTRI;
			push_back(NewElement);

			NewElement = new TElement(Node2, Node3,	Node31);
			NewElement->Typ	= NEWTRI;
			push_back(NewElement);

			if (DebugLevel >= 9) write_fortschritt("Element	%d : %d	- %d - %d wird gendert\n", Count, Node1->Nr, Node2->Nr, Node3->Nr);

			Element->Typ = NO_ELEMENT;
			ChangeCount++;
		}
		else if	(Schnitt23 && Schnitt31)
		{
			NewElement = new TElement(Node1, Node2,	Node31);
			NewElement->Typ	= NEWTRI;
			push_back(NewElement);

			NewElement = new TElement(Node2, Node23, Node31);
			NewElement->Typ	= NEWTRI;
			push_back(NewElement);

			NewElement = new TElement(Node23, Node3, Node31);
			NewElement->Typ	= NEWTRI;
			push_back(NewElement);

			if (DebugLevel >= 9) write_fortschritt("Element	%d : %d	- %d - %d wird gendert\n", Count, Node1->Nr, Node2->Nr, Node3->Nr);

			Element->Typ = NO_ELEMENT;
			ChangeCount++;
		}
		else if	(Schnitt12)
		{
			NewElement = new TElement(Node1, Node12, Node3);
			NewElement->Typ	= NEWTRI;
			push_back(NewElement);

			NewElement = new TElement(Node12, Node2, Node3);
			NewElement->Typ	= NEWTRI;
			push_back(NewElement);

			if (DebugLevel >= 9) write_fortschritt("Element	%d : %d	- %d - %d wird gendert\n", Count, Node1->Nr, Node2->Nr, Node3->Nr);

			Element->Typ = NO_ELEMENT;
			ChangeCount++;
		}
		else if	(Schnitt23)
		{
			NewElement = new TElement(Node1, Node2,	Node23);
			NewElement->Typ	= NEWTRI;
			push_back(NewElement);

			NewElement = new TElement(Node1, Node23, Node3);
			NewElement->Typ	= NEWTRI;
			push_back(NewElement);

			if (DebugLevel >= 9) write_fortschritt("Element	%d : %d	- %d - %d wird gendert\n", Count, Node1->Nr, Node2->Nr, Node3->Nr);

			Element->Typ = NO_ELEMENT;
			ChangeCount++;
		}
		else if	(Schnitt31)
		{
			NewElement = new TElement(Node1, Node2,	Node31);
			NewElement->Typ	= NEWTRI;
			push_back(NewElement);

			NewElement = new TElement(Node2, Node3,	Node31);
			NewElement->Typ	= NEWTRI;
			push_back(NewElement);

			if (DebugLevel >= 9) write_fortschritt("Element	%d : %d	- %d - %d wird gendert\n", Count, Node1->Nr, Node2->Nr, Node3->Nr);

			Element->Typ = NO_ELEMENT;
			ChangeCount++;
		}
	}
	write_fortschritt("%d von %d Elemente, %d Elemente angepasst\n", Count,	size(),	ChangeCount);

	return (ChangeCount);
}

//---------------------------------------------------------------------------
void TElementList::CleanChangedElements(int DebugLevel)
{
	write_fortschritt("->Suche nach	entarteten Elementen gestartet\n");

	int Count = 0;
	for (unsigned int i = 0; i < size(); i++)
	{
		TElement* Element = (*this)[i];

		Count++;

		if (DebugLevel >= 9 || (DebugLevel >= 1	&& Count % 5000	== 0) || (Count	% 50000	== 0))
		{
			write_fortschritt("%d von %d Elementen berprft\n", Count, size());
		}

		// Hier	werden die zum Lschen markierten Elemente ignoriert.
		if (Element->Typ == NO_ELEMENT)	continue;

		// Hier	werden die neuen  Elemente wieder auf einen korrekten Typ gesetzt und dann ignoriert.
		if (Element->Typ == NEWTRI)
		{
			Element->Typ = TRI;
		}

		if (Element->Typ == QUAD)
		{
			dump_error(__FILE__, __LINE__, "Viereckselemente werden	im Moment noch nicht untersttzt");
		}

		TNode* Node1 = Element->Node1;
		TNode* Node2 = Element->Node2;
		TNode* Node3 = Element->Node3;

		if (Node1->X ==	Node2->X && Node1->Y ==	Node2->Y)
		{
			Element->Typ = NO_ELEMENT;
			continue;
		}

		if (Node2->X ==	Node3->X && Node2->Y ==	Node3->Y)
		{
			Element->Typ = NO_ELEMENT;
			continue;
		}

		if (Node3->X ==	Node1->X && Node3->Y ==	Node1->Y)
		{
			Element->Typ = NO_ELEMENT;
			continue;
		}

	}
	write_fortschritt("%d von %d Elementen berprft\n", Count, size());

	write_fortschritt("<-Suche nach	entarteten Elementen beendet\n");
}

//---------------------------------------------------------------------------
void TElementList::ChangeElements(TNodeList *NodeList, TProfilList *SperrenList, TProfilList *GrabenList, int DebugLevel)
{
	write_fortschritt("->Anpassung der Elemente an Dmme und Grben gestartet\n");

	NodeList->SortByXY();

	if (GrabenList && GrabenList->size() > 0)
	{
		write_fortschritt("->Anpassung der Elemente an Grben gestartet\n");
		HelpChangeElements(NodeList, GrabenList, DebugLevel);
		write_fortschritt("<-Anpassung der Elemente an Grben beendet\n");
	}

	if (SperrenList	&& SperrenList->size() > 0)
	{
		write_fortschritt("->Anpassung der Elemente an Dmme gestartet\n");
		HelpChangeElements(NodeList, SperrenList, DebugLevel);
		write_fortschritt("<-Anpassung der Elemente an Dmme beendet\n");
	}

	CleanChangedElements(DebugLevel);

	write_fortschritt("<-Anpassung der Elemente an Dmme und Grben beendet\n");
}

//---------------------------------------------------------------------
// TErgebnisPolygon
//---------------------------------------------------------------------

//---------------------------------------------------------------------
TErgebnisPolygon::TErgebnisPolygon(double diff)
{
	Diff = diff;
}

//---------------------------------------------------------------------
long double TErgebnisPolygon::Area(void)
{
	if (size() < 3)
	{
		// Wenn	weniger	als 3 Punkte da	sind, um das Polygon zu	bilden
		// ist die Flche immer	0
		return (0.0);
	}

	long double Area = 0.0;
	for (unsigned int i = 1; i < size(); i++)
	{
		TXYZ *Xy1 = (*this)[i-1];
		double X1 = Xy1->X;
		double Y1 = Xy1->Y;

		TXYZ *Xy2 = (*this)[i];
		double X2 = Xy2->X;
		double Y2 = Xy2->Y;

		Area = Area + X1 * Y2 -	X2 * Y1;
	}

	TXYZ *LastXy = (*this)[size()-1];
	double Lx = LastXy->X;
	double Ly = LastXy->Y;

	TXYZ *FirstXy = (*this)[0];
	double Fx = FirstXy->X;
	double Fy = FirstXy->Y;

	Area = Area + Lx * Fy -	Fx * Ly;

	Area = Area / 2.0;

	return (Area);
}

//---------------------------------------------------------------------
// TErgebnisPolygonList
//---------------------------------------------------------------------

//---------------------------------------------------------------------------
TErgebnisPolygonList::~TErgebnisPolygonList(void)
{
	Clear();
}

//---------------------------------------------------------------------
void TErgebnisPolygonList::Clear(void)
{
	for (TErgebnisPolygonList::iterator i=begin(); i!=end(); i++)
	{
		TErgebnisPolygon *ErgebnisPolygon = *i;
		delete ErgebnisPolygon;
		ErgebnisPolygon	= 0;
	}
	clear();
}

//---------------------------------------------------------------------
TErgebnisPolygonList* TErgebnisPolygonList::Copy(void)
{
	TErgebnisPolygonList *NewErgebnisPolygonList = new TErgebnisPolygonList;

	for (unsigned int i=0; i<size(); i++)
	{
		TErgebnisPolygon *OldErgebnisPolygon = (*this)[i];

		TErgebnisPolygon *NewErgebnisPolygon = new TErgebnisPolygon(OldErgebnisPolygon->Diff);

		for (unsigned int j=0; j<OldErgebnisPolygon->size(); j++)
		{
			TXYZ *OldXy = (*OldErgebnisPolygon)[j];
			TXYZ *NewXy = new TXYZ(OldXy);

			NewErgebnisPolygon->push_back(NewXy);
		}

		NewErgebnisPolygon->MinX = OldErgebnisPolygon->MinX;
		NewErgebnisPolygon->MaxX = OldErgebnisPolygon->MaxX;
		NewErgebnisPolygon->MinY = OldErgebnisPolygon->MinY;
		NewErgebnisPolygon->MaxY = OldErgebnisPolygon->MaxY;

		NewErgebnisPolygonList->push_back(NewErgebnisPolygon);
	}

	return(NewErgebnisPolygonList);
}

//---------------------------------------------------------------------
void TErgebnisPolygonList::Append(TErgebnisPolygonList *AddErgebnisPolygonList)
{
	for (TErgebnisPolygonList::iterator i =	AddErgebnisPolygonList->begin(); i != AddErgebnisPolygonList->end(); i++)
	{
		TErgebnisPolygon *AddErgebnisPolygon = *i;
		push_back(AddErgebnisPolygon);
	}
}

//---------------------------------------------------------------------
TErgebnisPolygon* TErgebnisPolygonList::First(void)
{
	if (size() <= 0) return	(0);

	TErgebnisPolygon *ErgebnisPolygon = this->front();

	return (ErgebnisPolygon);
}

//---------------------------------------------------------------------
TInsideTyp TErgebnisPolygonList::IsInsidePolygonList(long X, long Y)
{

	for (TErgebnisPolygonList::iterator i =	begin(); i != end(); i++)
	{
		TErgebnisPolygon* Polygon = *i;

		int alpha = IsInside(Polygon, X, Y);

		if (alpha == 0)		continue;			// not inside -> versuche das nchste Polygon
		if (alpha == 4)		return (INSIDE);    // inside
		if (alpha == -4)	return (INSIDE);    // inside
		if (alpha == -1)	return (ON_LINE);   // on line
		if (alpha == 2)	    return (ON_LINE);	// on line twisted
		if (alpha == -2)    return (ON_LINE);	// on line twisted

		dump_error(__FILE__, __LINE__, "Alpha (%d) ungltig\n",	alpha);
	}

	return (NOT_INSIDE);
}

//---------------------------------------------------------------------
// TEdge
//---------------------------------------------------------------------

//---------------------------------------------------------------------
TEdge::TEdge(TNode *node1, TNode *node2)
{
	if (node1->Nr <	0)
	{
		dump_error (__FILE__, __LINE__,	"Beim Erzeugen einer Kante ist die Knotennummer	1 negativ (%d)\n", node1->Nr);
	}

	if (node2->Nr <	0)
	{
		dump_error (__FILE__, __LINE__,	"Beim Erzeugen einer Kante ist die Knotennummer	2 negativ (%d)\n", node2->Nr);
	}

	Node1 =	node1;
	Node2 =	node2;

	double Dx = Node2->X - Node1->X;
	double Dy = Node2->Y - Node1->Y;
	Distance = sqrt	(Dx*Dx +Dy*Dy);

	if (Distance < MinNodeDistanceKonst)
	{
		dump_error (__FILE__, __LINE__,	"Beim Erzeugen einer Kante ist der Abstand der Knoten zu dicht\nAbstand: %f\nKnoten 1: (%d,%d,%d)\nKnoten 2: (%d,%d,%d)\n", Distance, Node1->Nr, Node1->X, Node1->Y, Node2->Nr,	Node2->X, Node2->Y);
	}

	Arc = acos (Dx / Distance) * 180.0 / M_PI;
	if (Dy < 0) Arc	= 360.0	- Arc;

	X = 0;
	Y = 0;
	Ready =	false;
	IsBoundary = false;
}

//---------------------------------------------------------------------
TEdge::TEdge(TEdge *Edge)
{
	Node1 =	Edge->Node1;
	Node2 =	Edge->Node2;
	Distance = Edge->Distance;
	Arc = Edge->Arc;
	X = Edge->X;
	Y = Edge->Y;
	Ready =	Edge->Ready;
	IsBoundary = Edge->IsBoundary;
}


//---------------------------------------------------------------------

//---------------------------------------------------------------------
// TEdgeList
//---------------------------------------------------------------------

//---------------------------------------------------------------------
TEdgeList::TEdgeList(void)
{
	Anz = 0;
	AvgDistance = 0;

	// Nur als Hllen fr das Suchen nach Kanten
	SearchNode1 = new TNode(0, 1, 1, 1);
	SearchNode2 = new TNode(0, 2, 2, 2);
	SearchEdge = new TEdge(SearchNode1, SearchNode2);
}

//---------------------------------------------------------------------
TEdgeList::~TEdgeList(void)
{
	Clear();

	delete SearchEdge;
	delete SearchNode1;
	delete SearchNode2;

	SearchEdge = 0;
	SearchNode1 = 0;
	SearchNode2 = 0;
}

//---------------------------------------------------------------------
void TEdgeList::Clear(void)
{
	if (EdgeVector.size() >	0)
	{
		int Count = 0;
		for (TEdgeVector::const_iterator i = EdgeVector.begin(); i != EdgeVector.end();	i++)
		{
			TEdge *Edge = *i;
			delete Edge;
			Edge = 0;
			Anz--;

			Count++;

			if (Count % 100000 == 0)	write_fortschritt("%d von %d Objekte freigegeben\n", Count, EdgeVector.size());
		}
		write_fortschritt("%d Objekte freigegeben\n", Count);
	}

	if (Anz	!= 0)
	{
		dump_error(__FILE__, __LINE__, "Anz der	Kanten ist nicht 0\n");
	}

	AvgDistance = 0;

	EdgeVector.clear();
}

//---------------------------------------------------------------------
void TEdgeList::Empty(void)
{
	for (TEdgeVector::const_iterator i = EdgeVector.begin(); i != EdgeVector.end();	i++)
	{
		Anz--;
	}

	if (Anz	!= 0)
	{
		dump_error(__FILE__, __LINE__, "Anz der	Kanten ist nicht 0\n");
	}

	AvgDistance = 0;

	EdgeVector.clear();
}

//---------------------------------------------------------------------
void TEdgeList::Add(TEdge *edge)
{
	if (EdgeVector.capacity() < EdgeVector.size() +	1)
	{
		EdgeVector.reserve((size_t)(EdgeVector.capacity() * 1.1	+ 200000));
		write_fortschritt("Kapazitt der Kantenliste wurde auf %d erhht\n", EdgeVector.capacity());
	}

	EdgeVector.push_back(edge);
	AvgDistance = AvgDistance * Anz;
	Anz++;
	AvgDistance = (unsigned	int)((AvgDistance + edge->Distance) / Anz + 0.5);
}

//---------------------------------------------------------------------
TEdge* TEdgeList::AddWithIndex(TNode *Node1, TNode *Node2)
{
	TEdge *NewEdge = new TEdge(Node1, Node2);

	Add(NewEdge);

	Node1->EdgeIndex = new TEdgeIndex (Node1->EdgeIndex, NewEdge);

	Node2->EdgeIndex = new TEdgeIndex (Node2->EdgeIndex, NewEdge);

	return (NewEdge);
}

//---------------------------------------------------------------------
TEdge* TEdgeList::FindWithIndex(TNode *Node1, TNode *Node2)
{
	if (Anz	== 0)	return(0);

	TEdgeIndex *AktEdgeIndex = Node1->EdgeIndex;

	while (AktEdgeIndex)
	{
		TEdgeIndex *NextEdgeIndex = AktEdgeIndex->NextEdgeIndex;

		TEdge *Edge = AktEdgeIndex->Edge;

		if (Edge->Node1	== Node1 && Edge->Node2	== Node2)
		{
			return (Edge);
		}

		AktEdgeIndex = NextEdgeIndex;
	}

	return (0);
}

//---------------------------------------------------------------------
void TEdgeList::BuildEdgeIndex(void)
{
	for (TEdgeVector::const_iterator i=EdgeVector.begin(); i != EdgeVector.end(); i++)
	{
		TEdge *Edge = *i;

		TNode *Node1 = Edge->Node1;
		Node1->EdgeIndex = new TEdgeIndex (Node1->EdgeIndex, Edge);

		TNode *Node2 = Edge->Node2;
		Node2->EdgeIndex = new TEdgeIndex (Node2->EdgeIndex, Edge);
	}
}

//---------------------------------------------------------------------
// TEdgeIndex
//---------------------------------------------------------------------

//---------------------------------------------------------------------
TEdgeIndex::TEdgeIndex(TEdgeIndex *nextedgeindex, TEdge	*edge)
{
	NextEdgeIndex =	nextedgeindex;
	Edge = edge;
}

//---------------------------------------------------------------------
// TProfil
//---------------------------------------------------------------------

//---------------------------------------------------------------------
TProfil::TProfil(int station, int wsp)
{
	Station	= station;
	Wsp = wsp;

	PointList = 0;

	MinX = 0;
	MaxX = 0;
	MinY = 0;
	MaxY = 0;
}

//---------------------------------------------------------------------
TProfil::TProfil(TProfil *profil)
{
	Station	= profil->Station;
	Wsp = profil->Wsp;

	PointList = new	TPointList;
	for (TPointList::iterator i=profil->PointList->begin();	i != profil->PointList->end(); i++)
	{
		TPoint *OldPoint = *i;

		TPoint *NewPoint = new TPoint(OldPoint);

		PointList->insert(NewPoint);
	}

	MinX = 0;
	MaxX = 0;
	MinY = 0;
	MaxY = 0;

}

//---------------------------------------------------------------------
TProfil::~TProfil(void)
{
	if (PointList)
	{
		for (TPointList::iterator i=PointList->begin();	i != PointList->end(); i++)
		{
			TPoint *Point =	*i;
			delete Point;
			Point =	0;
		}
		delete PointList;
        PointList = 0;
	}

	MinX = 0;
	MaxX = 0;
	MinY = 0;
	MaxY = 0;
}

//---------------------------------------------------------------------
void TProfil::AddPoint(int meter)
{
	if (0 == PointList || PointList->size()	< 2)
	{
		dump_error(__FILE__, __LINE__, "Weniger	als 2 Punkte im	Profil vorhanden\n");
	}

	long X = 0;
	long Y = 0;
	GetXY(meter, &X, &Y);

	TPoint*	NewPoint = new TPoint(X, Y, 0, meter);

	unsigned int Anz = PointList->size();

	PointList->insert(NewPoint);

	// Wenn	sich die Anzahl	nicht erhht hat,
	// so wurde der	Punkt auch nicht hinzugefgt
	// -> er muss wieder gelscht werden

	if (PointList->size() == Anz)
	{
		delete NewPoint;
		NewPoint = 0;
	}
	else
	{
		if (PointList->size() == 1)
		{
			MinX = NewPoint->X;
			MaxX = NewPoint->X;
			MinY = NewPoint->Y;
			MaxY = NewPoint->Y;
		}
		else
		{
			if (MinX > NewPoint->X)	MinX = NewPoint->X;
			if (MaxX < NewPoint->X)	MaxX = NewPoint->X;
			if (MinY > NewPoint->Y)	MinY = NewPoint->Y;
			if (MaxY < NewPoint->Y)	MaxY = NewPoint->Y;
		}
	}
}

//---------------------------------------------------------------------
void TProfil::AddPoint(long x, long y, int z)
{
	if (0 == PointList)	PointList = new	TPointList;

	TPointList::iterator i = PointList->begin();

	TPoint *LastPoint = 0;
	while (i != PointList->end())
	{
		LastPoint = *i;
		i++;
	}

	double MeterD =	0.0;

	if (LastPoint)
	{
		long X = LastPoint->X;
		long Y = LastPoint->Y;
		int M =	LastPoint->Meter;

		double Dx = X -	x;
		double Dy = Y -	y;

		double D = sqrt(Dx * Dx	+ Dy * Dy);

		MeterD = M + D;
	}

	int Meter = (int)(MeterD + 0.5);
	
	TPoint*	NewPoint = new TPoint(x, y, z, Meter);

	PointList->insert(NewPoint);

	// Wenn	sich die Anzahl	nicht erhht hat,
	// so wurde der	Punkt auch nicht hinzugefgt
	// -> er muss wieder gelscht werden

	if (PointList->size() == 1)
	{
		MinX = NewPoint->X;
		MaxX = NewPoint->X;
		MinY = NewPoint->Y;
		MaxY = NewPoint->Y;
	}
	else
	{
		if (MinX > NewPoint->X)	MinX = NewPoint->X;
		if (MaxX < NewPoint->X)	MaxX = NewPoint->X;
		if (MinY > NewPoint->Y)	MinY = NewPoint->Y;
		if (MaxY < NewPoint->Y)	MaxY = NewPoint->Y;
	}
}

//---------------------------------------------------------------------
bool TProfil::GetXY(int	meter, long *x,	long *y)
{
	if (0 == PointList || PointList->size()	< 2)
	{
		dump_error(__FILE__, __LINE__, "Weniger	als 2 Punkte im	Profil vorhanden\n");
	}

	TPointList::iterator i = PointList->begin();

	TPoint *LastPoint = 0;
	TPoint *Point =	0;
	while (i != PointList->end())
	{
		Point =	*i++;

		if (Point->Meter == meter)
		{
			*x = Point->X;
			*y = Point->Y;
			return(true);
		}

		if (LastPoint != 0)
		{
			if (meter < LastPoint->Meter)
			{
				// Der gesuchte	Punkt liegt vor	dem ersten Punkt des Profils
				double Dx = Point->X - LastPoint->X;
				double Dy = Point->Y - LastPoint->Y;
				double Dm = Point->Meter - LastPoint->Meter;

				double Faktor =	(meter - LastPoint->Meter) / Dm;

				*x = (long)(LastPoint->X + Dx *	Faktor + 0.5);
				*y = (long)(LastPoint->Y + Dy *	Faktor + 0.5);

				if (Faktor >= 0.0)
				{
					dump_error(__FILE__, __LINE__,	"Faktor	ist nicht negativ, aber	der gesuchte Punkt liegt vor dem Profil\n");
				}

				return (false);
			}
			else if	(meter >= LastPoint->Meter && meter <= Point->Meter)
			{
				// Der gesuchte	Punkt liegt zwischen Punkten des Profils
				double Dx = Point->X - LastPoint->X;
				double Dy = Point->Y - LastPoint->Y;
				double Dm = Point->Meter - LastPoint->Meter;

				double Faktor =	(meter - LastPoint->Meter) / Dm;

				*x = (long)(LastPoint->X + Dx *	Faktor + 0.5);
				*y = (long)(LastPoint->Y + Dy *	Faktor + 0.5);

				if (Faktor <= 0.0 || Faktor >= 1.0)
				{
					dump_error(__FILE__, __LINE__,	"Faktor	ist nicht zwischen 0 und 1, aber der gesuchte Punkt liegt im Profil\n");
				}

				return (true);
			}
			else if	(i == PointList->end())
			{
				// Der gesuchte	Punkt liegt hinter dem letzten Punkt des Profils
				double Dx = Point->X - LastPoint->X;
				double Dy = Point->Y - LastPoint->Y;
				double Dm = Point->Meter - LastPoint->Meter;

				double Faktor =	(meter - LastPoint->Meter) / Dm;

				*x = (long)(LastPoint->X + Dx *	Faktor + 0.5);
				*y = (long)(LastPoint->Y + Dy *	Faktor + 0.5);

				if (Faktor <= 1.0)
				{
					dump_error(__FILE__, __LINE__,	"Faktor	ist nicht grer als 1,	aber der gesuchte Punkt	liegt hinter dem Profil\n");
				}

				return (false);
			}
		}

		LastPoint = Point;
	}

	dump_error(__FILE__, __LINE__,	"So kann der Punkt nicht ermittelt werden\n");

	return (false);
}

//---------------------------------------------------------------------------
void TProfil::EqualizeProfil(TProfil *Profil)
{
	TPointList::iterator OneVon = PointList->begin();
	TPointList::reverse_iterator OneBis = PointList->rbegin();
	TPointList::iterator TwoVon = Profil->PointList->begin();
	TPointList::reverse_iterator TwoBis = Profil->PointList->rbegin();

	TPoint *PointOneVon = *OneVon;
	TPoint *PointOneBis = *OneBis;
	TPoint *PointTwoVon = *TwoVon;
	TPoint *PointTwoBis = *TwoBis;

	double MeterOneVon = PointOneVon->Meter;
	double MeterOneBis = PointOneBis->Meter;
	double MeterTwoVon = PointTwoVon->Meter;
	double MeterTwoBis = PointTwoBis->Meter;

	double LaengeOne  = MeterOneBis	- MeterOneVon;
	double LaengeTwo  = MeterTwoBis	- MeterTwoVon;


	TPointList::iterator i = ++TwoVon;
	for (unsigned int Anz =	0; Anz < Profil->PointList->size()-2; Anz++)
	{
		TPoint *PointTwoI = *i++;
		double MeterTwoI = PointTwoI->Meter;

		double Prozent = (MeterTwoI - MeterTwoVon) / LaengeTwo;
		int Meter = (int)(MeterOneVon +	LaengeOne * Prozent + 0.5);

		int OldAnz = PointList->size();
		int NewAnz = PointList->size();
		while (OldAnz == NewAnz)
		{
			AddPoint(Meter);
			NewAnz = PointList->size();
			Meter =	Meter +	1;
		}
	}
}

//---------------------------------------------------------------------------
void TProfil::InterpolateProfil(TProfil	*vorprofil, TProfil *nachprofil, double	Prozent)
{
	if (vorprofil->PointList->size() != nachprofil->PointList->size())
	{
		dump_error(__FILE__, __LINE__, "Angeglichene Profile ('%.2f' und '%.2f') haben verschieden viele Sttzstellen\n", vorprofil->Station / 10000.0,	nachprofil->Station / 10000.0);
	}

	TPointList::iterator V = vorprofil->PointList->begin();
	TPointList::iterator N = nachprofil->PointList->begin();

	for (unsigned int i=0; i<vorprofil->PointList->size(); i++)
	{
		TPoint *VPoint = *V++;
		TPoint *NPoint = *N++;

		double VX = VPoint->X;
		double NX = NPoint->X;
		long X = (long)(VX + Prozent * (NX - VX) + 0.5);

		double VY = VPoint->Y;
		double NY = NPoint->Y;
		long Y = (long)(VY + Prozent * (NY - VY) + 0.5);

		AddPoint(X, Y);
	}
}

//---------------------------------------------------------------------------
bool TProfil::CalcSchnitt(TGewaesserAchseList *GewaesserAchseList, double* X, double* Y)
{
	for (TGewaesserAchseList::iterator g=GewaesserAchseList->begin(); g != GewaesserAchseList->end(); g++)
	{
		TGewaesserAchse* GewaesserAchse	= *g;

		if (MaxX < GewaesserAchse->MinX)	continue;
		if (MinX > GewaesserAchse->MaxX)	continue;
		if (MaxY < GewaesserAchse->MinY)	continue;
		if (MinY > GewaesserAchse->MaxY)	continue;

		TPoint*	P0 = 0;
		TPoint*	P1 = 0;
		for (TPointList::iterator p = PointList->begin(); p != PointList->end(); p++)
		{
			if (P1 == 0)
			{
				P1 = *p;
				continue;
			}
			else
			{
				P0 = P1;
				P1 = *p;
			}

			long P0x = P0->X;
			long P0y = P0->Y;
			long P1x = P1->X;
			long P1y = P1->Y;

			if (P0x	< GewaesserAchse->MinX && P1x <	GewaesserAchse->MinX)	continue;
			if (P0x	> GewaesserAchse->MaxX && P1x >	GewaesserAchse->MaxX)	continue;
			if (P0y	< GewaesserAchse->MinY && P1y <	GewaesserAchse->MinY)	continue;
			if (P0y	> GewaesserAchse->MaxY && P1y >	GewaesserAchse->MaxY)	continue;

			long PDx = P1x - P0x;
			long PDy = P1y - P0y;
			if (PDx	== 0 &&	PDy == 0)
			{
				write_error(3207, "Profil-Punktabstand (%.3f %.3f) - (%.3f %.3f) ist zu	klein\n", P0x /	100.0, P0y / 100.0, P1x	/ 100.0, P1y / 100.0);
			}

			// Es heisst hier s fr	spur und nicht g fr gewsserachse

			TPoint*S0 = 0;
			TPoint*S1 = 0;
			for (TPointList::iterator s = GewaesserAchse->PointList->begin(); s != GewaesserAchse->PointList->end(); s++)
			{
				if (S1 == 0)
				{
					S1 = *s;
					continue;
				}
				else
				{
					S0 = S1;
					S1 = *s;
				}

				long S0x = S0->X;
				long S0y = S0->Y;
				long S1x = S1->X;
				long S1y = S1->Y;

				long SDx = S1x - S0x;
				long SDy = S1y - S0y;
				if (SDx	== 0 &&	SDy == 0)
				{
					write_error(3208, "Gewsserachsen-Sttzpunkteabstand (%.3f %.3f) - (%.3f %.3f) ist zu klein\n",	S0x / 100.0, S0y / 100.0, S1x /	100.0, S1y / 100.0);
				}

				bool Found = Calc2Schnitt(P0x, P0y, P1x, P1y, S0x, S0y,	S1x, S1y, X, Y);

				if (Found)	return (true);
			}
		}
	}

	return (false);
}

//---------------------------------------------------------------------------
bool TProfil::CalcSchnitt(TEdge* Edge)
{
	TPoint*	P0 = 0;
	TPoint*	P1 = 0;
	for (TPointList::iterator p = PointList->begin(); p != PointList->end(); p++)
	{
		if (P1 == 0)
		{
			P1 = *p;
			continue;
		}
		else
		{
			P0 = P1;
			P1 = *p;
		}

		long P0x = P0->X;
		long P0y = P0->Y;
		long P1x = P1->X;
		long P1y = P1->Y;

		long PDx = P1x - P0x;
		long PDy = P1y - P0y;
		if (PDx	== 0 &&	PDy == 0)
		{
			write_error(3207, "Profil-Punktabstand (%.3f %.3f) - (%.3f %.3f) ist zu	klein\n", P0x /	100.0, P0y / 100.0, P1x	/ 100.0, P1y / 100.0);
		}

		// Es heisst hier N fr	Node und nicht e fr edge

		TNode* N0 = Edge->Node1;
		TNode* N1 = Edge->Node2;

		long N0x = N0->X;
		long N0y = N0->Y;
		long N1x = N1->X;
		long N1y = N1->Y;

		long NDx = N1x - N0x;
		long NDy = N1y - N0y;
		if (NDx	== 0 &&	NDy == 0)
		{
			write_error(3220, "Kanten-Knotenabstand (%.3f %.3f) - (%.3f %.3f) ist zu klein\n",	N0x / 100.0, N0y / 100.0, N1x /	100.0, N1y / 100.0);
		}

		double DummyX, DummyY;
		bool Found = Calc2Schnitt(P0x, P0y, P1x, P1y, N0x, N0y,	N1x, N1y, &DummyX, &DummyY);

		if (Found)	return (true);
	}

	return (false);
}

//---------------------------------------------------------------------
// TProfilList
//---------------------------------------------------------------------

//---------------------------------------------------------------------
TProfilList::~TProfilList(void)
{
	Clear();
}

//---------------------------------------------------------------------
void TProfilList::Clear(void)
{
	if (size() > 0)
	{
		int Count = 0;
		for (TProfilList::iterator i=begin(); i	!= end(); i++)
		{
			TProfil	*Profil	= *i;

			delete Profil;
			Profil = 0;

			Count++;

			if (Count % 10000 == 0)	write_fortschritt("%d von %d Objekte freigegeben\n", Count, size());
		}
		write_fortschritt("%d Objekte freigegeben\n", Count);
	}

	clear();
}

//---------------------------------------------------------------------
TProfil* TProfilList::Find(int Station)
{
	for (iterator i	= begin(); i !=	end(); i++)
	{
		TProfil	*Profil	= *i;

		if (Station == Profil->Station)
		{
			return (Profil);
		}
	}
	return (0);
}

//---------------------------------------------------------------------
void TProfilList::InterpoliereProfile(double SollAbstand)
{
	write_fortschritt("->Interpolation der Profile gestartet\n");

	SollAbstand = SollAbstand / 100.0;

	write_fortschritt("Gewnschter Abstand der Profile: %.1f m\n", SollAbstand);

	TProfil	*FirstProfil = 0;
	TProfil	*NextProfil = 0;

	TProfilList* NewProfilList = new TProfilList();

	for (iterator i	= begin(); i !=	end(); i++)
	{
		if (NextProfil == 0)
		{
			NextProfil = *i;
			continue;
		}
		else
		{
			FirstProfil = NextProfil;
			NextProfil = *i;
		}

		double Diff = (NextProfil->Station - FirstProfil->Station) / 10.0;
		int Anz	= (int)(Diff / SollAbstand + 0.5) - 1;
		if (Anz	<= 0) continue;

		double Abstand = Diff /	(Anz + 1);

		TProfil	*FirstProfilNeu	= new TProfil(FirstProfil);
		TProfil	*NextProfilNeu = new TProfil(NextProfil);

		FirstProfilNeu->EqualizeProfil(NextProfil);
		NextProfilNeu->EqualizeProfil(FirstProfil);

		unsigned int Anz1 = FirstProfilNeu->PointList->size();
		unsigned int Anz2 = NextProfilNeu->PointList->size();

		if (Anz1 != Anz2)
		{
			dump_error(__FILE__, __LINE__, "Angeglichene Profile sind nicht	identisch\n");
		}

		for (int j=1; j<=Anz; j++)
		{
			double Prozent = j * Abstand / Diff;
			int NewStation = (int)(FirstProfil->Station + j	* Abstand * 10 + 0.5);
			int NewWsp = (int)(FirstProfil->Wsp + Prozent *	(NextProfil->Wsp - FirstProfil->Wsp) + 0.5);

			TProfil* NewProfil = new TProfil(NewStation, NewWsp);
			NewProfil->InterpolateProfil(FirstProfilNeu, NextProfilNeu, Prozent);

			NewProfilList->insert(NewProfil);

		}

		delete FirstProfilNeu;
		delete NextProfilNeu;

		FirstProfilNeu = 0;
		NextProfilNeu =	0;
	}

	for (TProfilList::iterator i = NewProfilList->begin(); i != NewProfilList->end(); i++)
	{
		insert(*i);
	}

	NewProfilList->clear();
	delete NewProfilList;
	NewProfilList =	0;

	write_fortschritt("<-Interpolation der Profile beendet\n");
}

//---------------------------------------------------------------------
void TProfilList::FillProfile(double SollAbstand, int SollAnz, int DebugLevel)
{
	write_fortschritt("->Auffllen der Profile mit Sttzstellen gestartet\n");

	int PointCount = 0;
	int Count = 0;
	for (iterator i	= begin(); i !=	end(); i++)
	{
		TProfil* Profil	= *i;

		TPoint*	LastPoint = *Profil->PointList->rbegin();
		double Laenge =	LastPoint->Meter;

		double Abstand = 0.0;

		if (SollAnz > 0)
		{
			Abstand	= Laenge / SollAnz;
		}
		else
		{
			Abstand	= SollAbstand;
		}


		TPoint*	FirstPoint = 0;
		TPoint*	NextPoint = 0;
		for (TPointList::iterator j = Profil->PointList->begin(); j != Profil->PointList->end(); j++)
		{
			FirstPoint = NextPoint;
			NextPoint = *j;

			if (FirstPoint == 0)	continue;

			double Diff = NextPoint->Meter - FirstPoint->Meter;


			int Anz	= (int)(Diff / Abstand + 0.5) -	1;
			if (Anz	< 0) Anz = 0;
			double Abstand = Diff /	(Anz + 1);

			for (int k=1; k<=Anz; k++)
			{
				int Meter = (int)(FirstPoint->Meter + k	* Abstand + 0.5);

				Profil->AddPoint(Meter);

				PointCount++;

				if (DebugLevel >= 9 || (DebugLevel >= 1	&& PointCount %	100 == 0) || (PointCount > 0 &&	PointCount % 1000 == 0))
				{
					write_fortschritt("Profil %d von %d aufgefllt, %d Punkte hinzugefgt\n", Count, size(), PointCount);
				}
			}

			if (Anz	>= 1)
			{
				// Jetzt wieder	von vorne
				j = Profil->PointList->find(FirstPoint);
				FirstPoint = 0;
				NextPoint = 0;
			}
		}

		Count++;
	}

	write_fortschritt("Profil %d von %d aufgefllt\n", Count, size());

	write_fortschritt("<-Auffllen der Profile mit Sttzstellen beendet\n");
}

//---------------------------------------------------------------------
// TGewaesserAchse
//---------------------------------------------------------------------

//---------------------------------------------------------------------
TGewaesserAchse::TGewaesserAchse(void)
{
	MinX = 999999999;
	MaxX = -999999999;
	MinY = 999999999;
	MaxY = -999999999;

	PointList = 0;
}

//---------------------------------------------------------------------
TGewaesserAchse::TGewaesserAchse(TGewaesserAchse *gewaesserAchse)
{
	MinX = 999999999;
	MaxX = -999999999;
	MinY = 999999999;
	MaxY = -999999999;

	PointList = new	TPointList;
	for (TPointList::iterator i=gewaesserAchse->PointList->begin();	i != gewaesserAchse->PointList->end(); i++)
	{
		TPoint *OldPoint = *i;

		TPoint *NewPoint = new TPoint(OldPoint);

		PointList->insert(NewPoint);

		if (i == gewaesserAchse->PointList->begin())
		{
			MinX = NewPoint->X;
			MaxX = NewPoint->X;
			MinY = NewPoint->Y;
			MaxY = NewPoint->Y;
		}
		else
		{
			if (MinX > NewPoint->X)	MinX = NewPoint->X;
			if (MaxX < NewPoint->X)	MaxX = NewPoint->X;
			if (MinY > NewPoint->Y)	MinY = NewPoint->Y;
			if (MaxY < NewPoint->Y)	MaxY = NewPoint->Y;
		}
	}
}

//---------------------------------------------------------------------
TGewaesserAchse::~TGewaesserAchse(void)
{
	if (PointList)
	{
		for (TPointList::iterator i=PointList->begin();	i != PointList->end(); i++)
		{
			TPoint *Point =	*i;

			delete Point;
			Point =	0;
		}

		delete PointList;
		PointList = 0;
	}
	MinX = 999999999;
	MaxX = -999999999;
	MinY = 999999999;
	MaxY = -999999999;
}

//---------------------------------------------------------------------
void TGewaesserAchse::AddPoint(int meter)
{
	if (0 == PointList || PointList->size()	< 2)
	{
		dump_error(__FILE__, __LINE__, "Weniger	als 2 Punkte im	Gewsserachse vorhanden\n");
	}

	long X = 0;
	long Y = 0;
	GetXY(meter, &X, &Y);

	TPoint*	NewPoint = new TPoint(X, Y, 0, meter);

	PointList->insert(NewPoint);

	if (MinX > NewPoint->X)	MinX = NewPoint->X;
	if (MaxX < NewPoint->X)	MaxX = NewPoint->X;
	if (MinY > NewPoint->Y)	MinY = NewPoint->Y;
	if (MaxY < NewPoint->Y)	MaxY = NewPoint->Y;
}

//---------------------------------------------------------------------
void TGewaesserAchse::AddPoint(long x, long y, int z)
{
	if (0 == PointList)	PointList = new	TPointList;

	TPointList::iterator i = PointList->begin();

	TPoint *LastPoint = 0;

	if (PointList->size() > 0)	LastPoint = *PointList->rbegin();

	double MeterD =	0.0;

	if (LastPoint)
	{
		double X = LastPoint->X;
		double Y = LastPoint->Y;
		double M = LastPoint->Meter;

		double Dx = X -	x;
		double Dy = Y -	y;

		double D = sqrt(Dx * Dx	+ Dy * Dy);

		MeterD = M + D;

		if (D == 0)	return;	// Den Punkt hatten wir schon
	}

	int Meter = (int)(MeterD + 0.5);
	
	TPoint*	NewPoint = new TPoint(x, y, z, Meter);

	unsigned int Anz = PointList->size();

	PointList->insert(NewPoint);

	if (Anz == PointList->size())
	{
		delete NewPoint;
		NewPoint = 0;
	}
	else
	{
		if (MinX > NewPoint->X)	MinX = NewPoint->X;
		if (MaxX < NewPoint->X)	MaxX = NewPoint->X;
		if (MinY > NewPoint->Y)	MinY = NewPoint->Y;
		if (MaxY < NewPoint->Y)	MaxY = NewPoint->Y;
	}
}

//---------------------------------------------------------------------
bool TGewaesserAchse::GetXY(int	meter, long *x,	long *y)
{
	if (0 == PointList || PointList->size()	< 2)
	{
		dump_error(__FILE__, __LINE__, "Weniger	als 2 Punkte im	Gewsserachse vorhanden\n");
	}

	TPointList::iterator i = PointList->begin();

	TPoint *LastPoint = 0;
	TPoint *Point =	0;
	while (i != PointList->end())
	{
		Point =	*i;

		if (Point->Meter == meter)
		{
			*x = Point->X;
			*y = Point->Y;
			return(true);
		}

		if (LastPoint != 0)
		{
			if (meter < LastPoint->Meter)
			{
				// Der gesuchte	Punkt liegt vor	dem ersten Punkt des Profils
				long Dx	= Point->X - LastPoint->X;
				long Dy	= Point->Y - LastPoint->Y;
				int Dm = Point->Meter -	LastPoint->Meter;

				double Faktor =	(meter - LastPoint->Meter) / Dm;

				*x = (long)(LastPoint->X + Dx *	Faktor + 0.5);
				*y = (long)(LastPoint->Y + Dy *	Faktor + 0.5);

				if (Faktor >= 0.0)
				{
					dump_error(__FILE__, __LINE__,	"Faktor	ist nicht negativ, aber	der gesuchte Punkt liegt vor dem Profil\n");
				}

				return (false);
			}
			else if	(meter > LastPoint->Meter && meter < Point->Meter)
			{
				// Der gesuchte	Punkt liegt zwischen Punkten des Profils
				long Dx	= Point->X - LastPoint->X;
				long Dy	= Point->Y - LastPoint->Y;
				int Dm = Point->Meter -	LastPoint->Meter;

				double Faktor =	(meter - LastPoint->Meter) / Dm;

				*x = (long)(LastPoint->X + Dx *	Faktor + 0.5);
				*y = (long)(LastPoint->Y + Dy *	Faktor + 0.5);

				if (Faktor <= 0.0 || Faktor >= 1.0)
				{
					dump_error(__FILE__, __LINE__,	"Faktor	ist nicht zwischen 0 und 1, aber der gesuchte Punkt liegt im Profil\n");
				}

				return (true);
			}

		}

		LastPoint = Point;

		i++;
	}

	// Der gesuchte	Punkt liegt hinter dem letzten Punkt des Profils
	long Dx	= Point->X - LastPoint->X;
	long Dy	= Point->Y - LastPoint->Y;
	int Dm = Point->Meter -	LastPoint->Meter;

	double Faktor =	(meter - LastPoint->Meter) / Dm;

	*x = (long)(LastPoint->X + Dx *	Faktor + 0.5);
	*y = (long)(LastPoint->Y + Dy *	Faktor + 0.5);

	if (Faktor <= 1.0)
	{
		dump_error(__FILE__, __LINE__,	"Faktor	ist nicht grer als 1,	aber der gesuchte Punkt	liegt hinter dem Profil\n");
	}

	return (false);
}

//---------------------------------------------------------------------
// TGewaesserAchseList
//---------------------------------------------------------------------

//---------------------------------------------------------------------
TGewaesserAchseList::~TGewaesserAchseList(void)
{
	Clear();
}

//---------------------------------------------------------------------
void TGewaesserAchseList::Clear(void)
{
	for (TGewaesserAchseList::iterator i=begin(); i	!= end(); i++)
	{
		TGewaesserAchse	*GewaesserAchse	= *i;

		delete GewaesserAchse;
		GewaesserAchse = 0;
	}

	clear();
}




